# Simplified version of a Tcl/Tk DEMO - patched by Paul Tarau 1992
# All copyrigtht on this file goes to the original holders
# Possible new errors are mine...

# mkItems w
#
# Create a top-level window containing a canvas that displays the
# various item types and allows them to be selected and moved.  This
# demo can be used to test out the point-hit and rectangle-hit code
# for items.
#
# Arguments:
#    w -	Name to use for new top-level window.

proc mkItems {{w .citems}} {
    catch {destroy $w}
    toplevel $w
    dpos $w
    wm title $w "BinProlog-Tcl/Tk Canvas Demonstration"
    wm iconname $w "Items"
    wm minsize $w 100 100
    set c $w.frame2.c

    message $w.msg -font -Adobe-Times-Medium-R-Normal--*-180-* -width 13c \
	    -bd 2 -relief raised -text "This window contains a canvas widget with examples of the various kinds of items supported by canvases.  The following operations are supported:\n  Button-1 drag:\tmoves item under pointer.\n  Button-2 drag:\trepositions view.\n  Button-3 drag:\tstrokes out area.\n  Ctrl+f:\t\tprints items under area."
    frame $w.frame2 -relief raised -bd 2
    button $w.ok -text "OK" -command "destroy $w"
    pack $w.msg -side top -fill x
    pack $w.frame2 -side top -fill both -expand yes
    pack $w.ok -side bottom -pady 5 -anchor center

    canvas $c -scrollregion {0c 0c 20c 14c} -width 15c -height 10c \
	    -xscroll "$w.frame2.hscroll set" -yscroll "$w.frame2.vscroll set"
    scrollbar $w.frame2.vscroll  -relief sunken -command "$c yview"
    scrollbar $w.frame2.hscroll -orient horiz -relief sunken -command "$c xview"
    pack $w.frame2.hscroll -side bottom -fill x
    pack $w.frame2.vscroll -side right -fill y
    pack $c -in $w.frame2 -expand yes -fill both

    set font1 -Adobe-Helvetica-Medium-R-Normal--*-120-*
    set font2 -Adobe-Helvetica-Bold-R-Normal--*-240-*
    if {[tk colormodel $c] == "color"} {
	set blue DeepSkyBlue3
	set red red
	set bisque bisque3
	set green SeaGreen3
    } else {
	set blue black
	set red black
	set bisque black
	set green black
    }

    # Set up demos within each of the areas of the grid.

    $c create oval 1c 2c 3c 6c -outline $red -width 2m -tags item
    $c create text 6c 6c -anchor n -font $font1 -width 4c \
	    -text "A short string of text, word-wrapped, justified left, and anchored north (at the top)." -tags item
    $c create text 10c 7c -anchor w -font $font1 -fill $blue \
	    -text "Several lines,\n each centered\nindividually,\nand all anchored\nat the left edge." \
	    -justify center -tags item

    button $c.button -text "Press Me" -command "butPress $c $red"
    $c create window 6c 2c -window $c.button -anchor nw -tags item
    entry $c.entry -width 50 -relief sunken
    $c.entry insert end "Edit this text If you want to"
    $c create window 1c 1c -window $c.entry -anchor nw -tags item

    # Set up event bindings for canvas:

    $c bind item <Any-Enter> "itemEnter $c"
    $c bind item <Any-Leave> "itemLeave $c"
    bind $c <2> "$c scan mark %x %y"
    bind $c <B2-Motion> "$c scan dragto %x %y"
    bind $c <3> "itemMark $c %x %y"
    bind $c <B3-Motion> "itemStroke $c %x %y"
    bind $c <Control-f> "itemsUnderArea $c"
    bind $c <1> "itemStartDrag $c %x %y"
    bind $c <B1-Motion> "itemDrag $c %x %y"
    bind $w <Any-Enter> "focus $c"
}

# Utility procedures for highlighting the item under the pointer:

proc itemEnter {c} {
    global restoreCmd

    if {[tk colormodel $c] != "color"} {
	set restoreCmd {}
	return
    }
    set type [$c type current]
    if {$type == "window"} {
	set restoreCmd {}
	return
    }
    if {$type == "bitmap"} {
	set bg [lindex [$c itemconf current -background] 4]
	set restoreCmd [list $c itemconfig current -background $bg]
	$c itemconfig current -background SteelBlue2
	return
    }
    set fill [lindex [$c itemconfig current -fill] 4]
    if {(($type == "rectangle") || ($type == "oval") || ($type == "arc"))
	    && ($fill == "")} {
	set outline [lindex [$c itemconfig current -outline] 4]
	set restoreCmd "$c itemconfig current -outline $outline"
	$c itemconfig current -outline SteelBlue2
    } else {
	set restoreCmd "$c itemconfig current -fill $fill"
	$c itemconfig current -fill SteelBlue2
    }
}

proc itemLeave {c} {
    global restoreCmd

    eval $restoreCmd
}

# Utility procedures for stroking out a rectangle and printing what's
# underneath the rectangle's area.

proc itemMark {c x y} {
    global areaX1 areaY1
    set areaX1 [$c canvasx $x]
    set areaY1 [$c canvasy $y]
    $c delete area
}

proc itemStroke {c x y} {
    global areaX1 areaY1 areaX2 areaY2
    set x [$c canvasx $x]
    set y [$c canvasy $y]
    if {($areaX1 != $x) && ($areaY1 != $y)} {
	$c delete area
	$c addtag area withtag [$c create rect $areaX1 $areaY1 $x $y \
		-outline black]
	set areaX2 $x
	set areaY2 $y
    }
}

proc itemsUnderArea {c} {
    global areaX1 areaY1 areaX2 areaY2
    set area [$c find withtag area]
    set items ""
    foreach i [$c find enclosed $areaX1 $areaY1 $areaX2 $areaY2] {
	if {[lsearch [$c gettags $i] item] != -1} {
	    lappend items $i
	}
    }
    puts stdout "Items enclosed by area: $items"
    set items ""
    foreach i [$c find overlapping $areaX1 $areaY1 $areaX2 $areaY2] {
	if {[lsearch [$c gettags $i] item] != -1} {
	    lappend items $i
	}
    }
    puts stdout "Items overlapping area: $items"
}

set areaX1 0
set areaY1 0
set areaX2 0
set areaY2 0

# Utility procedures to support dragging of items.

proc itemStartDrag {c x y} {
    global lastX lastY
    puts "c = $c" 
    set lastX [$c canvasx $x]
    set lastY [$c canvasy $y]
}

proc itemDrag {c x y} {
    global lastX lastY
    set x [$c canvasx $x]
    set y [$c canvasy $y]
    $c move current [expr $x-$lastX] [expr $y-$lastY]
    set lastX $x
    set lastY $y
    set this [$c find closest $x $y]
    inform_prolog $this lastX $x
    inform_prolog $this lastY $y
}

# Procedure that's invoked when the button embedded in the canvas
# is invoked.

proc butPress {w color} {
    set i [$w create text 5c 5c -text "Hello!!" -fill $color -anchor n]
    after 500 "$w delete $i"
    p bb
}

# Position a dialog box at a reasonable place on the screen.

proc dpos w {
    wm geometry $w +300+300
}

proc inform_prolog {a b c} {
  p "bb_let('$a','$b',$c)"
}

mkItems .cw

source server.tcl

start_prolog


