/*
From dsmith@lucy.cs.waikato.ac.nz Thu Jul 11 00:10 ADT 1996
Date: Thu, 11 Jul 1996 15:00:33 +1200
From: dsmith@lucy.cs.waikato.ac.nz (Don Smith)
Subject: AND-threads via assumel
To: tarau@info.umoncton.ca
Content-Transfer-Encoding: 7BIT
Content-Type: text
Content-Length: 2579

Paul,

This implementation of threads via first-class continuations uses
assumel, as you suggested, to save and consume the continuation during 
task switching.  Hence it will only work for a single AND continuation:
but that should be fine!      So, assumel does NOT invoke copying?

When I do

	| ?- go.

it runs for a while, and then suddenly returns no. It should continue
forever.  It sill consumes increasing amounts of memory.

Maybe it's a problem with gc or bb_gc: when I explicitly call
cc and bb_gc in p1/2 below, it causes an error and dies with the message:

$assumed/0 ??? warning: *** bad stamp or empty slot in set/3 or rm/2 ***
>>> blackboard overflow, left only(128680): culprit(bb_put_1($assumed,...,...))
*** discarding inconsistent blackboard content

When I comment out the calls to gc and bb_gc, it runs for a while and then
just returns no.

             Don
*/


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
init_processes(Processes):- 
	bb_def(process,count,0), 
	load_processes(Processes).

load_processes([]):- !.  % Normally, this should never be called.
load_processes([Last]):- !,load_last_process(Last).
load_processes([H|T]):- load_process(H),load_processes(T).

load_process(P):- 
	bb_val(process,count,Count),
	NewCount is Count+1,
	bb_set(process,count,NewCount),
%	bb_def(process,Count,call(P,switch)).
	assumel(process(Count,call(P,switch))).

load_last_process(Last):- 
	bb_val(process,count,Count), 
	NewCount is Count+1,
	bb_set(process,count,NewCount),
	bb_def(process,current_pid,Count),
%	bb_def(process,Count,[]),  % Dummie holder
	switch,  % The continuation is the call to Last below.
	call(Last).


switch:- 
	bb_val(process,current_pid,OLD_PID), 
	bb_val(process,count,Count),
	NEW_PID is (OLD_PID+1) mod Count,
	bb_set(process,current_pid,NEW_PID), 
	switch_aux(OLD_PID,NEW_PID).

switch_aux(OLD_PID,NEW_PID,Cont)::- 
	assumel(process(OLD_PID,Cont),
		process(NEW_PID,NewCont,  % consume the old process continuation.
					NewCont)).
/*
switch_aux(OLD_PID,NEW_PID,Cont)::- 
	bb_set(process,OLD_PID,Cont,
		bb_val(process,NEW_PID,NewCont,
			bb_set(process,NEW_PID,[],  % To enable gc.
					NewCont))).
*/
	
% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

p1(N,0):- 0 is N mod 50000,statistics,nl,
	bb_gc,gc,    % CAUSES ERROR!
	statistics,nl,nl,fail.
p1(N,PID):- 0 is N mod 50000,write(PID),nl,fail.
p1(N,PID):- 0 is N mod 1000,!, 
%	write(PID),tab(1),flush_output,
	switch, 
	M is N+1,p1(M,PID).
p1(N,PID):- M is N+1,p1(M,PID).

go:- init_processes([ p1(0,0),p1(0,1), p1(0,2) ]).

