/*
 * interface module with the standard Prolog parser
 *
 */

#include "interf.h"
#include <stdio.h>

extern struct specsyms g;
extern string *atomtable;
extern cell new_func();

#define PUT_FLOAT(F) \
double_to_ints((F),&i1,&i2,&i3); \
MAKE_FLOAT(i1,i2,i3)

#define MAKE_FLOAT(I,J,K) \
{  H[0]=g.bp_float; \
   H[1]=(I); H[2]=(J); H[3]=(K);\
   H+=4;\
}


termtype *makefloat(Hptr,f)
register heapcell *Hptr;
double f;
/*
 * internatizes the float value f
 * and return it
 */
{
  register term saveH = *Hptr, H = *Hptr;
  no i1, i2, i3;

  PUT_FLOAT(f);
  *Hptr = H;
  return (termtype *)saveH;
}

termtype *makeinteger(Hptr,i)
register heapcell *Hptr;
int i;
/*
 * internalizes the integer value i
 * and return it
 */
{
  return (termtype *)INPUT_INT(i);
}

termtype *makeatom(Hptr,s)
register heapcell *Hptr;
char *s;
/*
 * internalizes the string s and returns
 * it as atom
 */
{
  return (termtype *)new_func(s,0);
}

termtype *maketerm(Hptr,s,n)
register heapcell *Hptr;
char *s;
int n;
/*
 * internalizes the functor string s with arity n
 * and returns a pointer to it
 */
{
  register term H = *Hptr;
  int i;

  *H = new_func(s,n);
  *Hptr += n+1;
  return (termtype *)H;
}

termtype *makevar(Hptr) 
register heapcell *Hptr;
/*
 * creates a variable and returns it
 */
{
  register heapcell H = *Hptr;
  *H = (cell)H;
  *Hptr += 1;
  return (termtype *)H;
}

symboltype *getsymbol(s,a)
char *s;
int a;
/*
 * return the internalized version of the functor s with
 * arity a; create the symbol if it does not already exist
 */ 
{
  return (symboltype *)new_func(s,a);
}

termtype *deref(tp)
term tp;
/* no comment */
{
  cell xval;
  if (!GETTAG(tp)) {
    DEREF2(tp,xval);
  }
  if (ISVAR(tp) && (ISINT(*(term)tp) || ISATOM(*(term)tp)))
    return (termtype *)(*(term)tp);
  else
    return (termtype *)tp;
}

int fileread(buf, len)
register char *buf;
register int len;
/*
 * read at most len bytes from stdin into buf
 * returns the number of bytes read
 */
{
#if 1
  char *p;

  p  = fgets(buf, len, inputfile);
  if (p) return strlen(buf); else return 0;
#else
  if (inputfile) {
    return read(fileno(inputfile), buf, len); 
  } else {
    return 0;
  }
#endif
}

int operator(op, prepri, inpri, postpri, spec)
char *op;
int *prepri, *inpri, *postpri, *spec;
/*
 * operator interface.
 * given a string representation op,
 * it returns the priority as prefix operator (prepri),
 * infix operator (inpri) and postfix operator (postpri).
 * In addition it returns the specifier of the operator.
 * This is the OR of all the possible (overloaded)
 * specifiers. The function returns 1 is op is an
 * operator, 0 otherwise 
 */
{
  cell opsym = new_func(op, 0);
  cell xval;

  *spec = 0;
  xval=hget(opsym,g.prefixop); 
  if(xval && g.empty!=xval) {
    *prepri = OUTPUT_INT(xval);
    xval = hget(g.prefixop,opsym);
    if (xval == g.fx) *spec |= FX;
    else if (xval == g.fy) *spec |= FY;
    else fatal_error("no assoc for prefix operator");
  } else *prepri = 0;
  xval=hget(opsym,g.infixop); 
  if(xval && g.empty!=xval) {
    *inpri = OUTPUT_INT(xval);
    xval = hget(g.infixop,opsym);
    if (xval == g.xfx) *spec |= XFX;
    else if (xval == g.xfy) *spec |= XFY;
    else if (xval == g.yfx) *spec |= YFX;
    else fatal_error("no assoc for infix operator");
  } else *inpri = 0;
  xval=hget(opsym,g.postfixop); 
  if(xval && g.empty!=xval) {
    *postpri = OUTPUT_INT(xval);
    xval = hget(g.postfixop,opsym);
    if (xval == g.xf) *spec |= XF;
    else if (xval == g.yf) *spec |= YF;
    else fatal_error("no assoc for postfix operator");
  } else *postpri = 0;
  return *prepri + *inpri + *postpri;
}

void putvarname(buffer, t)
register char *buffer;
register termtype *t;
/*
 * this functions writes a null terminated string
 * in buffer, corresponding to the output format
 * for variable t
 */
{
  sprintf(buffer, "_%d", t);
}
