/*
 * This file contains the interface with the underlying
 * Prolog system. Put here all the include files you
 * need to complete the defines that follow
 */

#include "global.h"

extern struct specsyms g;
extern FILE *inputfile;
double ints_to_double();

typedef cell symboltype;
typedef term termtype;
typedef term heapcell;

/**************************************************
 * here starts the obligatory part of interface.h
 **************************************************/

/***************** type tests ********************/
/*   t is a term */

#define ISFLOAT(t)    (!GETTAG((term)(t)) && BP_FLOAT(*((term)(t))))
#define ISINT(t)      (INTEGER(t))
#define ISATOM(t)     (SYMCONST(t))
#define ISTERM(t)     (!GETTAG((term)(t)) && COMPOUND(*((term)(t))))
#define ISVAR(t)      (!GETTAG(t))

/***************** value definitions *************/
/* these definitions can be used as lvalue and
   as rvalues!  They will have to be redefined
   by the user.  **********************************/

/* floating point value */
#define __FLOAT(t)    (ints_to_double(((term)t)[1],((term)t)[2],((term)t)[3]))

/* integer value */
#define __INT(t)      (OUTPUT_INT(t))

/* internalized atom */
#define __ATOM(t)     (t)

/* string representation of an atom t */
#define __ATOMNAME(t) (NAME(t))

/* string representation of a compound term t  */
#define __TERMNAME(t) (NAME(*((term)(t))))

/* arity of a compound term  t */
#define __ARITY(t)    (GETARITY(*((term)(t))))

/* argument n of compound term t */
#define __ARG(t,n)    (((termtype **)(t))[n])

/* last argument of a compound term t
   some implementations (eg BinProlog) use a compaction
   scheme that treats the last argument of a compound
   term different from the other arguments */
#define __LASTARG(t,n) (&(((termtype *)(t))[n]))


/********************* fixed part of interface.h   
 This part should not be modified by the user  *******/

termtype *makefloat();
termtype *makeinteger();
termtype *makeatom();
termtype *maketerm();
termtype *makevar();
symboltype *getsymbol();
termtype *deref();
int fileread();
int operator();
void putvarname();


/************* defines used to deal with operators */

#define XFX       0x0001
#define XFY       0x0002
#define YFX       0x0004
#define XF        0x0010
#define YF        0x0020
#define FX        0x0040
#define FY        0x0080
#define DELIMITER 0x0100
#define TERM      0x1000
#define LTERM     0x3000

#define isprefix(X)    ((X) & (FX|FY))
#define ispostfix(X)   ((X) & (XF|YF))
#define isinfix(X)     ((X) & (XFX|XFY|YFX))

#define isfx(X)        ((X) & FX)
#define isfy(X)        ((X) & FY)
#define isxf(X)        ((X) & XF)
#define isyf(X)        ((X) & YF)
#define isxfx(X)       ((X) & XFX)
#define isxfy(X)       ((X) & XFY)
#define isyfx(X)       ((X) & YFX)

#define isop(X)        ((X) & (XFX|XFY|YFX|XF|YF|FX|FY))
#define isterm(X)      ((X) & TERM)

#define islterm(X)     (((X) & LTERM) == LTERM)
#define isdelimiter(X) ((X) & DELIMITER)
