% objects with backtrackable state
% based on intuitionistic assumption

% our primitive constructors
relation(proto_of).
relation(part_of).
relation(inst_of).
relation(state_of).

proto(X):-clone(X,proto).

part(Body,Part):-clone_rel(part,Body,Part).

inst(Body,Part):-clone_rel(inst,Body,Part).

state(Body,Part):-clone_rel(state,Body,Part).

% creating a new relation
clone_rel(Rel,Body,Part):-
  symcat(Body,Part,BodyPart),
  clone(BodyPart,proto),
  symcat(Rel,of,RelOf),
  Pred=..[RelOf,Body,Part],
  add_rel(Pred).

% creating a new link
clone(X,Name):-
  symcat(Name,of,Link), % creates something like `..._of'
  create_clone(Link,X,Name). %

create_clone(Op,Name,Parent):-
  get_op(Op),
  Rel=..[Op,Name,Parent],
  add_or_call(Name,Op,Rel).

add_or_call(Name,Link,Rel):-
  nonvar(Name),!, % create if Name is given
  add_rel(Rel),
  term_append(Name,arg(X),Pred),
  add_link(Pred,Link,Name,X).
add_or_call(_Name,_Link,Rel):- 
  % call otherwise and retrieve _Name and _Link
  Rel.

get_op(Op):-nonvar(Op),!.
get_op(Op):-relation(Op).

add_rel(R):-assumed(R),!,fail.
add_rel(R):-assumei(R).

add_link(Pred,_,_,_):-is_assumed(Pred),!. % assumes at most once
add_link(Pred,Link,Name,X):-
  % this makes Pred have the ability of replication
  assumei( ( Pred:-create_clone(Link,X,Name) ) ).

% example of inheritance
is_a(Super,Inst):-assumed(inst_of(Class,Inst)),inherits_from(Super,Class).

inherits_from(X,X).
inherits_from(X,Z):-assumed(proto_of(X,Y)),inherits_from(Y,Z).

  
go:-
  proto(animal),
    animal(mamal),
      mamal(dog),
      mamal(cat),
  proto(friendly),
    friendly(dog),
    friendly(cat),
  part(dog,leg),
  inst(dog,fido),
  inst(cat,minou),
  inst(dog_leg,no_3),
  state(no_3,broken)
  ,listing
  .


go1:-
 proto(animal),animal(dog),inst(dog,fido),part(dog,nose),listing.

/*
 add exceptions, overriding
 add methods: sink, source, arrow
*/


moo:-
  proto(room),
  proto(port),
  room(blue1),
  room(red1),
  port(north1),
  port(east1),
  part(blue,north1),
  listing.

/*

% assumed dog_leg/1:
dog_leg(A) :- 
        create_clone(proto_of,A,dog_leg).

% assumed mamal/1:
mamal(A) :- 
        create_clone(proto_of,A,mamal).

% assumed part_of/2:
part_of(dog,leg).

% assumed inst_of/2:
inst_of(dog,fido).
inst_of(cat,minou).
inst_of(dog_leg,no_3).

% assumed no_3_broken/1:
no_3_broken(A) :- 
        create_clone(proto_of,A,no_3_broken).

% assumed cat/1:
cat(A) :- 
        create_clone(proto_of,A,cat).

% assumed dog_fido/1:
dog_fido(A) :- 
        create_clone(proto_of,A,dog_fido).

% assumed dog/1:
dog(A) :- 
        create_clone(proto_of,A,dog).

% assumed animal/1:
animal(A) :- 
        create_clone(proto_of,A,animal).

% assumed dog_leg_no_3/1:
dog_leg_no_3(A) :- 
        create_clone(proto_of,A,dog_leg_no_3).

% assumed state_of/2:
state_of(no_3,broken).

% assumed proto_of/2:
proto_of(animal,proto).
proto_of(mamal,animal).
proto_of(dog,mamal).
proto_of(cat,mamal).
proto_of(friendly,proto).
proto_of(dog,friendly).
proto_of(cat,friendly).
proto_of(dog_leg,proto).
proto_of(dog_fido,proto).
proto_of(cat_minou,proto).
proto_of(dog_leg_no_3,proto).
proto_of(no_3_broken,proto).

% assumed cat_minou/1:
cat_minou(A) :- 
        create_clone(proto_of,A,cat_minou).

% assumed friendly/1:
friendly(A) :- 
        create_clone(proto_of,A,friendly).


The LogiMOO object model.

We have experimented with various object models during the development of
LogiMOO. A prototype and delegation based model similar to SELF 4.0 has
been finally retained.

*/
