:-op(960,xfx,<=).

create_task_loop(E):-
  create_engine(E),
  load_engine(E,task_loop(E,X),X).

task_loop(E,X):-
  retract1(task_on_engine(E,Goal,X))->once(Goal)
 ;!,asserta(empty_engine(E)),fail.
task_loop(E,X):-
   task_loop(E,X).

add_task(E,Goal,X):-
   ( retract1(empty_engine(E))->load_engine(E,task_loop(E,X),X)
   ; true
   ),
   assertz(task_on_engine(E,Goal,X)).

ask_task_loop(E,X):-ask_engine(E,X).

destroy_task_loop(E):-
   retractall(task_on_engine(E,_,_)),
   retractall(empty_engine(E)),
   destroy_engine(E).


ttest:-
   create_task_loop(E),
   write('creating engine'(E)),nl,
   add_task(E,A=1,A),
   add_task(E,A=2,A),
   add_task(E,A=3,A),
   ask_task_loop(E,X),write(X),nl,
   ask_task_loop(E,Y),write(Y),nl,
   add_task(E,A=4,A),
   ask_task_loop(E,Z),write(Z),nl,
   ask_task_loop(E,U),write(U),nl,
   (
   \+ ask_task_loop(E,_)->destroy_task_loop(E)
   ; write('unable to destroy'(E))
   ).


% executes Producer on an `orthogonal' OR-engine
Answer<=Producer=>Consumer:-!, % the consumer is explicit
  cross_run(Producer,Answer,Consumer).
Answer<=Producer:-             % the consumer is in the `future'
  cross_run_with_current_continuation(Producer,Answer).
  
cross_run(Producer,Answer,Consumer):-
  default_engine_params(H,S,T),
  cross_run(H,S,T,Producer,Answer,Consumer).

cross_run(H,S,T,Producer,Answer,Consumer):-
   compound(Answer),copy_term(Answer,A),
   new_engine(H,S,T,Producer,Answer,Handle),
   (A:-ask_or_kill(Handle,A))=>Consumer,
   destroy_engine(Handle).

cross_run_with_current_continuation(Producer,Answer):-
   default_engine_params(H,S,T),
   cross_run_with_current_continuation(H,S,T,Producer,Answer).

cross_run_with_current_continuation(H,S,T,Producer,Answer):-
   compound(Answer),copy_term(Answer,A),
   new_engine(H,S,T,Producer,Answer,Handle),
   assumei((A:-ask_or_kill(Handle,A))).

ask_or_kill(Handle,Answer):-ask_engine(Handle,Answer),!.
ask_or_kill(Handle,_):-destroy_engine(Handle),fail.

new_engine(H,S,T,Goal,Answer,Handle):-
  create_engine(H,S,T,Handle),
  load_engine(Handle,Goal,Answer).

find_all(X,Goal,Xs):-
  default_engine_params(H,S,T),
  find_all(H,S,T,X,Goal,Xs).

find_all(H,S,T,X,Goal,Xs):-
  new_engine(H,S,T,Goal,X,Handle),
  collect_answers(Handle,Xs),
  destroy_engine(Handle).

collect_answers(Handle,[X|Xs]):-ask_engine(Handle,A),!,
  copy_term(A,X),
  collect_answers(Handle,Xs).
collect_answers(_,[]).


call_and_fail(Goal):-
  default_engine_params(H,S,T),
  call_and_fail(H,S,T,Goal).

call_and_fail(H,S,T,Goal):-
  new_engine(H,S,T,Goal,X,Handle),
  (ask_engine(Handle,X),fail
   ;destroy_engine(Handle)
  ).

call_once(Goal):-
  default_engine_params(H,S,T),
  call_once(H,S,T,Goal).

call_once(H,S,T,Goal):-
  copy_term(Goal,G),
  new_engine(H,S,T,G,G,Handle),
  ask_engine(Handle,X),
  copy_term(X,Goal),
  destroy_engine(Handle).

% overridable default engine sizes for heap,stack,trail
default_engine_params(H,S,T):-assumed(engine_params(H,S,T)),!.
default_engine_params(128,32,32).

% tests
cross_test:-
  a(V)<=member(V,[s(A),s(B),c(A,B)]),
  a(X),a(Y),a(Z),
  write([V,A,B,X,Y,Z]),nl,
  listing,
  destroy(a(_)),
  write('engine closed'),nl.

destroy(AnswerPattern):-
  assumed_clause(
    AnswerPattern,
    ask_or_kill(Handle,AnswerPattern),
    intuitionistic(Scope)
  ),
  destroy_engine(Handle),
  Scope='$closed'. % close the scope of the assumption
  
cross_impl_test:-
  engine_params(100,50,50)=> % overrides the default engine parameters
  ( G=member(A,[a(B),1,s(B)]),
    a(A)<=G=> (a(X),a(Y),a(Z),write(G+[X,Y,Z]),nl)
  ).

find_all_test:-
  As=[A,B,B,A],
  find_all(X,member(s(X),As),Xs),write(As+Xs),nl.

engine_test:-
  find_all_test,
  call_and_fail((X=f(X),write(X),nl)),
  nl,write('engine dies gracefully'),nl,
  call_once(member(A,[B,B])),
  write(A+B),nl.

simple_test:-
  a(X)<=member(X,[1,2,3])=>(a(A),a(B),a(C),write([A,B,C]),nl).

exec_tests([]).
exec_tests([G|Gs]):-
  write(test=G),nl,
  (G->write(suceeds);write(fails)),
  nl,
  exec_tests(Gs).


% emulating assert on top of an extra engine

new_engine(E):-
    default_engine_params(H,S,T),
    new_engine(H,S,T,make_worker(X),X,E).

send_engine(E,Goal):-assumel(todo(Goal)),ask_engine(E,_).

make_worker(E):-current_engine(E),worker.

worker:-
  assumed(todo(X)),X\==done,
  copy_term(X,G),
  (G->true;true),
  co_worker.

co_worker.
co_worker:-worker.

eassert(E,A):-send_engine(E,assumei(A)).

assert_test:-
  new_engine(E),
     send_engine(E,ttyprint(hello)),
     eassert(E,a(1)),
     eassert(E,a(2)),
     listing,
  send_engine(E,done),
  destroy_engine(E).

gc_test(Max):-
  new_engine(E),
     send_engine(E,statistics),
     ( for(I,1,Max),
        eassert(E,a(I)),
        send_engine(E,statistics),
       fail
     ; ttyprint(asserts=Max)
     ),
  send_engine(E,done),all_tests,
  destroy_engine(E).

gc_test:-gc_test(200).

all_tests:-
  exec_tests([cross_test,cross_impl_test,find_all_test,simple_test,
  assert_test]).

%:-[boyer].
