/*
Hi Veronica,

I have worked out what we have discussed on order insensitive use of AGs.
Basically, 2 constructs came out:

=X which `joins' all assumptions matching a given pattern (= hints
towards the fact that +X and -X actually play symmetric roles in this case,
no need for 2 notations here).

Two `brodacast/consume' constructs:

?X registers a consumer of X
*X broadcasts to all consumers (using intuitionistic implication)

The distinction between ? and * cannot be avoided because
a clear producer/consumer dichotomy here.

I will attach the code.


Paul

-------------------------------------------------------------------------
*/

:-op(400,fx,(*)).
:-op(400,fx,(?)).
:-op(400,fx,(=)).
:-op(1150,xfx,(==>)).

% joins H with matching assumptions and executes action B
(H==>B):-assumed(H),!,B.
(H==>B):-assumel((H:-B)).

% joins X with matching facts
=X:-add_true(X,(H:-B)),(H==>B).

?X:-add_true(X,(H:-B)), register(H,B).
 
% registers interests of X for matching broadcast
register(H,B):-assumed(H),!,B.
register(H,B):-assumel((waiting(H):-B)).

% broadcasts X to all registered matching assumptions waiting for it
*X:-add_true(X,(H:-B)),apply_to_all_assumed(H,B),assumei((H:-B)).

apply_to_all_assumed(H,B):-
   copy_term((H:-B),(CH:-CB)),
   assumed(waiting(CH)),
   CB,
   !,
   apply_to_all_assumed(H,B).
apply_to_all_assumed(_,_).

% examples
go1:- =a(1), (a(X)==>write(evrika(X)),nl).

go2:-  (a(X)==>write(evrika(X)),nl), =a(2), =a(3).

go:-go1,go2.

xgo:- ?a(5), ?a(10), *((a(X):-write(evrika(X)),nl)), ?a(20), ?a(30).

/*
?- go1.
evrika(1)
yes
?- go2.
evrika(2)
yes
?- xgo.
evrika(5)
evrika(10)
evrika(20)
evrika(30)
yes
*/



/* possible IMPACT on coordination model:

- linda: client, server oriented, lack of symmetry between in/out
  => complex temporal semantics

we propose a unique join operation:


=a(X) if a(Y) is there, remove it and return X=Y, if not put a(X)

H ==> B

if (H:-B1) is there take it, exec B1 and do B
if not, put (H:-B)

more than the usual client/server:
multiple server arch: web of active worlds


RPC: call and wait for answer: wrong, creates unnecessary blocking
WOW (Web Of Worlds): call, schedule callback, leave
                     on callback execute on pear

*/
