blackboards(X)  :- bb_op(4,_,X).
blackboard(B)   :- bb_op(8,B,_).
blackboard(B,H,P,N) :- bb_op(9,bb(B,H,P,N),_).
bbkill(stdbb)   :- !, fail.
bbkill(B)       :- bb_op(10,B,_).
kill(P)         :- bb_op(13,P,_).
bbclean(B)      :- bb_op(11,B,_).
bbclean         :- bb_op(11,stdbb,_).
bbempty(B)      :- bblock(B), bbterms(B,[]), bbprocesses(B,[]), bbunlock(B).
bbempty         :- bbempty(stdbb).
bbterms(B,X)    :- bb_op(15,B,X).
bbterms(X)      :- bb_op(15,stdbb,X).
bbprocesses(B,X):- bb_op(16,B,X).
bbprocesses(X)  :- bb_op(16,stdbb,X).
bbinfo(B,X)     :- bb_op(17,B,X).
bbinfo(X)       :- bb_op(17,stdbb,X).
processinfo(P,I):- bb_op(3,P,I).
processes(P)    :- bb_op(19,_,P).
server(X)       :- bb_op(20,X,X).
killserver(X)   :- bb_op(21,X,X).
bbmem(B,L)      :- bb_op(18,B,L).
bbmem(L)        :- bb_op(18,stdbb,L).
bbmalloc(S)     :- bb_op(14,bb(stdbb,S),_).
bbmalloc(B,S)   :- bb_op(14,bb(B,S),_).
bbfree(S)       :- bb_op(12,bb(stdbb,S),_).
bbfree(B,S)     :- bb_op(12,bb(B,S),_).
poolmem(L)      :- bb_op(5,_,L).
tellp(P)        :- bb_op(7,P,_).
testmem         :- bb_op(6,1,_).

bbhelp :- 
  write('tellt/1, tellt/2'), nl,
  write('gett/1, gett/2, gett/3'), nl,
  write('readt/1, readt/2, readt/3'), nl,
  write('gett_nb/1, gett_nb/2, gett_nb/3'), nl,
  write('readt_nb/1, readt_nb/2, readt_nb/3'), nl,
  write('blackboard/1'), nl, %, blackboard/4'), nl,
  write('blackboards/1'),nl,
  write('processes/1'), nl,
  write('bbinfo/2, processinfo/2'), nl,
  write('bbshow/0, bbshow/1'), nl,
  write('bbterms/0, bbterms/1'), nl,
  write('bbprocesses/0, bbprocesses/1'), nl,
  write('bbkill/1'), nl,
  write('kill/1'), nl,
  write('bbclean/0, bbclean/1'), nl,
  write('bbempty/0, bbempty/1'), nl,
  write('bbmem/1, bbmem/2'), nl,
  write('bbmalloc/1, bbmalloc/2'), nl,
  write('bbfree/1, bbfree/2'), nl,
  write('poolmem/1'), nl,
  write('bblock/1, bbunlock/1'), nl.
%  write('server/1, killserver/1'), nl.
  
bbshow :- bbshow(stdbb).

bbshow(B) :- 
  bbinfo(B,I), write(I), nl,
  bbterms(B,T), write(terms=T), nl,
  bbprocesses(B,P), write(processes=P), nl.

totalbbmem(T) :- totalbbmem(stdbb,T).
totalbbmem(B,T) :- bbmem(B,L), sumprod(L,0,T).

sumprod([],T,T).
sumprod([buddy(N,S)|L],A,T) :- A1 is A + N*S, sumprod(L,A1,T).


bblock(B) :- bb_op(1,B,_).
bblock(B) :- bb_op(2,B,_), fail.

bbunlock(B) :- bb_op(2,B,_).
bbunlock(B) :- bb_op(1,B,_), fail.


bblock :- bblock(stdbb).
bbunlock :- bbunlock(stdbb).

tellt(T) :- put__(stdbb,T).
tellt(B,T) :- put__(B,T).

put__(B,T) :- nonvar(T), T={A}, !,
              make_list__(A,L), bblock(B), put_set__(L,B), bbunlock(B).
put__(B,T) :- bb_put(B,T).

put_set__([],_B).
put_set__([X|L],B) :- my_bb_put(B,X), put_set__(L,B).

my_bb_put(B,X) :- bb_put(B,X). /* should not be inline builtin */

kludge. /* to suppress bug */

get__(B,T,C,M) :- nonvar(T), T={A}, !,
                  make_list__(A,L), get_set__(M,B,L,C).
get__(B,T,C,M) :- get__(B,T,C,M,0).

get__(B,T,C,M,_X) :- bb_get(B,T,C,M).
get__(_B,_T,_C,_M,_X). /* special trick to create choice point */

gett(T)                :- get__(stdbb,T,true,3).
gett_nb(T)             :- get__(stdbb,T,true,2).
readt(T)               :- get__(stdbb,T,true,1).
readt_nb(T)            :- get__(stdbb,T,true,0).

gett(B,T)              :- get__(B,T,true,3).
gett_nb(B,T)           :- get__(B,T,true,2).
readt(B,T)             :- get__(B,T,true,1).
readt_nb(B,T)          :- get__(B,T,true,0).

gett(B,T,C)            :- get__(B,T,C,3).
gett_nb(B,T,C)         :- get__(B,T,C,2).
readt(B,T,C)           :- get__(B,T,C,1).
readt_nb(B,T,C)        :- get__(B,T,C,0).

check__([],_B) :- !.
check__([X|L],B) :- readt_nb(B,X), check__(L,B).

checkb__([],_).
checkb__([X|L],B) :-  readt_nb(B,X), checkb__(L,B).
checkb__([X|_],B) :- readt(B,X), !, fail.

removeb__([],_B).
removeb__([X|L],B) :- gett_nb(B,X), !, removeb__(L,B).
removeb__([_X|_],_) :- fail.

get_set__(3,B,S,C) :- bblock, repeat, checkb__(S,B), metacall(C),
                              removeb__(S,B), bbunlock.
get_set__(2,B,S,C) :- bblock, check__(S,B), metacall(C), 
                             removeb__(S,B), bbunlock.
get_set__(1,B,S,C) :- bblock, repeat, checkb__(S,B), metacall(C), bbunlock.
get_set__(0,B,S,C) :- bblock, check__(S,B), metacall(C), bbunlock.

make_list__(T,[X|L]) :- functor(T,',',_N), !, T=(X,Y), make_list__(Y,L).
make_list__(X,[X]).

