/*
 *  Memory allocator based on a buddy system algorithm
 *  
 *  28/02/93 Initial version: Koen De Bosschere
 */

#ifndef _BUDDYMALLOC_H
#define _BUDDYMALLOC_H

#define BUDDYMAX 32
#define BUDDYSIZE ((BUDDYMAX+1)<<2)

typedef struct buddyblock {
  struct buddyblock *next;
} buddyblock_t;

typedef buddyblock_t *buddycontrolblock_t[BUDDYMAX+1];

void buddycreate(buddyblock_t **buddy, int size);           
/* creation of a the buddy system */
void buddyinit(buddyblock_t **buddy, buddyblock_t *address, unsigned int size);             
/* initialization of an existing system */
unsigned int buddymax(buddyblock_t **buddy);          
/* size of biggest available block */
unsigned int buddymem();          
/* total available memory */
int buddyshow(buddyblock_t **buddy);              
/* show the contents of a buddy system */
buddyblock_t *buddymalloc(buddyblock_t **buddy, unsigned int size);  
/* malloc a block */
void buddyfree(buddyblock_t **buddy, buddyblock_t *address, unsigned int size);             
/* free a block */
unsigned int realbuddysize(unsigned int size);     
/* returns real power of two */

#endif
