% Linda-based parallel findall 
% Author: Paul Tarau
% Last uptdate: Sat May  6 11:44:49 ADT 1995

/*

This program shows how a lazy parallel logic engine can be implemented
using Multi-BinProlog. The background engine is seen as an abstract
data-type consisting of a (possibly infinite) lazily produced streams of
answers.  It can be opened with bb_open_engine/3 and closed either when no
more answers are available or because the master process thinks its
answers are not needed anymore.  Answers produced by the engine can
also be gathered through a re-entrant findall primitive.  On machines
with more than one CPU, our multiple parallel engines materialize as
threads scheduled on different processors and can give signifacant
speed-up over sequential execution.
We have measured an almost 4-times speed-up for a graph-coloring program
on a 4 processor Sparcstation 10.

Koen, please adapt mcolor.pl and the parsec script 
for your 8-processor SS1000.

*/


% master operations

% creates a parallel engine
bb_open_engine(G,X,Handle):-
  blackboard(Handle),
  tellt(Handle,fuel(go)),
  tellp(engine(Handle,X,G)).

% gets answer(C,X) or C=yes or no from the parallel engine
query_engine(Handle,Ok,Answer):- 
   gett(Handle,answer(Ok,X)),
   !,
   tellt(Handle,fuel(go)),
   Ok=yes,
   Answer=X.

close_engine(Handle):-
   bbkill(Handle).

% slave operations

engine(Handle,X,G):-
  G,
  getfuel(Handle),
  tellt(Handle,answer(yes,X)),
  fail.
engine(Handle,_,_):-
  tellt(Handle,answer(no,[])),
  getfuel(Handle).

getfuel(Handle) :- gett(Handle,fuel(_)), !.

% application: parallel findall

query_answers(Handle,As):-
   gett(Handle,answer(C,A)),
   !,
   query_answers0(C,A,Handle,As).

query_answers0(no,End,_Handle,End).
query_answers0(yes,X,Handle,[X|As]):-
   tellt(Handle,fuel(go)),
   query_answers(Handle,As).

par_findall(X,G,Xs):-
   bb_open_engine(G,X,Handle),
   query_answers(Handle,Xs),
   close_engine(Handle).

go:-par_findall(X,member(X,[1,2,3,4,5]),Xs),write(Xs),nl.
