% Emulating orthogonal OR-engines in Multi-BinProlog
% Author: Paul Tarau
% Last uptdate: Sun May 21 01:20:45 ADT 1995

:-op(960,xfx,<=).

:-[all].

% executes Producer on an `orthogonal' OR-engine
Answer<=Producer=>Consumer:-!,        % the consumer is explicit
   cross_run(Producer,Answer,Consumer).
Answer<=Producer:-                    % the consumer is in the `future'
   cross_run_with_current_continuation(Producer,Answer).

cross_run(Producer,Answer,Consumer):-
   compound(Answer),copy_term(Answer,A),
   bb_open_engine(Producer,Answer,Handle),
   (A:-ask_or_kill(Handle,A))=>Consumer,
   close_engine(Handle).

cross_run_with_current_continuation(Producer,Answer):-
   compound(Answer),copy_term(Answer,A),
   bb_open_engine(Producer,Answer,Handle),
   assumei((A:-ask_or_kill(Handle,A))).

ask_or_kill(Handle,Answer):-
   query_engine(Handle,Ok,Answer),
   ( Ok=yes->true
   ; close_engine(Handle)
   ).

cross_test:-
  a(V)<=member(V,[s(A),s(B),c(A,B)]),
  a(X),a(Y),a(Z),
  write([V,A,B,X,Y,Z]),nl,
  destroy(a(_)),
  write('engine closed'),nl.

cross_impl_test:-
  engine_params(30,10,10)=> % overrides the default engine parameters
  ( G=member(A,[a(B),1,s(B)]),
    a(A)<=G=> (a(X),a(Y),a(Z),write(G+[X,Y,Z]),nl)
  ).

destroy(AnswerPattern):-
  assumed_clause(
    AnswerPattern,
    ask_or_kill(Handle,AnswerPattern),
    intuitionistic(Scope)
  ),
  close_engine(Handle),
  Scope='$closed'. % close the scope of the assumption
