#include "global.h"
#include "bboard.h"
#include <stdio.h>
#include <signal.h>
#include <rpc/rpc.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <thread.h>
#include <assert.h>
#include "rpc.h"
#include <netconfig.h>

#if TRACE > 0 || PROF
extern void *calloc_trace(no nb,no size),
            *malloc_trace(no nb,no size),
            free_trace(void *ptr);
#endif

 
/*LINTLIBRARY*/

typedef struct {
  term  data;
  u_int datalen;
} rpcterm;

typedef struct {
   char         *bb_name;
   unsigned int bl_size;
} rpcmallocterm;  

typedef struct {
	char    *bb_name;
	rpcterm termlist;
} rpcallterm_node;


extern struct specsyms g;
extern string          *atomtable, sbuf, stop;

thread_t     serverthread = 0;
thread_t     killerthread = 0;
long         serverprog = 0;
struct stack globalwam[MaxStk];
sema_t       rpcserversema;
SVCXPRT      *transp;


bool_t
xdr_rpcterm(xdrsp, rpctrm)
  XDR     *xdrsp;
  rpcterm *rpctrm;
{ 
  return xdr_array(xdrsp, (caddr_t *)&rpctrm->data,
          &rpctrm->datalen, 1<<14,(u_int)sizeof(unsigned long), xdr_u_long); 
}


bool_t
xdr_rpcmalloc(xdrsp, mallocterm)
  XDR           *xdrsp;
  rpcmallocterm *mallocterm;
{
 if (!xdr_string(xdrsp, (char **)&mallocterm->bb_name,NAMESIZE) )
 	return (FALSE);
 if (!xdr_u_int(xdrsp, &mallocterm->bl_size) )	
 	return (FALSE);
 return (TRUE);	
}


bool_t
xdr_allboard(xdrsp, allterm)
  XDR             *xdrsp;
  rpcallterm_node *allterm;
{
 if (!xdr_string(xdrsp, (char **)&allterm->bb_name, NAMESIZE) )
 	return FALSE;
 if (!xdr_rpcterm(xdrsp, (rpcterm *)&allterm->termlist) )
 	return FALSE;
 return TRUE;
}


int term2reloc(t,buffer,wam,len)
  term         t, *buffer;
  struct stack *wam;
  int          len;
/******************************************************************************/
{
  int l,i;
  char *cp;
  buffer += 1;
  copytobuffer((term)buffer,t,wam,len>>2); 
  l = term_size((term)buffer) >> 2;
  cp = (char*)(buffer+l);
  for (i=0; i<l; i++) {

    if (VAR(buffer[i])) {
      buffer[i] = (term)((char *)(buffer[i])-(char *)buffer);
    } else if (IDENTIFIER(buffer[i])) {
      size_t n = strlen(NAME(buffer[i]));
      if ((cp-(char *)buffer) > len) {
        fprintf(stderr, "Heap overflow in term2reloc\n");
        thr_exit(0);
      }
   
      strcpy(cp,NAME(buffer[i]));
      buffer[i] = (term)(PUTARITY((PUTSYMNO(FUNTAG,(cp-(char*)buffer))),
                                  (GETARITY(((cell)(buffer[i]))))));
      cp += n+1;
    } else if (BP_FLOAT((T2C(buffer[i])))) {
      puts("Float not supported");
      i += 2;
    }
  } 
  buffer -= 1;
  buffer[0] = (term)l;
  
  return (int)(((cp-(char*)buffer)&(~3))+4);
}


term reloc2term(buffer,H,wam)
  term         H,buffer;
  struct stack *wam;
/******************************************************************************/
{
  int i;
  int n = (int)buffer[0];
  buffer += 1;
  for (i=0; i<n; i++) {
    if (VAR(buffer[i])) H[i] = (cell)((char*)(buffer[i])+(int)H);
    else if (IDENTIFIER(buffer[i])) {
      H[i] = new_func((((char*)buffer)+GETSYMNO((cell)(buffer[i]))), 
                             GETARITY((cell)(buffer[i])));
    } else if (BP_FLOAT(T2C(buffer[i]))) {
      puts("Float not supported");
      i += 2;
    } else
      H[i] = buffer[i];
  }
  return H+n;
}


/******************************************************************************
 *                                                                            *
 *                                     SERVER                                 *
 *                                                                            *
 ******************************************************************************/

instr start_put()
/******************************************************************************/
{ 
  instr p=(instr)hget(g.predmark, new_func("put__",3));
  if(!p) fatal_error("no definition for: put__/3");
  return p;
}


instr start_get()
/******************************************************************************/
{ 
  instr p=(instr)hget(g.predmark, new_func("get__",6));
  if(!p) fatal_error("no definition for: get__/6");
  return p;
}


void *putterm_1(arg)
  void *arg;
/******************************************************************************/
{
  register SVCXPRT *transp = (SVCXPRT *)arg;
  term             result;
  cell             RegisterBuffer[TEMPARGS+MAXREG];
  
  register term    regs=RegisterBuffer+TEMPARGS;
  register term    H;
  register instr   P;
  register term    *A;
  struct stack     *wam = (stack)XALLOC(MaxStk,struct stack);
  term             bp();
  rpcterm          argum;
  
  argum.data =(term) NULL;
  argum.datalen = 0;
     
/* puts("test1"); */
  if (!svc_getargs(transp, xdr_rpcterm, (caddr_t)&argum)) {
    svcerr_decode(transp);
    fprintf(stderr, "getargs unsuccesful\n");
    thr_exit(0);
    return NULL;
  }
  memcpy(wam,globalwam,sizeof(struct stack)*MaxStk);
  init_small_mem(wam);
  A=wam[ChoiceStk].base;
  addprocess(thr_self(),new_func("rpc_put",0),wam);
  TR_TOP=wam[TrailStk].base;
  H=(term)wam[HeapStk].base;
  P=start_put();
  regs[1] = (cell)(H+1);
  regs[2] = (cell)(H+2);
  regs[3] = g.true;
  H = (term)reloc2term(argum.data,H,wam);                       
  result = bp(regs,H,P,A,wam);
  if (!svc_sendreply(transp, xdr_int, (caddr_t)&result)) {
    svcerr_systemerr(transp);
    thr_exit(0);
    return NULL;
  }
/* puts("test2"); */
/* printf("%08x %08x %08x\n", &argum, argum.data, argum.datalen); */
  if (!svc_freeargs(transp, xdr_rpcterm, (caddr_t)&argum)) {
    svcerr_systemerr(transp);                                 
    thr_exit(0);
    return NULL;
  }
/* puts("test3"); */
  deleteprocess(thr_self());
  /* NOT REACHED */
  thr_exit(0);
  return NULL;
}


void *getterm_1(arg)
  void *arg;
/******************************************************************************/
{
  register SVCXPRT *transp = (SVCXPRT *)arg;
  term result;
  cell RegisterBuffer[TEMPARGS+MAXREG];
  register term regs=RegisterBuffer+TEMPARGS;
  register term H, oldH;
  register instr P;
  register term *A;
  struct stack *wam = (stack)XALLOC(MaxStk,struct stack);
  term bp();
  static rpcterm response;
  rpcterm argument ; 
  
  argument.data =(term) NULL;
  argument.datalen = 0;

  if (!svc_getargs(transp, xdr_rpcterm, (caddr_t)&argument)) {
    svcerr_decode(transp);
    thr_exit(0);
    return NULL;
  }
  memcpy(wam,globalwam,sizeof(struct stack)*MaxStk);
  init_small_mem(wam);
  A=wam[ChoiceStk].base;
  addprocess(thr_self(),new_func("rpc_get",0),wam);
  TR_TOP=wam[TrailStk].base;
  oldH = H = (term)wam[HeapStk].base;
  H = (term)reloc2term(argument.data,H,wam);
  P=start_get();
  regs[1] = (cell)oldH[1];
  regs[2] = (cell)oldH[2];
  regs[3] = (cell)oldH[3];
  regs[4] = (cell)oldH[4];
  regs[5] = (cell)oldH[5];
  regs[6] = g.true; 
/*  puts("getterm_1"); */
  result = bp(regs,H,P,A,wam);
/*  puts("getterm_2"); */
  if (result) {
    response.datalen = (term2reloc((term)oldH[2],(term *)result,wam,
                   (char *)wam[HeapStk].margin-(char *)result)+5) >> 2;
    response.data = result;
  } else {
    response.datalen = 0;
    response.data = (term)NULL;
  }
  if (!svc_sendreply(transp, xdr_rpcterm, (caddr_t)&response)) {
    svcerr_systemerr(transp);
    thr_exit(0);
  }
  if (!svc_freeargs(transp, xdr_rpcterm, (caddr_t)&argument)) {
    svcerr_systemerr(transp);
    thr_exit(0);
    return NULL;
  }
  response.data = NULL;
  xdr_free(xdr_rpcterm, (caddr_t)&response);
  deleteprocess(thr_self());
  /* NOT REACHED */
  thr_exit(0);
  return NULL;
}

void *bbterms_1(arg)
  void *arg;
/******************************************************************************/
{
  register SVCXPRT *transp = (SVCXPRT *)arg;
  char             *argument = (char *)NULL; 
  blackboard_t     *bb;
  term             H, t;
  rpcterm          response;
  struct stack     wam[MaxStk];
   
  if (!svc_getargs(transp, xdr_wrapstring, (caddr_t)&argument)) {
    svcerr_decode(transp);
    fprintf(stderr, "getargs unsuccesful");
    thr_exit(0);
    return NULL;
  } 
  bb = (blackboard_t *)hget(g.bbmark,new_func(argument,0));
    
  init_small_mem(wam);
  H = (term)wam[HeapStk].base;
  H = (term)bbterms(bb, H, &t, wam);
  if (H) {
    response.datalen = (term2reloc(t,wam[ChoiceStk].base,wam,wam[ChoiceStk].size) + 5) >> 2; 
    response.data = (term)(wam[ChoiceStk].base);
  } else {
    response.datalen = 0;
    response.data = (term)NULL;
  }
  if (!svc_sendreply(transp, xdr_rpcterm, (caddr_t)&response)) {
    svcerr_systemerr(transp);
    thr_exit(0);
    return NULL;
  }
  delete_small_mem(wam);
  if (!svc_freeargs(transp, xdr_wrapstring, (caddr_t)&argument)) {
    svcerr_systemerr(transp);
    thr_exit(0);
    return NULL;
  }
  response.data = (term)NULL;
  xdr_free(xdr_rpcterm, (caddr_t)&response);  
  return NULL;
  thr_exit(0); 
}


void *bbprocesses_1(arg)
  void *arg;
/******************************************************************************/
{
  register SVCXPRT *transp = (SVCXPRT *)arg;
  char             *argument = (char *)NULL;              
  blackboard_t     *bb;
  term             H, t;
  rpcterm          response;                         
  struct stack     wam[MaxStk];

  if (!svc_getargs(transp, xdr_wrapstring, (caddr_t)&argument)) {
    fprintf( stderr, "getargs unsuccesful ");
    svcerr_decode(transp);
    thr_exit(0);
    return NULL;
  } 
  bb = (blackboard_t *)hget(g.bbmark,new_func(argument,0));
  init_small_mem(wam);
  H = (term)wam[HeapStk].base;
  H = (term)bbprocesses(bb, H, &t, wam);
  if (H) {
    response.datalen = (term2reloc(t,wam[ChoiceStk].base,wam,wam[ChoiceStk].size) + 5) >> 2; 
    response.data = (term)(wam[ChoiceStk].base);
  } else {
    response.datalen = 0;
    response.data = (term)NULL;
  }
  if (!svc_sendreply(transp, xdr_rpcterm, (caddr_t)&response)) {
    svcerr_systemerr(transp);
    thr_exit(0);
    return NULL;
  }
  delete_small_mem(wam);
  if (!svc_freeargs(transp, xdr_wrapstring, (caddr_t)&argument)) {
    svcerr_systemerr(transp);
    thr_exit(0);
    return NULL;
  }
  response.data = NULL;
  xdr_free(xdr_rpcterm, (caddr_t)&response);
    thr_exit(0);
  return NULL;
}


void *bbclean_1(arg)
  void *arg;
/******************************************************************************/
{
  register SVCXPRT *transp = (SVCXPRT *)arg;
  char             *argument = (char *)NULL; 
  blackboard_t     *bb;
  int              result;

  sema_wait(&rpcserversema);
  if (!svc_getargs(transp, xdr_wrapstring, (caddr_t)&argument)) {
    fprintf( stderr, "getargs unsuccesful\n");
    svcerr_decode(transp);
    thr_exit(0);
    return NULL;
  } 
  sema_post(&rpcserversema);
  bb = (blackboard_t *)hget(g.bbmark,new_func(argument,0));
  result = bbclean(bb);
  if (!svc_sendreply(transp, xdr_int, (caddr_t)&result)) {
    svcerr_systemerr(transp);
    thr_exit(0);
    return NULL;
  }
  if (!svc_freeargs(transp, xdr_wrapstring, (caddr_t)&argument)) {
    svcerr_systemerr(transp);
    thr_exit(0);
    return NULL;
  }
    thr_exit(0);
    return NULL;
}

void *bbcheck_1(arg)
  void *arg;
/******************************************************************************/
{
  register SVCXPRT *transp = (SVCXPRT *)arg;
  char             *argument = (char *)NULL; 
  blackboard_t     *bb;
  int              result;

  if (!svc_getargs(transp, xdr_wrapstring, (caddr_t)&argument)) {
    svcerr_decode(transp);
    return NULL;
  } 
  bb = (blackboard_t *)hget(g.bbmark,new_func(argument,0));
  result = bb && bb != (blackboard_t *)g.empty;
  if (!svc_sendreply(transp, xdr_int, (caddr_t)&result)) {
    svcerr_systemerr(transp);
    return NULL;
  }
  if (!svc_freeargs(transp, xdr_wrapstring, (caddr_t)&argument)) {
    svcerr_systemerr(transp);
    return NULL;
  }
  return NULL;
}
 
 
void *blackboards_1(arg)  
  void *arg;
/******************************************************************************/
{
  register SVCXPRT *transp = (SVCXPRT *)arg;
  char             *argument = (char *)NULL; 
  term             H, t;
  rpcterm          response;
  struct stack     wam[MaxStk];
      
     
  if (!svc_getargs(transp, xdr_wrapstring, (caddr_t)&argument)) {
    svcerr_decode(transp);
    fprintf( stderr, "getargs unsuccesful");
    thr_exit(0);
    return NULL;
  } 
  
  init_small_mem(wam);
  H = (term)wam[HeapStk].base;
  H = (term)blackboards(H, &t);
  
  if (H) {
    response.datalen = (term2reloc(t,wam[ChoiceStk].base,wam,wam[ChoiceStk].size) + 5) >> 2; 
    response.data = (term)(wam[ChoiceStk].base);
  } else {
    response.datalen = 0;
    response.data = (term)NULL;
  }
  if (!svc_sendreply(transp, xdr_rpcterm, (caddr_t)&response)) {
    svcerr_systemerr(transp);
    thr_exit(0);
    return NULL;
  }
  delete_small_mem(wam);
  if (!svc_freeargs(transp, xdr_wrapstring, (caddr_t)&argument)) {
    svcerr_systemerr(transp);
    thr_exit(0);
    return NULL;
  }
  response.data = NULL;
  xdr_free(xdr_rpcterm, (caddr_t)&response);
    thr_exit(0);
  return NULL;
}


void *bbmem_1(arg)  
  void *arg;
/******************************************************************************/
{
  register SVCXPRT *transp = (SVCXPRT *)arg;
  char             *argument = (char *)NULL; 
  blackboard_t     *bb;
  term             H, t;
  rpcterm          response;
  struct stack     wam[MaxStk];
      
  if (!svc_getargs(transp, xdr_wrapstring, (caddr_t)&argument)) {
    svcerr_decode(transp);
    fprintf( stderr, "getargs unsuccesful");
    thr_exit(0);
    return NULL;
  } 
  bb = (blackboard_t *)hget(g.bbmark,new_func(argument,0));
  init_small_mem(wam);
  H = (term)wam[HeapStk].base;
  H = (term)bbmem(bb,H, &t);
  
  if (H) {
    response.datalen = (term2reloc(t,wam[ChoiceStk].base,wam,wam[ChoiceStk].size) + 5) >> 2; 
    response.data = (term)(wam[ChoiceStk].base);
  } else {
    response.datalen = 0;
    response.data = (term)NULL;
  }
  if (!svc_sendreply(transp, xdr_rpcterm, (caddr_t)&response)) {
    svcerr_systemerr(transp);
    thr_exit(0);
    return NULL;
  }
  delete_small_mem(wam);
  if (!svc_freeargs(transp, xdr_wrapstring, (caddr_t)&argument)) {
    svcerr_systemerr(transp);
    thr_exit(0);
    return NULL;
  }
  response.data = NULL;
  xdr_free(xdr_rpcterm, (caddr_t)&response);
  thr_exit(0);
  return NULL;
}


void *bbmalloc_1(arg)  
  void *arg;
/******************************************************************************/

{
  register SVCXPRT *transp = (SVCXPRT *)arg;
  int result;
  rpcmallocterm argum;
  blackboard_t  *bb;
  
  argum.bb_name =(char*) NULL;
  argum.bl_size = 0;
   
  if (!svc_getargs(transp, xdr_rpcmalloc, (caddr_t)&argum)) {
    svcerr_decode(transp);
    fprintf( stderr, "getargs unsuccesful\n");
    thr_exit(0);
    return NULL;
  }

  bb = (blackboard_t *)hget(g.bbmark,new_func(argum.bb_name,0));
  result = bbmalloc(bb,argum.bl_size);  
  
  if (!svc_sendreply(transp, xdr_int, (caddr_t)&result)) {
    svcerr_systemerr(transp);
    thr_exit(0);
    return NULL;
  }
  if (!svc_freeargs(transp, xdr_rpcterm, (caddr_t)&argum)) {
    svcerr_systemerr(transp);                                 
    thr_exit(0);
    return NULL;
  }
  /* deleteprocess(thr_self()); */
  /* NOT REACHED */
  thr_exit(0);
  return NULL;
}


void *bbfree_1(arg)  
  void *arg;
/******************************************************************************/
{
  register SVCXPRT *transp = (SVCXPRT *)arg;
  int result;
  rpcmallocterm argum;
  blackboard_t  *bb;
  
  argum.bb_name =(char*) NULL;
  argum.bl_size = 0;
   
  if (!svc_getargs(transp, xdr_rpcmalloc, (caddr_t)&argum)) {
    svcerr_decode(transp);
    fprintf( stderr, "getargs unsuccesful\n");
    thr_exit(0);
    return NULL;
  }

  bb = (blackboard_t *)hget(g.bbmark,new_func(argum.bb_name,0));
  result = bbfree(bb,argum.bl_size);  
  
  if (!svc_sendreply(transp, xdr_int, (caddr_t)&result)) {
    svcerr_systemerr(transp);
    thr_exit(0);
    return NULL;
  }
  if (!svc_freeargs(transp, xdr_rpcterm, (caddr_t)&argum)) {
    svcerr_systemerr(transp);                                 
    thr_exit(0);
    return NULL;
  }
  /* deleteprocess(thr_self()); */
  /* NOT REACHED */
  thr_exit(0);
  return NULL;
}


void *bbmove_1(arg)  
  void *arg;		
/******************************************************************************/
{   
  register SVCXPRT    *transp = (SVCXPRT *)arg;
  rpcallterm_node     argument;
  remotebb_t          *bb; 
  blackboard_t        *auxbb;
  localbb_t           *newbb;
  term                H; 
  term                t;
  rpcterm             response;
  struct stack        wam[MaxStk]; 
  int                 result,alreadyowner;
  register int        size;
  register unsigned   datasize;
  register termnode_t *tp;
  no                  fname;
  register pool_t     *pool = (pool_t *)hget(g.bbmark,g.bbmark);
  u_long n;
      
  argument.termlist.data = (term) NULL;
  argument.termlist.datalen = 0;
  argument.bb_name = (char *)NULL;   
  if (!svc_getargs(transp, xdr_allboard, (caddr_t)&argument)) {
    svcerr_decode(transp);
    fprintf( stderr, "getargs unsuccesful");
    thr_exit(0);
    return NULL;
  } 
  
  fname = new_func(argument.bb_name,0);
  bb = (remotebb_t *)hget(g.bbmark,fname); 
  init_small_mem(wam);
  H = (term)wam[HeapStk].base;
  t = H;
  H = (term) reloc2term(argument.termlist.data,H,wam);
  bblock(bb);
  remove_bbclient(bb); /* delete the list of client handles */

  assert(pool);
  mutex_lock(&(pool->mutex));
  /* make the new (local) blackboard structure */
  newbb = (localbb_t *)buddymalloc(pool->buddy, BLACKBOARDSIZE);
  memcpy(newbb,bb,sizeof(blackboard_t));
  /* put new blackboard in pool */
  newbb->nextbb = pool->blackboards;
  pool->blackboards = (blackboard_t *)newbb;
  /* remove old blackboard from the pool */
  auxbb = (blackboard_t *)newbb;
  while (auxbb->nextbb != (blackboard_t *)bb) 
    auxbb = auxbb->nextbb;
  auxbb->nextbb = bb->nextbb;
  /* delete old blackboard structure */
  buddyfree(pool->buddy,bb,sizeof(remotebb_t)); 
  mutex_unlock(&(pool->mutex));
  
  hset(g.bbmark,fname,(no)newbb); /* symbol table points to new structure */
  newbb->type = LOCALBB;
  buddycreate(newbb->buddy, BUDDYSIZE);
  buddyinit(newbb->buddy, 
            (char *)bb+realbuddysize(sizeof(localbb_t)), 
            BLACKBOARDSIZE-realbuddysize(sizeof(localbb_t)));
  newbb->firstterm = (termnode_t *)NULL;
  newbb->lastterm = (termnode_t *)NULL;
  newbb->firstprocess = (processnode_t *)NULL;
  newbb->lastprocess = (processnode_t *)NULL;
  newbb->timestamp = (term)INPUT_INT(0);

  /* put terms and timestamps on blackboard newbb */
  *(H-1) = g.DOT;
  t++;
  newbb->timestamp = (term)(*t);   
  while ( *t != g.DOT)
    t++;
  while ( t < (H-1)) {
    t++;
    size = term_size(t);
    datasize = sizeof(termnode_t)-4+size;
    tp =(termnode_t *) buddymalloc(newbb->buddy, datasize);
    if (!tp) {
      warnmes("Blackboard memory exhausted---");
      return 0; /* not really elegant */
    }
    tp->next = (termnode_t *)NULL;
    tp->size = datasize;
    copytobb((term)tp->data,t,wam);
    while ( *t != g.DOT)
      t++;
    t++;
    tp->timestamp = (term)(*t);
    if (newbb->firstterm) {
      newbb->lastterm->next = tp;
      newbb->lastterm = tp;
    } else {
      newbb->firstterm = newbb->lastterm = tp;
    }
    while ( *t != g.DOT)
      t++;     
  } 
  
  result =0;  
  if (!svc_sendreply(transp, xdr_int, (caddr_t)&result)) {
    svcerr_systemerr(transp);
    thr_exit(0);
    return NULL;
  }
  if (!svc_freeargs(transp, xdr_allboard, (caddr_t)&argument)) {
    svcerr_systemerr(transp);
    thr_exit(0);
    return NULL;
  }

  bbunlock(newbb); /* mutex lock is now on newbb */
  delete_small_mem(wam); 
  xdr_free(xdr_allboard, (caddr_t)&argument);
  thr_exit(0);
  return NULL;
}


static void
rpcmessageserver(rqstp, transp)
  struct svc_req *rqstp;
  register SVCXPRT *transp;
/******************************************************************************/
{ 
  thread_t thread;
  

  {/*char *p = malloc(1); */
/*   int thrmode, thrnum, thract, thrcre, threrr; */
/*    printf("-1- %08x\n", p); free(p); */
/*
   rpc_control(RPC_SVC_MTMODE_GET,&thrmode);
   rpc_control(RPC_SVC_THRMAX_GET,&thrnum);
   rpc_control(RPC_SVC_THRTOTAL_GET,&thract);
   rpc_control(RPC_SVC_THRCREATES_GET,&thrcre);
   rpc_control(RPC_SVC_THRERRORS_GET,&threrr);
   printf("%d %d %d %d %d\n", thrmode, thrnum, thract, thrcre, threrr); */
  }

  switch (rqstp->rq_proc) {
  case NULLPROC:
    svc_sendreply(transp, xdr_void, NULL);
    return;

  case PUTTERM: 
    thr_create(NULL,0,putterm_1,transp,THR_DETACHED,&thread);
    return;

  case GETTERM: 
   thr_create(NULL,0,getterm_1,transp,THR_DETACHED,&thread);
   return;
  
  case BBTERMS:
    bbterms_1(transp); return;
    thr_create(NULL,0,bbterms_1,transp,THR_DETACHED,&thread);
    return;
  
  case BBPROCESSES:
    thr_create(NULL,0,bbprocesses_1,transp,THR_DETACHED,&thread);
    return;

  case BBCLEAN:
    thr_create(NULL,0,bbclean_1,transp,THR_DETACHED,&thread);
    return;
  
  case BBCHECK: 
    thr_create(NULL,0,bbcheck_1,transp,THR_DETACHED,&thread);
    return;

  case BLACKBOARDS:
    thr_create(NULL,0,blackboards_1,transp,THR_DETACHED,&thread);
    return;
    
  case BBMEM:
    thr_create(NULL,0,bbmem_1,transp,THR_DETACHED,&thread);
    return;
    
  case BBMALLOC:
    thr_create(NULL,0,bbmalloc_1,transp,THR_DETACHED,&thread);
    return;
    
  case BBFREE:
    thr_create(NULL,0,bbfree_1,transp,THR_DETACHED,&thread);
    return;

  case BBMOVE:  
    thr_create(NULL,0,bbmove_1,transp,THR_DETACHED,&thread);
    return;
  
  default:
    svcerr_noproc(transp);
    return;
  }
}


void stopserver(i)
  int i;
/******************************************************************************/
{
  if ( serverthread != thr_self()) {
    fprintf(stderr, "wrong thread (%d) was to be killed\n", (long)thr_self());
    return;
    }
  sema_post(&rpcserversema); /* signal killserver() that thread is exiting */
  thr_exit(0);
}


void killserver(void)
/******************************************************************************/
{
  CLIENT *cl;
  
  if (serverprog) {
    thr_kill(serverthread,SIGUSR2);
    thr_continue(serverthread);
    sema_wait(&rpcserversema); /* wait for execution of stopserver() */  
    fflush(stdout);
    svc_done(transp); /* works only in User MT mode */
    transp = (SVCXPRT *)NULL;
    serverprog = 0;
  }
}


long nametoprog(s)
  char *s;
/******************************************************************************/
{
  if (strlen(s) > 4) {
    fprintf(stderr, "server name <%s> too long (>4)\n", s);
    thr_exit(0);
    return 0;
  } else {
    long l[2], i;
    char *buffer = (char *)l;

    l[0] = l[1] = 0;
    strcpy(buffer,s);
    return l[0];
  }
}


void *rpcserver(arg)
  void *arg;
/******************************************************************************/
{
  sigset_t         set;
  struct netconfig *config;
 
  if (IDENTIFIER((term)arg)) {
    serverprog = nametoprog(NAME((term)arg)); 
  } else if (VAR(arg))
    serverprog = thr_self();
  else {
    fprintf(stderr, "Error: server name must be constant");
    thr_exit(0);
  }
  
  sigemptyset(&set);
  sigaddset(&set,SIGUSR2);
  thr_sigsetmask( SIG_UNBLOCK, &set, NULL);
  signal(SIGUSR2,stopserver);
  
  /* Creation of the server. We are using the intermediate level because
     we need the variable "transp" to end the service afterwards. */
  config = (struct netconfig *)getnetconfigent( "tcp");
  transp = svc_tp_create(rpcmessageserver, serverprog, BBVERS, config);
  freenetconfigent(config);
  if (!transp) {
    fprintf(stderr, "Cannot create server\n");
    thr_exit(0);
    return NULL;
  } 
    
  sema_post(&rpcserversema);
  svc_run();
  thr_exit(0); /* NOT REACHED */
  return NULL;
}


term initserver(name,wam)
  term  name;
  stack wam;
/******************************************************************************/
{
  int      info;
  sigset_t set;
  
  memcpy(globalwam,wam,sizeof(struct stack)*MaxStk);
  sema_init(&rpcserversema, 0, USYNC_THREAD, NULL);
    
  /* set USER MT Mode */
  info = RPC_SVC_MT_USER; 
  rpc_control(RPC_SVC_MTMODE_SET, &info);
  
  thr_create(NULL, 0, rpcserver, (void *)name, 
                   THR_DETACHED|THR_NEW_LWP|THR_BOUND, &serverthread);
  sema_wait(&rpcserversema);
  
  if (IDENTIFIER(name)) 
    return name;
  else if (VAR(name)) 
    return (term)INPUT_INT(serverthread);
  else return (term)NULL;
}


/******************************************************************************
 *                                                                            *
 *                                     CLIENT                                 *
 *                                                                            *
 ******************************************************************************/

/* Default timeout can be changed using clnt_control() */
static struct timeval TIMEOUT = { 25000, 0 }; 

term rpc_putterm(bb,wam,A,H,regs)
  remotebb_t     *bb;
  register stack wam; 
  register term  *A;
  term           H;
  term           regs; 
/******************************************************************************/
{ 
  rpcterm data;
  int     result;
  term    temp[3];
  term    buffer = H;
  CLIENT  *client;

  temp[0] = (term)g.DOT; temp[1]=(term)bb->bboard;
  temp[2] = (term)regs[2];
  data.data = buffer;
  data.datalen = (term2reloc((term)temp,(term *)buffer,wam,
                     (char*)wam[HeapStk].margin-(char *)buffer)+5) >> 2;
  client = get_bbclient(bb);
  if (clnt_call(client, PUTTERM, xdr_rpcterm, (caddr_t)&data,
                xdr_int, (caddr_t)&result, TIMEOUT) == RPC_SUCCESS) { 
    
    int i, res = result;
    return_bbclient(bb,client);
    return res ? H : 0;
  } else
    return_bbclient(bb,client);
    return 0;
}


term rpc_getterm(bb, wam, A, H, timestamp, regs)
  remotebb_t     *bb;
  register stack wam; 
  register term  *A;
  term           H;
  term           *timestamp;
  term           regs;
/******************************************************************************/
{
  term    temp[6];
  term    buffer = H;
  rpcterm data, result;
  CLIENT  *client;

  temp[0] = (term)new_func("rpc_get",5);
  temp[1] = (term)bb->bboard; 
  temp[2] = (term)regs[2];
  temp[3] = (term)regs[3]; 
  temp[4] = (term)regs[4];
  temp[5] = (term)regs[5];
  data.datalen = (term2reloc((term)temp,(term*)buffer,wam,
                    (char*)wam[HeapStk].margin-(char *)buffer)+5) >> 2; 
  data.data = buffer;
  result.data = (term)NULL;
  result.datalen = 0;

  client = get_bbclient(bb);
  if (clnt_call(client, GETTERM, xdr_rpcterm, (caddr_t)&data, 
      xdr_rpcterm, (caddr_t)&result, TIMEOUT)==RPC_SUCCESS && result.datalen) {
    term oldH = H;
    return_bbclient(bb,client);
    H = (term)reloc2term(result.data,H,wam);
    xdr_free(xdr_rpcterm,(caddr_t)&result); 
    return unify(regs[2],oldH,wam,A) ? H : 0;
  } else  {
    return_bbclient(bb,client);
    return 0;
  }
}


term rpc_bbterms(bb,H,t,wam)
  remotebb_t *bb;
  term       H, *t;
  stack      wam;
/******************************************************************************/
{
  rpcterm result; 
  CLIENT  *client;
  
  result.data = (term)NULL;
  result.datalen = 0;

  client = get_bbclient(bb);
  if (clnt_call(client, BBTERMS, xdr_wrapstring, (caddr_t)&NAME(bb->bboard),
      xdr_rpcterm, (caddr_t)&result, TIMEOUT)==RPC_SUCCESS && result.datalen) {

    return_bbclient(bb,client);
    *t = H;
    H = (term)reloc2term(result.data,H,wam);
    xdr_free(xdr_rpcterm,(caddr_t)&result);
    return H;
  } else {
    return_bbclient(bb,client);
    fprintf( stderr, "RPC failed\n");
    return (term)NULL;
    }
}


term rpc_bbprocesses(bb,H,t,wam)
  remotebb_t *bb;
  term       H, *t;
  stack      wam;
/******************************************************************************/
{
  rpcterm result; 
  CLIENT  *client;

  result.data = (term)NULL;
  result.datalen = 0;
  client = get_bbclient(bb);
  if (clnt_call(client, BBPROCESSES, xdr_wrapstring, (caddr_t)&NAME(bb->bboard),
        xdr_rpcterm, (caddr_t)&result, TIMEOUT)==RPC_SUCCESS && result.datalen) {
    return_bbclient(bb,client);
    *t = H;
    H = (term)reloc2term(result.data,H,wam);
    xdr_free(xdr_rpcterm,(caddr_t)&result); 
    return H;
  } else 
    return_bbclient(bb,client);
    return (term)NULL;
}


int rpc_bbclean(bb)
  remotebb_t *bb;
/******************************************************************************/
{
  int    result; 
  char   *bbname;
  CLIENT *client;

  assert(bb);
  assert(bb->type == REMOTEBB);
  bbname = NAME(bb->bboard);
  client = get_bbclient(bb);
  if (clnt_call(client, BBCLEAN, xdr_wrapstring, (caddr_t)&bbname,
                xdr_int, (caddr_t)&result, TIMEOUT)==RPC_SUCCESS) {
    return_bbclient(bb,client);
    return result;
  } else 
    return_bbclient(bb,client);
    return 0;
}


int rpc_bbcheck(cl,bbname)
  CLIENT *cl;
  char   *bbname;
/******************************************************************************/
{
  int result; 

  if (clnt_call(cl, BBCHECK, xdr_wrapstring, (caddr_t)&bbname,
                xdr_int, (caddr_t)&result, TIMEOUT)==RPC_SUCCESS) {
    return result;
  } else 
    return 0;
}


term rpc_blackboards(host,prognum,H,t,wam)
  register cell   host;
  register u_long prognum;
  term            H, *t;
  stack           wam;
/******************************************************************************/
{
  rpcterm         result; 
  register CLIENT *cl;
  char            *argument = NULL;

  if (cl = clnt_create(NAME(host),prognum,BBVERS,"tcp") ) {
    
   } else { 
     return (term) NULL;
     }
     
  result.data = (term)NULL;
  result.datalen = 0;

  if (clnt_call(cl, BLACKBOARDS, xdr_wrapstring, (caddr_t)&argument,
        xdr_rpcterm, (caddr_t)&result, TIMEOUT)==RPC_SUCCESS && result.datalen) {
    
    *t = H;
    H = (term)reloc2term(result.data,H,wam);
    xdr_free(xdr_rpcterm,(caddr_t)&result);
    clnt_destroy(cl);  
    return H;
  } else {
    fprintf( stderr, "RPC failed\n");
    clnt_destroy(cl);
    return (term)NULL;
    }
}


term rpc_bbmem( bb, H, t, wam)
  register remotebb_t *bb;
  register term H, *t;
  stack    wam;
/******************************************************************************/
{
  rpcterm result; 
  CLIENT  *client;
  
  result.data = (term)NULL;
  result.datalen = 0;

  client = get_bbclient(bb);
  if (clnt_call(client, BBMEM, xdr_wrapstring, (caddr_t)&NAME(bb->bboard),
      xdr_rpcterm, (caddr_t)&result, TIMEOUT)==RPC_SUCCESS && result.datalen) {
  
    return_bbclient(bb,client);
    *t = H;
    H = (term)reloc2term(result.data,H,wam);
    xdr_free(xdr_rpcterm,(caddr_t)&result);
    return H;
  } else {
    return_bbclient(bb,client);
    fprintf( stderr, "RPC failed\n");
    return (term)NULL;
    }
}


int rpc_bbmalloc(bb, blocksize)
  remotebb_t *bb;
  unsigned int blocksize;
/******************************************************************************/
  
{
  int           result;
  rpcmallocterm mterm;
  char          *name;
  CLIENT        *client;
  
  name = NAME(bb->bboard);
  mterm.bb_name = (char *)XALLOC(strlen(name) + 1,char);
  strcpy(mterm.bb_name, name); 
  mterm.bl_size = blocksize;
  
  client = get_bbclient(bb);
  if (clnt_call(client, BBMALLOC, xdr_rpcmalloc, (caddr_t)&mterm,
                xdr_int, (caddr_t)&result, TIMEOUT)==RPC_SUCCESS) {
    return_bbclient(bb,client);
    XFREE(mterm.bb_name);
    return result; 
  } else {
    return_bbclient(bb,client);
    XFREE(mterm.bb_name);
    return 0;    
  }
}


int rpc_bbfree(bb, blocksize)
  remotebb_t *bb;
  unsigned int blocksize;
/******************************************************************************/
  
{
  int           result;
  rpcmallocterm mterm;
  char          *name;
  CLIENT        *client;
  
  name = NAME(bb->bboard);
  mterm.bb_name = (char *)XALLOC(strlen(name) + 1,char);
  strcpy(mterm.bb_name, name); 
  mterm.bl_size = blocksize;
    
  client = get_bbclient(bb);
  if (clnt_call(client, BBFREE, xdr_rpcmalloc, (caddr_t)&mterm,
                xdr_int, (caddr_t)&result, TIMEOUT)==RPC_SUCCESS) {
    return_bbclient(bb,client);
    XFREE(mterm.bb_name);
    return result; 

  } else {
    return_bbclient(bb,client);
    XFREE(mterm.bb_name);
    return 0;
  }
}


int rpc_bbmove(bb_l,host,prognum,bb_r) 
  register cell   bb_l, host;   
  register u_long prognum;
  register cell   bb_r;
/**********************************************
 changes bb_l to remote blackboard, and bb_r
 from remote blackboard to a local one.
 **********************************************/
{
  int                 result; 
  CLIENT              *cl;
  register            cell xval, name;
  register localbb_t  *lbb;
  register remotebb_t *newbb;
  blackboard_t        *auxbb;
  int                 alreadyowner;
  termnode_t          *tp;
  register pool_t     *pool = (pool_t *)hget(g.bbmark,g.bbmark);
  term                H,t;
  struct stack        wam[MaxStk];
  rpcallterm_node     allterm;

  result=0;
  lbb = (localbb_t *)hget(g.bbmark,(no)bb_l);
  if (!(cl = clnt_create(NAME(host),prognum,BBVERS,"tcp"))) {
    fprintf( stderr, "clnt_create unsuccesful\n");	
    return 0;
  }   
  bblock(lbb);

  /* Prepare for sending the terms of the blackboard */
  init_small_mem(wam);
  H = (term)wam[HeapStk].base;
  t = H;
  /* making a list of terms and timestamps */
  PUSH_LIST((T2C(lbb->timestamp)));
  tp = lbb->firstterm;
  while (tp) {
    PUSH_LIST((T2C(tp->data)));
    PUSH_LIST((T2C(tp->timestamp)));
    tp = tp->next;
  }
  PUSH_NIL();

  allterm.termlist.datalen = (term2reloc(t,wam[ChoiceStk].base,
    				wam,wam[ChoiceStk].size) + 5) >> 2;
  allterm.termlist.data = (term)(wam[ChoiceStk].base);
  allterm.bb_name = NAME(bb_r);
    
  if (!clnt_call(cl, BBMOVE, xdr_allboard, (caddr_t)&allterm,
                xdr_int, (caddr_t)&result, TIMEOUT)==RPC_SUCCESS ) {
   fprintf( stderr, "RPC failed\n");
   bbunlock(lbb);
   return 0;
  }  
  delete_small_mem(wam);
  if (!rpc_bbcheck(cl,NAME(bb_r))) {
    fprintf(stderr,"couldn't change remote blackboard in a local one\n");
    bbunlock(lbb);
    return 0;
  }
  /* Now bb_r is a complete local blackboard */
  /* creating a (new) remote bb instead of the local one */
  assert(pool);
  mutex_lock(&(pool->mutex));
  /* make the new (remote) blackboard structure */
  newbb = (remotebb_t *)buddymalloc(pool->buddy, sizeof(remotebb_t));
  memcpy(newbb,lbb,sizeof(blackboard_t));
  /* put new blackboard in pool */
  newbb->nextbb = pool->blackboards;
  pool->blackboards = (blackboard_t *)newbb;
  /* remove old blackboard from the pool */
  auxbb = (blackboard_t *)newbb;
  while (auxbb->nextbb != (blackboard_t *)lbb) 
    auxbb = auxbb->nextbb;
  auxbb->nextbb = lbb->nextbb;
  /* delete old blackboard structure */
  buddyfree(pool->buddy,lbb,BLACKBOARDSIZE);   
  mutex_unlock(&(pool->mutex));
  
  hset(g.bbmark,bb_l,(no)newbb); /* symbol table points to new structure */
  newbb->type = REMOTEBB;
  strncpy(newbb->host,NAME(host),HOSTNAMESIZE);
  newbb->host[HOSTNAMESIZE-1] = 0; /* 0-terminate */
  newbb->prognum = prognum;
  newbb->bboard = bb_r;
  newbb->cl_list = (clnt_node_t *)XALLOC(1,clnt_node_t);
  newbb->cl_list->inuse = FALSE;
  newbb->cl_list->cl = cl;
  newbb->cl_list->next = NULL; 

  bbunlock(newbb);         
  return 1;
}





