#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
#
# agtProcess
#
#  This is the main script initializing an agent.
#
# Arguments:
#
#  initial information for the agent in form of a keyedlist including
#  the following keys:
#
#  NAME, TYPE, STATUS, HOST, PORT, HOMEPATH, XDEFAULT
#
#

# global tk_library

# Redirect stderr to stdout to see errors of agent at UAM
#
 dup stdout stderr

# get the creation time of the agent:
 set creationTime [getclock] 

# get the agents basic information form arguments in global
# array 'agents':
 set agents(SELF) $argv

# 'agents()' is an global array which stores information about all
# agents known by this agent in form of a keyedlist.
#
# possible keys: NAME	  - name of the agent format: idf.host.port or UAM
#		 TYPE	  - type of the agent, name of file
#		 STATUS   - status of the agent (SELF only)
#		 ADDR	  - address of the agent: host.port
#                HOSTIP   - IP number of agents host
#		 INFID	  - fileIdf for incoming pipe to this agent
#		 OUTFID	  - fileIdf for outgoing pipe to this agent
#		 SOCKET	  - fileIdf for base socket (SELF only)
#		 LOGFID   - fileIdf for standard Logfile (SELF only)
#		 MLGFID   - fileIdf for mesg-eval Logfile (SELF only)
#		 CLGFID   - fileIdf for communication Logfile (SELF only)
#		 FILENO   - file counter for the Logfiles (SELF only)
#		 HOMEPATH - just for transferring the ideas_HomePath from UAM
#                XDEFAULT - decision to read/not read Xdefault file   
#
#

# copy the HOMEPATH information in global variable
# 'ideas-HomePath' and remove the key HOMEPATH:
 keylget agents(SELF) HOMEPATH ideas_HomePath
 keyldel agents(SELF) HOMEPATH 

# read the IDEAS init file:
 if {[catch {source init.tcl} errorMesg]} {
   puts $errorMesg
   exit
 }

# read the .ideasXdefaults file
 if {[keylget agents(SELF) XDEFAULT]} {
   catch {option readfile .ideasXdefaults}
 }
 keyldel agents(SELF) XDEFAULT

# set the path for auto-loading procedures:
 set auto_path "$ideas_HomePath $auto_path"

 cd $ideas_HomePath

# The global variable 'agt_VarTimeOffset' is a special hack
# to work with UNIX-datestrings, which are long integers,
# in BinProlog:
# BinProlog cannot handle that long integers, so always when
# a UNIX-datestring is copied to BinProlog, the offset is subtracted
# before, to decrement the integer to a BinProlog-readable value.
# On the other side a date coming from BinProlog is incremented 
# with this offset.
#
 set agt_VarTimeOffset 790000000



# Initial setting about the agent itself and his UAM:

# get the agents host out of the name: name.host
#
 set agtHost [crange [keylget agents(SELF) NAME] \
                     [string first . [keylget agents(SELF) NAME]]+1 \
                     end]

 keylset agents(SELF) HOSTIP [lindex [server_info addresses $agtHost] 0]

 keylset agents(UAM) NAME UAM TYPE SUPER-AGENT \
                     ADDR [keylget agents(SELF) HOSTIP].none \
                     INFID stdin OUTFID stdout


# create socket for listening to connections
# and register its fileId
#
 agt_openSocket 

 catch {unset agtHost}

 # set agents name to global variable SELF, so that it can be used in actions
 # defined by user
 keylget agents(SELF) NAME SELF



 wm withdraw .
 wm title . "Observation: $SELF"
 wm iconname . "Agent $SELF" 

 wm iconbitmap . questhead
 wm minsize . 500 260


 frame .f0
 pack .f0 -side top -fill x

 # create the main menu
 frame .mbar -relief ridge -bd 2

 pack .mbar -in .f0 -side top -fill x 

 # MENU view:
 menubutton .mbar.view -text View -underline 0 \
		-menu .mbar.view.menu

 menu .mbar.view.menu

 .mbar.view.menu add checkbutton -label "Logfile" -underline 0 \
		-variable agt_VarLogView \
		-command {if {$agt_VarLogView} {
			    pack .f1 -padx 3 -pady 3 -fill both -expand yes
			  } else {
			    pack forget .f1
			  }
			  wm geometry . ""
			 }

 .mbar.view.menu add checkbutton -label "Message-Evaluation" -underline 0 \
		-variable agt_VarEvalView \
		-command {if {$agt_VarEvalView} {
			    pack .f4 -padx 3 -pady 3 -fill both -expand yes
			  } else {
			    pack forget .f4
			  }
			  wm geometry . ""
			 }

 .mbar.view.menu add checkbutton -label "Communication" -underline 0 \
		-variable agt_VarCommView \
		-command {if {$agt_VarCommView} {
			    pack .f5 -padx 3 -pady 3 -fill both -expand yes
			  } else {
			    pack forget .f5
			  }
			  wm geometry . ""
			 }

 .mbar.view.menu add checkbutton -label "Communication-Paths" -underline 14 \
		-variable agt_VarPathsView \
		-command {if {$agt_VarPathsView} {
			    pack .f2 -padx 3 -pady 3 -fill both -expand yes
			  } else {
			    pack forget .f2
			  }
			  wm geometry . ""
		 }

 .mbar.view.menu add checkbutton -label "Belief-Base" -underline 0 \
		-variable agt_VarBelBaseView \
		-command {if {$agt_VarBelBaseView} {
			    pack .f3 -padx 3 -pady 3 -fill both -expand yes
			  } else {
			    pack forget .f3
			  }
			  wm geometry . ""
			 }


 # Button trace in menu - NO menubutton !

 button .mbar.trace -text Trace -relief flat \
    -underline 0 \
    -command {
              if {!$agt_VarTraceEnable} {
                set agt_VarTraceEnable 1
                pack .trace -in .f0 -side bottom \
                  -padx 3 -pady 3 -fill both -expand yes
                .mbar.trace configure -text Untrace
		bind . <Alt-Key-t> ""
		bind . <Alt-Key-u> {.mbar.trace invoke}
              } else {
                set agt_VarTraceEnable 0
                pack forget .trace
                set agt_VarWakeUp 1
                .mbar.trace configure -text Trace
		bind . <Alt-Key-u> ""
		bind . <Alt-Key-t> {.mbar.trace invoke}
              }
             }

 bind . <Alt-Key-t> {.mbar.trace invoke}

 # MENU exit - only shown if lost UAM
 menubutton .mbar.exit -text Exit -underline 0 -menu .mbar.exit.menu

 menu .mbar.exit.menu
 .mbar.exit.menu add command -label "Quit   " -underline 0 \
                -accelerator "Ctrl+q" -command agt_exit



 # MENU help:
 menubutton .mbar.help -text Help -underline 0 -menu .mbar.help.menu

 menu .mbar.help.menu
 .mbar.help.menu add command -label "ShortHelp" -underline 0\
                -command "agt_showShortHelp .shortHelp -1 -1"

 pack .mbar.view .mbar.trace -side left -padx 2m

 pack .mbar.help -side right -padx 2m

 tk_menuBar .mbar .mbar.view .mbar.help


 # create frame for 'Wake Up'-button
 #
 frame .trace -relief ridge -bd 2

 button .trace.wakeup -bitmap @alarmclock -width 20 -state disabled \
                  -command {set agt_VarWakeUp 1}

 pack .trace.wakeup -pady 3 -ipadx 10m 


 # create view for logfile
 #
 frame .f1 -relief ridge -bd 2
 pack .f1 -padx 3 -pady 3 -fill both -expand yes

 label .f1.label -text "Logfile:" \
		-relief ridge -bd 2 
		

 frame .f1.mbar -relief ridge -bd 2

 checkbutton .f1.mbar.scroll -text "Auto-Scroll" \
	-relief flat \
	-variable agt_VarLogScroll


 text .f1.text -relief ridge -bd 2 \
            -yscrollcommand ".f1.scroll set" \
            -wrap word \
	    -height 15 \
	    -width 65
 scrollbar .f1.scroll -relief ridge -command ".f1.text yview"

 pack .f1.label -side top -fill x
 pack .f1.mbar  -side top -fill x 
 pack .f1.mbar.scroll -side right -padx 2m
 pack .f1.scroll -side right -fill y
 pack .f1.text -side left -fill both -expand yes

 # set checkbutton variable for message view to on
 set agt_VarLogView 1

 # create view for mesg-evaluation
 #
 frame .f4 -relief ridge -bd 2

 label .f4.label -text "Message-Evaluation:" \
		-relief ridge -bd 2 

 frame .f4.mbar -relief ridge -bd 2

 checkbutton .f4.mbar.scroll -text "Auto-Scroll" \
	-relief flat \
	-variable agt_VarMesgEvalScroll

 checkbutton .f4.mbar.save -text "Save" \
	-relief flat \
	-variable agt_VarMesgEvalSave
		
 text .f4.text -relief ridge -bd 2 \
            -yscrollcommand ".f4.scroll set" \
            -wrap word \
	    -height 15 \
	    -width 65
 scrollbar .f4.scroll -relief ridge -command ".f4.text yview"

 pack .f4.label -side top -fill x
 pack .f4.mbar  -side top -fill x 
 pack .f4.mbar.scroll .f4.mbar.save -side right -padx 2m
 pack .f4.scroll -side right -fill y
 pack .f4.text -side left -fill both -expand yes

 # create view for communication observation
 #
 frame .f5 -relief ridge -bd 2

 label .f5.label -text "Communication:" \
		-relief ridge -bd 2 

 # logwidget for communication observation
 frame .f5.log

 frame .f5.mbar -relief ridge -bd 2

 button .f5.mbar.filter -text Filter -relief flat \
	-command {
          .f5.filter.list1 delete 0 end
	  foreach elem $agt_VarComNotFilter {
	    .f5.filter.list1 insert end $elem
          }
	  set agt_VarTempComNotFilter $agt_VarComNotFilter
	  .f5.filter.list1 selection set 0

          .f5.filter.list2 delete 0 end
	  foreach elem $agt_VarComFilter {
	    .f5.filter.list2 insert end $elem
          }
	  set agt_VarTempComFilter $agt_VarComFilter
	  .f5.filter.list2 selection set 0

	  pack forget .f5.log
          pack .f5.filter -fill both -expand no
         }

 checkbutton .f5.mbar.scroll -text "Auto-Scroll" \
	-relief flat \
	-variable agt_VarCommScroll

 checkbutton .f5.mbar.save -text "Save" \
	-relief flat \
	-variable agt_VarCommSave
		
 text .f5.text -relief ridge -bd 2 \
            -yscrollcommand ".f5.scroll set" \
            -wrap word \
	    -height 15 \
	    -width 65
 scrollbar .f5.scroll -relief ridge -command ".f5.text yview"

 pack .f5.label -side top -fill x

 pack .f5.mbar  -in .f5.log -side top -fill x 
 pack .f5.mbar.filter -side left -padx 2m
 pack .f5.mbar.scroll .f5.mbar.save -side right -padx 2m
 pack .f5.scroll -in .f5.log -side right -fill y
 pack .f5.text -in .f5.log -side left -fill both -expand yes


 # filter widget for communication observation
 frame .f5.filter -relief ridge -bd 2 \
	    -height 200 \
	    -width 480

 label .f5.filter.a \
	-text "Select agents to watch communication with." \
	-anchor c
 pack .f5.filter.a -side top -fill x -pady 5

 frame .f5.filter.b
 pack  .f5.filter.b \
	-side top -fill x -padx 1 -pady 5

 frame .f5.filter.b.l -relief flat
 pack .f5.filter.b.l -side left

 frame .f5.filter.b.r -relief flat
 pack .f5.filter.b.r -side right

 label .f5.filter.lbl1 \
	-text "Don't watch to:" \
	-anchor w 

 listbox .f5.filter.list1 \
	-yscroll ".f5.filter.scroll1 set" \
	-relief ridge \
	-bd 2\
	-height 6 -width 21 \
	-exportselection no

 bind .f5.filter.list1 <Double-Button-1> {
	 set index [.f5.filter.list1 near %y]
	 .f5.filter.list1 selection set $index
	 .f5.filter.add flash
	 .f5.filter.add invoke
	}

 scrollbar .f5.filter.scroll1 \
	-relief ridge \
	-command ".f5.filter.list1 yview"
 pack .f5.filter.lbl1 \
	-in .f5.filter.b.l \
	-side top \
	-fill x
 pack .f5.filter.scroll1 \
	-in .f5.filter.b.l\
	-side right \
	-fill y
 pack .f5.filter.list1 \
	-in .f5.filter.b.l \
        -fill both 

 label .f5.filter.lbl2 \
	-text "Watch to:" \
	-anchor w 

 listbox .f5.filter.list2 \
	-yscroll ".f5.filter.scroll2 set" \
	-relief ridge \
	-bd 2\
	-height 6 -width 21 \
	-exportselection no

 bind .f5.filter.list2 <Double-Button-1> {
	 set index [.f5.filter.list2 near %y]
	 .f5.filter.list2 selection set $index
	 .f5.filter.rem flash
	 .f5.filter.rem invoke
	}


 scrollbar .f5.filter.scroll2 \
	-relief ridge \
	-command ".f5.filter.list2 yview"

 pack .f5.filter.lbl2 \
	-in .f5.filter.b.r \
	-side top \
	-fill x
 pack .f5.filter.scroll2 \
	-in .f5.filter.b.r\
	-side right \
	-fill y
 pack .f5.filter.list2 \
	-in .f5.filter.b.r \
        -fill both


 button .f5.filter.add \
	-text "Add ->" \
	-width 8 \
	-command {
	  set sel [.f5.filter.list1 curselection]
	  foreach index $sel {
            set elem [.f5.filter.list1 get $index]
            .f5.filter.list1 delete $index
	    .f5.filter.list1 selection set $index
	    .f5.filter.list2 insert end $elem
	    .f5.filter.list2 selection set $index
	    lappend agt_VarTempComFilter $elem
	    set agt_VarTempComNotFilter \
              [lreplace $agt_VarTempComNotFilter $index $index]
	    .f5.filter.rem configure -state normal
	    if {[.f5.filter.list1 size] == 0} {
	      .f5.filter.add configure -state disabled
	    }
	  }
	  update
	}

 button .f5.filter.rem \
	-text "<- Remove" \
	-width 8 \
	-command {
	  set sel [.f5.filter.list2 curselection]
	  foreach index $sel {
            set elem [.f5.filter.list2 get $index]
            .f5.filter.list2 delete $index
	    .f5.filter.list2 selection set $index
	    .f5.filter.list1 insert end $elem
	    .f5.filter.list1 selection set $index
            lappend agt_VarTempComNotFilter $elem
	    set agt_VarTempComFilter \
              [lreplace $agt_VarTempComFilter $index $index]
	    .f5.filter.add configure -state normal
	    if {[.f5.filter.list2 size] == 0} {
	      .f5.filter.rem configure -state disabled
	    }
	  }
	  update
	}

 .f5.filter.rem configure -state disabled

 pack  .f5.filter.rem .f5.filter.add \
	-side bottom \
	-in .f5.filter.b \
	-padx 10 -pady 5 -ipadx 2m -ipady 0m 

 button .f5.filter.ok \
	-text OK \
	-width 8 \
	-command {
	  if {![cequal [lsort $agt_VarComFilter] \
	               [lsort $agt_VarTempComFilter]]} {
	    agt_logComm "Logging communication with the\
              following agents now:\n\{$agt_VarTempComFilter\}\n"
	  }
          set agt_VarComFilter $agt_VarTempComFilter
	  set agt_VarComNotFilter $agt_VarTempComNotFilter
	  pack forget .f5.filter
	  pack .f5.log -fill both -expand yes}

 button .f5.filter.cancel \
	-text Cancel \
	-width 8 \
	-command {pack forget .f5.filter
	          pack .f5.log -fill both -expand yes}
 
 pack .f5.filter.cancel .f5.filter.ok \
	-side right \
	-padx 5 -pady 10 -ipadx 2m -ipady 0m


 # initialize the global filter variables
 #
 set agt_VarComFilter ""
 set agt_VarComNotFilter "SELF UAM"
  
 .f5.mbar.filter invoke

 # create view for communication paths
 #
 frame .f2 -relief ridge -bd 2

 label .f2.label -text "Communication Paths:" \
		-relief ridge -bd 2 
		
 canvas .f2.c -relief ridge -bd 2 \
            -yscrollcommand ".f2.scroll set" \
	    -height 200 
 scrollbar .f2.scroll -relief ridge -command ".f2.c yview"

 pack .f2.label -side top -fill x
 pack .f2.scroll -side right -fill y
 pack .f2.c -side left -fill both -expand yes

 agt_displAgtBitmap SELF
 agt_displAgtBitmap UAM
 agt_drawSolid UAM


 # create view for belief-base
 #
 frame .f3 -relief ridge -bd 2

 label .f3.label -text "Belief-Base:" \
		-relief ridge -bd 2 
		
 text .f3.text -relief ridge -bd 2 \
            -yscrollcommand ".f3.scroll set" \
            -wrap word \
	    -height 15 \
	    -width 65
 scrollbar .f3.scroll -relief ridge -command ".f3.text yview"

 pack .f3.label -side top -fill x
 pack .f3.scroll -side right -fill y
 pack .f3.text -side left -fill both -expand yes

 bind all <Button-3> {agt_showShortHelp %W %X %Y}

 update idletasks

 wm deiconify .
 
 focus .mbar

 tkwait visibility .f1



# open logfiles for agent
#

 # get a free number for file counting
 #
 for {set fileNo 0} {[file exists "logfiles/$SELF.$fileNo.log"]} \
    {incr fileNo} {}

 keylset agents(SELF) FILENO $fileNo

 # open the standard logfile
 #
 keylset agents(SELF) LOGFID [open "logfiles/$SELF.$fileNo.log" w]

 agt_puts "Logfile for agent '$SELF'\ncreated at\
                  [fmtclock $creationTime "%a %b %d %Y, %T"]\n"

 # open the logfile for mesg-evaluation
 #
 keylset agents(SELF) MLGFID [open "logfiles/$SELF.$fileNo.msg" w]

 agt_logMesgEval "Message Evaluation Logfile for agent '$SELF'\ncreated at\
                  [fmtclock $creationTime "%a %b %d %Y, %T"]"

 # open the logfile for communication
 #
 keylset agents(SELF) CLGFID [open "logfiles/$SELF.$fileNo.com" w]

 agt_logComm "Communication Logfile for agent '$SELF'\ncreated at\
                  [fmtclock $creationTime "%a %b %d %Y, %T"]\n"



# initialize variable for trace enabling with 'no'
 set agt_VarTraceEnable 0


# start agents prolog process
#
 set agt_VarPrologError [agt_startProlog]


# initialize counter for outgoing messages, 
# used for :REPLY-WITH key in messages 
#
 set agt_MesgCounter 0

# initialize list for messages to send
#
 set agt_SendMesgList "" 

# initialize buffers for received messages
#
# agt_RecMesgList(-1) is for UAM-command-mesgs which have highest priority
# agt_RecMesgList(0) is for intern messages of SELF, which have next priority
# agt_RecMesgList(1), 
# agt_RecMesgList(2), 
# agt_RecMesgList(3) are for all the other messages; 3 has lowest priority
#
 set agt_RecMesgList(-1) ""
 set agt_RecMesgList(0) ""
 set agt_RecMesgList(1) ""
 set agt_RecMesgList(2) ""
 set agt_RecMesgList(3) ""

# use the procedure 'agt_chgStatus' to send an initial status
# message to UAM; because 'agt_chgStatus' is changing the status
# set it to active for this little moment now and it is changed
# to deactive by the procedure - so it is correct for the start of
# agent
#
 agt_puts "STATUS: $agents(SELF)\n"

 keylset agents(SELF) STATUS act
 agt_chgStatus none


# initialize global array OTHER that includes the 
# names of agents SELF if actually connected to
#  (special hack 8-)
#
 set OTHER(dummy) ""
 unset OTHER(dummy)

# set a trace on OTHER to mention a change on it and
# updating the filter for watching special communications
#
 trace variable OTHER w agt_updateComFilter
 trace variable OTHER u agt_updateComFilter
 trace variable agents(UAM) u agt_updateComFilter


# do some more initials:
# includes reading/evaluation of script file, inits prolog database
#
 if {$agt_VarPrologError} {
   keylset mesg :RECEIVER UAM :TYPE prolog-error
   agt_send $mesg
   agt_sourceStdMesgRules
 } else {
   agt_doTypeInits
 }

# set communication control on connection to UAM
#
 fileevent stdin readable {agt_commCtrl stdin UAM}

# open pipe to send messages to the agent himself
# and set commCtrl on it
#
 pipe in out

 keylset agents(SELF) INFID $in \
                      OUTFID $out

 fileevent $in readable "agt_commCtrl $in $SELF"

# start detecting the socket for connections to other agents
#
 fileevent [keylget agents(SELF) SOCKET] readable {agt_conDetected [keylget agents(SELF) SOCKET]}


# send messages appended till now:
# there is the statusinfo and maybe an ASL-error from agt_doTypeInits
#
 agt_dispatchMesgs 

# enter main loop
#
 while {1} {

  if {([set agt_VarBufferEmpty(-1) [lempty $agt_RecMesgList(-1)]]) && \
      ([keylget agents(SELF) STATUS] == "deact")} {
 
      # there is no UAM-command-mesg and agent is deact, so wait
      # for an incomming mesg that could change this;
      # then start loop again

      tkwait variable agt_RecMesgList

      set agt_VarWorkMesg ""

  } elseif {!$agt_VarBufferEmpty(-1)} {
        # there is an UAM-command-mesg so eval it 
        # because of highest priority
        #
        set agt_VarWorkMesg [lvarpop agt_RecMesgList(-1)]

  } elseif {![lempty $agt_RecMesgList(0)]} {
        # there is no UAM-command-mesg but a message from SELF 
        # so eval this one next
        #
        set agt_VarWorkMesg [lvarpop agt_RecMesgList(0)]

  } elseif {![lempty $agt_RecMesgList(1)]} {
        # 
        #
        set agt_VarWorkMesg [lvarpop agt_RecMesgList(1)]

  } elseif {![lempty $agt_RecMesgList(2)]} {
        #
        #
        set agt_VarWorkMesg [lvarpop agt_RecMesgList(2)]

  } elseif {![lempty $agt_RecMesgList(3)]} {
        # 
        #
        set agt_VarWorkMesg [lvarpop agt_RecMesgList(3)]

  } else {
        # here the agent is active but there are no mesg in buffers,
        # so wait for an incoming message and then start loop again

        tkwait variable agt_RecMesgList
        
        set agt_VarWorkMesg ""
  }   

  if {$agt_VarWorkMesg != ""} {
        # the message is evaluated here
        #
        agt_evalMesg [lindex $agt_VarWorkMesg 0] [lindex $agt_VarWorkMesg 1]

        # dispatch messages built up in evaluation
        #
        agt_dispatchMesgs

        update
  }
 }



