#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# agt_catchPrologCrash
#
#  Tries to restart Prolog process and rebuilt the actual 
#  Belief-Base after Prolog process break down.
#  Called on write error to Prolog from agt_writeProlog.
# 
# Arguments:
#
# 	prologFd	- filedescriptor of crashed process
#
# Results:
#
#  0, if Prolog process is re-established, 
#     registers new Prolog process descriptor in agents(SELF);
#  1, if process could not be establishes
#
proc agt_catchPrologCrash {prologFd} {
global agents

 # wait for process to terminate
 catch {wait [pid $prologFd]}

 # close old Prolog filedescriptor
 catch {close $prologFd}

 agt_puts "Prolog-Process crashed !\nRebuilding now..." 


 # start new Prolog process and register its fid
 #
 if {[agt_startProlog]} {
   keylset mesg :RECEIVER UAM :TYPE prolog-error
   agt_send $mesg
   agt_puts "Rebuilding failed!\n"
   return 1
 }

 # save Belief-Base to a temp-file and read it back to Prolog
 #
 if {([agt_saveBelBase "agt_tmpFile"]) || \
     ([agt_loadBelBase "agt_tmpFile"])} {
   keylset mesg :RECEIVER UAM :TYPE prolog-error
   agt_send $mesg
   agt_puts "Rebuilding failed!\n"
   return 1
 }

 agt_puts "Belief-Base has been rebuilt now."

 catch {close $tmpFid}

 return 0
}

