#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# agt_closecon 
#
#  Closes a connection to a specified address.
#
# Arguments:
#
#	addr	-	addr of agent to close connection to
#
# Results:
#
#  Returns
#
#   0 - if connection is closed
#   1 - if no connection to addr is existing or waiting for response
#   2 - if specified addr belongs to SELF
#   3 - if specified addr belongs to UAM
#
#  Removes all entries about the agent belonging to addr in arrays 'agents'
#  and 'OTHER'
#
proc agt_closecon {addr} {
 global agents OTHER
 global agt_VarComFilter

 # transform address to TCP/IP-integer form:
 set addr [agt_hostport2addr $addr]

 # search for an entry with this address in array 'agents':
 set found 0

 foreach agtName [array name agents] {

   if {[cequal [keylget agents($agtName) ADDR] $addr]} {
     set found 1
     break
   }
 }

 if {!$found} {
   agt_puts "No connection to address '$addr' !\n"
   return 1
 }

 # do not remove SELF !
 if {[cequal $agtName SELF]} {
   agt_puts "Cannot remove connection to SELF !\n"
   return 2
 }

 # do not remove connection to UAM !
 if {[cequal $agtName UAM]} {
   agt_puts "Cannot remove connection to UAM !\n"
   return 3
 }

 #  remove pipes from/to agent
 #
 catch {set inFid [keylget agents($agtName) INFID]}
 catch {removeinput $inFid}
 catch {close $inFid}
 catch {close [keylget agents($agtName) OUTFID]}
 
 agt_puts "Connection to agent '$agtName' closed!\n"

 if {[lsearch $agt_VarComFilter $agtName] != -1} {
     agt_logComm "Communication with agent '$agtName' closed!\n"
 }
 
 # forget all about this agent
 #
 catch {unset OTHER($agtName)}

 agt_delAgtBitmap $agtName
 catch {unset agents($agtName)}

 return 0
}

