#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

# agt_commCtrl
#
#  Procedure for controlling incomming pipes.
#
#  agt_commCtrl is called by addinput whenever an event occurs
#  on an incoming pipe.
#  The event is analysed and it is checked if pipe is broken.
#  If not, message is read from pipe and put in the FIFO for
#  incoming messages or the agent deals with it directly, if it
#  is a command-message from his UAM.
#
# Arguments:
# 
# 	fid	-	pipe identifier
#	agtName -	name of agent pipe comes from
#			(may be 'unknown')
#
# Results:
#
#  None
#

proc agt_commCtrl {fid agtName} {
 global agents agt_RecMesgList SELF


 # it follows a special hack:
 #
 # there were problems when building connections across the net to
 # other hosts; socket building works, but when a message is sent
 # reading causes an error, 'Resource temporarily unavailable'
 #
 # it seems to me that it is an delay error:
 # the signal for addinput seems to be faster as the mesg
 #
 # the following catches these error until the mesg can be read...
 # 
 # not nice but it works
 #
 while {[catch {set readResult [lgets $fid mesg]}]} {}  

 # if the empty string is read the connection is broken
 #
 if {$readResult == -1} {
 # broken connection detected  
   agt_conBroken $fid $agtName
   return
 }

 if {[catch {keylget mesg :SENDER sender}]} {
   # the mesg has a currupted format and will be
   # dismissed by 'agt_evalMesg' later;
   # put it in mesg-buffer with priority 1 for that
   lappend agt_RecMesgList(1) [list $fid "$mesg"]
   return
 }


 if {[cequal $sender UAM]} {

   # in this case the received message has :SENDER UAM
   #
   if {![cequal $fid stdin]} {

     # in this case the mesg was not received on UAM's pipe - stdin, 
     # so forget it
     #
     agt_puts "Received a message with sender 'UAM' not from UAM:\n$mesg \
               \nMessage ignored!\n"
     return

   } elseif {[cequal [keylget mesg :TYPE] command]} {

     # this is an UAM-command-mesg, so append it to the mesg-buffer
     # agt_RecMesgList(-1) which has highest priority:

     lappend agt_RecMesgList(-1) [list $fid "$mesg"]

   } else {

     # this is a 'normal' mesg from UAM, so append it to the mesg-buffer
     # according to the :PRIORITY of the mesg or agt_RecMesgList(1)
     # by default:

     if {(![keylget mesg :PRIORITY priority]) || \
         (![regexp ^1$|^2$|^3$ $priority])} {
         keylset mesg :PRIORITY 1
         set priority 1
     } 

     lappend agt_RecMesgList($priority) [list $fid "$mesg"] 
 
   }

 } elseif {[cequal $sender $SELF]} {

   if {![cequal $fid [keylget agents(SELF) INFID]]} {
     agt_puts "Received a message with sender '$SELF' not from SELF:\n$mesg \
               \nMessage ignored!\n"
     return
   }
   
   # this message is from the agent himself, so put it in the mesg-buffer
   # agt_RecMesgList(0), which is for SELF mesgs only, if no lower priority
   # is specified:
   #
   if {(![keylget mesg :PRIORITY priority]) || \
       (![regexp ^0$|^1$|^2$|^3$ $priority])} {
       keylset mesg :PRIORITY 0
       set priority 0
   } 

   lappend agt_RecMesgList($priority) [list $fid "$mesg"] 

 } else {

   # the mesg is from an other agent, so put it in the mesg-buffer
   # that belongs to the mesg :PRIORITY, agt_RecMesgList(1) by
   # default
   #

   if {(![keylget mesg :PRIORITY priority]) || \
       (![regexp ^1$|^2$|^3$ $priority])} {
       keylset mesg :PRIORITY 3
       set priority 3
   } 

   lappend agt_RecMesgList($priority) [list $fid "$mesg"] 
 }

}


