#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# agt_conBroken  
#
#  Handles a broken connection on fid,
#  closes pipes to/from agent. 
#
# Arguments:
#
#	fid	-	name of broken pipe
#	agtName	-	name of agent pipe 'fid' comes from
#
# Results:
#
#  None
#

proc agt_conBroken {fid agtName} {
 global agents SELF OTHER
 global agt_VarComFilter

 catch {fileevent $fid readable ""}
  
 #  remove pipes from/to agent
 #
 catch {close $fid}
 catch {close [keylget agents($agtName) OUTFID]}
 
 agt_puts "Connection $fid to agent '$agtName'\
          broken!\n"

 if {[lsearch $agt_VarComFilter $agtName] != -1} {
     agt_logComm "Communication with agent '$agtName' broken!\n"
 }
 
 if {[cequal $agtName $SELF]} {
   # if the pipe leads from/to SELF build up a new one:
   keyldel agents(SELF) OUTFID
   keyldel agents(SELF) INFID

   agt_puts "Pipe to SELF broken !\nRebuilding now ... " 0

   # open new pipe to send messages to the agent himself
   # and set commCtrl on it

   pipe in out

   keylset agents(SELF) INFID $in \
                        OUTFID $out

   fileevent $in readable "agt_commCtrl $in $SELF"

   agt_puts "ready\n"

 } else {
   # if the pipe not leads to SELF forget everything about the agent
   # the pipe belongs to:
   catch {unset OTHER($agtName)}

   agt_delAgtBitmap $agtName
   catch {unset agents($agtName)}
 } 

 # if the file is 'stdin' the connection to UAM is broken;
 # display the 'Exit'-menu in the menuebar of the agent to enable
 # a correct quit:

 if {[cequal $fid stdin]} {
   pack .mbar.exit -side left -padx 2m

   tk_menuBar .mbar .mbar.view .mbar.exit .mbar.help

   bind all <Control-q> {agt_exit}
 }

}

