#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# agt_evalMesg
#
#  Tries to match an received message with agents mesg-rules
#  and to evalute a matching rule.
#
# Arguments:
#
#	agt_VarFid	-	file descriptor mesg arrived on
#	agt_VarMesg	-	mesg just received
#
# Results:
#  
#  Side effects belonging to evaluation of the mesg rules.
#
proc agt_evalMesg {agt_VarFid agt_VarMesg} {
global agents agt_MesgRules SELF OTHER
global agt_VarBreakEval agt_VarInitialization
global agt_VarComFilter
global agt_VarStdMesgRuleNo

 keylget agents(SELF) NAME SELF

 # set the actual time in Prolog
 agt_timeSet

 # display received message
 if {[catch {set agt_VarText [agt_formatMesg $agt_VarMesg]} errorMesg]} {

  # Format error in received mesg
  # send back an error to sender
  #
  # search for the agent belonging to the fid:

  # added by Rene Schubert 12/03/97
  # tries to prevent sudden crash of agent
  # handles completely unformatted messages
  # message is discarded without further action

  if {[catch {set agt_VarAgtName [keylkeys agents]}]} {
      agt_puts "FORMAT ERROR when receiving :\n$agt_VarMesg\n"
      return
  }
  # end of hack

  foreach agt_VarAgtName [keylkeys agents] {
    if {[keylget agents($agt_VarAgtName) INFID] == $agt_VarFid]} {
      break
    }
  }
   
  keylset agt_VarRespMesg :RECEIVER $agt_VarAgtName :TYPE error \
       :CODE 1 :COMMENT "message format error" 

  if {[keylget agt_VarMesg :REPLY-WITH replyWith]} {
    keylset agt_VarRespMesg :REPLY-TO $replyWith
  }

  agt_send $agt_VarRespMesg
  
  agt_puts "FORMAT ERROR when receiving :\n$agt_VarMesg\nFailure:\
            '$errorMesg'\n"
  return
 }

 agt_puts "Message received\n$agt_VarText"

 keylget agt_VarMesg :SENDER mesgSender

 regsub $SELF $mesgSender SELF mesgSender

 foreach name $agt_VarComFilter {
   if {[lsearch $mesgSender $name] != -1} {
     agt_logComm "Message received from '$name'\n$agt_VarText"
   }
 } 
 

 # initialize a variable that shows if a rule matched this mesg:
 set agt_VarRuleMatch 0


 #run through all MesgRules
 #
 set agt_VarMaxRule [array size agt_MesgRules]

 for {set agt_VarRuleNo 1} {$agt_VarRuleNo <= $agt_VarMaxRule} \
     {incr agt_VarRuleNo} {
   # initialize the evaluation break variable with NO
   set agt_VarBreakEval 0

   # get actual rule
   set agt_VarActRule $agt_MesgRules($agt_VarRuleNo)
   
   # get the pattern for the mesgForm from actRule
   set agt_VarMesgForm [lindex $agt_VarActRule 0]

   # test if listlengths of form is less than that of mesg,
   # if not take next MesgRule
   if {[llength $agt_VarMesgForm] > [llength $agt_VarMesg]} {
     continue
   }
  
   # get the keylist of received mesg
   set agt_VarMesgKeys [keylkeys agt_VarMesg]

   # get the keylist of mesgForm
   set agt_VarFormKeys [keylkeys agt_VarMesgForm]

   # test if keylists of form is included in that of mesg,
   # if not take next MesgRule
   set agt_VarKeyError 0

   foreach agt_VarKey $agt_VarFormKeys {
     if {[lsearch $agt_VarMesgKeys $agt_VarKey] == -1} {
       set agt_VarKeyError 1
       break
     }
   }

   if {$agt_VarKeyError} {continue}

  # try to match the rule with the message:

   set agt_VarMatchError 0

   set agt_VarKeyValue ""

   # run through all mesgForm keys
   foreach agt_VarKey $agt_VarFormKeys {
     keylget agt_VarMesgForm $agt_VarKey agt_VarPattern
     keylget agt_VarMesg $agt_VarKey agt_VarMatch

     # try to match the actual key of mesg with the agt_VarPattern
     # in mesgForm

     if {![agt_keyPattMatches $agt_VarPattern $agt_VarMatch]} {
       set agt_VarMatchError 1
       break
     }

     keylset agt_VarKeyValue $agt_VarKey $agt_VarMatch
   }

   # if the matching failed take next MesgRule
   if {$agt_VarMatchError} {
     continue
   }

   # in this case the rule matches the message, so eval it:

   set agt_VarRuleMatch 1

   # get the MesgCond of the actRule and eval it:
   set agt_VarMesgCond [lindex $agt_VarActRule 1]

   # eval MesgCond and catch errors
   if {[catch {set agt_VarCondResult [expr $agt_VarMesgCond]} errorMesg]} {
     agt_puts "Error when evaluating Message-Rule No. $agt_VarRuleNo,\
               message condition:\n'$agt_VarMesgCond'"
     agt_logMesgEval "\nError when evaluating Message-Rule No. $agt_VarRuleNo,\
               message condition:\n'$agt_VarMesgCond'"
     if [info exists errorMesg] {
       agt_puts "Error: $errorMesg\n"
       agt_logMesgEval "Error: $errorMesg"
     }
     agt_tracepoint
     continue
   }

   if {$agt_VarCondResult != 0} {
     set index 2
   } else {
     set index 3
   }

   # depending on result of MesgCond choose the action from rule,
   # eval it and catch errors
   set agt_VarAction [lindex $agt_VarActRule $index] 

  #  mesg-rule control output
   if {$agt_VarRuleNo > $agt_VarStdMesgRuleNo} {
     agt_logMesgEval "\nEvaling Message-Rule No. $agt_VarRuleNo:\nMesg-Form:\
        \t$agt_VarMesgForm\nVar-Values:\t$agt_VarVariableValues\nMesg-Cond:\
        \t($agt_VarMesgCond) evals to\
        $agt_VarCondResult\nAction:\t\t($agt_VarAction)"
   }
  
   if {[catch {eval $agt_VarAction} errorMesg]} {
     agt_puts "Error when evaluating Message-Rule No. $agt_VarRuleNo,\
               action:\n'$agt_VarAction'"
     agt_logMesgEval "\nError when evaluating Message-Rule No. $agt_VarRuleNo,\
               action:\n'$agt_VarAction'"
     if [info exists errorMesg] {
       agt_puts "Error: $errorMesg\n"
       agt_logMesgEval "Error: $errorMesg"
     }
     agt_tracepoint
   }

   # if 'agt_break' is called in the action break the evaluation:
   #
   if {$agt_VarBreakEval} {
     break
   }

   # eval the standard action of the rule which is
   # done independent of MesgCond
   set agt_VarStdAction [lindex $agt_VarActRule 4]

   if {$agt_VarRuleNo > $agt_VarStdMesgRuleNo} {
     agt_logMesgEval "Std-Action:\t($agt_VarStdAction)"
   }

   if {[catch {eval $agt_VarStdAction} errorMesg]} {
     agt_puts "Error when evaluating Message-Rule No. $agt_VarRuleNo,\
               standard-action:\n'$agt_VarStdAction'"
     agt_logMesgEval "\nError when evaluating Message-Rule No. $agt_VarRuleNo,\
               standard-action:\n'$agt_VarStdAction'"
     if [info exists errorMesg] {
       agt_puts "Error: $errorMesg\n"
       agt_logMesgEval "Error: $errorMesg"
     }
     agt_tracepoint
   }

   # if 'agt_break' is called in the action break the evaluation:
   #
   if {$agt_VarBreakEval} {
     break
   }

 }

 # if no rule matches the message send back an error message 
 # to the sender of the message:
 #
 if {!$agt_VarRuleMatch} {
  keylset agt_VarRespMesg :RECEIVER [keylget agt_VarMesg :SENDER] :TYPE error \
       :CODE 2 :COMMENT "no match for message" \
       :REPLY-TO [keylget agt_VarMesg :REPLY-WITH] 

  agt_send $agt_VarRespMesg
 }

 agt_puts "End of message evaluation!\n"

}


#
# agt_break
#
#  Sets global variable agt_VarBreakEval to 1. This is used to break evaluation
#  of a message in agt_evalMesg
#
# Arguments: None.
#
# Results:
#
#  global agt_VarBreakEval is 1
#
proc agt_break {} {
global agt_VarBreakEval

 set agt_VarBreakEval 1
}



