#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# agt_getHandshake
#
#  Reacts to a handshake received on an incoming pipe.
#
# Arguments:
#
#	inFid	-	identifier of incomming pipe
#	agtName	-	name of agent that sent rehandshake
#	agtIdfk	-	identifikation information of that agent
#	replyTo	-	number to reply with

proc agt_getHandshake {inFid agtName agtIdfk replyTo} {
 global agents OTHER

 catch {removeinput $inFid}

 # initialize flags for registering agents identifikation 
 # and sending a confirmation mesg
 #
 set register 0
 set send_handshake 0

 # check if connection to senders address exists 
 #

 # get agents address from received identifikation:
 #
 keylget agtIdfk ADDR recAddr

 set fid_found 0

 foreach name [array name agents] {

   if {[keylget agents($name) ADDR] == $recAddr} {
     set fid_found 1
     break
   }
 }

 if {!$fid_found} {

   # this is a connection setup:
   # create an outgoing pipe, register agent identifikation
   # and response with a handshake message
   #
   if {[catch {set outFid [agt_createCon $recAddr]}]} {
     agt_puts "Connection to $agtName failed!\n"

     catch {unset agents($agtName)}
     catch {close $inFid}
     return
   }

   set register 1
   set send_handshake 1

 } else {

   # there is an entry for this address, so this is the
   # confirmation of the connection:
   #
   # if the connection is fully established do nothing,
   # else register agents idetifikation under his name
   # 

   if {[keylget agents($name) INFID {}] == 0} {
     # there must be a filedesc for outgoing mesgs
     #
     keylget agents($name) OUTFID outFid

     # remove old array entry
     #
     unset agents($name)

     # delete old agent bitmap
     #
     agt_delAgtBitmap $name

     set register 1
   }
   
 }


 # set communication control back on the incoming pipe - if not broken
 #
 if {[catch {
        fileevent $inFid readable "agt_commCtrl $inFid $agtName"
    }]} {
   agt_conBroken $inFid $agtName
   return
 }

 if {$register} {
   # register agents identifikation
   #
   set agents($agtName) $agtIdfk

   # display the agent bitmap with a solid connection
   #
   agt_displAgtBitmap $agtName
   agt_drawSolid $agtName

   keylset agents($agtName) INFID $inFid \
                            OUTFID $outFid

   set OTHER($agtName) $recAddr

   agt_puts "Registered: $agents($agtName)\n"
 }

 if {$send_handshake} {
   # send a handshake for confirmation to sender
   #
   keylset idfk NAME [keylget agents(SELF) NAME] \
                TYPE [keylget agents(SELF) TYPE] \
                ADDR [keylget agents(SELF) ADDR] 

   keylset response :RECEIVER $agtName :TYPE handshake \
                    :IDFK $idfk :REPLY-TO $replyTo

   agt_send $response
 }
 
}


