#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# agt_handshake
#
#  Tries to connect to a specified agent and initializes
#  a handshake.
#
# Arguments:
#
#	agtAddr	-	address of the agent to connect to
#
# Results:
#
#	0	-	connection established, handshake send
#	1	-	waiting for response already (i.e. for re-handshake)
#	2	-	connection exists already 
#	3	-	connection refused
#
 
proc agt_handshake {hostportAddr} {
 global agents

 # transform the address in TCP/IP-integer form:
 set agtAddr [agt_hostport2addr $hostportAddr]

 # check if there is a connection to specified host and port
 # already existing
 #
 set agtList [array name agents]

 foreach name $agtList {

   if {[keylget agents($name) ADDR] == $agtAddr} {

   if {[catch {keylget agents($name) INFID}]} {
       agt_puts "Waiting for response of '$hostportAddr' already\
               :\n\n$agents($name)\n"
       return 1
     } else {
       agt_puts "Connection to '$hostportAddr' exists already\
               :\n\n$agents($name)\n"
       return 2
     }
   }
 }


 keylset agents($agtAddr) NAME $agtAddr \
                          ADDR $agtAddr
     
 # try to connect to specified address
 #
 if {[catch {keylset agents($agtAddr) OUTFID [agt_createCon $agtAddr]} errorMesg]} {
   # failure in creating connection
   #
   agt_puts "Connection to '$agtAddr' failed:\n$errorMesg\n"
   catch {unset agents($agtAddr)}
   return 3
 }
    

 agt_puts "register: $agents($agtAddr)\n"


 # set SELF identifikation to send with handshake
 #
   keylset idfk NAME [keylget agents(SELF) NAME] \
                TYPE [keylget agents(SELF) TYPE] \
                ADDR [keylget agents(SELF) ADDR]

   keylset mesg :RECEIVER $agtAddr :TYPE handshake :IDFK $idfk
   agt_send $mesg

 agt_displAgtBitmap $agtAddr
 agt_drawWeak $agtAddr

#agt_puts "Handshake to: $agents($agtAddr)\n"
 return 0
}

