#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# agt_keyPattMatches
#
#  Tries to match a key pattern of a mesg rule with a 
#  key content of a received mesg.
#  Called by agt_evalMesg.
#
# Arguments:
#
#	pattern	-	pattern to match with
#	match	-	string that should be matched
#
# Results:
#
#  Returns 1 if the match was successful. In this case the
#  variables of the pattern are bounded to values of the
#  match in the context of the calling procedure, so that 
#  they can be used for evalling mesg cond and actions.
#  If matching fails 0 is returned.
#
proc agt_keyPattMatches {pattern match} {
upvar agt_VarVariableValues variableValues

 # this is a variable used to print the actual values of
 # variables in mesg-form, initialize it as an empty string
 set variableValues ""

 # indexcounter for pattern
 set ctp 0      

 # indexcounter for match
 set ctm 0     

 # counter for number of variables in pattern
 set ctVar 0   

 # flag that causes variable recognition in pattern
 set varCheck 1

 # get length of pattern and match
 set lenp [string length $pattern]
 set lenm [string length $match]

 # start matching 
 while {$ctp <= $lenp} {
   set chp [cindex $pattern $ctp]
   set chm [cindex $match $ctm] 

   if {[cequal $chp "~"]} {
       if {[cindex $pattern $ctp+1] == "~"} {
         incr ctp
         set chp [cindex $pattern $ctp]

       } else {
         # set varCheck off
         set varCheck 0
        incr ctp
        continue
       } 
   }


   if {$varCheck && [ctype upper $chp]}  {
   # prefix of variable in pattern detected

     incr ctVar
     
     # read name of variable, 
     # i. e. char of form a-z, A-Z, 0-9, _
     while {([ctype alnum $chp] || [cequal $chp _])} {
       append varName($ctVar) $chp
       incr ctp
       set chp [cindex $pattern $ctp]
     }

     # $chp is now separator after varName

     # initialize variable named $varName($ctVar) at
     # one level up
     upvar $varName($ctVar) localVar$ctVar
     catch {unset localVar$ctVar}

#puts "varName:($ctVar) >$varName($ctVar)<"; flush stdout

     set varCont($ctVar) ""     

     # read char from match until separator or end of match
     # is reached
     while {(![cequal $chm $chp]) && ($ctm <= $lenm)} {
       append varCont($ctVar) $chm
       incr ctm
       set chm [cindex $match $ctm]
     }

#puts "varCont($ctVar): >$varCont($ctVar)<"; flush stdout  

     # if end of match is reached befor separator is read
     # matching failed
     if {$ctm > $lenm} { return 0 }

     # else content for variable is examined and saved in
     # varCont($ctVar), so continue matching

     #set varCheck on
     set varCheck 1

     continue
   } 
 
   # in case of no variable in pattern 
   # just compare the actual characters

   if [cequal $chp $chm] {
     incr ctp
     incr ctm
   } else {
     return 0
   } 

   # set varCheck on
    set varCheck 1

 }

 # if strings are completely matched set varCont to its
 # variable at upper level
 for {set ct 1} {$ct <= $ctVar} {incr ct} {
   set localVar$ct $varCont($ct)
   lappend variableValues "$varName($ctVar) $varCont($ct)"
 }
  
 return 1
}


