#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# agt_loadDescPart
#
#  Loads a part of agent description from a file as
#  specified in a 'load <fileName>' statement in type file
#
# Arguments:
#
#	typeFile (referenz)	- list of lines of agt type file not
#				  scanned till now
#	fileName		- file to read
#
# Results:
#
#  Set 'typeFile' as the concatenation of the lines list read
#  from fileName with the old 'typeFile'.
#  Shows an error message, if 'fileName' is not readable.
#
	
proc agt_loadDescPart {typeFile fileName} {
upvar $typeFile file

# try to open file 'fileName' in directory 'desc'
 if {[catch {set fid [open desc/$fileName]} errorMesg]} {
   agt_puts "Error when loading '$fileName':"
   if {[info exists errorMesg]} {
     agt_puts "$errorMesg\n"
   }
   return
 }

 # read file and split it in lines
 set data [split [read $fid] \n]
 close $fid

# concat the just read part and the old one
 set file [concat $data $file]

 agt_puts "Loaded: $fileName\n"

}


