#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# agt_parseActSrc
#
#  Used to read the ACTION-C-SOURCE specification from type file.
#
# Arguments:
# 	typeFile (referenz)	- list of lines of agt type file not
#				  scanned till now
#	actLine	(referenz)	- actual line to scan
#
# Results:
#
#  'typeFile' is scanned for the specification in balanced
#  brackets, read lines are popped.
#  'actLine' contains the unscanned rest of the actual line to scan. 
#  Returns an error with an errorMesg if a scan error occurs
#
proc agt_parseActSrc {typeFile actLine} {
upvar $typeFile file \
      $actLine line

 # variable 'actSrcFile' will contain the name of the source file
 set actSrcFile ""

 # The next character in this line must be an open bracket
 #
 if {[cindex $line 0] != "\{"} {
   return -code error "'\{' expected at '$line'"
 }

 # initialize counter for open brackets 
 set brackets 1

 # first character in $line is now a \{-bracket
 # remove the bracket and following spaces
 set line [string trimleft [csubstr $line 1 end]]

# initialize a character counter for 'line'
 set chCt -1

 # run through file until brackets are balanced
 while {$brackets != 0} {
   incr chCt
   set char [cindex $line $chCt]

   switch -glob -- $char {
     \{  {incr brackets 1}

     \}  {incr brackets -1}

     "" {
         # if last character in line is "\" remove it before
         # appending the line to actSrcFile and poping next line 
         # from file;
         #
         if {[cequal "\\" [cindex $line $chCt-1]]} {
           if {$chCt >= 2} {
             append actSrcFile [crange $line 0 $chCt-2] 
           }
           append actSrcFile " "
         }

         if {[cequal $file ""]} {break}

         # get the next line from file
         set line [string trim [lvarpop file]]
         set chCt -1
        }
    
     #  {
         # this is a remark, so forget this line and
         # get the next line from file
         set line [string trim [lvarpop file]]
         set chCt -1}
   }
 }

 if {$chCt} {
   append actSrcFile [crange $line 0 $chCt-1]
 } 
    
 if {$brackets} {
   return -code error "'\}' expected at '$line'."
 }

 # character under 'chCt' is now a \}-bracket;
 # remove the chars till 'chCt' inclusive the bracket and following
 # white spaces
 #
 set line [string trimleft [crange $line $chCt+1 end]]

}


