#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# agt_parseOneAction
#
#  Parses the type file for one correct action and put this
#  to the Tcl-Interpreter.
#
# Arguments:
#
#	typeFile (referenz)	- list of lines of agt type file not
#				  scanned till now
#	actLine	(referenz)	- actual line to scan next;
#				  starts with the begin of a action
#
# Results:
#
#  Removes the action from typeFile, set actline so that its first character 
#  is the first one after the end of action.
#  Returns an error with error message if syntax error is detected. 
# 
proc agt_parseOneAction {typeFile actLine} {
upvar $typeFile file \
      $actLine line
global agents

 # the first character of line is here the first one _AFTER_
 # the (-bracket and line has no succeeding white spaces

 # the scanning itself is done in a subroutine that puts the read
# characters of the action in variable 'action' or 
 # returns with error if it fails
 #
 if {[catch {agt_parseSubPart file line action "\n"} errorMesg]} {
   return -code error $errorMesg
 }


 # check if the action is really a Tcl-procedure declaration
 #
 if {![regexp ^proc $action]} {
   return -code error \
     "proc-declaration expected in '[csubstr $action 0 20] ...'"
 }

 # eval the action so that it can be called by Tcl
 #
 if {[catch {eval $action} errorMesg]} {
   return -code error \
     "$errorMesg in '[csubstr $action 0 20] ...'"
 }
 
 agt_puts "Read Action: >$action<\n"
}


