#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# agt_parseOneBeliefRule
#
#  Parses the type file for one correct belief-rule and inserts it in BeliefBase.
#
# Arguments:
#
#	typeFile (referenz)	- list of lines of agt type file not
#				  scanned till now
#	actLine	(referenz)	- actual line to scan next;
#				  starts with the begin of a belief-rule
#
# Results:
#
#  Removes the belief-rule from 'typeFile',
#  set 'actline' so that its first character is the first one 
#  after the end of rule and adds the rule to the belief-base.
#  Returns an error with error message if syntax error is 
#  detected. 
#

proc agt_parseOneBeliefRule {typeFile actLine} {
upvar $typeFile file \
      $actLine line
global agents

 # the first character of line is here the first one _AFTER_
 # the (-bracket and line has no succeeding white spaces

 # the scanning itself is done in a subroutine that puts the read
 # characters of the action in variable 'beliefRule' or
 # returns with error if it fails
 #
 if {[catch {agt_parseSubPart file line beliefRule} errorMesg]} {
   return -code error $errorMesg
 }


 if {![string match *:-* $beliefRule]} {
   return -code error "':-' expected in '$beliefRule'"
 }
   

 # substitute all occurenses of $SELF in the scanned
 # beliefRule with agents name
 regsub -all {\$SELF} $beliefRule [keylget agents(SELF) NAME] beliefRule 
 
 # insert the rule in the BeliefBase
 if {[catch {agt_quietQuery assert(($beliefRule))} errorMesg]} {
   return -code error "Prolog - $errorMesg"
 }

 agt_puts "Read Belief-Rule: >$beliefRule<\n"
}


