#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# agt_parseOneMesgRule
#
#  Parses the type file for one correct message-rule and puts it in global array agt_MesgRules.
#
# Arguments:
#
#	typeFile (referenz)	- list of lines of agt type file not
#				  scanned till now
#	actLine	(referenz)	- actual line to scan next;
#				  starts with the begin of a message-rule
#
# Results:
#
#  Removes the message-rule from 'typeFile',
#  set 'actline' so that its first character is the first one 
#  after end of rule and add the rule to the global array
#  agt_MesgRules.
#  Returns and error with error message if syntax error is 
#  detected. 
#

proc agt_parseOneMesgRule {typeFile actLine} {
upvar $typeFile file \
      $actLine line

global agents agt_MesgRules

 # the first character of line is here the first one _AFTER_
 # the (-bracket and line has no succeeding white spaces

 # initialize a counter for ()-brackets
 #
 set brackets 1

 # initialize a counter for the 5 parts of a mesg-rule
 #
 set index 0

 # go on till brackets are balanced
 while {$brackets} {

   switch -regexp -- $line {
    ^$|^[\\]$  {
                # if not end of file get next line
                #
                if {[cequal $file ""]} {
                  return -code error \
                    "')' expected"
                }

                set line [string trim [lvarpop file]]
               }

    ^[(] {
          # not more than 5 parts in a mesg-rule
          #
          if {$index == 5} {
             return -code error \
             "')' expected after '($rulePart(4))'"
          }

          # remove the (-bracket and following white spaces
          # from line
          #
          set line [string trimleft [csubstr $line 1 end]]

          # on index 2, 3 and 4 Tcl-commands are read, so we
          # have to add a ";" after each command
          #
          if {[string match *$index* "2 3 4"]} {
            set eofl ";"
          } else {
            set eofl ""
          }

          # scan for expression in ()-brackets with subroutine,
          # return on error
          #
          if {[catch {agt_parseSubPart file line rulePart($index) $eofl} \
               errorMesg]} {
            return -code error $errorMesg
          }

          # line starts now with first character after )-bracket;
          # read "," after parts with index less or equal 3
          #
          if {$index <= 3} {
            if {[cindex $line 0] != ","} {
              return -code error \
                "',' expected after '($rulePart($index))'"
            }
            
            # remove ',' and following white spaces from line
            #
            set line [string trimleft [csubstr $line 1 end]]
          }

          # increment the part index
          incr index
         }

     ^[)]  {incr brackets -1}
    
    default {if {$index == 5} {
               set errorMesg "')' expected at '$line'"
             } else {
               set errorMesg "'(' expected at '$line'"
             }
             return -code error $errorMesg
            }   

   }
 }

 # first character in $line is now a )-bracket,
 # remove the bracket and following white spaces
 #
 set line [string trimleft [csubstr $line 1 end]]

 # check now the message-form:
 #
 # check for correct message-keys
 #  
 regsub -all "\}\{" $rulePart(0) "\} \{" rulePart(0)

 foreach key [keylkeys rulePart(0)] {
   if {[cindex $key 0] != ":"} {
     return -code error \
     "in '$rulePart(0)': expected ':' as first character of key '$key'"
   }

   if {[catch {keylget rulePart(0) $key keyCont} errorMesg]} {
     return -code error \
      "in '$rulePart(0)': $errorMesg"
   }

   # convert key to upper characters
   #
   keyldel rulePart(0) $key
   keylset rulePart(0) [string toupper $key] $keyCont   
   
 }


 # replace all occurences of $SELF in mesg-form with agents name
 #
 regsub -all {\$SELF} $rulePart(0) [keylget agents(SELF) NAME] \
    rulePart(0)


 # built up the new rule as a list of the subparts
 #
 for {set index 0} {$index <= 4} {incr index} {
   lappend newMesgRule $rulePart($index)   
 }

 # set the new mesg-rule in the global array agt_MesgRules
 #
 set ruleNo [expr [array size agt_MesgRules]+1]

 set agt_MesgRules($ruleNo) $newMesgRule


 agt_puts "Read Message-Rule $ruleNo: >$agt_MesgRules($ruleNo)<\n"

}


