#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# agt_parseSubPart
#
#  Parses the type file for one expression in balanced ()-bracket.
#
# Arguments:
#
#	typeFile (referenz)	- list of lines of agt type file not
#				  scanned till now
#	actLine	(referenz)	- actual line to scan next;
#				  starts with the begin of the subPart
#	subPart (referenz)	- contains the parsed expression
#
# Results:
#
#  Returns the characters in balanced ()-brackets in 'subPart',
#  removes them from 'typeFile' and sets 'actline' so that its first
#  character is the first one afer )-bracket.
#  Returns an error with error message if brackets are not 
#  balanced. 
#

proc agt_parseSubPart {typeFile actLine subPart {eofl ""}} {
upvar $typeFile file \
      $actLine line \
      $subPart part

 # the first character of line is here the first one _AFTER_
 # the (-bracket and line has no succeeding white spaces

 # initialize a counter for ()-brackets and one for double-quotes
 #
 set brackets 1
 set dblquote 0

 # initialize a counter for parsing position in line
 #
 set chCt -1

 set part ""

 # go on till brackets are balanced
 while {$brackets} {

   incr chCt

   set char [cindex $line $chCt]

   switch -glob -- $char {
     (  {if {[cindex $line $chCt-1] != "\\"} {
            incr brackets
         }
        }

     )  {if {[cindex $line $chCt-1] != "\\"} {
            incr brackets -1
         }
        }

     \" {if {[cindex $line $chCt-1] != "\\"} {
           incr dblquote [expr ($dblquote) ? -1 : 1]
         }
        }

     "" {
         # if last character in line is "\" remove it before
         # appending the line to part and poping next line from
         # file;
         # => white spaces before "\" are save
         #
         if {[cequal "\\" [cindex $line $chCt-1]]} {

           if {$chCt >= 2} {
             append part [crange $line 0 $chCt-2] 
           }

           append part " "

         } else {
           
           # eofl is used to add ";" or "\n" after Tcl-commands
           if {$chCt} {
             append part [crange $line 0 $chCt-1] $eofl
           }

         }

         if {[cequal $file ""]} {
             break
         }

         set line [string trim [lvarpop file]]
         set chCt -1
        }
    
     #  {
         # this is a comment if the '#'-character occurs 
         # at the beginning of the line
         if {$chCt == 0} {
           set line [string trim [lvarpop file]]
           set chCt -1
         }
        }


   }
 }

 if {$chCt} {
   append part [crange $line 0 $chCt-1]
 }

 if {$brackets} {
   return -code error "')' missing for '($part'"
 }

 # catch unbalanced double-quotes, because test had shown that
 # an unbalanced double quote at the end of a belief makes Prolog hang
 #
 if {$dblquote} {
   return -code error "unbalanced double-quotes in '$part'"
 }

 # first character in $line is now a )-bracket,
 # remove the bracket and following white spaces
 #
 set line [string trimleft [crange $line $chCt+1 end]]

}

