%
% Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
% Computer Science Department,
% Christian-Albrechts-University of Kiel,
% Olshausenstr. 40, 24118 Kiel, Germany
%
% All rights reserved.
% No warranties will be given on any issues arising out of the use
% of this software product.
% Permission to use, copy, modify and distribute this software
% product for non-commercial purposes is hereby granted, provided
% that the above copyright notice appears in all copies and
% respective publications.
% All commercial trades with this product outside the CAU Kiel
% without specific written prior permission are prohibited.
%

%
% agt_prologServ.pl
%
% BinProlog side of a Tcl/BinProlog interface 
% used in IDEAS.
%
% Based on 'server.pl' by Paul Tarau.
% Modified for IDEAS by Olaf Scheew, 1995
%

% The BinProlog process is in IDEAS a child of the
% Tcl process of an agent and is subordinated to Tcl.
% BinProlog gets the input from Tcl on stdin and
% writes the output to stdout.
% Tcl sends queries integrated in one of three
% special string which are 'call_prolog', 'query_prolog'
% and 'query_prolog_list'. Prolog reacts on this strings
% as defined later.

:-module(prolog).
:-['myprogs/wam'].
:-end_module.


:-op(200,fx,('$')).

% start predicate for the Prolog Tcl-server
%
main:-prolog_server.

% main loop of the server:
% initializes the server when starting, repeats
% reading input from stdin and evaling it by using
% the predicate 'react'
%
prolog_server:-
	init_server,
	repeat,
	  tcl_in(X,Vs),
        react(X,Vs),
	!.

% initialization of server:
% shuts of interactivity and sends a EOPA
% - End Of Prolog Answer - to Tcl for initialization
% confirmation
%
init_server:-
	interactive(no),
	end_symbol.

% the predicate 'react':
react(end_of_file,_):-!.

% 'call_prolog' is used to do a quite Prolog evaluation,
% i.e. there is no result given back to Tcl except errors
%
react(call_prolog(X),_):-!,X,!,end_symbol,fail.

% 'query_prolog' gives back all the results of the Prolog
% evaluation in the standard BinProlog format
%
react(query_prolog(X),Vs):-query_prolog(X,Vs),fail.

% 'query_prolog_list' gives back the result in a special form:
% all values for variables in the query X are divided from each 
% other by a '|' symbol.
% If the result of query is TRUE, a single '|' is sent. 
% If the result is FALSE, just the EOPA is sent.
% This format is catched on the Tcl side of this interface and 
% used to build up a Tcl list of the result.
% 'query_prolog_list' is the standard form used for all queries
% to the Belief-Base.
% 
react(query_prolog_list(X),Vs):-query_prolog_list(X,Vs),fail.

query_prolog(X,Vs):-X,nl,member(A,Vs),write(A),nl,fail.
query_prolog(_,_):-end_symbol.

query_prolog_list(X,Vs):-X,write('|'),nl,member(A,Vs),write(A),nl,fail.
query_prolog_list(_,_):-end_symbol.

% 'EOPA' marks the end of the prolog answer and is sent at
% the end of each query evaluation to Tcl.
end_symbol :- cwrite('EOPA'),nl,flush.

% predicates 'in' and 'in1' for reading input from
% stdin into a Prolog variable
tcl_in(T):-in(T,_).

tcl_in(T,Vs):-read_term(L,Vs),!,in1(L,T).

in1([X|Xs],T):-!,member(T,[X|Xs]).
in1(T,T).
