#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# agt_showShortHelp
#
#  This procedure is invoked by pressing the right mouse button
#  on widgets in UAM.
#  It creates a shortHelp window and shows a helptext registered
#  for this widget in global array 'agt_ShortHelp'
#  done in 'agt_initShortHelp.tcl'.
#
# Arguments:
#
#	helpItem	-	widget name to show help for
#	xPos	  	-	x-coordinate for help window
#	yPos	  	-	y-coordinate for help window
#
# Results:
#
#  None.
#

proc agt_showShortHelp {helpItem xPos yPos} {
 global agt_ShortHelp

 # initialize shortHelp at first call
 #
 if {![info exists agt_ShortHelp]} {
   source agt_initShortHelp.tcl
 }

 if {![info exists agt_ShortHelp($helpItem)]} {
   return
 }

 # create the ShortHelp-window
#
 set w .shortHelp
 catch {destroy $w}

 toplevel $w
 wm withdraw $w
 wm title $w "IDEAS - ShortHelp"
 wm iconname $w "ShortHelp"

 frame $w.f1 
 frame $w.f2
 pack $w.f1 -padx 5 -pady 5 -fill both -expand yes
 pack $w.f2 -side bottom -padx 5 -pady 5 -fill x

 button $w.but1 -text Close \
		-command "destroy $w" \
		-width 10
 pack $w.but1 -in $w.f2 -expand 1 \
		-padx 3m -ipadx 2m -ipady 1m


 text $w.text -relief raised -bd 2 \
            -yscrollcommand "$w.scroll set" \
	    -font -Adobe-Times-Medium-R-Normal-*-180-* \
	    -wrap word \
	    -height 8 \
	    -width 35
 scrollbar $w.scroll -command "$w.text yview"
 pack $w.scroll -in $w.f1 -side right -fill y
 pack $w.text -in $w.f1 -side left -fill both -expand yes
 
 bind $w <Enter> "set agt_VarSafeFocus [focus]; focus $w"
 bind $w <Leave> {focus $agt_VarSafeFocus}
 bind $w <Return> "$w.but1 invoke"
 bind $w <Escape> "$w.but2 flash; $w.but2 invoke"
 
 update idletasks

 if {$xPos < 0} {
    set dad [winfo parent $w]
    set xPos [expr [winfo x $dad] + \
		 ([winfo reqwidth $dad] - [winfo reqwidth $w])/2]
    set yPos [expr round([winfo y $dad] + [winfo reqheight $dad]*0.3)]
 }

 wm geometry $w +$xPos+$yPos
 wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
 wm deiconify $w
 
 $w.text insert end $agt_ShortHelp($helpItem)

 $w.text configure -state disabled

}

