#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# These are the standard message-rules known by all agents.
# They control the basic actions of the agent, as initialization
# status change, exit, handshakes etc.
#
# The message-rules are stored in global array 'agt_MesgRules'
# and this piece of Tcl-Code is 'sourced' by 'agt_doTypeInits'
# when initializing or reseting an agent.
#
set agt_MesgRules(1) [list {{:SENDER UAM} \
                            } \
                            {[.f1.text index end] >= 500} \
                            {.f1.text configure -state normal
                             .f1.text delete 1.0 100.0
                             .f1.text configure -state disabled
                             if {[.f4.text index end] >= 500} {
                               .f4.text configure -state normal
                               .f4.text delete 1.0 100.0
                               .f4.text configure -state disabled
                             }
                             if {[.f5.text index end] >= 500} {
                               .f5.text configure -state normal
                               .f5.text delete 1.0 100.0
                               .f5.text configure -state disabled
                             }
                            } \
                            {} \
                            {set agt_VarRuleMatch 0}]

set agt_MesgRules(2) [list {{:TYPE command} \
                            {:COMMAND handshake} \
                            {:ADDRESS Addr} \
                            {:REPLY-TO ReplyTo} \
                            {:REPLY-WITH ReplyWith} \
                            {:SENDER Send} \
                            {:RECEIVER Recv}} \
                            {$Send == "UAM"} \
                            {foreach agtName $Addr { \
                               agt_handshake $agtName \
                             } \
                            } \
                            {agt_puts "No command messages from agent\
                             '$Send' accepted !\nMessage ignored!\n"
                             agt_break
                            } \
                            {}]
                                           
set agt_MesgRules(3) [list {{:TYPE command} \
                            {:COMMAND chg_status} \
                            {:REPLY-TO ReplyTo} \
                            {:REPLY-WITH ReplyWith} \
                            {:SENDER Send} \
                            {:RECEIVER Recv}} \
                            {$Send == "UAM"} \
                            {agt_chgStatus $ReplyWith} \
                            {agt_puts "No command messages from agent\
                             '$Send'accepted !\nMessage ignored!\n"
                             agt_break
                            } \
                            {}]

set agt_MesgRules(4) [list {{:TYPE command} \
                            {:COMMAND exit} \
                            {:REPLY-TO ReplyTo} \
                            {:REPLY-WITH ReplyWith} \
                            {:SENDER Send} \
                            {:RECEIVER Recv}} \
                            {$Send == "UAM"} \
                            {agt_exit} \
                            {agt_puts "No command messages from agent\
                             '$Send' accepted !\nMessage ignored!\n"} \
                            {agt_break}]

set agt_MesgRules(5) [list {{:TYPE command} \
                            {:COMMAND reset} \
                            {:REPLY-TO ReplyTo} \
                            {:REPLY-WITH ReplyWith} \
                            {:SENDER Send} \
                            {:RECEIVER Recv}} \
                            {$Send == "UAM"} \
                            {agt_puts "Doing a RESET now!\n\nReseting\
                                       Prolog process:\n"; 
                             agt_doTypeInits} \
                            {agt_puts "No command messages from agent\
                             '$Send' accepted !\nMessage ignored!\n"} \
                            {agt_break}]

set agt_MesgRules(6) [list {{:TYPE command} \
                            {:COMMAND save} \
                            {:FILE FileName} \
                            {:REPLY-TO ReplyTo} \
                            {:REPLY-WITH ReplyWith} \
                            {:SENDER Send} \
                            {:RECEIVER Recv}} \
                            {$Send == "UAM"} \
                            {
                             if {[agt_saveBelBase $FileName]} { \
                               keylset mesg :TYPE failed \
                             } else { \
                               keylset mesg :TYPE ok \
                             }; \
                             keylset mesg :REPLY-TO $ReplyWith \
                                          :RECEIVER $Send ; \
                             agt_send $mesg \
                            } \
                            {agt_puts "No command messages from agent\
                             '$Send' accepted !\nMessage ignored!\n"
                             agt_break
                            } \
                            {}]


set agt_MesgRules(7) [list {{:TYPE command} \
                            {:COMMAND load} \
                            {:FILE FileName} \
                            {:REPLY-TO ReplyTo} \
                            {:REPLY-WITH ReplyWith} \
                            {:SENDER Send} \
                            {:RECEIVER Recv}} \
                            {$Send == "UAM"} \
                            { \
                             if {[agt_loadBelBase $FileName]} { \
                               keylset mesg :TYPE failed \
                             } else { \
                               keylset mesg :TYPE ok \
                             } ;\
                             keylset mesg :REPLY-TO $ReplyWith \
                                          :RECEIVER $Send ; \
                             agt_send $mesg \
                            } \
                            {agt_puts "No command messages from agent\
                             '$Send' accepted !\nMessage ignored!\n"
                             agt_break
                            } \
                            {}]

set agt_MesgRules(8) [list {{:TYPE handshake} \
                            {:IDFK Idfk} \
                            {:REPLY-TO ReplyTo} \
                            {:REPLY-WITH ReplyWith} \
                            {:SENDER Send} \
                            {:RECEIVER Recv}} \
                            {1} \
                            {agt_getHandshake $agt_VarFid \
                                $Send $Idfk $ReplyWith} \
                            {} \
                            {}]


set agt_MesgRules(9) [list {{:TYPE error} \
                            {:CODE Code} \
                            {:COMMENT Comment} \
                            {:REPLY-TO ReplyTo} \
                            {:REPLY-WITH ReplyWith} \
                            {:SENDER Send} \
                            {:RECEIVER Recv}} \
                            {1} \
                            {agt_puts "Received an ERROR-MESSAGE!\n"} \
                            {} \
                            {}]

set agt_MesgRules(10) [list {{:TYPE initialization} \
                            {:REPLY-TO ReplyTo} \
                            {:REPLY-WITH ReplyWith} \
                            {:SENDER Send} \
                            {:RECEIVER Recv}} \
                            {$agt_VarInitialization} \
                            {set agt_VarInitialization 0} \
                            {agt_break} \
                            {}]



set agt_MesgRules(11) [list {{:TYPE command} \
                            {:COMMAND testcon} \
                            {:ADDRESS Addr} \
                            {:REPLY-TO ReplyTo} \
                            {:REPLY-WITH ReplyWith} \
                            {:SENDER Send} \
                            {:RECEIVER Recv}} \
                            {$Send == "UAM"} \
                            {foreach addr $Addr { \
       			      if {[set agtName [agt_checkaddr $addr]] != 0} {\
                               keylset mesg :RECEIVER $agtName \
                                 :TYPE  information \
                                 :MESSAGE "This is just a Test!" 
                               agt_send $mesg
                              } else {\
                               agt_puts "Not connected to address '$addr'.\n"\
                              }\
			     }\
                            } \
                            {agt_puts "No command messages from agent\
                             '$Send' accepted !\nMessage ignored!\n"
                             agt_break
                            } \
                            {}]



