#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# agt_updateComFilter
#
#  This function is called automatically by Tcl-command 'trace'
#  whenever the global arra OTHER is changed, i.e. is written
#  or unset.
#  agt_updateComFilter checks which element of the array is new
#  or deleted and if the communication to the agent with this name
#  is filtered for communication logfile, i.e. if this name is in 
#  global list agt_VarComFilter or in agt_VarComNotFilter.
#  These lists and their two help lists, agt_VarTempComFilter and
#  agt_VarTempComNotFilter, are updated.
#  Also the entry in the filter selection widget in agents observation
#  window are updated. 
#   
# Arguments:
#
#	varname  - name of the traced variable (here always OTHER)
#	element  - element of array that is modified
#	operator - operations that happend to the element:
#	            w for writing,
#		    u for unsetting the element
# Results: None.
#
proc agt_updateComFilter {varname element operator} {
global OTHER agt_VarComFilter agt_VarTempComFilter
global agt_VarComNotFilter agt_VarTempComNotFilter

 if {[cequal $operator w]} {
   # when operation is writing it is always a new element;
   # append it always to the NotFilter list
   #
   .f5.filter.list1 insert end $element
   lappend agt_VarComNotFilter $element
   lappend agt_VarTempComNotFilter $element
 } else {
  if {[cequal $operator u]} {
    # when a element is unset, check in which list it appears
    # and remove it from there and from the list in filter widget
    #
    if {[set index [lsearch $agt_VarComFilter $element]] != -1} {
      set agt_VarComFilter [lreplace $agt_VarComFilter $index $index]
    } else {
      if {[set index [lsearch $agt_VarComNotFilter $element]] != -1} {
        set agt_VarComNotFilter [lreplace $agt_VarComNotFilter $index $index]
      }
    }

    if {[set index [lsearch $agt_VarTempComFilter $element]] != -1} {
      set agt_VarTempComFilter [lreplace $agt_VarTempComFilter $index $index]
      .f5.filter.list2 delete $index
    } else {
      if {[set index [lsearch $agt_VarTempComNotFilter $element]] != -1} {
        set agt_VarTempComNotFilter \
          [lreplace $agt_VarTempComNotFilter $index $index]
        .f5.filter.list1 delete $index
      }
    }
  }
 }

 update
}
