#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# agt_valid
#
#  Returns the valid time interval of a belief.
#
# Arguments:
#
#  belief - belief to check
#  option - if specified as 'log' a human readable time is printed to Logfile,
#           by default not done
#
# Results:
#
#  -1          - if Prolog error occurs
#  0           - if belief is not in Belief-Base
#  {from till} - if belief is valid from 'from' and till 'till'
#                where 'from' and 'till' are UNIX datestrings
#
#
proc agt_valid {belief {option no}} {
global agt_VarTimeOffset

 if {[set result [agt_listQuery "validity($belief,FROM, TILL)"]] == -1} {
   return -1
 }

 if {$result == 0} {
   if {[cequal $option "log"]} { 
     agt_puts "No Belief '$belief'.\n"
   }
   return 0
 }

 # the result is a list where each element is a KeyedList;
 # for an unique result only deal with the first list element:
 #
 set result [lindex $result 0]

 keylget result FROM timeFrom
 keylget result TILL timeTill

 # correct the Prolog-time with the time offset used to make
 # a UNIX datestring Prolog readable:
 incr timeFrom $agt_VarTimeOffset
 incr timeTill $agt_VarTimeOffset

 if {[cequal $option "log"]} { 
   agt_puts "Belief '$belief' valid\nfrom\
             ([fmtclock $timeFrom])\ntill\
             ([fmtclock $timeTill]).\nCurrent time:\
             ([fmtclock [getclock]])\n"
 }

 return "$timeFrom $timeTill"

}

#
# agt_time
#
# Returns the actual time of the agent by reading
# the standard Prolog-fact time/1 of Belief-Base.
#
# Arguments: None.
#
# Results:
#
# The actual time in integer format or
# -1 if an error occured.
# 
proc agt_time {} {
global agt_VarTimeOffset

 if {[set result [agt_listQuery "time(Time)"]] == -1} {
   return -1
 }

 set actTime [lindex $result 0]

 keylget actTime Time actTime

 return [incr actTime $agt_VarTimeOffset]
}
