% C-instruction coder
wcode_c(Op,Reg,F,N):-
  patch_specials(F,NewF),
  make_ccode(Op,Reg,NewF,N).

make_ccode(Op,Reg,F,N):- % writes C-macro to *.h file
  c_instr(Op,Reg,F,N,Instr),!,
  c_trace(Instr),
  wchunk(Instr).
make_ccode(Op,Reg,F,N):- % C-ified bytecode
  make_bcode(Op,Reg,F,N).

% makes the bytcode in *.c file
make_bcode(Op,Reg,F,N):-
  write('{'),write(Op),write(','),
  write(Reg),write(','),
  write(N),write(',"'),write(F),write('"},'),nl.

make_comment(Cmd):-
  currpred(Pred/Arity-No/Offs),NL=[10],
  make_cmd0(
     [NL," /* ",Pred," / ",Arity," [",No," / ", Offs, "] */"],
  Cmd).

% writes hint about starting/ending a C-chunk
wspec(Op,Reg,_,N):- % begin chunk
  c_spec(Op,FunSym,Instr),!,
  make_comment(Cmd),
  lwchunk(Cmd),
  wchunk(Instr),
  lwrite(Cmd),
  c_trace(Instr),
  write(' {'),write(Op),write(','),
  write(Reg),write(','),
  write(N),write(',(void *)'),write(FunSym),write('},'),nl.
wspec(Op,Reg,F,N):- % end chunk:
  make_ccode(Op,Reg,F,N).
  
lwrite(L):-member(X,L),put_code(X),fail.
lwrite(_).

% deals with info about current predicate
newpred0(H):-
  functor(H,F,N),
  bb_val(predicate,name,F),
  bb_val(predicate,arity,N),!,
  bb_val(predicate,clause,I),
  I1 is I+1,
  bb_set(predicate,clause,I1),
  bb_let(predicate,offset,0).
newpred0(H):-
  functor(H,F,N),
  bb_let(predicate,name,F),
  bb_let(predicate,arity,N),
  bb_let(predicate,clause,1),
  bb_let(predicate,offset,0).

currpred(F/N-I/Offs):-
  bb_val(predicate,name,F),
  bb_val(predicate,arity,N),
  bb_val(predicate,clause,I),
  bb_val(predicate,offset,Offs).

patch_specials(Name,NewName):-
  name(Name,Xs),B=92,Q=34, % i.e. = \ and "
  findall(Y,
     ( member(X,Xs), patch_it(X,B,Q,Y)
     ),
  Ys),
  name(NewName,Ys).

patch_it(B,B,_,Y):-!,(Y=B;Y=B).
patch_it(Q,B,Q,Y):-!,(Y=B;Y=Q).
patch_it(X,_,_,X).

c_chunk_file(F_H):-val(c_chunk,file_name,F_H).

write_c_chunk(G):-!,
  c_chunk_file(CF),
  telling(F),
  tell(CF),
  G,
  tell(F).

wchunk(C):-write_c_chunk((write(C),nl)).

lwchunk(C):-atomic(C),!,write_c_chunk((write(C),nl)).
lwchunk([X|Xs]):-write_c_chunk((lwrite([X|Xs]),nl)).

% starts and ends *.c file
c_decorate_file(header):-!,
  init_gensym(xx),
  wchunk('/* do not edit: generated by co.pl */'),
  write('#include "../src/global.h"'),nl,
  write('#include "../src/c_defs.h"'),nl,
  c_chunk_file(F_H),write('#include "'),write(F_H),write('"'),
  nl,nl,
  write('struct bp_instr {unsigned char op,reg; unsigned short arity; char *name;} wam_bp[] = {'),
  nl.
c_decorate_file(footer):-
  write('{255,0,0,"c"}};'),nl,nl,
  write('long wam_bp_size=sizeof(wam_bp);'),nl.



% chunk generation

% flag: true if in chunk, counts FunSym
c_check:-c_check0(_).

c_check0(No):-bb_val(c_chunk,in,No),integer(No).

c_find_fun(F,_,_,_):-integer(F),!,
  errmes(unexpected,integer(F)).
c_find_fun(F,N,No,K):-
  for(I,0,No),
    bb_val(fun_symbol,I,F),
    bb_val(fun_arity,I,N),
  !,
  K=I.
c_find_fun(F,N,No,No1):-
  No1 is No+1,
  bb_set(c_chunk,in,No1),
  bb_let(fun_symbol,No1,F),
  bb_let(fun_arity,No1,N),
  make_bcode(6,0,F,N). % looks like write_constant 

c_chunk_begin:-
  bb_let(c_chunk,in, -1).

c_chunk_end:-
  bb_let(c_chunk,in,no).

c_inc(F,N,K):- % increments code ofset (P-area)
  bb_val(c_chunk,in,No), 
  integer(No),
  c_find_fun(F,N,No,K).

% trace=>generates code which prints out C-ified operations
% begin_end=>generates begin...end c_chunk for byte code on file

c_trace(Instr):-bb_val(c_flag,trace,yes),!,
  write_c_chunk((
  write('c_trace(" '),write(Instr),write(' ")'),nl
  )).
c_trace(_).


% computes/retrieves offsets from P-at-begin-of-the C-chunk
% used in c_instr.pl

new_offset(O):-
  bb_val(predicate,offset,O),
  O1 is O+1,
  bb_set(predicate,offset,O1).

get_offset(O):-
  bb_val(predicate,offset,O).
