% BinProlog x.xx Copyright (C) 1992 Paul Tarau. All rights reserved.
% COMPILER: dcgs --> prolog --> binary progs --> code

%----------PREPROCESSOR TO BINARY LOGIC PROGRAMS-------------

% Transforms definite metaprograms to binary meta-programs
% definite meta clause -> definite clause+ some replacements

mdef_to_def((H:-B),(H:-NewB)):-repl_body(B,NewB).
mdef_to_def((:-B),(:-NewB)):-repl_body(B,NewB),!,
	NewB,  % called here and not propagated to the next step
	fail.

% replaces metavars and some builtins in clause bodies
			 
repl_body(MetaVar,call(MetaVar)):-var(MetaVar),!.
repl_body(M,ExpM):-repl_macro(M,ExpM),!.
repl_body(A,NewA):-split_op(A,L),!,strip_nil(L,NewA).
repl_body(A,A).

cutp(X):-name(X,"$cut").

repl_macro('!','$bin_cut'(X)):-cutp(X).
repl_macro(var(X),Known):-nonvar(X),repl_known(var(X),Known).
repl_macro(nonvar(X),Known):-nonvar(X),repl_known(nonvar(X),Known).
repl_macro(atomic(X),Known):-nonvar(X),repl_known(atomic(X),Known).
repl_macro(float(X),Known):-nonvar(X),repl_known(float(X),Known).
repl_macro(atomic(X),Known):-nonvar(X),repl_known(atomic(X),Known).
repl_macro((A,B),NewAB):-repl_conj(A,B,NewAB).
repl_macro((A;B),NewAB):-repl_disj(A,B,NewAB).
repl_macro((A->B),(NewA->NewB)):-
	repl_body(A,NewA),
	repl_body(B,NewB).
repl_macro(compare(R,A,B),compare0(A,B,R)).
repl_macro(A==B,compare0(A,B,=)).
repl_macro(A@<B,compare0(A,B,<)).
repl_macro(A@>B,compare0(A,B,>)).
repl_macro(':=:'(K,X),lval(K1,K2,X)):-repl_lval(K,K1,K2).
repl_macro('#'(X),dcg_connect(X)).
% you can speed up read.pl, write.pl with this if you do not use the C-parser
%repl_macro(put_code(N),put(N)):-which_prolog(binprolog).
%repl_macro(nl,cnl):-which_prolog(binprolog).
%repl_macro(get_code(N),get0(N)):-which_prolog(binprolog). 

repl_conj(A,B,NewAB):-nonvar(A),split_op(A,NewA),!,
	repl_body(B,NewB),
	app_body(NewA,NewB,NewAB).
repl_conj(A,B,(NewA,NewB)):-
	repl_body(A,NewA),
	repl_body(B,NewB).

repl_disj(If,C,if(NewA,NewB,NewC)):-nonvar(If),If=(A->B),!,
	repl_body(A,NewA),
	repl_body(B,NewB),
	repl_body(C,NewC).
repl_disj(A,B,or(NewA,NewB)):-
	repl_body(A,NewA),
	repl_body(B,NewB).

repl_lval('#'(K1,K2),K1,K2):-!.
repl_lval(K,K,K).

repl_known(X,true):-X,!.
repl_known(_,fail).
	
split_op(X is B,R):-split_is_rel(X,B,R).
split_op(A < B,R):-split_rel(less,A,B,R).
split_op(A > B,R):-split_rel(greater,A,B,R).
split_op(A =< B,R):-split_rel(less_eq,A,B,R).
split_op(A >= B,R):-split_rel(greater_eq,A,B,R).
split_op(A =:= B,R):-split_rel(arith_eq,A,B,R).
split_op(A =\= B,R):-split_rel(arith_dif,A,B,R).
% split_op(tab(Expr),R):-split_is(X,Expr,R,[tab(Y)]). % faster static tab

split_is_rel(X,B,[expr(B,X)]):-var(B),!.
split_is_rel(X,B,[+(B,0,X)]):-atomic(B),!.
split_is_rel(X,B,[+(B,0,X)]):-float(B),!.
split_is_rel(X,B,[+(R,0,X)]):-which_prolog(binprolog),ground(B),!,expr(B,R).
split_is_rel(X,B,R):-split_is(X,B,R,[]).

app_body([],Bs,Bs):-!.
app_body([A|As],Bs,(A,Cs)):-app_body(As,Bs,Cs).

strip_nil([A],A):-!.
strip_nil([A|As],(A,Bs)):-strip_nil(As,Bs).

split_rel(Op,A,B,Res):-split_rel_1(Op,A,B,Res,[]).

split_rel_1(Op,A,B)-->
	split_is(X,A),
	split_is(Y,B),
	{OpXY=..[Op,X,Y]},
	emit_is(OpXY).

%split_is(X,A)-->{var(A)},!,emit_is(expr(A,X)). % very slow but flexible is/2
split_is(X,A)-->{var(A)},!,{X=A}.               % fast but static is/2
split_is(X,A)-->{atomic(A)},!,{X=A}.
split_is(X,A)-->{float(A)},!,{X=A}.
split_is(R,OpAB)-->
	{OpAB=..[Op,A,B]},!,
	split_is(VA,A),
	split_is(VB,B),
	{OpArgs=..[Op,VA,VB,R]},
	emit_is(OpArgs).
split_is(R,OpA)-->
	{OpA=..[Op,A]},
	split_is(VA,A),
	{OpArgs=..[Op,VA,R]},
	emit_is(OpArgs).

emit_is(X,[X|Xs],Xs).

% converts a definite clause to a binary metaclause
%    where each metavariable Cont represents a "continuation"
%    and a goal G is represented by a clause :- G.

def_to_mbin((H:-B),M):-!,def_to_mbin0(H,B,M).
def_to_mbin(H,M):-def_to_mbin0(H,true,M).

def_to_mbin0('@@'(H,Upper),B,(HC:-BC)) :- nonvar(H),!,
	termcat(H,ContH,HC),
	add_upper_cont(B,Upper,ContH,BC).
def_to_mbin0(H,B,(HC:-BC)) :-
	termcat(H,Cont,HC),
	add_cont(B,Cont,BC).		

add_upper_cont(B,Upper,ContH,BC):-nonvar(Upper),!,
	add_cont(Upper,ContU,ContH),
	add_cont(B,ContU,BC).
add_upper_cont(B,Upper,ContH,BC):-
	add_cont((strip_cont(ContH,Upper,ContU),B),ContU,BC).

% adds a continuation to a term

add_cont((true,Gs),C,GC):-!,add_cont(Gs,C,GC).
add_cont((fail,_),C,fail(C)):-!.
add_cont((G,Gs1),C,GC):-!,
		 add_cont(Gs1,C,Gs2),
		 term_append(G,cont(Gs2),GC).
add_cont(G,C,GC):-term_append(G,cont(C),GC).

termcat(T,C,TC):-term_append(T,cont(C),TC).


% simple WAM-code lister

show_code(IIs):-
	findall(I,show_code0(IIs,I),Is),
	length(Is,Len),
	write('WAM-ASSEMBLER:'(length=Len)),nl,
	nth_member(I,Is,K),
          show_instr(K,I),
	fail.
show_code(_):-nl.


show_code0(IIs,I):-
	member(Is,IIs),
	member(I,Is),
	show_or_skip(I).

show_or_skip(ii(get,variable,arg(I),var(I-_,_))):-!,fail.
show_or_skip(ii(put,value,arg(I),var(I-_,_))):-!,fail.
show_or_skip(_).

show_instr(K,I):- I=ii(Op,T,X,Y),
	write([K]),write(' '),write(Op),write('_'),
	write(T),write(' '),
	show_fun(X),
        show_info(Y),
	nl.

show_fun(put):-!.
show_fun(get):-!.
show_fun(T):-write(T).

show_info(Var):-Var=var(I-_,_),!,
  write(' '),write(var(I)),
  put_code(9),put_code(9),put_code(9),write('% '),write(Var).
show_info(Var):-write(' '),write(Var).
  
show_steps(asm,M,D):-
  nl,M\==D,write('EXPANDED:'),nl,portray_clause(M),nl,fail
; write('DEFINITE:'),nl,portray_clause(D),nl,fail.
show_steps(_,_,_).

% ------------------------------------------------------------------
% BYTE CODE GENERATOR

maxarity(256).

% #if defined TEMPS_AT_END
% temparg(Arg,Temp):-maxarity(Max),Temp is (Max-1)-Arg.

temparg(Arg,Arg).

gen_instr(ii(Op,Type,Arg,Var),Enc):-beautify(Arg,Op,Type,Var,Enc),!.

write_instr(empty_op,_,yes):-!.
write_instr(ii(O,T,X,Y),To,yes):-symcat(O,T,OT),encode(OT,X,Y,To),!.
write_instr(_,_,no).
  
gen_code(To,IIs):-
	member(Is,IIs),mc_peephole(Is,Enc),
        write_instr(Enc,To,Ok),
	Ok=no,
	!,
	fail.
gen_code(_,_).

% eliminates redundancies and beautifies the code
beautify(arg(X),Op,Type,V,Enc):-!,encode_arg(X,Op,Type,V,Enc).
beautify(temp(X),Op,Type,V,Enc):-!,temparg(X,Temp),
   encode_arg(Temp,Op,Type,V,Enc).
beautify(cutarg(X),Op,Type,V,Enc):-!,encode_arg(X,Op,Type,V,Enc).
beautify(Val,Op,Type,var(Xn-_,_),Enc):-!,encode2(Op,Type,Val,Xn,Enc).
beautify(put,write,constant,X,Enc):-cutp(X),!,
  encode2(push,cut,?,?,Enc).
beautify(Y,Op,X,Z,Enc):-encode2(Op,X,Y,Z,Enc).

encode_arg(I,get,variable,var(I-_,_),empty_op):-!.
encode_arg(I,put,value,var(I-_,_),empty_op):-!.
encode_arg(An,Op,Type,var(Xn-_,_),Enc):-!,encode2(Op,Type,Xn,An,Enc).
encode_arg(1,Op,constant,X,Enc):-cutp(X),!,encode2(Op,cut,?,?,Enc).
encode_arg(An,Op,constant,S,Enc):-encode2(Op,constant,S,An,Enc).

encode2(Op,Type,X,Y,ii(Op,Type,X,Y)).

% compacts voids, terms and useless c_chunks
mc_peephole(Xs,X):-mc_top(Xs,X).

mc_input(O,Is,Js):-mc_input0(O,Is,Js). % ,ttyprint(O).

mc_input0(Y,[X|Xs],Ys):-
  gen_instr(X,I),
  mc_empty(I,Y,Xs,Ys).

mc_empty(empty_op,Y,Xs,Ys):-!,mc_input0(Y,Xs,Ys).
mc_empty(I,I,Xs,Xs).

mc_top(Ys,X):-mc_input(X0,Ys,Xs),mc0(Xs,X0,X).

mc0(Xs,X0,X):-op_type(X0,Op,Type),mc1(Op,Type,Xs,X0,X).

mc1(write,variable,Xs,X0,R):-mc_push(Xs,Ys,X0,Y),!,mc5(Ys,Y,R).
mc1(c_chunk,variable,Xs,X0,R):-!,mc_chunk(Xs,Ys,X0,Y),mc5(Ys,Y,R).
mc1(_,void,Xs,X0,R):-!,
 mc_sel(Xs,Ys,1,N),
 mc3(Ys,X0,N,R).
mc1(_,_,Xs,X0,R):-
 mc5(Xs,X0,R).

mc_push(Zs,Xs,X0,Y):-mc_input(End,Zs,Xs),
  op_type(End,push,structure),
  mc_transform_push(X0,End,R),!,Y=R.
mc_push(Xs,Xs,X0,X0).

mc_chunk(Zs,Xs,_,After):-
   mc_input(End,Zs,Ys),
   op_type(End,c_chunk,value),!,
   mc_input(After,Ys,Xs).
mc_chunk(Xs,Xs,X0,X0).

mc_sel(Zs,Ys,N1,N3):-
  mc_input(Y,Zs,Xs),
  op_type(Y,_,void),!,
  N2 is N1+1,
  mc_sel(Xs,Ys,N2,N3).
mc_sel(Xs,Xs,N,N).

mc3(_,X0,N,R):-mc4(N,X0,R).
mc3(Zs,_,_,R):-mc_input(Y,Zs,Ys),mc0(Ys,Y,R).

mc4(N,X0,R):-mc_transform_void(N,X0,R),!.
mc4(1,X0,X0).

mc5(_,R,R).
mc5(Zs,_,R):-mc_input(Y,Zs,Ys),mc0(Ys,Y,R).

op_type(ii(Op,T,_,_),X,Y):-!,X=Op,Y=T.
op_type(Bad,_,_):-errmes(expected(ii/4),Bad).

mc_transform_void(N,ii(Op,void,X,_),ii(Op,void,X,N)).

mc_transform_push(
  ii(write,variable,_,X),ii(push,structure,FN,X),
  ii(push,constant,FN,X)).

encode(unify_variable,get,Ai,To):-        wcode(To,1,Ai,?,0).
encode(write_variable,put,Ai,To):-        wcode(To,2,Ai,?,0).

encode(unify_value,get,Ai,To):-           wcode(To,3,Ai,?,0).
encode(write_value,put,Ai,To):-           wcode(To,4,Ai,?,0).

encode(unify_constant,get,Const,To):-     wcode(To,5,0,Const,0).
encode(write_constant,put,Const,To):-     wcode(To,6,0,Const,0).

encode(push_constant,F/N,Ai,To):-         wcode(To,6,Ai,F,N).

encode(get_constant,Const,Ai,To):-        wcode(To,7,Ai,Const,0).
encode(get_structure,Func/Arity,Ai,To):-  wcode(To,8,Ai,Func,Arity).

encode(put_constant,Const,Ai,To):-        wcode(To,9,Ai,Const,0).
encode(put_structure,Func/Arity,Ai,To):-  wcode(To,10,Ai,Func,Arity).

encode(get_variable,Xn,Ai,To):-           wcode(To,11,Xn,?,Ai). %move_reg

encode(put_value,Xn,Ai,To):-              wcode(To,11,Ai,?,Xn). %move_reg

encode(put_variable,Xn,Ai,To):-           wcode(To,12,Xn,?,Ai).

encode(get_value,Xn,Ai,To):-              wcode(To,13,Xn,?,Ai).

encode(push_cut,?,?,To):-                 wcode(To,14,0,?,0).
encode(put_cut,?,?,To):-                  wcode(To,15,0,?,0).
encode(get_cut,?,?,To):-                  wcode(To,16,0,?,0).

encode('execute_?',Pred,Arity,To):-       wcode(To,17,0,Pred,Arity).
% proceed ==> 18 (by emulator)

encode(load_constant,AReg,Const,To):-     wcode(To,28,AReg,Const,0).

encode(load_variable,AReg,V,To):-         wcode(To,50,AReg,?,V).

encode(load_value,AReg,V,To):-            wcode(To,29,AReg,?,V).

encode(push_structure,Func/Arity,Ai,To):-  wcode(To,51,Ai,Func,Arity).
encode(push_offset,Func/Arity,Ai,To):-  wcode(To,52,Ai,Func,Arity).

encode(unify_void,_,Ai,To):-        wcode(To,61,Ai,?,0).
encode(write_void,_,Ai,To):-        wcode(To,62,Ai,?,0).
encode(c_chunk_variable,Len,Ai,To):-  write_or_skip(To,63,Ai,?,Len).
encode(c_chunk_value,Len,Ai,To):-     write_or_skip(To,64,Ai,?,Len).

encode('clause_?',Pred,N,To):-n_nop(X1), X is X1+1,          
	wcode(To,X,0,Pred,N).
encode('firstarg_?',F/N,MaxRegs,To):-
        n_nop(X1), X is X1+2,
	wcode(To,X,MaxRegs,F,N).

encode('end_?',Reg,Mode,To):-             wcode(To,0,Reg,Mode,0).

encode(inline_variable,No,Arg,To):-n_inline(N),  wcode(To,N,Arg,?,No).

encode(arith_variable,No,Arg,To):-n_arith(N),   wcode(To,N,Arg,0,No).
encode(arith_value,No,Arg,To):-n_arith(N),      wcode(To,N,Arg,1,No).

encode('builtin_?',No,Arg,To):-n_builtin(N),    wcode(To,N,Arg,?,No).

encode_op(P,A,N,Mode):-n_nop(X1),  
  X is X1+3,
  make_cmd([P," ",A," ",N],Name),
  wcode(Mode,X,0,Name,0).

% we put metacall at right to be able to override them
% by reconsulting builtins.pl
n_inline(N):-metacall(bu_ctr(n_inline,N)).
n_arith(N):-metacall(bu_ctr(n_arith,N)).
n_builtin(N):-metacall(bu_ctr(n_builtin,N)).
n_nop(N):-metacall(bu_ctr(n_nop,N)).

bu(Pred,Type,Where):-metacall(bu0(Pred,Type,Where)).

% compiler to C

c_threshold(6).
c_skip(1).

write_or_skip(To, Op,Ai,Nop,Len):-c_threshold(Min),Len>Min,!,
   write_or_skip0(To,Op,Ai,Nop,Len).
write_or_skip(_, _,_, _,_).

write_or_skip0(wam,Op,Ai,Nop,Len):-
   bb_val(c_flag,begin_end,yes),!,
   wcode(wam,Op,Ai,Nop,Len).
write_or_skip0(c,Op,Ai,Nop,_Len):-!,
   wspec(Op,Ai,Nop,0). % ## _Len not used in C.
write_or_skip0(c,Op,Ai,Nop,Len):-Op=64,!,
   wcode(c,Op,Ai,Nop,Len).
write_or_skip0(_,_Op,_Ai,_Nop,_Len).

wcode(mem,Op,Reg,F,N):-add_instr(Op,Reg,F,N).
wcode(wam,Op,Reg,F,N):-
  put_code(Op),put_code(Reg),put_code(N),write(F),put_code(0).
wcode(c,  Op,Reg,F,N):-wcode_c(Op,Reg,F,N). % ## When compiling to C

% related to newpred

newpred(c,C):-!,C=(H:-_),newpred0(H).
newpred(_,_).

% Compiler interface ------------------------------------------------------
compile_mem(File):-
  see(File),compile_mem1(mem,File),seen. % do not change this

% KDB: can probably be replaced by:
% compile_mem(File):-compile_mem1(mem,File).
% Paul Tarau: do not replace, it gets wrong un compiling files which
% cannot be found !!!

% Mode must be mem
compile_mem1(Mode,File):-
	ttyprint(compiling(to(Mode),File,'...')),
	ctime(T1),
	compile_mem2(Mode,File),
	!,
	ctime(T2),T is T2-T1,
	ttyprint(compile_time(T)),
	abort.
compile_mem1(_,_):-
	ttyprint('compilation aborted'),
	restart,
	abort.

/* dead code ?
mcompile(File):-
	seeing(X),
	see(File),
	(compile_mem2(mem,File)->true;true),
	seen,
	see(X).
*/

compile_mem2(Mode,File):-
	mcomp_file(Mode,File),
	terminate_file(mem,'$end1',1).

% modes: wam,c,asm,bin
compile0(Mode,[F|Fs],OutFile):-!,
  init_cmd,
  ctime(T1),
  xcompile(Mode,[F|Fs],OutFile),
  ctime(T2),
  T is T2-T1,
  write(total_compile_time(T)),nl.
compile0(Mode,File,OutFile):-xcompile(Mode,[File],OutFile).

xcompile(Mode,InFiles,OutFile):-
	tell(OutFile),
	decorate_file(header,Mode),
	compile_builtins(Mode),
	member(InFile,InFiles),
	ttyprint(compiling(to(Mode),InFile,'...')),
	ctime(T1),
	comp_file(Mode,InFile),
	ctime(T2),T is T2-T1,
	ttyprint(compile_time(T)),
	fail.
xcompile(Mode,_,_):-
	terminate_file(Mode,'$end0',0),
	decorate_file(footer,Mode),
	fail.
xcompile(_,_,_):-
	told.

decorate_file(Where,c):-!,c_decorate_file(Where).
decorate_file(_,_).

terminate_file(Mode,Dummy,Level):-
	maincomp(Mode,Dummy),
	gg_emit(Mode,[[ii(end,?,Level,Mode)]]),
	close_c_chunk_file(Mode).

close_c_chunk_file(c):-!,write_c_chunk(told).
close_c_chunk_file(_).

gg_emit(mem,C):-gen_code(mem,C).
gg_emit(wam,C):-gen_code(wam,C).
gg_emit(c,C):-gen_code(c,C).
gg_emit(asm,C):-show_code(C).

survive_cleanup(F0,F):-
	name(F0,Survivor),  % Survivor is now a list on the heap
	restart,     % total cleanup of name-spaces, strings, files etc...
	call_ifdef(init_io,true),
	name(F,Survivor),   % F is now a valid name
        init_cmd.

% compiles to memory	
mcomp_file(Mode,InFile):-
	survive_cleanup(InFile,F),
	translate_all(F,Mode).

comp_file(Mode,InFile):-
	member(Mode,[wam,asm,bin,c]),!,
	translate_all(InFile,Mode).

xolder_file(_F,OldF):- \+ exists_file(OldF).

translate_all(F,Mode):-translate_all0(F,Mode).

memoable_file(F):-
  Suf=[_,_,_],
    name(F,L),
    append(_,Suf,L),
    !,
  Suf=".pl".

memoing_translate_all(F,Mode):-
  Mode=wam,which_prolog(binprolog),memoable_file(F),!,
  symcat(F,Mode,OldF),
  (   \+ older_file(OldF,F) -> true
    ; make_memo_file(F,Mode,OldF)
  ),
  include_memoed_file(OldF).
memoing_translate_all(F,Mode):-
  translate_all0(F,Mode).

make_memo_file(F,Mode,MemoF):-
  Mode=wam,
  telling(CF),
    tell(MemoF),
      translate_all0(F,Mode),
    told,
  tell(CF).

include_memoed_file(OldF):-
     seeing(CF),
   		see(OldF),
			repeat,
				get_code(C),
                                put_to_eof(C),
			!,
		seen,
     see(CF).

put_to_eof(-1):-!.
put_to_eof(C):-put_code(C),fail.

translate_all0(F,Mode):-
	seeing(F0),
		see(F),
			repeat,
				read(C),
				translate(C,Mode),
			!,
		seen,
	see(F0).

translate(end_of_file,_):-!.
translate(':-'(C),Mode):-!,translate_cmd(C,Mode),fail.
translate('::-'(H,B),Mode):-!,compile_binary(Mode,(H:-B)),fail.
translate(C,Mode):-is_dynamic_clause(C),!,translate_cmd(assert(C),Mode),fail.
translate(C,Mode):-maincomp(Mode,C),fail.

is_dynamic_clause((H:-_)):-!,is_dynamic(H).
is_dynamic_clause(H):-is_dynamic(H).

maincomp(Mode,C):-
	preprocess(C,M,D,B),
	show_steps(Mode,M,D),
	compile_binary(Mode,B).

translate_cmd([F],Mode):-!,include_file(F,Mode).
translate_cmd(compile(F),Mode):-!,include_file(F,Mode).
translate_cmd(ensure_loaded(F),Mode):-!,include_file(F,Mode).
translate_cmd(consult(F),Mode):-!,include_file(F,Mode).   % can be consult
translate_cmd(reconsult(F),Mode):-!,include_file(F,Mode). % can be consult

translate_cmd(op(X,Y,Z),Mode):-member(Mode,[wam,c]),!,
	op(X,Y,Z), 
	encode_op(X,Y,Z,Mode).
translate_cmd(C,Mode):-
        exec_cmd(C,Mode).

exec_cmd(C,mem):-
  telling(F),
  tell(user),
    (C->true;true),
  told,
  tell(F),
  fail.
exec_cmd(Body,Mode):-member(Mode,[wam,c]),!,
  cmd2clause(Body,C),
  maincomp(Mode,C),
  fail.

% execution of `:-' runtime commands
%
% it expects: 
% - init_gensym in decorate_file(header...
% - exec_run_time_commands/0 in main/1

cmd_root('$run_time_command').

init_cmd:-cmd_root(Root),init_gensym(Root).

cmd2clause(Body,(Head:-Body)):-
  cmd_root(Root),
  gensym(Root,Head).

exec_run_time_commands:-
   cmd_root(Root),
   exec_run_time_commands0(Root),
   fail.
exec_run_time_commands.

exec_run_time_commands0(Root):-
   MaxCmd=10000,
   for(I,1,MaxCmd),
     symcat(Root,I,Cmd),
     (is_compiled(Cmd)->Cmd,fail;true),
   !.

include_file(IFile,Mode):-
	seeing(CF1),
	find_file(IFile,F),
	ttyprint(begin(including(F),in(CF1))),
	memoing_translate_all(F,Mode),
	seeing(CF2),ttyprint(end(including(F),in(CF2))).

% preprocess(Clause,BinClause):-preprocess(Clause,_,_,BinClause).

preprocess(C,M,D,B):-
	std_expand_term(C,E),
	fact2rule(E,M),    
	mdef_to_def(M,D),
	def_to_mbin(D,B).

fact2rule((:-B),(:-B)):-!.
fact2rule((H:-B),(H:-B)):-!.
fact2rule(H,(H:-true)).


% --------------------------------------------------------------------
% Compiler interface

make_cmd(Cs,Cmd):-
  make_cmd0(Cs,Xs),
  name(Cmd,Xs).

make_cmd0(Cs,Xs):-
  findall(X,char_in_cmd(X,Cs),Xs).

char_in_cmd(X,Cs):-member(C,Cs),listify(C,Xs),member(X,Xs).

listify([X|Xs],[X|Xs]):-!.
listify(X,L):-name(X,L).

make_appl(File):-
	compile0(wam,
        [File,'init.pl','oper.pl','lib.pl','read.pl','write.pl','dcg.pl'],
        'newappl.bp').
 
% ?- make_executable_unix_appl('/opt/bin/ru','progs/hello.pl','hello').

make_executable_unix_appl(AbsolutePathToEmulator,SourceF,ExecF):-
	make_appl(SourceF),
	make_cmd([
	"(echo '#! /bin/sh '",
	"; echo 'exec ", 
	AbsolutePathToEmulator, " -q1 $* $0'",
	" ; cat newappl.bp) > ",ExecF,
	" ; chmod +x ", ExecF
	], Cmd),
	system(Cmd).

cmake:-cmake(wam).

cmake(Project):-make(c,Project,'.c').
	
make:-make(wam).

bmake:-bmake('BMARK_bootstrap',wam).

bmake(Mes,Project):-make(wam,Project,'.bp',Mes).

make(Project):-make(wam,Project,'.bp').

make(Mode,Project,Suf):-make(Mode,Project,Suf,Project).

make(Mode,Project,Suf,Mes):-
	ctime(T1),
	find_file(Project,F),make_cmd([Project,Suf],NewF),
        make_include(Mode,Project),
	compile0(Mode,F,NewF),
	ctime(T2),T is T2-T1,
	write(Mes=time(T)),nl.

make_include(c,Project):-!,make_cmd([Project,".h"],F_H),
   let(c_chunk,file_name,F_H).
make_include(_,_).

asm:-comp_file(asm,user).

asm(Project):-make(asm,Project,'.asm','BinWAM intermediate code').

ls:-system('ls -tF').
dir:-system(dir).

t(File):-edit(textedit,File).
e(File):-edit(emacs,File).

edit(Editor,File):-
	find_file(File,F),
        make_cmd([Editor," ",F],R),
        system(R),
	compile_mem(F).

boot:-make.

s:-statistics.

