#ifndef _GLOBAL_H
#define _GLOBAL_H

#ifndef SAFETY
#define SAFETY -1
#endif

#if SAFETY < 0 && TRACE < 1
#define NDEBUG
#endif

/* SIZEs of various DATA AREAS: can be replaced with
variables as they will be dynamically allocated */

#define MAXDCG 256

#ifdef STATIC_TABLES
#define HBITS 13
#define ABITS 12
#define HMAX (1<<HBITS)
#define MAXATOM (1<<ABITS)
#else
#define HMAX (max.DICT)
#define MAXATOM (max.ATOMS)
#endif

#define MAXLEX (8*MAXATOM)

#ifndef TRACE
#define TRACE 0
#endif

#ifndef PROF
#define PROF 0
#endif

/* 
Tracing (debugging) and profiling information:

-----TRACE------
0==>default: debug.c, prof.h are possibly not compiled, ASSERT2 is off
1==>ASSERT2 is on
2==>EXECS AND JUMPS are traced
>=3==>ALL INSTRUCTIONS AND DATA AREAS ARE SHOWN

----PROF-------
0==>default: no profiler, debug.c, prof.h possibly ignored
1==>instructions are counted but not timed
2==>instructions are probabilistically timed (chance 1/13...)
3==>all instructions are timed
*/

#if TRACE > 0
#include <assert.h>
#define ASSERT2(Cond,Term) \
  if(!(Cond)) \
    {(void)local_error((Term),"ASSERT2 failed",wam); \
    assert(0);} 
#else
#define ASSERT2(Cond,Term)
#endif

#include <stdio.h>

typedef char *string;
typedef unsigned char byte;
typedef unsigned long no;
typedef no cell;
typedef cell *term;
typedef cell *instr; /* to circumvent bug in insight ! KDB */

extern string *atomtable;

#define EMPTY

#define VARTAG PUTTAG(0L,0L)
#define FUNTAG PUTTAG(0L,2L)
#define INTTAG PUTTAG(0L,1L)

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

typedef struct limit
{
  long HEAP,CHOICE,TRAIL,CODE,BOARD, DICT, ATOMS,SBUF,QUIET;
}
*limit;

/* RESTRICTED SET OF WAM INSTRUCTIONS
an easy optimisation is to specialize them for lists*/

#define MAXOP 256

#define END 0

#define UNIFY_VARIABLE 1
#define WRITE_VARIABLE 2

#define UNIFY_VALUE 3
#define WRITE_VALUE 4

#define UNIFY_CONSTANT 5
#define WRITE_CONSTANT 6

#define GET_CONSTANT 7
#define GET_STRUCTURE 8

#define PUT_CONSTANT 9
#define PUT_STRUCTURE 10

#define MOVE_REG 11

#define PUT_VARIABLE 12
#define GET_VALUE 13

#define PUSH_CUT 14
#define PUT_CUT 15
#define GET_CUT 16

#define EXECUTE 17
#define PROCEED 18

#define TRY_ME_ELSE 19
#define RETRY_ME_ELSE 20
#define TRUST_ME_ELSE 21

#define TRY_ME_ONLY 22
#define NONDET 23

#define EXEC_SWITCH 24
#define SWITCH 25

#define EXEC_JUMP_IF 26 
#define JUMP_IF 27

#define LOAD_CONSTANT 28
#define LOAD_VALUE 29

#define GET_UNIFY_VAR_VAR 30
#define UNIFY_VAR_VAR 31
#define PUT_WRITE_VAR_VAR 32
#define WRITE_VAR_VAR 33

#define GET_UNIFY_VAL_VAL 34
#define UNIFY_VAL_VAL 35
#define PUT_WRITE_VAL_VAL 36
#define WRITE_VAL_VAL 37

#define GET_UNIFY_VAR_VAL 38
#define UNIFY_VAR_VAL 39
#define PUT_WRITE_VAR_VAL 40
#define WRITE_VAR_VAL 41

#define GET_UNIFY_VAL_VAR 42
#define UNIFY_VAL_VAR 43
#define PUT_WRITE_VAL_VAR 44
#define WRITE_VAL_VAR 45

#define MOVE_REGx2 46
#define LOAD_VALUEx2 47
#define LOAD_VAL_SHORT 48

#define EXEC_TRY 49
#define LOAD_VARIABLE 50

#define PUSH_STRUCTURE 51
#define PUSH_OFFSET 52

#define PUSH_VAR_VAR 53
#define CONSTANT_VAR_VAR 54
#define PUSH_VAL_VAR 55
#define CONSTANT_VAL_VAR 56
#define PUSH_VAL_VAL 57
#define CONSTANT_VAL_VAL 58
#define PUSH_VAR_VAL 59
#define CONSTANT_VAR_VAL 60

#define UNIFY_VOID 61
#define WRITE_VOID 62

#define BEGIN_C_CHUNK 63
#define END_C_CHUNK 64
/* normally the last hardwired instruction: generated by load.c */
#define APPLY 65

#define NOP LAST_BUILTIN+1
#define CLAUSE NOP+1
#define FIRSTARG NOP+2
#define OPERATOR NOP+3

#define CodeStk 0
#define HeapStk 1
#define TrailStk 2
#define ChoiceStk 3
#define BoardStk 4
#define MaxStk 5

typedef struct stack
  {
    long size;
    long over;
    term *top;
    term *oldtop;
    term *base;
    term *margin;
    term *end;
    term *maxused;
    string name;
  } 
  *stack;

#define cbase (wam[CodeStk].base)
#define ctop (wam[CodeStk].top)
#define bak_ctop (wam[CodeStk].oldtop)
#define codend (wam[CodeStk].end)

/* GLOBAL REGISTERS */

#define SAVED_H (*(A-2))
#define SAVED_TR (*(A-1))
#define SAVED_P (*A)

#define TR_TOP wam[TrailStk].top

/* HASING TABLE ENTRY */

typedef struct hentry
{
  no pred;
  no fun;
  no val;
}
*hentry;

/* DATA AREA FORMATS & INSTRUCTION FORMATS */

#define LSCALE (sizeof(long)/sizeof(int))

#define TAGBITS (1+LSCALE)
#define ARITYBITS 8

#define REGBITS 11
#define ARGBITS 11
#define OPBITS (sizeof(no)*8-REGBITS-ARGBITS)

/* if OPBITS > 10 it will be slow on SPARCs */

#define MAXARITY (1L<<ARITYBITS)
#define MAXREG ((1L<<(ARGBITS))-1)

/**********************LOW-LEVEL TERM OPERATIONS*****************/

#define LBITS ARITYBITS
#define RBITS TAGBITS
#define MBITS (sizeof(no)*8-LBITS-RBITS)

#define RMASK (((no)(1L)<<RBITS)-1L)
#define LMASK ((no)~(0L)<<(MBITS+RBITS))
#define MMASK (~LMASK&~RMASK)

#define LGET(W) ((no)(W)>>(MBITS+RBITS)) 
#define LPUT(W,Val) ((no)(W)&~LMASK) | (no)(Val)<<(MBITS+RBITS)
#define LSET(W,Val)     ((no)(W)= LPUT(W,Val))

#define MGET(W) ((no)(W)<<LBITS>>(LBITS+RBITS))
#define MPUT(W,Val) (((no)(W)&~MMASK) |((no)(Val)<<(LBITS+RBITS)>>LBITS))
#define MSET(W,Val) ((no)(W)=MPUT(W,Val))

#define RGET(W) ((no)(W)&RMASK)
#define RPUT(W,Val) (((no)(W) & ~RMASK) | (no)(Val)) 
#define RSET(W,Val) ((no)(W)=RPUT(W,Val)) 


/*******************INTERFACE TERM OPERATIONS ******************/

#define SETTAG(W,Val) RSET(W,Val)
#define PUTTAG(W,Val) RPUT(W,Val)
#define GETTAG(W) RGET(W)

#define SETSYMNO(W,Val) MSET(W,Val)
#define PUTSYMNO(W,Val) MPUT(W,Val)
#define GETSYMNO(W) MGET(W)

/* #define SETARITY(W,Val) LSET(W,Val) */
#define PUTARITY(W,Val) LPUT(W,Val)

/*************** HIGH-LEVEL TERM OPERATIONS************************/

#define HIER Comment("Hier")
#define T2T(t) C2T(T2C(t))

#if SAFETY > 1
#define Comment(C) fprintf(stderr,"%s:%d %s\n",__FILE__,__LINE__,C)
#else
#define Comment(C)
#endif

#if SAFETY <= 0
#define T2C(t) ((cell)(t))
#define C2C(c) (c)
#define C2T(c) ((term)(c))
#else
#include "safety.h"
#endif

#define GETREF(t) (C2C(*(T2T(t))))
#define GETCELL(t) (C2C(*(C2T(t))))
#define GETARITY(W) LGET(C2C(W))

#define SETREF(t,Ref) (*(T2T(t)) = T2C(Ref))
#define SETCELL(t,Cell) (*(T2T(t)) = C2C(Cell))

#define INITREF(t,Ref) SETREF(t,Ref)
#define INITCELL(t,Ref) SETCELL(t,Ref)

#define NONVARREF(t) (NONVAR(GETREF(t)))
#define NAME(s) (atomtable[GETSYMNO(C2C(s))])

#define INPUT_INT(Extval) ((no)((long)(Extval)<<TAGBITS) | INTTAG)
#define OUTPUT_INT(Intval) ((long)(C2C(Intval))>>TAGBITS)

#define NONVAR(ACell) GETTAG(C2C(ACell))
#define VAR(ACell) (!GETTAG(C2C(ACell)))

#define INTEGER(ACell) (GETTAG(C2C(ACell)) == INTTAG)
#define IDENTIFIER(ACell) (GETTAG(C2C(ACell)) == FUNTAG)
#define SYMCONST(ACell) (IDENTIFIER(ACell) && !GETARITY(ACell))
#define COMPOUND(ACell) (IDENTIFIER(ACell) && GETARITY(ACell))
#define SIMPLE(ACell) (!COMPOUND(ACell))
#define BP_FLOAT(Cell) (g.bp_float==(C2C(Cell)))
#define NUMERIC(Cell) (INTEGER(Cell) || BP_FLOAT(Cell))

#define ATOMIC(ACell) \
(NONVAR(ACell) && (INTEGER(ACell) || !GETARITY(ACell)))


/*************** LOW-LEVEL INSTRUCTION OPERATIONS *****************/

#define LCBITS REGBITS
#define MCBITS ARGBITS
#define RCBITS OPBITS

#define RCMASK (((no)(1L)<<RCBITS)-1)
#define LCMASK ((no)~(0L)<<(MCBITS+RCBITS))
#define MCMASK (~LCMASK&~RCMASK)

/* Scale must be 0 or TAGBITS */

#define LCGET(W,Scale) ((no)(W)>>(MCBITS+RCBITS+Scale)) 
#define LCPUT(W,Val) ((no)(W)&~LCMASK) | (no)(Val)<<(MCBITS+RCBITS+TAGBITS)
#define LCSET(W,Val)    ((W)= LCPUT(W,Val))

#define MCGET(W,Scale) ((no)(W)<<(LCBITS)>>(LCBITS+RCBITS+Scale))
#define MCPUT(W,Val) \
  (((no)(W)&~MCMASK)|((no)(Val)<<(LCBITS+RCBITS+TAGBITS)>>LCBITS))
#define MCSET(W,Val) ((W)=MCPUT(W,Val))

#define RCGET(W) ((no)(W)&RCMASK)
#define RCPUT(W,Val) (((no)(W) & ~RCMASK) | (no)(Val)) 
#define RCSET(W,Val) ((W)=RCPUT(W,Val)) 

/*******INTERFACE INSTRUCTION OPERATIONS ********************/

#define SETOPCELL(W,Val) RCSET(W,Val)
#define PUTOPCELL(W,Val) RCPUT(W,Val)
#define GETOPCELL(W) RCGET(W)

#define SETLEFTCELL(W,Val) MCSET(W,Val)
#define PUTLEFTCELL(W,Val) MCPUT(W,Val)
#define GETLEFTCELL(W) MCGET(W,TAGBITS)

#define SETREGCELL(W,Val) LCSET(W,Val)
#define PUTREGCELL(W,Val) LCPUT(W,Val)
#define GETREGCELL(W) LCGET(W,TAGBITS)

/********* HIGH LEVEL INSTRUCTION OPERATIONS*****************/

#define REGFIELD GETREGCELL(fields)
#define LEFTFIELD GETLEFTCELL(fields)
#define OPFIELD GETOPCELL(fields)

#define On REGFIELD
#define Oi LEFTFIELD

#define An (*(term)((no)regs+ LCGET(fields,0) )) /* cell */
#define Ai (*(term)((no)regs+ MCGET(fields,0) )) /* cell */

#define GETFIELDS(Step) fields=((term)P)[Step]

/* cell instr; An, Ai, Op */

#define GETOP(Ip) GETOPCELL(*(instr)(Ip))
#define SETOP(Ip,Val) SETOPCELL(*(term)(Ip),(term)(Val))

#define GETREG(Ip) GETREGCELL(*(instr)(Ip))
#define SETREG(Ip,Val) SETREGCELL(*(term)(Ip),(term)(Val))

#define GETLEFT(Ip) GETLEFTCELL(*(instr)(Ip))
#define SETLEFT(Ip,Val) SETLEFTCELL(*(term)(Ip),(term)(Val))

#define GETFUN(Ip) (*(term)((Ip)+1))
#define SETFUN(Ip,Val) (*(term)(++(Ip)))=(Val)

#define GETLABEL(Ip) (*(instr *)((Ip)+1))
#define SETLABEL(Ip,Val) (*(instr *)((Ip)+1))=(Val)

#define COPY_CELLS(To,From,Arity) \
{ register no k; /* for k=1 to Arity */ \
  for(k=(Arity)<<(TAGBITS); k; k-= 1<<(TAGBITS) ) \
    { ASSERT2(!COMPOUND(GETREF((term)((no)(From)+k))),\
        GETREF((term)((no)(From)+k)));\
      SETCELL((term)((no)(To)+k),GETREF((term)((no)(From)+k)) ); \
    } \
}

#define COPY_ARGS(To,From,Arity)  \
 { COPY_CELLS(To,From,(Arity-1)); \
   SETREF(To+Arity,From+Arity);   \
 }

#define COPY_REGS()\
if(fields)\
{ COPY_CELLS(regs,xref,fields-1);\
  xval=xref[fields];\
  if(COMPOUND(xval)) xval=T2C(xref+fields);\
  regs[fields]=xval;\
}

/* OTHER MACROS */
#define NEWLN() fprintf(g.tellfile,"\n"); fflush(g.tellfile)

#define INSTR_LEN(Instr) instr_len[GETOP(Instr)]
#define ON(stk,Ob) \
  (wam[stk].base<=(term*)(Ob) && (term*)(Ob)<wam[stk].end)

#define MOD(x,y) ((x) & ((y)-1))

#define SYSTIME 0
#define LOADTIME 1
#define RUNTIME 2
#define BBOARDTIME 3
#define VARTIME 4

typedef struct specsyms
{
  byte timestamp;
  cell NIL,DOT,DIF,empty,
  fx, fy, xf, yf, xfx, xfy, yfx,prefixop,infixop,postfixop,
  true,prolog_main,metacall,metatrue,predmark,seemark,tellmark,
#ifdef BLACKBOARD
  bbmark, stdbb,
#endif
  end_of_file,
  opmark,
  user,closed_file,bp_float,
  seefunc,tellfunc;
  string sbuf,stop; /* character buffer */
  long linking;
  cell compare_vals[3];
  string bak_newlex;
  no bak_newatom;
  long bak_opcount;
  no bak_hcount;
  long stime,rtime;
  term connect;
  long dcg;
  long self;
  
  FILE *seefile, *tellfile;
 } *specsyms;

#define NEWVAR(R) SETREF(H,H); (R)=T2C(H++)

#define NEWVAR2(R1,R2) SETREF(H,H); (R1)=(R2)=(cell)(H++)

#define HNEWVAR(N,Val) xref=H+(N); (Val)=T2C(xref); SETREF(xref,xref)
#define HNEWVAL(N,Val) SETCELL((H+N),(Val))
#define HNEXT(N) H += (N)

#define PUSH_TO_REG() \
      xref=C2T(An); \
      SETREF(xref,H)

#define IS_LIST(Term) (g.DOT==(Term))

#define PUSHVAL(Val) SETCELL(H++,(Val))
#define MAKE_LIST() PUSHVAL(g.DOT)

#define PUSH_LIST(Elem) {  \
  MAKE_LIST();             \
  PUSHVAL((Elem));         \
}

#define PUSH_NIL() PUSHVAL(g.NIL)

#define DEREF2(x,v) \
{ while(VAR((v)=GETREF(x)) && (cell)(x)!=(v)) (x)=(term)(v); \
}

#define FDEREF(Expr) \
xref=C2T(Expr); \
if(NONVAR(T2C(xref))) xval=T2C(xref); \
else DEREF2(xref,xval)

/* this gives the maximal arity of a builtin */
#define TEMPARGS 8

#define X(I) regs[-(I)]

#define RX(I) ((term) (X(I)))

#define TEMP(I) regs[-(I)]

#define IN(I,Expr) \
{ ASSERT2(!COMPOUND(Expr),(Expr)); \
  xref=C2T(Expr); \
  if( NONVAR(T2C(xref)) ) X(I)=T2C(xref); \
  else \
  { DEREF2(xref,xval); \
    X(I)=(COMPOUND(xval))?T2C(xref):xval; \
  }\
}

#define TRAIL_IT(V) \
{IF_OVER("trailing in unify",TR_TOP,TrailStk,BYE()); *TR_TOP++=(V);}

#define TRAIL_IF(V) if((V)<SAVED_H) TRAIL_IT(V)

#ifndef NO_VALUE_TRAIL
#define VTRAIL_IT(V,Val) \
{IF_OVER("value trailing",TR_TOP,TrailStk,BYE()); \
TR_TOP[0]=(term)(Val); TR_TOP[1]= (void *)(1+(cell)(V)); TR_TOP+=2;}
#define VTRAIL_IF(V,Val) if((V)<SAVED_H) VTRAIL_IT(V,Val)
#define FUSSY_VAR_CHECK(Xval,Xref) \
  if(Xval==(cell)Xref) \
    LOCAL_ERR(Xval, \
      "Destructive assignment on unbound variables is unsafe")

#define SMART_VTRAIL_IF(Xref,Xval) \
  /* FUSSY_VAR_CHECK(Xval,Xref); */ \
  if((Xref<SAVED_H) && (NONVAR(Xval) || Xval<(cell)SAVED_H)) \
    VTRAIL_IT(Xref,Xval)
#endif

#define DEFPRED(Name,Val) hdef(g.predmark,(Name),(Val),g.timestamp)
#define SETPRED(Name,Val) hset(g.predmark,(Name),(Val))
#define GETPRED(Name) (instr)hget(g.predmark,(Name))

#define ATOMIZE(Fun) if(!ATOMIC((Fun))) (Fun)=GETREF((term)(Fun))

#define ABORTING() goto abort_0

#define ABORT(Mess) {can_recover(Mess); ABORTING();}

#define RECOVER() link_halt(); ABORTING();

#define NO() return 0

#define BYE() bp_halt(1)

#define OVER(Culprit,Top,LimitNo,Action) \
{ overflow_by((Top),LimitNo,wam,(Culprit)); Action; \
}

#define IF_OVER(Culprit,Top,LimitNo,Action) \
if(((term *)(Top)) >= wam[LimitNo].margin) OVER(Culprit,Top,LimitNo,Action)

#define WARNING(errname) {warnmes(errname); FAILURE()}

#define ERREXIT(Mes) {warnmes(Mes); return 0;}

#define WARFUN(Sym,Msg) {local_error((Sym),(Msg),wam); FAILURE()}

#define LOCAL_ERR(Sym,Msg) local_error((Sym),(Msg),wam)

#endif
