% which_prolog(binprolog). =>moved to read.pl and ../cparser/stdio.pl

% -----------------------------------------------------------------
% LIBRARY of basic predicates

init_gensym(Root):-let(gensym,Root,0).

gensym(Root,Symbol):-gensym_no(Root,N),symcat(Root,N,Symbol).

gensym_no(Root,N1):-
        val(gensym,Root,N),!,
        N1 is N+1,
        set(gensym,Root,N1).
gensym_no(Root,N):-N=1,
        def(gensym,Root,N).

current_op(Pri,Assoc,Name):-nonvar(Name),!,
  get_op(Name,Assoc,Pri).
current_op(Pri,Assoc,Name):-
  for(I,0,10000),
  (val(opmark,I,Name)->get_op(Name,Assoc,Pri);!,fail).

op(Pri,Assoc,Name):-op0(Name,Assoc,Pri).

get_op(Name,Assoc,Pri):- % nonvar(Name),
	op_type(Assoc,Cls),
	get_op0(Name,Cls,Assoc,Pri).

get_op0(Name,Cls,Assoc,Pri):- % nonvar(Name), nonvar(Cls),
	val(Cls,Name,Assoc),
	val(Name,Cls,Pri),
	Pri>0.

op_type(xfy,infixop).
op_type(xfx,infixop).
op_type(yfx,infixop).

op_type(fx,prefixop).
op_type(fy,prefixop).

op_type(xf,postfixop).
op_type(yf,postfixop).

is_dynamic(P):-val(P,'$first',_).
is_assumed(H):-val('$assumed',H,Cs),nonvar(Cs).

is_builtin(T):-nonvar(T),!,termcat(T,cont(_),TC),bu(TC,_,_).
is_builtin(T):-var(T),bu(TC,_,_),strip_cont(TC,T,_).

is_interactive:-current_op(1200,fx,('?-')).

interactive(yes):-op(1200,fx,('?-')).
interactive(no):-op(1199,fx,('?-')).

expand_term(C,E):-portable_expand_term(C,E).
std_expand_term(C,D):-portable_expand_term(C,D).

/*
ustrip_cont(TC,T,C):-TC=..LC,append(L,[C],LC),!,T=..L.
strip_cont(TC,T,C):- ustrip_cont(TC,T,C).
*/

strip_cont(TC,T,C):- strip_cont0(TC,C-T).
% NASTY BUG IN BUILTIN strip_cont0... => SOLVED !!!

metatrue(TC)::-strip_cont(TC,T,C,metacall(T,C)).

metacall(G):-
  call_ifdef(meta_interpreter(G),
    call_ifdef(G,
      user_error('no meta_interpreter for undefined goal',G)
  )).

% true.
% call(X):-X.

call_ifdef(G,_):-is_compiled(G),!,G.
call_ifdef(_,Else):-Else.

X=X.

A->B :- A,!,B.

A->B ; C :- !,if(A,B,C).
X ; _ :-X.
_ ; Y :-Y.

or(A,_):-A.
or(_,B):-B.

if(A,B,_):-A,!,B.
if(_,_,C):-C.

(X,Y):-X,Y.

\+(X):-X,!,fail.
\+(_).

halt:-halt(0).

repeat.
repeat:-repeat.

findall_workhorse(X,G,_):-
	lift_heap('.',0), % cons,answer_size
	G,
	findall_store_heap(X).
findall_workhorse(_,_,Xs):-
	findall_load_heap(Xs).	

findall(X,G,Xs,End):-findall_workhorse(X,G,[End|Xs]).

findall(X,G,Xs):-findall_workhorse(X,G,[[]|Xs]).

% findall(G,Gs):-findall(G,G,Gs).

gc_read(R):-findall(X,read(X),[R]).

gc_call(G):-findall(G,G,Gs),member(G,Gs).

for(Min,Min,Max):-Min=<Max.
for(I,Min,Max):-
        Min<Max,
        Min1 is Min+1,
        for(I,Min1,Max).

numbervars('$VAR'(N0), N0, N) :- !, N is N0+1.
numbervars(X, N0, N) :- atomic(X), !, N0=N.
numbervars([X|Xs], N0, N) :- !,
        numbervars(X, N0, N1),
        numbervars(Xs, N1, N).
numbervars(X, N0, N) :-
        functor(X, _, A),
        numbervars(0, A, X, N0, N).

numbervars(A, A, _, N0, N) :- !, N0=N.
numbervars(A0, A, X, N0, N) :-
        A1 is A0+1,
        arg(A1, X, X1),
        numbervars(X1, N0, N1),
        numbervars(A1, A, X, N1, N).

ground(T):-numbervars(T,0,N),N>0,!,fail.
ground(_).

copy_term(T,CT):-copy_term(0,T,CT).
save_term(T,CT):-copy_term(1,T,CT).

flush:-seen_told(2).

see_or_fail(File):-see_tell(0,File).

exists_file(File):-see_or_fail(File),seen.

sread(I,O):-string_op(0,I,O).
swrite(I,O):-string_op(1,I,O).

system([]):-!.
system([X|Xs]):-!,name(Command,[X|Xs]),shell(Command).
system(Command):-atomic(Command),shell(Command).

[File]:-compile(File).

compile(File):-
	find_file(File,F),
	compile_mem(F),fail.

file_search_path(["","progs/","myprogs/","../src/"]).

file_extension_list([".pl",".pro",""]).

file_library(["../library/"],[".pl"]).

make_file_name(Prefs,File,Sufs,NewFile):-
        member(Suf,Sufs),
        member(Pref,Prefs),
        name(File,L),
        det_append(L,Suf,R),
        det_append(Pref,R,Fname),
        name(NewFile,Fname).

statistics(Area,[Used,Free]):-stat_dict(Area,No),stat0(Used,Free,No).

statistics:-
	statistics(Name,Data),
	cwrite(Name),cwrite(=),cwrite(Data),nl,
	fail.
statistics.	

atom(X):-integer(X),!,fail.
atom(X):-atomic(X).

float(X):-nonvar(X),X='$float'(_,_,_).

number(X):-integer(X),!.
number(X):-float(X).

compound(X):-nonvar(X), \+(atomic(X)).

=..(T,L):-var(T),!,list2term(L,T).
=..(T,L):-term2list(T,('.'),[],L).

term2list(T,L):-term2list(T,('.'),[],L).

term2dlist(T,Head,Queue):-term2list(T,('.'),Queue,Head).

% det_append0 works with an arbitrary constructor
% it guesses it from the first constructor in Xs

det_append(Xs,Ys,Zs):-det_append0(Xs,Ys,_-Zs).

det_append(Xs,Ys,Zs,End):-det_append0(Xs,Ys,End-Zs).

append_body(C1,C2,C3,True):-
  functor(C1,F,2),
  det_append0(C1,T,End-C3),
%  ( End==True->C2=T ;
     T=..[F,End,C2]
%  )
   .

% this makes about 2 Megalips on a Sparc 10-20:
append(Xs,Ys,Zs):-
  det_append0(Xs,NewXs,EndXs-Zs),
  general_append(EndXs,Ys,NewXs).

general_append([],Ys,Ys).
general_append([A|Xs],Ys,[A|Zs]):-
	general_append(Xs,Ys,Zs).

length(L,N):-var(N),!,get_length(L,0,N).
length(L,N):-make_length(L,0,N).

get_length([],I,I).
get_length([_|L],I0,I):-I1 is I0+1,get_length(L,I1,I).

make_length([],I,I):-!.
make_length([_|L],I0,I):-I0<I,I1 is I0+1,make_length(L,I1,I).

tab(Expr) :- expr(Expr,N),integer(N),!,tab0(N).
tab(Bad) :- user_error('in tab/1: should evaluate to an integer', Bad).

tab0(N):-for(_,1,N),put_code(32),fail.
tab0(_).

get(R):-repeat,get_code(X),(X>32;X<0),!,R=X.

% !! use something like Y=3+4, X is 1+expr(Y) instead of
% !! Y=3+4, X is 1+Y which will not work in compiled code

X is E:-expr(E,R),!,X=R.
X is E:-errmes(error_in_is,X is E).

expr(E,R):-atomic(E),!,R=E.
expr(E,R):-float(E),!,R=E.
expr(E,_Error):-var(E),!,user_error(variable_in_is,E).
expr(E,R):-functor(E,Op,2),!,
	arg(1,E,E1),arg(2,E,E2),
	expr(E1,X1),
	expr(E2,X2),
	G=..[Op,X1,X2,R],
	G,!.
expr(E,R):-functor(E,Op,1),
	arg(1,E,E1),
	expr(E1,X1),
	G=..[Op,X1,R],
	G,!.

compare(R,X,Y):-compare0(X,Y,R).

+(X,X).
-(X,N):- -(0,X,N).
\(X,N):- \(0,X,N).

A==B :- compare0(A,B,=).
A\==B :- compare0(A,B,R),'$noteq'(R).

A @< B :- compare0(A,B,<).
A @> B :- compare0(A,B,>).
A @=< B :- compare0(A,B,R),'$lesseq'(R).
A @>= B :- compare0(A,B,R),'$gteq'(R).

A < B :-expr(A,X), expr(B,Y), less(X,Y).

A > B :-expr(A,X), expr(B,Y), greater(X,Y).

A =< B :-expr(A,X), expr(B,Y), less_eq(X,Y).

A >= B :-expr(A,X), expr(B,Y), greater_eq(X,Y).

A =:= B :-expr(A,X), expr(B,Y), arith_eq(X,Y).

A =\= B :-expr(A,X), expr(B,Y), arith_dif(X,Y).

'$lesseq'(<).
'$lesseq'(=).

'$gteq'(>).
'$gteq'(=).

'$noteq'(<).
'$noteq'(>).

% BinProlog USER LEVEL naming primitives
% DO NOT rely on def/3, set/3, val/3,rm/2. Thay are unsafe
% and their implementation may change in the future...

/********* BASIC operations: safe with respect to bb_gc ***********/

% WARNING: may be moved to C, do not use their componnents directly

bb_def(A,B,X):-saved(X,S),def(A,B,S).
bb_set(A,B,X):-saved(X,S),set(A,B,S).
bb_val(A,B,CX):-val(A,B,X),copy_term(X,CX).
bb_rm(A,B):-rm(A,B).

/******************DERIVED OPERATIONS ***********************/

% safe bboard operations
bb_let(A,B,X):-val(A,B,_),!,bb_set(A,B,X).
bb_let(A,B,X):-bb_def(A,B,X).

bb_def(A,X):-bb_def(A,A,X).
bb_set(A,X):-bb_set(A,A,X).
bb_val(A,X):-bb_val(A,A,X).
bb_rm(A):-bb_rm(A,A).
bb_let(A,X):-bb_let(A,A,X).

saved(X,S):-save_term(X,NewX),!,S=NewX.
saved(X,S):-bb_gc,save_term(X,NewX),!,S=NewX.
saved(_,_):-bb_fail(saved/2).
saved(_,_):-
	restart,
	abort.

% unsupported & unsafe

let(X,Y,V):-def(X,Y,V),!.
let(X,Y,V):-set(X,Y,V).

def(A,X):-def(A,A,X).
set(A,X):-set(A,A,X).
val(A,X):-val(A,A,X).
let(A,X):-let(A,A,X).
rm(A):-rm(A,A).

% INVISIBLE DCG connect operation: normally macro-expanded

'#'(Word):-dcg_connect(Word).

% example: dcg_phrase(1,(#a,#b,#c),X).
dcg_phrase(DcgStream,Axiom,Phrase):-
  dcg_telling(X),dcg_tell(DcgStream),
    dcg_def(Phrase),
      Axiom,
    dcg_val([]),
  dcg_tell(X).

% GLOBAL LOGICAL VARIABLES

':=:'(K1#K2,X):-!,lval(K1,K2,X).
':=:'(K,X):-lval(K,K,X).

/* bboard garbage collector */

bb_gc:-	bb_list0(3,1,B),
	copy_term(B,NewB),
	bb_reset,
	bb_put_back(NewB).

bb_put_back([]).
bb_put_back([P,N,F,K,V|Xs]):-
        functor(O,P,N),
        functor(M,F,K),
	bb_put_1(O,M,V),
	bb_put_back(Xs).

bb_put_1(O,M,V):-save_term(V,NewV),set(O,M,NewV),!,NewV=V.
bb_put_1(O,M,_):-bb_fail(O+M).

bb_fail(X):-
	statistics(bboard,[_,Z]),
	user_error('blackboard overflow, left only'(Z),culprit(X)).

% An easy patch to read.pl for basic float input - Paul Tarau, Oct. 1993

try_float(_Ds,[atom('.'),integer(box(_Fraq,FDs))|Ts],I,[atom(Float)|Ts1]):-!,
	try_float_exp(Ts,Ts1,Exp),name(NFraq,FDs),
	input_float(I,NFraq,Exp,Float).
try_float(Ds,Tokens,I,[integer(box(I,Ds))|Tokens]).

try_float_exp([atom(e)|Ts],Ts1,Exp):-try_float_exp1(Ts,Ts1,Exp),!.
try_float_exp([atom(EN)|Ts],Ts,Exp):- [E]="e",
	name(EN,[E|NL]),name(Exp,NL),integer(Exp),!.
try_float_exp(Ts,Ts,0).

try_float_exp1([atom('-'),integer(box(Exp,_))|Ts],Ts,Exp1):-!,
	Exp1 is 0-Exp.
try_float_exp1([atom('+'),integer(box(Exp,_))|Ts],Ts,Exp).

float_minus(F,NegF):-NegF is 0.0-F.

pow(A,X,R):-float_fun2(pow,A,X,R).
log(A,X,R):-float_fun2(log,A,X,R).

sqrt(X,R):-float_fun2(pow,X,0.5,R).

% F(X+Y) --> R used for Y=0
exp(X,R):-float_fun(exp,X,R).
log(X,R):-float_fun(log,X,R).
sin(X,R):-float_fun(sin,X,R). 
cos(X,R):-float_fun(cos,X,R).
tan(X,R):-float_fun(tan,X,R).

% F(X)+Y --> R used for Y=0
atan(X,R):-float_fun('T',X,R).
asin(X,R):-float_fun('S',X,R).
acos(X,R):-float_fun('C',X,R).

integer(X,R):-float_fun(integer,X,R).
float(X,R):-R is 0.0+X.

% for portability...
ctime(T):-statistics(runtime,[T,_]).

is_predicate(H):-is_compiled(H),!.
is_predicate(H):-is_dynamic(H),!.
is_predicate(H):-is_assumed(H).

predicate_property(H,Prop):-nonvar(H),!,predicate_property0(H,Prop).
predicate_property(H,Prop):-
  generate_a_predicate(F,N),
  functor(H,F,N),
  predicate_property0(H,Prop).

predicate_property0(H,Prop):-is_dynamic(H),!,
  (Prop=(dynamic);Prop=(interpreted)).
predicate_property0(H,Prop):-is_assumed(H),!,
  (Prop=(assumed);Prop=(interpreted)).
predicate_property0(H,(built_in)):-is_builtin(H),!.
predicate_property0(H,(compiled)):-is_compiled(H).

current_predicate(F/N):-nonvar(F),nonvar(N),!,
   functor(H,F,N),is_predicate(H),!.
current_predicate(F/N):-
  generate_a_predicate(F,N).

predicate_key0('$assumed'/0,PN,V,PN):-nonvar(V).
predicate_key0(PN,'$first'/0,_V,PN).
predicate_key0(predmark/0,P/N,predmark,P/N1):-N1 is N-1.

generate_a_predicate(P,N):-
	bb_list0(0,0,Xs),
 	bb_element(K1+K2=V,Xs),
        predicate_key0(K1,K2,V,P/N).
