#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# uam_aboutIdeas
#
# This procedure displays a window with information about IDEAS.
#
#
proc uam_aboutIdeas {} {
    global tk_priv

    # 1. Create the top-level window and divide it into top
    # and bottom parts.
    
    set w .dialog
    catch {destroy $w}
    toplevel $w -class Dialog
    wm title $w "About IDEAS"
    wm iconname $w Dialog
    wm minsize $w 460 565
    wm maxsize $w 460 565
    frame $w.top 
    pack $w.top -side top -fill both -padx 5m -pady 5m
    frame $w.bot
    pack $w.bot -side bottom -fill both -expand yes

    # 2. Fill the top part with message.

    message $w.msg1 \
       -text "IDEAS\n\nInteractive Development Environment\nfor Agent\
              Systems\n\nVersion 1.0a, 1995\n\nComputer Science\
              Department\nUniversity of Kiel, Kiel, Germany\n" \
       -font -Adobe-Times-Medium-R-Normal-*-180-* \
       -justify center
    pack $w.msg1 -in $w.top -side top -expand 1 -fill x

    message $w.msg2 \
       -font -Adobe-Times-Medium-R-Normal-*-140-* \
       -justify left
    pack $w.msg2 -in $w.top -ipadx 5m -expand 1 -fill both

    # 3. Create a button at the bottom of the dialog.
	button $w.button1 -text OK \
			   -command "set tk_priv(button) 1; destroy $w" \
			   -width 10
	    frame $w.default -relief sunken -bd 1
	    raise $w.button1 $w.default
	    pack $w.default -in $w.bot -side left -padx 15m -pady 2m
	    pack $w.button1 -in $w.default -padx 1m -pady 1m \
		    -ipadx 2m -ipady 1m
	    bind $w <Return> "$w.button1 flash
			      $w.button1 invoke"
            bind $w <Escape> "$w.button1 flash
			      $w.button1 invoke"

	button $w.button2 -text Copyright \
			   -width 10
	    pack $w.button2 -padx 15m -pady 1m \
		    -ipadx 2m -ipady 1m -in $w.bot -side right

    # 4. Withdraw the window, then update all the geometry information
    # so we know how big it wants to be, then place the window relative
    # to its parent and de-iconify it.

    uam_showAbout

    wm withdraw $w
    update idletasks
    set dad [winfo parent $w]

    if {([set dadX [winfo x $dad]] == 0) || ([set dadY [winfo y $dad]] == 0)} {
      set x 300
      set y 200
    } else {
      set x [expr $dadX + ([winfo reqwidth $dad] - [winfo reqwidth $w])/2]
      set y [expr round($dadY + [winfo reqheight $dad]*0.3)]
    }

    wm geom $w +$x+$y
    wm deiconify $w

    # 5. Set a grab and claim the focus too.

    catch {grab $w}
    set oldFocus [focus]
    focus $w

    # 6. Wait for the user to respond, then restore the focus and
    # return the index of the selected button.

    set tk_priv(button) -1
    tkwait window $w
    grab release $w
    focus $oldFocus
    update idletasks
}


proc uam_showCopyright {} {
 .dialog.msg2 configure \
   -text "Copyright (c) 1995 by\n\tMatthias Klusch, Olaf Scheew, Bernd\
          Grossmann\n\tComputer Science Department,\n\tChristian-Albrechts-University\
          of Kiel,\n\tOlshausenstr. 40, 24118 Kiel, Germany\n\n\nAll rights\
          reserved.\nNo warranties will be given on any issues arising out\
          of the use\nof this software product.\nPermission to use, copy, modify\
          and distribute this software\nproduct for non-commercial purposes is\
          hereby granted, provided that the above copyright notice appears in\
          all copies and respective publications.\nAll commercial trades with\
          this product outside the CAU Kiel\nwithout specific written prior\
          permission are prohibited.\n"

 .dialog.button2 configure -text Back \
   -command "set tk_priv(button) 2; uam_showAbout"
}


proc uam_showAbout {} {
 .dialog.msg2 configure \
       -text "This tool has been developed within the FCSI-Project:\n'A\
              federative agent system for a recognition of interdatabase\
              dependencies' by Matthias Klusch.\n\nIt was implemented\
              using Tcl/Tk and BinProlog by Olaf Scheew.\nThe Agent\
              Specification Language ASL was designed by Bernd\
              Grossmann.\n\nPlease take any further requests to\n\n\
              \tDipl.-Inform. M. Klusch\n\
              \tComputer Science Dept.\ 
              \tCAU Kiel\n\
              \tOlshausenstr. 40, 24118 Kiel\n\
              \tGermany\n\n\
              \tE-Mail: mkl@informatik.uni-kiel.d400.de" 

 .dialog.button2 configure -text Copyright \
   -command "set tk_priv(button) 2; uam_showCopyright"
}
