#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# uam_compileActSrc
#
#  Compile the Action-C-Source file befor starting an agent,
#  if necessary.
#
# Arguments:
#
#	fileName - name of C-source file
#
# Results:
#
#  Shows a message in UAM logfile, generates an error,
#  if compilation failed.
#
proc uam_compileActSrc {fileName compileAdds} {
global ideas_Compiler ideas_ExtraCompilerOptions \
       ideas_TclIncludes ideas_X11Includes \
       ideas_TclLibraries ideas_CompLibraries

 # the current directory is set in calling procedure to
 #  $ideas_HomePath/sources
 #

 # if objectfile for 'agt_wish' does not exist, built it by compilation
 # of 'agt_wish.c' :
 #
 if {![file exists ../agt_wish.o]} {
   set CC "$ideas_Compiler ../agt_wish.c \
           -o ../agt_wish.o -c \
           -I$ideas_TclIncludes -I$ideas_X11Includes"

   # start the compilation of 'agt_wish.c':
   #
   uam_logText "Compiling agent base ...\n$CC" 

   # change the cursor to watch while compiling
   #
   foreach widget [winfo children .] {
     catch {$widget configure -cursor watch}
   }
   .log.text  configure -cursor watch

   update

   if {[catch {eval "exec $CC"} errorMesg]} {
     uam_logText "$errorMesg\nCompilation failed!\n\nWARNING:\
                  This is a basic error. You should check the entries in file\
                  'init.tcl'. If they are correct reinstall the files\
                  'agt_wish.c' in IDEAS home directory to\
                   make IDEAS working !\n"

     # reset the cursor
     # 
     foreach widget [winfo children .] {
       catch {$widget configure -cursor {}}
     }
     .log.text  configure -cursor {}
      .mbar  configure -cursor {}

     update

    return -code error
   } else {
     uam_logText "...ready.\n"
   }

 }

 # and now the Action-C-Source:

 set fileRootName [file rootname $fileName]

 set CC "$ideas_Compiler $fileName ../agt_wish.o \
         $compileAdds -o $fileRootName \
         -I$ideas_TclIncludes -I$ideas_X11Includes \
         -L$ideas_TclLibraries $ideas_CompLibraries \
         $ideas_ExtraCompilerOptions"

 # check if its necessary to compile the source, i. e. if the source is
 # change after the last compilation, if there was one, or the agents base
 # 'agt_wish' is compiled new:
 #
 if {[file exists $fileRootName] && \
     !([file mtime ../agt_wish.o] > [file mtime $fileRootName])} {
   if {[file mtime $fileRootName] > [file mtime $fileName]}  {
     uam_logText "No change in Action-C-Source '$fileName', not compiling.\n"
     return
   }
 }

 # start the compilation of Action-C-Source:
 #
 uam_logText "Compiling Action-C-Source '$fileName' ...\n$CC" 

 # change the cursor to watch during compilation
 #   
 foreach widget [winfo children .] {
   catch {$widget configure -cursor watch}
 }
 .log.text  configure -cursor watch

 update

 if {[set errorNo [catch {eval "exec $CC"} errorMesg]]} {
   uam_logText "$errorMesg\nCompilation failed!\n"
   set error 1
 } else {
   uam_logText "...ready.\n"
   set error 0
 }

 # reset the cursor
 # 
 foreach widget [winfo children .] {
   catch {$widget configure -cursor {}}
 }
 .log.text  configure -cursor {}
  .mbar  configure -cursor {}

 update

 if {$error} {
   return -code error
 }

 return
}

