#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# uam_edActSrcFile
#
#  This script generates a file browser, which lists the files of
#  type '.c' in the 'sources' directory, allows user to choose
#  one by typing, selecting or double-clicking and opens an editor to
#  edit it.
#
# Arguments:  None.
#
# Results:  None.
#
proc uam_edActSrcFile {} {
global ideas_HomePath

 catch {destroy .fl}
 toplevel .fl
 wm title .fl "Select Action-C-Source For Editing"
 wm iconname .fl "ActSrcSelect"
 wm withdraw .fl

 # Create two frames for managing the main window

 frame  .fl.frame1 -bd 7
 frame  .fl.frame2 -bd 7
 pack  .fl.frame1  .fl.frame2 -side left -fill y


 # Create a label in frame1 and a listbox with a scrollbar on the
 # right side

 label  .fl.frame1.lbl -text "C-Sources in directory:"
 pack  .fl.frame1.lbl -side top -anchor w

 scrollbar  .fl.frame1.scroll -command ".fl.frame1.list yview"
 pack  .fl.frame1.scroll -side right -fill y
 listbox  .fl.frame1.list -yscroll ".fl.frame1.scroll set" \
	-relief raised \
	-exportselection no

 pack  .fl.frame1.list -side left -fill both -expand no


 # Create in frame2 an entry for selected filename and
 # buttons for OK and CANCEL

 label  .fl.frame2.lbl -text "selected file:"
 entry  .fl.frame2.inpt -relief sunken
 pack  .fl.frame2.lbl  .fl.frame2.inpt -anchor w

button .fl.ok -text OK \
	-width 15 \
	-command "uam_edActSrcFileOK"

 frame .fl.default -relief sunken -bd 1
 raise .fl.ok .fl.default

 button .fl.cancel -text CANCEL \
	-width 15 \
	-command "destroy .fl; focus .mbar"
 pack .fl.cancel -in .fl.frame2 -side bottom \
	-pady 1 -ipadx 2m -ipady 1m 

 pack  .fl.default -in .fl.frame2 \
	-side bottom  \
	-pady 4m
 pack .fl.ok -in .fl.default \
	-padx 1m -pady 1m \
	-ipadx 2m -ipady 1m


# Set up bindings for the browser

 bindtags .fl.frame1.list {Listbox .fl.frame1.list}

 bind .fl.frame1.list <ButtonRelease-1> {
	uam_insertLbSel .fl.frame1.list .fl.frame2.inpt }

 bind .fl.frame1.list <Double-Button-1> {
	uam_insertLbSel .fl.frame1.list .fl.frame2.inpt 
	.fl.ok flash
	.fl.ok invoke}

 bind .fl.frame1.list <Tab> {focus .fl.frame2.inpt}
 bind .fl.frame1.list <Up>  {
	uam_insertLbSel .fl.frame1.list .fl.frame2.inpt }

 bind .fl.frame1.list <Down> {
	uam_insertLbSel .fl.frame1.list .fl.frame2.inpt }

 bind .fl.frame1.list <Return> {
	if {[set sel [.fl.frame1.list curselection]]!=""} {
	  uam_insertLbSel .fl.frame1.list .fl.frame2.inpt
	}
	.fl.ok flash
	.fl.ok invoke}

 bind .fl.frame1.list  <Escape> {.fl.cancel flash
	.fl.cancel invoke}

 bind .fl.frame2.inpt <Return> {.fl.ok flash
	.fl.ok invoke}

 bind .fl.frame2.inpt <Tab> {focus .fl.frame1.list}
 bind .fl.frame2.inpt <Escape> {.fl.cancel flash
	.fl.cancel invoke}


 bind .fl.frame1.list <Help> {uam_initHelp editACShlp}
 bind .fl.frame1.list <F1>   {uam_initHelp editACS.hlp}
 bind .fl.frame1.list <Control-q> {uam_exit}

 bind .fl.frame2.inpt <Help> {uam_initHelp editACS.hlp}
 bind .fl.frame2.inpt <F1>   {uam_initHelp editACS.hlp}
 bind .fl.frame2.inpt <Control-q> {uam_exit}

 # fill the listbox with files with extension '.c' in directory 'sources':
 uam_fillLb .fl.frame1.list "$ideas_HomePath/sources/*.c"

 uam_selectLbEntry .fl.frame1.list 0
 uam_insertLbSel .fl.frame1.list .fl.frame2.inpt 

 update idletasks

 # place the window on its parent window:
 set dad [winfo parent .fl]
 set x [expr [winfo x $dad] + \
   ([winfo reqwidth $dad] - [winfo reqwidth .fl])/2]
 set y [expr round([winfo y $dad] + [winfo reqheight $dad]*0.3)]
 wm geom .fl +$x+$y
 wm deiconify .fl

 focus .fl.frame2.inpt

 tkwait window .fl
}



# On OK do this:

proc uam_edActSrcFileOK {} {
global uam_EditProcs uam_Editor

  set file [.fl.frame2.inpt get]
  if {[lsearch [glob sources/*.c] sources/$file]>=0} {
   destroy .fl
   focus .mbar
   uam_startEditor sources/$file
  } else {
    uam_notExActSrc $file
    .fl.frame2.inpt delete 0 end
  }
}


