#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# uam_evalMesg
#
#  Tries to match a received message with UAM mesg-rules
#  and to evaluate a matching rule.
#
# Arguments:
#
#	fid	-	file descriptor mesg arrived on
#	mesg	-	mesg just received
#
# Results:
#  
#  Side effects belonging to evaluation of the mesg-rules.
#
proc uam_evalMesg {fid mesg} {
global uam_MesgRules


 # try to format received mesg, show it unformatted if this fails
 if {[catch {set text [uam_formatMesg $mesg]}]} {
  puts "Received without format: $mesg"
  return
 }

 # display received message
 uam_logText "Message received\n$text"

 set ruleMatch 0

 #run through all MesgRules
 for_array_keys ruleNo uam_MesgRules {
   # initialize the evaluaition break variable with NO
   set uam_VarBreakEval 0

   # get actual Rule
   set actRule $uam_MesgRules($ruleNo)

   # get the pattern for the mesgForm from actRule
   set mesgForm [lindex $actRule 0]

   # test if listlengths of form is less than that of mesg,
   # if not take next MesgRule
   if {[llength $mesgForm] > [llength $mesg]} {
     continue
   }

   # get the keylist of received mesg
   set mesgKeys [keylkeys mesg]

   # get the keylist of mesgForm
   set keylist [keylkeys mesgForm]

   # test if keylists of form is include in that of  mesg,
   # if not take next MesgRule
   set keyError 0

   foreach mesgFormKey $keylist {
     if {[lsearch $mesgKeys $mesgFormKey] == -1} {
       set agt_keyError 1
       break
     }
   }

   if {$keyError} {continue}

   # try to match the rule with the message:
   set matchError 0

   # run through all mesg keys
   foreach key $keylist {
     keylget mesgForm $key pattern
     keylget mesg $key match

     # try to match the actual key of mesg with the pattern
     # in mesgForm
     if {![uam_keyPattMatches $pattern $match]} {
       set matchError 1
       break
     }
   }

   # if the matching failed take next MesgRule
   if {$matchError} {
     continue
   }

   set ruleMatch 1

   # get the mesgCond of the actRule and eval it
   set mesgCond [lindex $actRule 1]

   if {[expr $mesgCond]} {
     set index 2
   } else {
     set index 3
   }
 

   # depending on mesgCond choose the action of rule
   # and eval it
   set action [lindex $actRule $index] 


   if {[catch {eval $action} errorInfo]} {
     uam_logText "Error when evaluating action: $action"
     if [info exists errorInfo] {
       uam_logText "errorInfo: $errorInfo"
     }
   }
   # if 'uam_break' is called in the action break the evaluation:
   #
   if {$uam_VarBreakEval} {
     break
   }

    # eval the standard action of the rule which is
   # done independent of mesgCond
   set stdAction [lindex $actRule 4]

   if {[catch {eval $stdAction} errorInfo]} {
     uam_logText "Error when evaluating stdAction: $stdAction"
     if [info exists errorInfo] {
       uam_logText "errorInfo: $errorInfo"
     }
   }

   # if 'uam_break' is called in the action break the evaluation:
   #
   if {$uam_VarBreakEval} {
     break
   }

 }

}


#
# uam_break
#
#  Sets global variable 'uam_VarBreakEval' to 1. This is used to break evaluation
#  of a message in 'uam_evalMesg'
#
# Arguments: None.
#
# Results:
#
#  global uam_VarBreakEval is 1
#
proc uam_break {} {
global uam_VarBreakEval

 set uam_VarBreakEval 1
}

