#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# uam_formatMesg
#
#  Formats a mesg so that it can be printed in a 'nice' form.
#
# Arguments:
# 
#   mesg	- mesg to format
#
# Results:
#  
#  Returns a string, that includes all parts of the mesg.
#  Keys are separated from their contents by Tabs and ':'.
#  Returns an error if format of keyedlist in mesg is wrong.

proc uam_formatMesg {mesg} {

  # first look for :SENDER, :RECEIVER, :TYPE and put this keys
  # in front of text;
  # then append all the other keys in alphabetical order
  #
  if {![keylget mesg :SENDER mesgSend]} {
    return -code error
  } else {
    append text ":SENDER  \t: $mesgSend\n"
    keyldel mesg :SENDER
  }

  if {![keylget mesg :RECEIVER mesgRecv]} {
    return -code error
  } else {
    append text ":RECEIVER\t: $mesgRecv\n"
    keyldel mesg :RECEIVER
  }

  if {![keylget mesg :TYPE mesgType]} {
    return -code error
  } else {
    append text ":TYPE\t\t: $mesgType\n" 
    keyldel mesg :TYPE 
  }

  if {[catch {set keyList [lsort [keylkeys mesg]]}]} {
    return -code error
  }


  foreach key $keyList {
    append text "[string toupper $key]\t"
    if {[string length $key] < 8} {
      append text "\t"
    }
    if {[catch {append text ": [keylget mesg $key]\n"}]} {
      return $mesg
    }
  }

 return $text
}

