#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# uam_helpSystemProcs
#
#  This file contains the procedures for the IDEAS HelpSystem.
#

#
# uam_initHelp
#
# Initializes the window for the help system.
#
#  Arguments:
#
#	fileName - name of helpfile to show in widget (standard: "")
#
#  Results: None.
#
proc uam_initHelp {{fileName ""}} {
global uam_VarHelpTopic uam_VarHelpIndex ideas_HomePath

 set w .help

 # create a new toplevel:
 #
 catch {destroy $w}
 toplevel $w
 wm withdraw $w
 wm title $w "IDEAS - Help System"
 wm iconname $w "IDEAS - Help" 

 # create a listbox for the list of available help topics:
 #
 frame $w.f0 -relief ridge -bd 2
 pack $w.f0 -padx 3 -pady 3 -side top -fill x -expand yes

 label $w.f0.lbl -text "Topics:" \
     -relief ridge -bd 2  

 listbox $w.f0.list -relief ridge \
     -yscroll "$w.f0.scroll set" \
     -height 8 -width 65 \
     -exportselection no

 scrollbar $w.f0.scroll -command "$w.f0.list yview"

 pack $w.f0.lbl -side top -fill x 
 pack $w.f0.scroll -side right -fill y
 pack $w.f0.list -side left -fill both -expand yes

 bind $w.f0.list <Double-Button-1> "$w.f1.show flash; $w.f1.show invoke"
 bind $w.f0.list <Return> "$w.f1.show flash; $w.f1.show invoke"

 bind $w <Escape> "$w.f1.close flash; $w.f1.close invoke"

 # create two buttons:
 #  'Show' to watch a topic
 #  'Close' to exit the HelpSystem
 # 
 frame $w.f1 -relief ridge -bd 2
 pack $w.f1 -padx 3 -pady 3 -side top -fill x -expand yes

 button $w.f1.show -text Show \
     -width 8 \
     -command "uam_showHelpFile"

 button $w.f1.close -text Close  \
     -width 8 \
     -command "destroy $w"

 pack $w.f1.show -side left -expand yes\
     -padx 15m -pady 1m -ipadx 2m -ipady 1m 

 pack $w.f1.close -side left -expand yes\
     -padx 15m -pady 1m -ipadx 2m -ipady 1m 


 # create a text widget for displaying the content of a topic:
 #
 frame $w.f2 -relief ridge -bd 2
 pack $w.f2 -padx 3 -pady 3 -side top -fill x -expand yes

 frame $w.f2.fr -relief ridge -bd 2
 pack $w.f2.fr -side top -fill x -expand yes

 label $w.f2.fr.lbl1 -relief flat\
     -padx 2m

 label $w.f2.fr.lbl2 -textvariable uam_VarHelpTopic \
     -relief flat -anchor w

 pack $w.f2.fr.lbl1 -side left
 pack $w.f2.fr.lbl2 -side left -fill x -expand yes

 text $w.f2.text -relief ridge -bd 2 \
            -yscrollcommand "$w.f2.scroll set" \
            -wrap word \
	    -height 25 \
	    -width 65 \
            -padx 1m \
            -state disabled \
            -font "-adobe-new century schoolbook-medium-r-*--14-100-*"

 scrollbar $w.f2.scroll -relief ridge -command "$w.f2.text yview"

 pack $w.f2.scroll -side right -fill y
 pack $w.f2.text -side left -fill both -expand yes

 # read the HelpIndex:
 #
 source $ideas_HomePath/help/uam_helpIndex.tcl

 # fill the list of topics:
 #
 foreach topic $uam_VarHelpIndex {
   $w.f0.list insert end [lindex $topic 0]
 }

 # show the specified HelpFile in the text widget:
 #
 uam_showHelpFile $fileName

 wm deiconify $w

}


#
# uam_showHelpFile
#
#  Shows a specified HelpFile in the text widget of the IDEAS
#  HelpSystem window.
#  The file is specified as an argument to this proc. If argument is
#  unspecified or empty , the current selection in topics-listbox of
#  the HelpSystem is got.
#  If no HelpFile is specified, it shows the standard inforamtion
#  for the HelpSystem.
#
# Arguments:
#
#	fileName - name of the HelpFile to show (standard: empty)
#
# Results: None.
#
proc uam_showHelpFile {{fileName ""}} {
global uam_VarHelpTopic uam_VarHelpIndex ideas_HomePath

 if {$fileName == ""} {

   # get the current selection in listbox 'Topics' of HelpSystem:
   set sel [.help.f0.list curselection]

   if {$sel == ""} {
     set uam_VarHelpTopic "Help: The IDEAS - Help System"
     set fileName helpSystem.hlp
   } else {
     # get the HelpFile for the selected topic:
     #
     set uam_VarHelpTopic [.help.f0.list get $sel]
     set fileName [lindex [lindex $uam_VarHelpIndex $sel] 1]
   }
 } else {
  set uam_VarHelpTopic \
    [lindex [lindex $uam_VarHelpIndex \
               [lsearch $uam_VarHelpIndex *$fileName*]] 0]
 }

 set error 0

 if {$fileName == ""} {
   set errorMesg "No help file specified for '$uam_VarHelpTopic' in\
       HelpIndex."
   set error 1
 } 

 # open the HelpFile and read the content:
 #
 if {$error || ([catch {set fid [open $ideas_HomePath/help/$fileName r]} \
                 errorMesg])} {
   set content "Sorry, this topic is not available ...\n\nError: $errorMesg"
 } else {
   set content [read $fid]
 }

 catch {close $fid}

 # show the content of HelpFile in text widget:
 #
 .help.f2.text configure -state normal
 .help.f2.text delete 1.0 end
 .help.f2.text insert end $content
 .help.f2.text configure -state disabled

 update
}

