#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# uam_lbTools.tcl
#
# This file includes some procedures to handle with listboxes used
# in 'ideas', 'uam_edActSrcFile', 'uam_edSystemFile', 'uam_edTypeFile'
# and 'uam_newAgent'.
#

# uam_fillLb
#
#  Fills the listbox 'lb' with a list of all the files matching 'filter'
#  in the actual directory .
#
proc uam_fillLb {lb filter} {
  $lb delete 0 [$lb size]
  catch { 
    foreach i [lsort [eval "glob $filter"]] {
        $lb insert end [file tail $i]
    }
  }
}

# uam_insertLbSel
#
#  Copies actual selected entry in listbox 'lb' into 
#  entry-field 'inpt'
#
proc uam_insertLbSel {lb inpt} {
  $inpt delete 0 end
  if {![catch {set sel [$lb get active]}]} {
    $inpt insert 0 $sel
  }
}

# uam_selectLbEntry
#
# Selects an entry in listbox 'lb' relativ to argument 'i':
#  $i == #-1  prior entry
#  $i == #+1  next entry
#  else: select entry with index $i 
#
proc uam_selectLbEntry {lb i} {
  set curInd 0
  if {($i == "#-1") || ($i == "#+1")} {
   set curInd [lindex [$lb curselection] 0]
   incr curInd [crange $i 1 end]
  } else {
   set curInd $i
  }

  $lb selection clear 0 end
  $lb selection set $curInd
}


