#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# uam_initMesgTo
#
#  Initializes the widget .mesgTo for defining mesges to agents.
#  The widget is shown when button 'Message To' in UAM is invoked.
#
# Arguments: None.
#
# Results: None.
#
proc uam_initMesgTo {} {
global ideas_KeyNumber

  # create a new toplevel widget .MESGTO
  #
  catch {destroy .mesgTo}
  toplevel .mesgTo
  wm title .mesgTo "Message To"
  wm iconname .mesgTo "MesgTo"
  wm withdraw .mesgTo

  # divide the widget in three frames:
  # one at the bottom for some buttons
  #
  frame .mesgTo.bot 
  pack .mesgTo.bot -side bottom -fill both -pady 5

  # and two big vertical frames, the left one for mesg keys
  # the right one for the content of each key
  #
  frame .mesgTo.left -width 15 -bd 3
  pack .mesgTo.left -side left
  frame .mesgTo.right -bd 3
  pack .mesgTo.right -side left -fill x -expand 1

  # create OK, Shortcuts and Cancel buttons at bottom of toplevel
  #
  button .mesgTo.ok -text OK \
		-width 8 \
		-command {uam_mesgToOK}

  frame .mesgTo.default -relief sunken -bd 1
  raise .mesgTo.ok .mesgTo.default

  pack .mesgTo.default -in .mesgTo.bot -side left \
                       -expand 1 -padx 3m -pady 2m

  pack .mesgTo.ok -in .mesgTo.default -padx 1m -pady 1m \
		    -ipadx 2m -ipady 1m

  button .mesgTo.short -text Shortcuts \
    -width 8 \
    -command {
       set dad .mesgTo
       set x [expr [winfo x $dad] - ([winfo reqwidth .shorts]+50)]
       if {$x < 0} {
         set x 0
       }
       set y [expr round([winfo y $dad])]
       wm geom .shorts +$x+$y
       if {[winfo ismapped .shorts]} {
         raise .shorts
       } else {
         wm deiconify .shorts
       }
       update
     }

  pack .mesgTo.short -in .mesgTo.bot -side left \
	-padx 2m -pady 1m -ipadx 2m -ipady 1m 

  button .mesgTo.cancel -text Cancel \
    -width 8 \
    -command {wm withdraw .mesgTo; focus .mbar}

  pack .mesgTo.cancel -in .mesgTo.bot -side left \
	-padx 2m -pady 1m -ipadx 2m -ipady 1m 

  bind .mesgTo.ok <Return> {.mesgTo.ok flash
                            .mesgTo.ok invoke}
  bind .mesgTo.ok <Escape> {.mesgTo.cancel flash
                            .mesgTo.cancel invoke}

  # at top of toplevel show headline
  # 'Key      Content' above the frames for this

  label .mesgTo.l1 -text Key -anchor w
  pack .mesgTo.l1 -in .mesgTo.left -side top -fill x

  label .mesgTo.l2 -text Content -anchor w
  pack .mesgTo.l2 -in .mesgTo.right -side top -fill x

  # create entry widgets in rows for input of keys .k
  # and content .c

  set keyNumber $ideas_KeyNumber

  for {set i 1} {$i <= $keyNumber} {incr i} {
    global uam_VarKey uam_VarCont
    catch {unset uam_VarKey($i); unset uam_VarCont($i)}

    entry .mesgTo.k$i -textvariable uam_VarKey($i) \
                -width 15 \
                -relief sunken 
                
    pack .mesgTo.k$i -in .mesgTo.left -side top -pady 1

    entry .mesgTo.c$i -textvariable uam_VarCont($i)\
		-exportselection 1 \
                -relief sunken 
    pack .mesgTo.c$i -in .mesgTo.right -side top -pady 1 \
        -fill x

    # set up the bindings for the entries:
    # jump from field to field with Tab or arrows, end with
    # Return or Escape
 
    bind .mesgTo.k$i <Help> {uam_initHelp mesgTo.hlp}
    bind .mesgTo.k$i <F1>   {uam_initHelp mesgTo.hlp}
    bind .mesgTo.c$i <Help> {uam_initHelp mesgTo.hlp}
    bind .mesgTo.c$i <F1>   {uam_initHelp mesgTo.hlp}

    bind .mesgTo.k$i <Escape> {.mesgTo.cancel flash
                               .mesgTo.cancel invoke}
    bind .mesgTo.c$i <Button-2> {catch {%W insert insert [selection get]}}
    bind .mesgTo.c$i <Escape> {.mesgTo.cancel flash
                               .mesgTo.cancel invoke}
    bind .mesgTo.c$i <Return> {.mesgTo.ok flash
                               .mesgTo.ok invoke}
    bind .mesgTo.c$i <Up> "focus .mesgTo.c[expr $i-1]"

    bind .mesgTo.k$i <Tab> "focus .mesgTo.c$i"
    bind .mesgTo.k$i <Right> "focus .mesgTo.c$i"
    bind .mesgTo.k$i <Return> "focus .mesgTo.c$i"
    bind .mesgTo.k$i <Left> "focus .mesgTo.c[expr $i-1]"
    bind .mesgTo.k$i <Up> "focus .mesgTo.k[expr $i-1]"
 
    bind .mesgTo.c$i <Tab> "focus .mesgTo.k[expr $i+1]"
    bind .mesgTo.c$i <Right> "focus .mesgTo.k[expr $i+1]"
    bind .mesgTo.c$i <Down> "focus .mesgTo.c[expr $i+1]"

    bind .mesgTo.k$i <Down> "focus .mesgTo.k[expr $i+1]"

    bind .mesgTo.c$i <Left> "focus .mesgTo.k[expr $i]"

    bind .mesgTo.k$i <Control-q> {uam_exit}
    bind .mesgTo.c$i <Control-q> {uam_exit}

  }

  bind .mesgTo.c2 <Tab> "focus .mesgTo.c3"
  bind .mesgTo.c2 <Right> "focus .mesgTo.c3"
  bind .mesgTo.c2 <Left> "focus .mesgTo.c$keyNumber"
  bind .mesgTo.c2 <Up> "focus .mesgTo.c$keyNumber"

  bind .mesgTo.c3 <Left> "focus .mesgTo.c2"

  bind .mesgTo.c$keyNumber <Tab> "focus .mesgTo.c2"
  bind .mesgTo.c$keyNumber <Right> "focus .mesgTo.c2"
  bind .mesgTo.c$keyNumber <Down> "focus .mesgTo.c2"

  bind .mesgTo.k4 <Up> "focus .mesgTo.k$keyNumber"
  bind .mesgTo.k$keyNumber <Down> "focus .mesgTo.k4"

  # make some entries unvisible that user should not edit 
  #
  set uam_VarKey(1) ":SENDER"
  set uam_VarCont(1) "UAM"
  .mesgTo.k1 configure -state disabled -relief flat
  .mesgTo.c1 configure -state disabled -relief flat

  set uam_VarKey(2) ":RECEIVER"
  .mesgTo.k2 configure -state disabled -relief flat

  set uam_VarKey(3) ":TYPE"
  .mesgTo.k3 configure -state disabled -relief flat

  update
}


#
# uam_mesgTo
#
#  Provides an UAM-tool to send messages to agents selected
#  from agent list.
#  It displays the toplevel widget .mesgTo created in uam_initMesgTo
#  
# Arguments:
#
#  None.
#
# Results:
#
#  None.
#
proc uam_mesgTo {} {
global uam_VarKey uam_VarCont

  # see how big the widgets want to be and place the toplevel 
  # on the screen right in the middle of its parent widget
  #
  set dad [winfo parent .mesgTo]
  set x [expr [winfo x $dad] + ([winfo reqwidth $dad] - [winfo reqwidth .mesgTo])/2]
  set y [expr round([winfo y $dad] + [winfo reqheight $dad]*0.3)]
  wm geom .mesgTo +$x+$y


  if {[winfo ismapped .mesgTo]} {
      raise .mesgTo
      update
  }

  set sel [.frame1.list curselection]

  set targetList ""

  # build a list of selected agent names
  #
  foreach i $sel {
    set listentry [.frame1.list get $i]
    lappend targetList [lindex $listentry 0]
  } 
  
  set uam_VarCont(2) $targetList
  set uam_VarCont(3) ""

  for {set i 4} {$i <= [array size uam_VarCont]} {incr i} {
    set uam_VarKey($i) ""
    set uam_VarCont($i) ""
  } 
  
  wm deiconify .mesgTo
 
  update

  if {$targetList == ""} {
    focus .mesgTo.c2
  } else {
    focus .mesgTo.c3
  }
}


#
# uam_mesgToOK
#
#  Called by uam_mesgTo when the OK-button is invoked.
#  Checks if each specified key has a content and builts
#  up the mesg for sending.
#
# Arguments:
#
#  None.
#
# Results:
#
#   None.
#

proc uam_mesgToOK {} {
global uam_VarKey uam_VarCont

 for {set i 1} {$i <= 6} {incr i} {
    if {![lempty $uam_VarKey($i)]} {
      if {![lempty $uam_VarCont($i)]} {
        # set a ":" before the key if necessary and convert
        # all chars to upper chars
        #
        set uam_VarKey($i) [string trim $uam_VarKey($i)]
        set uam_VarCont($i) [string trim $uam_VarCont($i)]


        if {![cequal [cindex $uam_VarKey($i) 0] ":"]} {
          set uam_VarKey($i) [string toupper :$uam_VarKey($i)]
        } else {
          set uam_VarKey($i) [string toupper $uam_VarKey($i)]
        }

        keylset mesg $uam_VarKey($i) $uam_VarCont($i)
      } else {
        BEEP
        focus .mesgTo.c$i
        return
      }
    }
 }

 uam_sendMesg $mesg

}


