#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# uam_newAgt
# 
# This script generates a file browser, which lists the files of
# type '.agt' in the 'desc'-directory. It allows the user to choose
# one file by typing its identifier or selecting it from the list
# and type a name for the new agent in an entry field.
#
# Further it provides a radiobutton to change the view of the
# browser: It then shows all files of type '.mas' and allows to
# select the description of a local Multi-Agent-System to start it.
#
# Arguments:  None.
#
# Results:  None.
#
proc uam_newAgt {} {
global uam_VarInitType

 # create a new toplevel:

 catch {destroy .init}
 global agents
 toplevel .init
 wm title .init "Initialize New Agents"
 wm iconname .init "InitNew"
 wm withdraw .init

 # Create three frames for managing the main window

 frame .init.top -bd 7
 frame .init.frame1 -bd 7
 frame .init.frame2 -bd 7

 pack .init.top -side top -fill x
 pack .init.frame1  .init.frame2 -side left \
	-fill y


 # Create the entry label for the agent name with a description on
 # top of the window
 
 label .init.top.lbl -text "Name for new Agent: "
 entry .init.top.inpt -relief sunken
 pack .init.top.lbl .init.top.inpt -side left -fill x

 # Create a label in frame1 and a listbox with a scrollbar on the
 # left side

 label .init.frame1.lbl 
 pack .init.frame1.lbl -side top -anchor w -fill x
 frame .init.frame1.listbox
 pack .init.frame1.listbox -side top -fill x

 scrollbar .init.frame1.scroll -command ".init.frame1.list yview"
 pack .init.frame1.scroll -in .init.frame1.listbox -side right -fill y

 listbox .init.frame1.list -yscroll ".init.frame1.scroll set" \
	-relief raised \
	-exportselection no

 pack .init.frame1.list -in .init.frame1.listbox -side left -fill both -expand no

 frame .init.frame1.choose
 pack .init.frame1.choose -side top -fill x -pady 3

 # create the radiobuttons to change the view of the browser:
 #
 radiobutton .init.frame1.agt -text Agents \
	-variable uam_VarInitType \
	-value 1 \
	-underline 0\
	-command uam_newAgtChgView

 radiobutton .init.frame1.mas -text Systems \
	-variable uam_VarInitType \
	-value 2 \
	-underline 0\
	-command uam_newAgtChgView

 pack .init.frame1.agt -side left -in .init.frame1.choose
 pack .init.frame1.mas -side left -fill x \
	-padx 5 -in .init.frame1.choose


 # Create in frame2 an entry for selected filename and
 # buttons for OK and CANCEL

 label .init.frame2.lbl
 entry .init.frame2.inpt -relief sunken
 pack .init.frame2.lbl .init.frame2.inpt -anchor w

 button .init.ok -text OK \
	-width 15 \
	-command "uam_initNewAgtOK"

 frame .init.default -relief sunken -bd 1
 raise .init.ok .init.default


 button .init.cancel -text CANCEL \
	-width 15 \
	-command "destroy .init"

 pack .init.cancel -in .init.frame2 -side bottom \
	-pady 1 -ipadx 2m -ipady 1m

 pack  .init.default -in .init.frame2 \
	-side bottom  \
	-pady 4m

 pack .init.ok -in .init.default \
	-padx 1m -pady 1m \
	-ipadx 2m -ipady 1m

 # Set up bindings for the browser

 bindtags .init.frame1.list {Listbox .init.frame1.list}

 bind .init <Alt-Key-a> {.init.frame1.agt invoke}
 bind .init <Alt-Key-s> {.init.frame1.mas invoke}

 bind .init.top.inpt <Return> {.init.ok flash; uam_initNewAgtOK}

 bind .init.top.inpt <Escape> {
	.init.cancel flash
	.init.cancel invoke}

 bind .init.top.inpt <Help> {uam_initHelp initNewAgt.hlp}
 bind .init.top.inpt <F1>   {uam_initHelp initNewAgt.hlp}
 bind .init.top.inpt <Control-q> {uam_exit}

 bind .init.frame1.list <Help> {uam_initHelp initNewAgt.hlp}
 bind .init.frame1.list <F1>   {uam_initHelp initNewAgt.hlp}
 bind .init.frame1.list <Control-q> {uam_exit}

 bind .init.frame1.agt <Help> {uam_initHelp initNewAgt.hlp}
 bind .init.frame1.agt <F1>   {uam_initHelp initNewAgt.hlp}
 bind .init.frame1.agt <Control-q> {uam_exit}

 bind .init.frame1.mas <Help> {uam_initHelp initNewAgt.hlp}
 bind .init.frame1.mas <F1>   {uam_initHelp initNewAgt.hlp}
 bind .init.frame1.mas <Control-q> {uam_exit}

 bind .init.frame2.inpt <Help> {uam_initHelp initNewAgt.hlp}
 bind .init.frame2.inpt <F1>   {uam_initHelp initNewAgt.hlp}
 bind .init.frame2.inpt <Control-q> {uam_exit}

 bindtags .init.frame1.list {Listbox .init.frame1.list}

 bind .init.frame1.list <ButtonRelease-1> {
	uam_insertLbSel .init.frame1.list .init.frame2.inpt }

 bind .init.frame1.list <Double-Button-1> {
	uam_insertLbSel .init.frame1.list .init.frame2.inpt
	.init.ok flash; .init.ok invoke}


 bind .init.frame1.list <Tab> {focus .init.frame2.inpt}

 bind .init.frame1.list <Up> {
	uam_insertLbSel .init.frame1.list .init.frame2.inpt }

 bind .init.frame1.list <Down> {
	uam_insertLbSel .init.frame1.list .init.frame2.inpt }

 bind .init.frame1.list <Return> {
	if {[set sel [.init.frame1.list curselection]]!=""} {
	  uam_insertLbSel .init.frame1.list .init.frame2.inpt
	}
	.init.ok flash; .init.ok invoke}

 bind .init.frame1.list  <Escape> {.init.cancel flash;.init.cancel invoke}

 bind .init.frame2.inpt <Return> {.init.ok flash; .init.ok invoke}

 bind .init.frame2.inpt <Escape> {.init.cancel flash; .init.cancel invoke}

 bind .init.frame1.agt <Return> {
	.init.frame1.agt configure -state normal
	.init.ok flash; .init.ok invoke}

 bind .init.frame1.agt <Escape> {.init.cancel flash; .init.cancel invoke}

 bind .init.frame1.mas <Return> {
	.init.frame1.mas deselect
	.init.ok flash; .init.ok invoke}

 bind .init.frame1.mas <Escape> {.init.cancel flash; .init.cancel invoke}
 bind .init.frame1.mas <space> {
	.init.frame1.mas deselect
	.init.frame1.mas invoke}



 # set globale variable 'uam_VarInitType' to 1,
 #   i. e. show agent type files,
 # and call proc 'uam_newAgtChgView' to do so:
 #
 set uam_VarInitType 1
 uam_newAgtChgView

 # update all the geometry information so we know how big
 # it wants to be, then place the window relative to its 
 # parent and de-iconify it.

 update idletasks
 set dad [winfo parent .init]
 set x [expr [winfo x $dad] + \
       ([winfo reqwidth $dad] - [winfo reqwidth .init])/2]
 set y [expr round([winfo y $dad] + [winfo reqheight $dad]*0.3)]

 wm geom .init +$x+$y
 wm deiconify .init


 tkwait window .init
}


 # On OK do this:

proc uam_initNewAgtOK {} {
global agents uam_VarInitType

 if {$uam_VarInitType == 1} {
   set newName [.init.top.inpt get]

   # test if newName is not empty
   #
   if {[string length $newName]} {
     set newName \
       [string tolower [string range [lindex $newName 0] 0 9]]
   } else { 
     uam_noAgtIdf 
     return
   }

   # test if newName includes non alphanumeric characters
   #
   if {![ctype alnum $newName]} {
     uam_illegalAgtName $newName
     return
   }

   # test if newName exists
   #
   set existAgts [array names agents]
   if {[lsearch $existAgts ${newName}.*] != -1} {
     uam_newNameExists $newName
     return
   }

   set typeFile [.init.frame2.inpt get]

   # test if selected filename exists
   #
   if {[lsearch [glob desc/*.agt] desc/$typeFile]==-1} {
     uam_notExType $typeFile
     return
   }

   destroy .init

   uam_startAgent $newName $typeFile

  } else {
   set systemFile [.init.frame2.inpt get]

   # test if selected filename exists
   #
   if {[lsearch [glob systems/*.mas] systems/$systemFile]==-1} {
     uam_notExSystem $systemFile
     focus .init.frame2.inpt
     return
   }

   destroy .init

   uam_startMAS $systemFile
 }
}


