#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# uam_sendMesg
# 
#  Sends a well formatted message to other agents.
#  If something is wrong, an error is displayed and
#  message is not sent.
#
# Arguments:
#
#	mesg	-	message to send
#
# Results:
#
#  None. 
#

proc uam_sendMesg {mesg} {
 global agents uam_MesgCounter


 # fill in return sender of message
 #
 keylset mesg :SENDER UAM

 keylset mesg :REPLY-WITH $uam_MesgCounter

 # check format of message
 #
 if {[catch {set text [uam_formatMesg $mesg]} errorMesg]} {
   uam_logText "FORMAT-ERROR when trying to send:\n$mesg\nFailure:\
                '$errorMesg'\n"
   break
 }

 if {[catch {keylget mesg :REPLY-TO}]} {
   keylset mesg :REPLY-TO none
 }

 incr uam_MesgCounter

 keylget mesg :RECEIVER receiverList

 foreach agtName $receiverList {
   if {([catch {keylget agents($agtName) FID agtFid}]) || \
       ([cequal $agtName UAM])} {

     uam_logText "UNKNOWN-ADDRESS-ERROR when trying to send:\n$text"
     uam_sendMesgError "Unknown-Address-Error:\n\n'$agtName' \
                        in Message\n\n$text"

   } elseif {[catch {puts $agtFid $mesg; flush $agtFid}]} {

     uam_logText "WRITE-ERROR when trying to send:\n$text"
    
   } else {
     uam_logText "Message sent to $agtName\n$text"

     wm withdraw .mesgTo
     focus .mbar
   }
 }

} 

