#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# uam_initShortcuts
#
#  Creates a toplevel widget for shortcuts in UAM when sending 
#  a mesg to an agent.
#
# Arguments: None.
#
# Results: 
#
#  A widget .shorts is created but not shown.
#
proc uam_initShortcuts {} {
global uam_Shortcuts ideas_KeyNumber
global uam_VarShortKey uam_VarShortCont

  # create a new toplevel widget .shorts
  #
  catch {destroy .shorts}
  toplevel .shorts
  wm title .shorts "Shortcuts"
  wm iconname .shorts "Shorts"
  wm withdraw .shorts

  # divide .shorts vertical in two big frames
  # the right one is not shown yet; it will be shown on
  # buttonpress Edit or New defined next
  #
  frame .shorts.f1 -relief ridge -bd 3
  frame .shorts.f2 -relief ridge -bd 3
  pack .shorts.f1 -side left -fill both

  # create the left side first:
  # divide it horizontal in three more frames
  #
  frame .shorts.f1.t
  pack .shorts.f1.t -side top -fill x -pady 5

  frame .shorts.f1.b
  pack .shorts.f1.b -side bottom -fill x -pady 5

  frame .shorts.f1.m
  pack .shorts.f1.m \
	  -side top -fill both \
	  -padx 3m -expand 1

  # create three buttons at the left top:
  #  New, Edit, Remove
  #
  button .shorts.new -text New \
    -width 8 \
    -command {uam_newShortcut}

  button .shorts.edit -text Edit \
    -width 8 \
    -command {uam_editShortcut
              # show right frame in shorts
              #
              pack .shorts.f2 -side left -fill both
              focus .shorts.nameEnt
             }

  button .shorts.del -text Remove \
    -width 8 \
    -command {uam_removeShortcut}

  pack .shorts.new .shorts.edit .shorts.del \
      -in .shorts.f1.t \
      -side left \
      -padx 2m -pady 1m -ipadx 2m -ipady 1m 
 
  # create two buttons at the left bottom:
  #  Select, Close
  #
  button .shorts.sel -text Select \
      -width 8 \
      -command {uam_selectShortcut}

  pack .shorts.sel \
      -in .shorts.f1.b \
      -side left \
      -padx 2m -pady 2m -ipadx 2m -ipady 1m 

  button .shorts.close -text Close \
      -width 8 \
      -command {wm withdraw .shorts; \
                catch {pack forget .shorts.f2}
                uam_shortcutResetFocus}
  pack .shorts.close \
      -in .shorts.f1.b \
      -side right \
      -padx 2m -pady 2m -ipadx 2m -ipady 1m 

  # create two listboxes with a label at the left middle:
  # one for shortcuts and one for the hotkeys 
  #
  label .shorts.f1.m.title \
      -text Shortcuts: 
  pack .shorts.f1.m.title \
      -side top\
      -fill x

  frame .shorts.listb 
#      -bd 3 
#      -relief raised

  pack .shorts.listb \
      -in .shorts.f1.m \
      -fill both \
      -expand 1

  scrollbar .shorts.scroll \
      -command "uam_shortsScrollCmd" 
  pack .shorts.scroll \
      -in .shorts.listb \
      -side right \
      -fill y

 
  listbox .shorts.hotk \
      -yscroll ".shorts.scroll set" \
      -height 8 -width 8 \
-relief raised \
      -exportselection no

  pack .shorts.hotk \
      -in .shorts.listb \
      -side right \
      -fill y 

  listbox .shorts.name \
      -yscroll ".shorts.scroll set" \
-relief raised \
      -exportselection no

  pack .shorts.name \
      -in .shorts.listb \
      -side left \
      -fill both \
      -expand 1

  bindtags .shorts.name {.shorts.name}
  bindtags .shorts.hotk {.shorts.hotk}

  bind .shorts.name <Button-1> {
      uam_selectLbEntry .shorts.name [.shorts.name nearest %y]
      uam_selectLbEntry .shorts.hotk [.shorts.name curselection]
      uam_editShortcut
  }

  bind .shorts.name <Double-Button-1> {
      uam_selectLbEntry .shorts.name [.shorts.name nearest %y]
      uam_selectLbEntry .shorts.hotk [.shorts.name curselection]
      uam_selectShortcut
  }

  bind .shorts.name <Button1-Motion> {
      uam_selectLbEntry .shorts.name [.shorts.name nearest %y]
      uam_selectLbEntry .shorts.hotk [.shorts.name curselection]
      uam_editShortcut
  }

  bind .shorts.hotk <Button-1> {
      uam_selectLbEntry .shorts.hotk [.shorts.hotk nearest %y]
      uam_selectLbEntry .shorts.name [.shorts.hotk curselection]
      uam_editShortcut
  }

  bind .shorts.hotk <Double-Button-1> {
      uam_selectLbEntry .shorts.hotk [.shorts.hotk nearest %y]
      uam_selectLbEntry .shorts.name [.shorts.hotk curselection]
      uam_selectShortcut
  }

  bind .shorts.hotk <Button1-Motion> {
      uam_selectLbEntry .shorts.hotk [.shorts.hotk nearest %y]
      uam_selectLbEntry .shorts.name [.shorts.hotk curselection]
      uam_editShortcut
  }

  bindtags .shorts.close {.shorts.close Button}

  bind .shorts.close <Tab> {
      focus .shorts.new
  }

  # define global the variable uam_Shortcuts as an array
  # - special hack 8-)
  #
  set uam_Shortcuts(0) dummy
  unset uam_Shortcuts(0)

  # read file with saved shortcuts
  catch {source uam_shortcutsFile.tcl}

  # fill the listboxes and bind shortcuts
  #
  uam_fillShortcutLb  

  # create now the right side :
  # divide it horizontal in three more frames
  #
  frame .shorts.f2.t
  pack .shorts.f2.t -side top -fill x -pady 5

  frame .shorts.f2.b
  pack .shorts.f2.b -side bottom -fill x -pady 5

  frame .shorts.f2.m
  pack .shorts.f2.m -side top -fill both -padx 1m -pady 3m

  # devide the rigth top vertical in two frames : 
  # in each an entry widgets for the name and the hotkey of
  # the new shortcut
  #

  frame .shorts.f2.t.r
  pack .shorts.f2.t.r -side right -fill y -padx 2m -pady 2m

  frame .shorts.f2.t.l
  pack .shorts.f2.t.l -side left -fill both -expand 1 \
      -padx 2m -pady 2m

  label .shorts.nameLbl -text Name:
  pack .shorts.nameLbl \
      -in .shorts.f2.t.l \
      -side left

  entry .shorts.nameEnt \
      -textvariable uam_VarShortName \
      -width 15 \
      -relief sunken
  pack .shorts.nameEnt\
      -in .shorts.f2.t.l \
      -side left \
      -fill x

  bind .shorts.nameEnt <Tab> {focus .shorts.hotkEnt}
  bind .shorts.nameEnt <Down> {focus .shorts.c1}
  bind .shorts.nameEnt <Return> {.shorts.add flash
                                 .shorts.add invoke}
  bind .shorts.nameEnt <Escape> {.shorts.done flash
                                 .shorts.done invoke
                                }

  bind .shorts.nameEnt <Help> {uam_initHelp shortcuts.hlp}
  bind .shorts.nameEnt <F1>   {uam_initHelp shortcuts.hlp}
  bind .shorts.nameEnt <Control-q> {uam_exit}

  label .shorts.hotkLbl -text "Hotkey: Alt+"
  pack .shorts.hotkLbl \
      -in .shorts.f2.t.r \
      -side left

  entry .shorts.hotkEnt \
      -textvariable uam_VarShortHotk \
      -width 2 \
      -relief sunken
  pack .shorts.hotkEnt\
      -in .shorts.f2.t.r \
      -side left \
      -fill x

  bindtags .shorts.hotkEnt {.shorts.hotkEnt}

  # the hotkey is only one char defined by user
  #
  bind .shorts.hotkEnt <Any-KeyPress> {
    if {[clength %K] == 1} {
      set uam_VarShortHotk [string tolower %K]
    }
  }

  bind .shorts.hotkEnt <BackSpace> {
   set uam_VarShortHotk ""
  }

  bind .shorts.hotkEnt <Help> {uam_initHelp shortcuts.hlp}
  bind .shorts.hotkEnt <F1>   {uam_initHelp shortcuts.hlp}

  bind .shorts.hotkEnt <Tab> {focus .shorts.c1}
  bind .shorts.hotkEnt <Down> {focus .shorts.c1}
  bind .shorts.hotkEnt <Left> {focus .shorts.nameEnt}
  bind .shorts.hotkEnt <Return> {.shorts.add flash
                                 .shorts.add invoke}
  bind .shorts.hotkEnt <Escape> {.shorts.done flash
                                 .shorts.done invoke
                                }
  bind .shorts.hotkEnt <Control-q> {uam_exit}

  # create two buttons at the right bottom:
  #  Add to List, Done Edit
  #
  button .shorts.add -text "Add to List" \
      -width 8 \
      -command {uam_addNewShortcut}

  frame .shorts.f2.default -relief sunken -bd 1
  raise .shorts.add .shorts.f2.default


  button .shorts.done -text "Done Edit" \
      -width 8 \
      -command {uam_shortcutResetFocus
                pack forget .shorts.f2
               }

  pack .shorts.f2.default \
      -in .shorts.f2.b \
      -side left \
      -expand 1 \
      -padx 3m -pady 2m

  pack .shorts.add \
      -in .shorts.f2.default \
      -padx 1m -pady 1m \
      -ipadx 2m -ipady 1m

  pack .shorts.done \
      -in .shorts.f2.b \
      -side right \
      -padx 2m -pady 1m -ipadx 2m -ipady 1m \
      -expand 1

  bindtags .shorts.add {.shorts.add  Button}
  bindtags .shorts.done {.shorts.done  Button}
  bind .shorts.add <Tab> {focus .shorts.done}
  bind .shorts.done <Tab> {focus .shorts.new}

  # create at the right middle the rows for the new shortcut:
  #
  # divide it vertical in two frames, the left one for the keys
  # the right one for the contents of the keys of then new shortcut
  #
  frame .shorts.f2.m.l -width 15 
  pack .shorts.f2.m.l -side left -padx 1m

  frame .shorts.f2.m.r 
  pack .shorts.f2.m.r -side left -padx 1m -fill x -expand 1

  label .shorts.keyLbl -text Key -anchor w
  pack .shorts.keyLbl -in .shorts.f2.m.l -side top -fill x

  label .shorts.contLbl -text Content -anchor w
  pack .shorts.contLbl -in .shorts.f2.m.r -side top -fill x

  # create entry widgets in rows for input of keys .k
  # and content .c

  set keyNumber [expr $ideas_KeyNumber-1]

  for {set i 1} {$i <= $keyNumber} {incr i} {

    catch {unset uam_VarShortKey($i); unset uam_VarShortCont($i)}

    entry .shorts.k$i -textvariable uam_VarShortKey($i) \
                -width 15 \
                -relief sunken 
                
    pack .shorts.k$i -in .shorts.f2.m.l -side top -pady 1

    entry .shorts.c$i -textvariable uam_VarShortCont($i)\
		-exportselection 1 \
                -relief sunken 
    pack .shorts.c$i -in .shorts.f2.m.r -side top -pady 1 \
        -fill x

    # set up the bindings for the entries:
    # jump from field to field with Tab or arrows, end with
    # Return or Escape
 
    bind .shorts.k$i <Help> {uam_initHelp shortcuts.hlp}
    bind .shorts.k$i <F1>   {uam_initHelp shortcuts.hlp}
    bind .shorts.c$i <Help> {uam_initHelp shortcuts.hlp}
    bind .shorts.c$i <F1>   {uam_initHelp shortcuts.hlp}

    bind .shorts.k$i <Escape> {.shorts.done flash
                               .shorts.done invoke
                              }
    bind .shorts.c$i <Button-2> {catch {%W insert insert [selection get]}}
    bind .shorts.c$i <Escape> {.shorts.done flash
                               .shorts.done invoke
                              }
    bind .shorts.c$i <Up> "focus .shorts.c[expr $i-1]"

    bind .shorts.k$i <Tab> "focus .shorts.c$i"
    bind .shorts.k$i <Right> "focus .shorts.c$i"
    bind .shorts.k$i <Return> {.shorts.add invoke
                               .shorts.add flash}
    bind .shorts.k$i <Left> "focus .shorts.c[expr $i-1]"
    bind .shorts.k$i <Up> "focus .shorts.k[expr $i-1]"
    bind .shorts.k$i <Down> "focus .shorts.k[expr $i+1]"
    bind .shorts.k$i <Control-q> {uam_exit}

    bind .shorts.c$i <Tab> "focus .shorts.k[expr $i+1]"
    bind .shorts.c$i <Right> "focus .shorts.k[expr $i+1]"
    bind .shorts.c$i <Down> "focus .shorts.c[expr $i+1]"
    bind .shorts.c$i <Return> {.shorts.add invoke
                               .shorts.add flash}
    bind .shorts.c$i <Left> "focus .shorts.k[expr $i]"
    bind .shorts.c$i <Control-q> {uam_exit}

  }
  bind .shorts.c1 <Left> "focus .shorts.c$keyNumber"
  bind .shorts.c1 <Up> "focus .shorts.nameEnt"

  bind .shorts.c$keyNumber <Tab> "focus .shorts.nameEnt"
  bind .shorts.c$keyNumber <Right> "focus .shorts.c1"
  bind .shorts.c$keyNumber <Down> "focus .shorts.nameEnt"

  bind .shorts.k2 <Up> "focus .shorts.k$keyNumber"
  bind .shorts.k$keyNumber <Down> "focus .shorts.k2"


  # make entry :TYPE user should not edit unvisible 
  #
  set uam_VarShortKey(1) ":TYPE"
  .shorts.k1 configure -state disabled -relief flat

  update idletasks

}


#
# uam_shortsScrollCmd
#
#  Used to adjust the view in the two listboxes in .shorts with
#  one scrollbar.
#
# Arguments:
#     a, b, c - possible arguments of the scrollbar
#
# Results: None.
#
proc uam_shortsScrollCmd {{a ""} {b ""} {c ""}} {
 eval ".shorts.name yview $a $b $c"
 eval ".shorts.hotk yview $a $b $c"
}


#
# uam_fillShortcutLb
#
#  Fills the shortcut listbox in .shorts and bind the hotkeys
#  to widget .mesgTo 
#
# Arguments: None.
#
# Results: None.
#
proc uam_fillShortcutLb {} {
global uam_Shortcuts ideas_KeyNumber

 .shorts.name delete 0 end
 .shorts.hotk delete 0 end
 
 # sort the listbox content
 #
 foreach key [lsort [array names uam_Shortcuts]] {
   keylget uam_Shortcuts($key) hotkey keyHotk
   .shorts.name insert end $key

   if {$keyHotk == ""} { 
    .shorts.hotk insert end " "
   } else {
    .shorts.hotk insert end <Alt+${keyHotk}>
   }

   # the shortcuts hotkeys are of the form Alt+char
   # bind them to .mesgTo
   #
   for {set i 1} {$i <= $ideas_KeyNumber} {incr i} {
     catch {bind .mesgTo.k$i <Alt-Key-$keyHotk> \
      	   "uam_shortcut2mesgTo $key"} errorM
     catch {bind .mesgTo.c$i <Alt-Key-$keyHotk> \
      	   "uam_shortcut2mesgTo $key"}
   }
 }

 update
}

#
# uam_newShortcut
#
#  Shows the right side of widget .shorts with empty entries;
#  allows the user to define a new shortcut.
#
# Arguments: None.
#
# Results: None.
#
proc uam_newShortcut {} {
global uam_VarShortName uam_VarShortHotk
global uam_VarShortKey uam_VarShortCont
global ideas_KeyNumber

 # show right frame in shorts
 #
# raise .shorts
 pack .shorts.f2 -side left -fill both

 # clear the entry widgets
 #
 set uam_VarShortName ""
 set uam_VarShortHotk ""

 set uam_VarShortCont(1) ""

 # attention: uam_VarShortKey(1) is :TYPE - do not edit
 
 set keyNumber [expr $ideas_KeyNumber-2]

 for {set i 2} {$i <= $keyNumber} {incr i} {
   set uam_VarShortKey($i) ""
   set uam_VarShortCont($i) ""
 }

 focus .shorts.nameEnt
}


#
# uam_editShortcut
#
#  Shows right frame of .shorts with entries filled with the
#  shortcut current selected in listbox .shorts, empty if no
#  selection is made.
#
# Arguments: None.
#
# Results: None.
#
proc uam_editShortcut {} {
global uam_VarShortName uam_VarShortHotk
global uam_VarShortKey uam_VarShortCont
global uam_Shortcuts


 # get only the first of the selected entries
 #
 set index [.shorts.name curselection]

 # if no selection is made handle as new shortcut
 #
 if {$index == ""} {
   set index [lindex [.shorts.hotk curselection] 0]
   if {$index == ""} {
     uam_newShortcut
     return
   }
 }

 # get name of the shortcut and fill the entries on the right side
 #
 set name [.shorts.name get $index]

 set uam_VarShortName [string trimright [csubstr $name 0 20]]
 keylget uam_Shortcuts($uam_VarShortName) hotkey uam_VarShortHotk 

 keylget uam_Shortcuts($uam_VarShortName) content list

 set i 1

 while {$list != ""} { 
   set uam_VarShortKey($i) [lvarpop list]
   set uam_VarShortCont($i) [lvarpop list]
   incr i
 }
}


#
# uam_selectShortcut
#
#  Handles a invoke on button 'Select'.
#  Gets current selection of listbox .shorts and calls a proc
#  to insert this shortcut in widget .mesgTo
#
# Arguments: None.
#
# Results: None.
#
proc uam_selectShortcut {} {

 # get only the first of the selected entries
 #
 set index [.shorts.name curselection]

 # if no selection is made do nothing
 #
 if {$index == ""} {
   set index [lindex [.shorts.hotk curselection] 0]
   if {$index == ""} {
     return
   }
 }

 # get name of the selected shortcut
 #
 set name [string trimright [csubstr [.shorts.name get $index] 0 20]]

 uam_shortcut2mesgTo $name
}


#
# uam_shortcut2mesgTo 
#
#  Inserts a specified shortcut in widget .mesgTo.
#
# Arguments:
#
#  	name - name of shortcut to insert
#
# Results: None.
#
proc uam_shortcut2mesgTo {name} {
global uam_Shortcuts uam_VarKey uam_VarCont

 keylget uam_Shortcuts($name) content mesgForm

 set i 3

 while {$mesgForm != ""} {
   set key [lvarpop mesgForm]
   set uam_VarKey($i) $key

   set cont [lvarpop mesgForm]
   set uam_VarCont($i) $cont

   incr i
 }

# if {[winfo ismapped .shorts.f2]} {
#   raise .mesgTo
# }

 focus .mesgTo.c3

 update
}


#
# uam_removeShortcut
#
#  Removes shortcuts selected in listbox from listbox and 
#  global variable uam_Shortcuts.
#
# Arguments: None.
#
# Results: None.
#
proc uam_removeShortcut {} {
global uam_Shortcuts ideas_KeyNumber

 set indexList [.shorts.name curselection]

 # if no selection is made do nothing
 #
 if {$indexList == ""} {
   set indexList [lindex [.shorts.hotk curselection] 0]
   if {$indexList == ""} {
     return
   }
 }

 while {$indexList != ""} {

   set index [lvarpop indexList end]

   set name [string trimright [csubstr [.shorts.name get $index] 0 20]]

   # remove a possibly existing binded hotkey to the shortcut
   #
   keylget uam_Shortcuts($name) hotkey keyHotk
   
   for {set i 1} {$i <= $ideas_KeyNumber} {incr i} {
     catch {bind .mesgTo.k$i <Alt-Key-$keyHotk> {}}
     catch {bind .mesgTo.c$i <Alt-Key-$keyHotk> {}}
   }
  
   .shorts.name delete $index
   .shorts.hotk delete $index
  
   unset uam_Shortcuts($name)
 }

}


#
# uam_addNewShortcut
#
#  Adds a specified shortcut to global variable uam_Shortcut,
#  inserts it in the listbox .shorts and bind it to .mesgTo
#
# Arguments: None.
#
# Results: None.
#
proc uam_addNewShortcut {} {
global uam_VarShortName uam_VarShortHotk
global uam_VarShortKey uam_VarShortCont
global uam_Shortcuts ideas_KeyNumber

 # the new shortcut must have a name
 #
 if {$uam_VarShortName == ""} {
   focus .shorts.nameEnt
   BEEP
   return
 }

 # bound the name of the new shortcut to limits:
 # there must be no space in it and its length must be lower 20
 #
 set uam_VarShortName [csubstr $uam_VarShortName  \
     0 [string first " " $uam_VarShortName]]
 set uam_VarShortName [csubstr $uam_VarShortName 0 20]

 # check if there is no binding for the specified hotkey on another
 # shortcut name
 #
 set bindList [bind .mesgTo.c3]

 if {[set index [lsearch $bindList <Alt-Key-$uam_VarShortHotk>]] != -1} {

   if {![cequal [bind .mesgTo.c3 [lindex $bindList $index]] \
               "uam_shortcut2mesgTo $uam_VarShortName"]} {
   focus .shorts.hotkEnt
   BEEP
   return
   }
 }

 set keyNumber [expr $ideas_KeyNumber-2]

 # read out the elements of the new shortcut ;
 # built up a list mesgForm
 #
 for {set i 1} {$i <= $keyNumber} {incr i} {
   if {$uam_VarShortKey($i) != ""} {
     if {![cequal [cindex $uam_VarShortKey($i) 0] ":"]} {
       set uam_VarShortKey($i) [string toupper :$uam_VarShortKey($i)]
     } else {
       set uam_VarShortKey($i) [string toupper $uam_VarShortKey($i)]
     }
   }

   lappend mesgForm $uam_VarShortKey($i) $uam_VarShortCont($i)
 }


 # built up the listbox entry for the shortcut
 #
 set nameLen [clength $uam_VarShortName]


 # remove a possible existing binded hotkey to a shortcut
 # with this name
 #
 catch {keylget uam_Shortcuts($uam_VarShortName) hotkey keyHotk}

 for {set i 1} {$i <= $ideas_KeyNumber} {incr i} {
   catch {bind .mesgTo.k$i <Alt-Key-$keyHotk> {}}
   catch {bind .mesgTo.c$i <Alt-Key-$keyHotk> {}}
 }
 
 # register the new shortcut in global uam_Shortcuts in this
 # special format:
 #
 keylset uam_Shortcuts($uam_VarShortName) hotkey $uam_VarShortHotk
 keylset uam_Shortcuts($uam_VarShortName) content $mesgForm


 # update the .shorts listbox
 #
 uam_fillShortcutLb

 focus .shorts.nameEnt

}


proc uam_shortcutResetFocus {} {
 
 if {[winfo ismapped .mesgTo]} {
   focus .mesgTo.c3
 } else {
   focus .mbar
 }
}

