#
# Copyright (c) 1995 by Matthias Klusch, Olaf Scheew, Bernd Grossmann
# Computer Science Department,
# Christian-Albrechts-University of Kiel,
# Olshausenstr. 40, 24118 Kiel, Germany
#
# All rights reserved.
# No warranties will be given on any issues arising out of the use
# of this software product.
# Permission to use, copy, modify and distribute this software
# product for non-commercial purposes is hereby granted, provided
# that the above copyright notice appears in all copies and
# respective publications.
# All commercial trades with this product outside the CAU Kiel
# without specific written prior permission are prohibited.
#

#
# uam_startAgent
#
#  Starts a new agent of a given type with a given name.
#
# Arguments:
#
#	agtName - name of new agent
#	agtType - name of type file to use
#
# Results:
#
#   Shows a message of the result in UAM logfile.
#
proc uam_startAgent {agtName agtType} {
global agents env ideas_HomePath uam_VarCompileAdds

 # get the Action-C-Source specification from the agents type file
 #
 if {[catch {set actSrcSpecif [uam_getActSrcFile $agtType]} errorMesg]} {
   uam_logText "ADL Syntax Error - ACTION-C-SOURCE:\n$errorMesg\n"
   uam_actSrcError $agtType
   return
 }

 set actSrcFile [lindex $actSrcSpecif 0]

 set compileAdds [lrange $actSrcSpecif 1 end]

 # change the working directory to
 # $ideas_HomePath/sources
 #
 set save_path [pwd]
 cd sources

 # compile the Action-C-Source if necessary:
 #
 if {[catch {uam_compileActSrc $actSrcFile $compileAdds}]} {
   uam_actSrcError $agtType
   cd $save_path
   return
 }

 cd $save_path

 # initialize new agents array entry
 #
 append agtName ".$env(HOST)"
 keylset agents($agtName) NAME $agtName \
         TYPE $agtType STATUS deact SOURCE $actSrcFile \
	 HOMEPATH $ideas_HomePath XDEFAULT [keylget agents(UAM) XDEFAULT]
   
 # create new agents process
 #
 set newFile [open "| sources/[file rootname $actSrcFile] -f agent.tcl \
                    $agents($agtName)" r+]

 set newPid [pid $newFile]
 keyldel agents($agtName) HOMEPATH
 keylset agents($agtName) PID $newPid FID $newFile

 uam_logText "Created: $agents($agtName)\n"
  
 # set communication control on pipe to/from new agent
 #
 fileevent $newFile readable "uam_commCtrl $newFile $agtName"

}

