#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/signal.h>
#include <sys/times.h>
#include <thread.h>
#include <assert.h>
#include <string.h>
#include "global.h"
#include "bboard.h"
#include "rpc.h"

#define THREADS 1
#define STACKSIZE 0

#define PREP_CALL(Goal) \
  { register term xref; \
    register cell xval; int arity; \
                  IF_OVER("PREP_CALL",H,HeapStk,"heap overflow"); \
                  FDEREF(Goal); \
                  arity=GETARITY(xval); \
                  P=(instr)hget(g.predmark,(no)(PUTARITY(xval,arity+1))); \
                  if(!P) \
                    { local_error(xval, \
                       "metacall to unknown predicate in recursive WAM\n", wam); \
                       return 0; \
                    } \
                  SETREF(regs+arity+1,g.true); \
                  COPY_CELLS(regs,xref,arity); \
  }

#define PUSH(top,a) (*++(top)=(cell)(a))
#define POP(top) (*(top)--)
#define NOTEMPTY(top,base) ((top)>(term)(base))
#define BIND(Var,Ref,Val) \
{SETREF(Var,COMPOUND(Val)?(cell)Ref:Val); TRAIL_IT(Var);}
#define VALID(v) ((v) && (cell)(v) != g.empty)

extern struct specsyms g;
extern string *atomtable,sbuf,stop;
extern term *unwind_trail();
extern struct limit max;
extern cell new_func();

extern bp();

pool_t *bb_pool_address = NULL;

struct {
  mutex_t mutex;
  wamprocess_t *process;
} allprocesses;


CLIENT * get_bbclient( bb)
 /* register*/ remotebb_t *bb;
/*************************************************************
 Gets a free client from the blackboard client list. If there
 is no free client, a new client is created and added to
 the list.
 Memory for list nodes is obtained with ordinary malloc !
 There is always at least 1 node in the list.
 *************************************************************/
{
#if 1
  clnt_node_t *cl_node;
  register int alreadyowner; /* for use of bb(un)lock */
  
  cl_node = bb->cl_list;
  bblock( bb);
  while (cl_node->inuse) { 
    if (cl_node->next != NULL)
      cl_node = cl_node->next;
    else break;
  }
  
  if ( cl_node->inuse) /* no free client in list */
  {
    cl_node->next = (clnt_node_t *) malloc( sizeof(clnt_node_t));
    cl_node = cl_node->next;
    cl_node->inuse = TRUE;
    cl_node->next = NULL;
    bbunlock(bb);
    if (!(cl_node->cl = clnt_create(bb->host,
                                    bb->prognum, BBVERS,"tcp"))) {
      printf("Creation of client in get_bbclient() has failed.\n");
      return NULL; /* error */
    }
  } 
  else /* free client has been found */
  {
    cl_node->inuse = TRUE;
    bbunlock(bb);
  }
  return cl_node->cl;  
#else
  return clnt_create(bb->host, bb->prognum, BBVERS,"tcp");
#endif
}


int return_bbclient( bb, client)
  remotebb_t *bb;
  CLIENT     *client;
/*************************************************************
 Makes the used client available for other remote processes.
 There is always at least 1 node in the list.
 *************************************************************/
{
#if 1
  clnt_node_t *cl_node;
  register int alreadyowner; /* for use of bb(un)lock */
  
  cl_node = bb->cl_list;
  bblock(bb);
  while (cl_node->cl != client) { 
    if (cl_node->next != NULL)
      cl_node = cl_node->next;
    else {
      printf("Error. Client is vanished in the deep depths");
      printf(" of memory...\n");
      return 1; /* error */
    }
  }
  cl_node->inuse = FALSE;
  bbunlock(bb);
  return 0; /* OK */
#else
  clnt_destroy(client);
#endif
}


int remove_bbclient( bb)
  remotebb_t *bb;
/*************************************************************
 Removes the list of clients out of memory.
 Blackboard should already be locked.
 Normally there is always at least one node in the list.
 *************************************************************/
{
  clnt_node_t *cl_node, *old_cl_node;
  
  cl_node = bb->cl_list;
  do {
    clnt_destroy(cl_node->cl);
    old_cl_node = cl_node; 
    cl_node = cl_node->next;
    free(old_cl_node);
    } while (cl_node != NULL);
  return 0;  
}


localbb_t *createlocalbb(name)
  register cell name;
/***************************************************
  allocates and initializes a blackboard of standard 
  size (BUDDYSIZE) from the shared memory pool 
***************************************************/
{
  register localbb_t *bb;
  register pool_t *pool = (pool_t *)hget(g.bbmark,g.bbmark);

  assert(pool);
  mutex_lock(&(pool->mutex));
  bb = (localbb_t *)buddymalloc(pool->buddy, BLACKBOARDSIZE);
  bb->nextbb = pool->blackboards;
  pool->blackboards = (blackboard_t *)bb;
  bb->type = LOCALBB;
  bb->owner = 0;
  bb->lockno = 1;
  strncpy(bb->name,NAME(name),NAMESIZE);
  bb->name[NAMESIZE-1] = 0; /* 0-terminate */
  mutex_init(&(bb->mutex),USYNC_THREAD,NULL);
  mutex_lock(&(bb->mutex));
  mutex_unlock(&(pool->mutex));
  /* variable part */
  buddycreate(bb->buddy, BUDDYSIZE);
  buddyinit(bb->buddy, 
            (char *)bb+realbuddysize(sizeof(localbb_t)), 
            BLACKBOARDSIZE-realbuddysize(sizeof(localbb_t)));
  bb->firstterm = (termnode_t *)NULL;
  bb->lastterm = (termnode_t *)NULL;
  bb->firstprocess = (processnode_t *)NULL;
  bb->lastprocess = (processnode_t *)NULL;
  bb->timestamp = (term)INPUT_INT(0);
  mutex_unlock(&(bb->mutex));
  return bb;
}

void cleanlocalbb(bb)
register localbb_t *bb;
/********************************************************
   auxiliary routine that removes the term list, 
   kills pending processes, and releases their
   resources; bb should be a valid blackboard and locked 
 ********************************************************/
{
  register termnode_t *tp;
  register processnode_t *pp;

  assert(bb);
  assert(bb->type == LOCALBB);
  tp = bb->firstterm;
  bb->firstterm = (termnode_t *)NULL;
  bb->lastterm = (termnode_t *)NULL;
  while (tp) {  /* free all memory in the term list */
    register termnode_t *ttp = tp->next;
    buddyfree(bb->buddy, tp, tp->size);
    tp = ttp;
  }
  pp = bb->firstprocess; 
  bb->firstprocess = (processnode_t *)NULL;
  bb->lastprocess = (processnode_t *)NULL;
  while (pp) { /* free all memory in processes list */
    register processnode_t *tpp = pp->next;
    thr_kill(pp->thread,SIGUSR1); /* first kill thread */
    thr_continue(pp->thread); /* to guarantee the processing of SIGUSR1 */
    buddyfree(bb->buddy, pp, pp->size);
    pp = tpp;
  }
  bb->timestamp = (term)INPUT_INT(0); /* reset timestamp counter */
}

int bbclean(bb)
register blackboard_t *bb;
/****************************************************
  empties the blackboard bb by removing all
  the terms, and all the pending processes;
  return 1 when successful, 0 when not.
 ***************************************************/
{
  assert(bb);

  if (bb->type == LOCALBB) {
    register int alreadyowner;
    bblock(bb); 
    cleanlocalbb((localbb_t *)bb);
    bbunlock(bb);
    return 1;
  } else {
    return rpc_bbclean(bb);
  }
}

int deletelocalbb(pool,bb)
register pool_t *pool;
register localbb_t *bb;
/***************************************************
  deletes all the terms and processes from the
  blackboard, and then deletes the blackboard itself
  and returns all the resources to the pool.
  assumes that pool is locked
 *************************************************/
{
  register int bi, alreadyowner;

  assert(bb);
  assert(bb->type == LOCALBB);
  bblock(bb);
  /* free all memory in the term and process list here */
  cleanlocalbb(bb);
  /* return all the remaining blocks to shared memory */
  for (bi=2; bi<BUDDYMAX; bi++)
    while (bb->buddy[bi]) { 
      register buddyblock_t *bp = bb->buddy[bi]->next;
      buddyfree(pool->buddy, bb->buddy[bi], 1 << bi);
      bb->buddy[bi] = bp;
    }
  /* free the buddy system */
  mutex_destroy(&(bb->mutex)); 
  buddyfree(pool->buddy, bb, sizeof(localbb_t));
  return 1;
}

int deleteremotebb(pool,bb)
register pool_t *pool;
register remotebb_t *bb;
/************************************************
  removes alias of remote blackboard.
  assumes that pool is locked
 *************************************************/
{ 
  clnt_node_t *cl_node1, *cl_node2;

  assert(bb);
  assert(bb->type == REMOTEBB);
  mutex_lock(&(bb->mutex));
  cl_node1 = bb->cl_list;
  cl_node2 = NULL;
  while ( cl_node1) {
    cl_node2 = cl_node1;
    cl_node1 = cl_node1->next;
    clnt_destroy(cl_node2->cl);
    free(cl_node2);
  }  
  mutex_destroy(&(bb->mutex));
  buddyfree(pool->buddy,bb,sizeof(remotebb_t)); 
  return 1;
}

int deleteblackboard(pool,bb)
register pool_t *pool;
register blackboard_t *bb;
/*********************************************
  delete blackboard, and return resources to
  the memory pool. Assume pool to be locked
 ********************************************/
{
  register blackboard_t *bp, *bq;
  register int result = 0;

  assert(pool);
  /* remove blackboard from blackboard list */
  bq = (blackboard_t *)NULL;
  bp = pool->blackboards;
  while (bp != bb && bp) {
    bq = bp; bp = bp->nextbb;
  }
  if (bp) {
    if (bq) 
      bq->nextbb = bp->nextbb;
    else
      pool->blackboards = bp->nextbb;
    /* remove blackboard */
    if (bb->type == LOCALBB) 
      result = deletelocalbb(pool,(localbb_t *)bb);
    else
      result = deleteremotebb(pool,(remotebb_t *)bb);
  } 
  return result;
}

int printable(l)
long l;
/* checks whether the 32-bit bit pattern of l can be displayed
   as string */
{
  register char *p = (char *)&l;
  return ('a' <= p[0] && p[0] <= 'z' &&  p[0]) &&
         ('a' <= p[1] && p[1] <= 'z' || !p[1]) &&
         ('a' <= p[2] && p[2] <= 'z' || !p[2]) &&
         ('a' <= p[3] && p[3] <= 'z' || !p[3]);
}

term blackboards(H,t)
register term H, *t;
/**************************************
   this routine puts all available blackboards
   into a prolog list (t).
 **************************************/
{
  register pool_t *pool = (pool_t *)hget(g.bbmark,g.bbmark);
  register blackboard_t *bb;

  assert(pool);
  mutex_lock(&(pool->mutex));
  bb = pool->blackboards; 
  *t = H;
  while (bb) {
    PUSH_LIST(new_func(bb->name,0));
    bb = bb->nextbb;
  }
  PUSH_NIL();
  mutex_unlock(&(pool->mutex));
  return H;
}

term bbmem(bb,H,t)
register term H, *t;
register localbb_t *bb;
/**************************************
   this routine puts the number of non-empty buddies
   of blackboard bb into a prolog list (t).
 **************************************/
{
  register int bi;
  register cell buddy2 = new_func("buddy",2);
  register term beginoflist;
  register int alreadyowner;

  assert(bb);
  bblock(bb);
  beginoflist = H;
  PUSH_NIL();
  for (bi=BUDDYMAX-1; bi>0; bi--) {
    if (bb->buddy[bi]) {
      register term buddyterm = H;
      PUSHVAL(buddy2);
      PUSHVAL(INPUT_INT(listlength(bb->buddy[bi])));
      PUSHVAL(INPUT_INT(1<<bi));
      PUSH_LIST(T2C(buddyterm));
      PUSHVAL(T2C(beginoflist));
      beginoflist = H-3;
    }
  }
  *t = beginoflist;
  bbunlock(bb);
  return H;
}

int bbmalloc(bb, blocksize)
register localbb_t *bb;
register unsigned int blocksize;
/***********************************
  this routines adds a block of blocksize
  to the buddy system of the blackboard bb
 **********************************/
{
  register pool_t *pool = (pool_t *)hget(g.bbmark,g.bbmark);
  register void *p;
  register int n = blocksize, alreadyowner;

  assert(pool);
  assert(bb);
  assert(bb->type == LOCALBB);
   
  /* blocksize must be a power of 2 */
  if (!n) return 0;
  while (!(n&1)) n >>= 1; 
  if (n != 1) return 0;

  mutex_lock(&(pool->mutex));
  bblock(bb);
  p = buddymalloc(pool->buddy, blocksize);
  mutex_unlock(&(pool->mutex));
  if (p) buddyinit(bb->buddy, p, blocksize);
  bbunlock(bb);
  return p != NULL;
}

int bbfree(bb, blocksize)
register localbb_t *bb;
register unsigned int blocksize;
/***********************************
  this routines removes a block of blocksize
  from the buddy system of the blackboard bb
  and returns it to the memory pool
 **********************************/
{
  register pool_t *pool = (pool_t *)hget(g.bbmark,g.bbmark);
  register void *p;
  register int n = blocksize, alreadyowner;

  assert(pool);
  assert(bb);
  assert(bb->type == LOCALBB);
   
  /* blocksize must be a power of 2 */
  if (!n) return 0;
  while (!(n&1)) n >>= 1; 
  if (n != 1) return 0;

  mutex_lock(&(pool->mutex));
  bblock(bb);
  p = buddymalloc(bb->buddy, blocksize);
  bbunlock(bb);
  if (p) buddyfree(pool->buddy, p, blocksize);
  mutex_unlock(&(pool->mutex));
  return p != NULL;
}

term poolmem(pool,H,t)
register term H, *t;
register pool_t *pool;
/**************************************
   this routine puts all non-empty buddies of
   the memory pool into a prolog list (t).
 **************************************/
{
  register int bi;
  register cell buddy2 = new_func("buddy",2);
  register term beginoflist;

  assert(pool);
  mutex_lock(&(pool->mutex));
  beginoflist = H;
  PUSH_NIL();
  for (bi=BUDDYMAX-1; bi>0; bi--) {
    if (pool->buddy[bi]) {
      register term buddyterm = H;
      PUSHVAL(buddy2);
      PUSHVAL(INPUT_INT(listlength(pool->buddy[bi])));
      PUSHVAL(INPUT_INT(1<<bi));
      PUSH_LIST(T2C(buddyterm));
      PUSHVAL(T2C(beginoflist));
      beginoflist = H-3;
    }
  }
  *t = beginoflist;
  mutex_unlock(&(pool->mutex));
  return H;
}

term my_recursive_copytobb(to, from, firstfree, wam)
register term from, to, firstfree;
register stack wam;
/**************************************************
  routine that recursively copies the term 'from'
  to the location 'to', and with 'firstfree'
  the first available location
**************************************************/
{  
  register cell val_t;

  /* I should check here fore duplication of bb terms */
  if (VAR(from)) DEREF2(from,val_t) else val_t = (cell)from;
  if(VAR(val_t)) {
    if (ON_BB(val_t)) {
      SETREF(to,val_t);
    } else {
      SETREF(to,to);
      SETREF(from,to);
      TRAIL_IT(from);
    }
  } else if(INTEGER(val_t) || !GETARITY(val_t)) {
    SETREF(to,(term)val_t);
  } else {
    register int i = GETARITY(val_t);
    SETREF(to,firstfree);
    to = firstfree;
    SETREF(to, (term)val_t);
    firstfree = to+i; 
    firstfree = my_recursive_copytobb(firstfree,(term)(from+i),firstfree,wam); 
    if (firstfree == to+i) firstfree++;
    while(--i) {
      firstfree = my_recursive_copytobb((term)(to+i),(term)from[i],firstfree,wam);
    }
  }
  return firstfree;
}

term my_recursive_copyfrombb(to, from, firstfree, wam)
register term from, to, firstfree;
register stack wam;
/*******************************************
  same as copytobb, but in the inverse direction
 ******************************************/
{  
  register cell val_t;

  if (VAR(from)) DEREF2(from,val_t) else val_t = (cell)from;
  if(VAR(val_t)) { 
    if (ON_BB(val_t)) { 
      SETREF(to,to);
      SETREF(from,to);
      TRAIL_IT(from);
    } else { 
      SETREF(to,val_t);
    }
  } else if(INTEGER(val_t) || !GETARITY(val_t)) {
    SETREF(to,(term)val_t);
  } else {
    register int i = GETARITY(val_t);
    SETREF(to,firstfree);
    to = firstfree;
    SETREF(to, (term)val_t);
    firstfree = to+i;
    firstfree = my_recursive_copyfrombb(firstfree,(term)(from+i),firstfree,wam); 
    if (firstfree == to+i) firstfree++;
    while(--i) {
      firstfree = my_recursive_copyfrombb((term)(to+i),(term)from[i],firstfree,wam);
    }
  }
  return firstfree;
}

void copytobb(bb,t,wam)
register term bb,t;
register stack wam;
/******************************
  routine that copies the term t onto the
  blackboard bb
 ******************************/
{
  register term *bakTR=TR_TOP;
  (void)my_recursive_copytobb(bb,t,bb,wam);
  TR_TOP=unwind_trail(TR_TOP,bakTR);
}

term copyfrombb(t,bb,wam)
register term t,bb;
register stack wam;
/**************************
  routine that copies the term 'bb' from
  the blackboard into t
 *************************/
{
  register term *bakTR=TR_TOP;
  t = my_recursive_copyfrombb(t,bb,t,wam);
  TR_TOP=unwind_trail(TR_TOP,bakTR);
  return t;
}

term *cleantrail(lowtrail, hightrail, H, wam)
term *lowtrail, *hightrail, *H;
register stack wam;
/*******************************************
   routine to clean up the temporary trail
   crated by bbunify
 ******************************************/
{
  register term *tr, *trq;

  for (tr=lowtrail; tr<hightrail; tr++) {
    if (!ON_BB(*tr) && ON_BB(**tr)) 
      *H = my_recursive_copyfrombb((term)*tr,(term)**tr,*H,wam);
  } 
  unwind_trail(TR_TOP,hightrail); 
  for (trq=tr=lowtrail; tr<hightrail; tr++) {
    if (ON_BB(*tr))
      **tr = (cell)*tr;
    else 
      *trq++ = *tr;
  }
  return trq;
}

no bbunify(t1,t2,wam,A,H,status)
register term t1,t2; register stack wam; register term *A, *H;
register int *status;
/*******************************************
  unification of the terms t1 and t2
  status = 1 if t1 gets instatiated
 ******************************************/
{
  register term U=(term)A; /* top of push down list */
  register v1 = (cell)t1, v2 = (cell)t2;
  term *bakTR=TR_TOP;
  register int answerneeded = 0;

  PUSH(U,v2); PUSH(U,v1);
  while(NOTEMPTY(U,A))
  { 
   register term t1=U--,t2=U--;
   DEREF2(t1,v1); DEREF2(t2,v2);
   if(t1!=t2)  /* var on bb */
     {   
      if(VAR(v1)) /* unb. var. v1 */
      {  
        answerneeded = 1;
        if(VAR(v2)) /*unb. var. v2 */
          {
            SETREF((term)v2,v1);
            TRAIL_IT((term)v2);
          }
        else
          BIND((term)v1,t2,v2)
      }  
      else if(VAR(v2)) { /* v1 is NONVAR */
        BIND((term)v2,t1,v1)  /* this requires copying */
      } 
      else if(v1!=v2) { /* both are NONVAR */
        TR_TOP=unwind_trail(TR_TOP,bakTR);
        *status = 0;
        return FALSE;
      } 
      else if(IDENTIFIER(v1) && (v1=GETARITY(v1)) )
        { /* they have the same FUNCTOR, v1==v2 */
          IF_OVER("bb_unify",U,ChoiceStk,"unification overflow");
 
          PUSH(U,t2+v1); PUSH(U,t1+v1);
          while(--v1)
            {
              PUSH(U,t2[v1]);
              PUSH(U,t1[v1]);
            }
        }
     }
  }
  if (*status)
    TR_TOP = cleantrail(bakTR,TR_TOP,H,wam);
  *status = answerneeded;
  return TRUE;
}  

term my_recursive_copytobuffer(to, from, firstfree, wam, buffer, len)
register term from, to, firstfree, buffer;
register stack wam;
register int len;
/************************************************
   same as copytobb but now to a buffer
 ***********************************************/
{  
  register cell val_t;

  if (VAR(from)) DEREF2(from,val_t) else val_t = (cell)from;
  if(VAR(val_t)) {
    if (val_t >= (cell)buffer && val_t < (cell)(buffer+len)) {
      SETREF(to,val_t);
    } else {
      SETREF(to,to);
      SETREF(from,to);
      TRAIL_IT(from);
    }
  } else if(INTEGER(val_t) || !GETARITY(val_t)) {
    SETREF(to,(term)val_t);
  } else {
    register int i = GETARITY(val_t);
    SETREF(to,firstfree);
    to = firstfree;
    SETREF(to, (term)val_t);
    firstfree = to+i;
    firstfree = my_recursive_copytobuffer(firstfree,(term)(from+i),firstfree,
                      wam,buffer,len); 
    if (firstfree == to+i) firstfree++;
    while(--i) {
      firstfree = my_recursive_copytobuffer((term)(to+i),(term)from[i],firstfree,
                      wam,buffer,len);
    }
  }
  return firstfree;
}


term copytobuffer(buf,t,wam,len)
register term buf,t;
register stack wam;
register int len;
/************************************
   copies a term into a buffer
 ***********************************/
{
  register term *bakTR=TR_TOP, firstfree;
  firstfree = my_recursive_copytobuffer(buf,t,buf,wam,buf,len);
  TR_TOP=unwind_trail(TR_TOP,bakTR);
  return firstfree;
}

int term_size(t)
register term t;
/*************************************
  calculates an upper bound for the term t
 *************************************/
{
  register cell val_t;

  if (VAR(t)) DEREF2(t,val_t) else val_t = (cell)t;
  if(VAR(val_t) || INTEGER(val_t) || !GETARITY(val_t)) {
    return sizeof(cell);
  } else {
    register int i = GETARITY(val_t), size = i*sizeof(cell);
    size += term_size((term)(t+i));
    while(--i) {
      register unsigned s = term_size((term)t[i]);
      if (s > sizeof(cell))
        size += s;
    }
    return size;
  }
}

term putterm(bb, wam, A, H, regs)
register localbb_t *bb;
register stack wam; 
register term *A;
term H;
register term regs; 
/************************************
  this routine puts a term on the local
  blackboard bb
 ************************************/
{
  register term t = (term)(regs[2]);
  register int size;
  register unsigned datasize;
  register termnode_t *tp;
  register processnode_t *pp, *pq = NULL;
  register term *bakTR=TR_TOP;
  register int alreadyowner;
  register term oldH = H;

  assert(bb);
  assert(bb->type == LOCALBB);
  bblock(bb);
  bb->timestamp += (1<<RBITS);

  pp = bb->firstprocess;
  while (pp) {
    int status = 0;
    register int cond;
    if (bbunify((term)pp->pattern,t,wam,A,&H,&status)) {
      register instr P;
      if(pp->condition == (term *)g.true) cond = 1;
      else {
#if 1
        H=(term)bp_prolog_call((pp->condition),regs,H,P,A,wam);
#else
        PREP_CALL((cell)(pp->condition));
        H = (term)bp(regs,H,P,A,wam);
#endif
        cond = (int)H;
      }
    } else cond = 0;
    if (cond) {
      if (pq) {
        pq->next = pp->next;
      } else {
        bb->firstprocess = pp->next;
      }
      if (pp == bb->lastprocess) bb->lastprocess = pq;
      if (status) {
        register int size = sizeof(termnode_t)-4+term_size(t);
        pp->answer  = (termnode_t *)buddymalloc(bb->buddy, size);
        copytobb((term)pp->answer->data, t, wam);
        pp->answer->size = size;
      }
      pp->timestamp = bb->timestamp;
      TR_TOP = unwind_trail(TR_TOP,bakTR);
      if (pp->modifier & DESTRUCTIVE) {
/*        thr_continue(pp->thread);  */
        cond_signal(&(pp->condvar));
        bbunlock(bb);
        return H;
      } else {
        register processnode_t *ppcont = pp;
        /* keep pq */
        pp = pp->next;
/*        thr_continue(ppcont->thread);  */
        cond_signal(&(ppcont->condvar));
      }
    } else {
      /* restore initial status */
      TR_TOP = unwind_trail(TR_TOP,bakTR);
      H = oldH;
      pq = pp;
      pp = pp->next;
    }
  }
  size = term_size(t);
  datasize = sizeof(termnode_t)-4+size;
  tp = (termnode_t *)buddymalloc(bb->buddy, datasize);
  if (!tp) {
    warnmes("Blackboard memory exhausted");
    bbunlock(bb);
    return 0;
  }
  tp->next = (termnode_t *)NULL;
  tp->size = datasize;
  copytobb((term)tp->data,t,wam);
  tp->timestamp = bb->timestamp;
  if (bb->firstterm) {
    bb->lastterm->next = tp;
    bb->lastterm = tp;
  } else {
    bb->firstterm = bb->lastterm = tp;
  }
  bbunlock(bb);
  return H;
}

term getterm(bb, wam, A, H, timestamp, regs)
register localbb_t *bb;
register stack wam; 
register term *A;
term H;
register term *timestamp;
register term regs;
/*********************************************
  this routine gets a term from blackboard bb
 *********************************************/
{
  register term t = (term)(regs[2]);
  register int modifier = (int)(regs[4]);
  register term condition = (term)(regs[3]);
  register termnode_t *tp, *tq;
  register int size;
  register unsigned datasize;
  register processnode_t *pp;
  register int alreadyowner;
  register term oldH = H;
  register term *bakTR;

  assert(bb);
  assert(bb->type == LOCALBB);
  bakTR = TR_TOP;
  bblock(bb);
  tq = (termnode_t *)NULL;
  tp = bb->firstterm;
  while (tp) {
    int status = 1;
    register int cond;
    if (*timestamp < tp->timestamp && 
        bbunify((term)tp->data,t,wam,A,&H,&status)) {
      if(condition == (term)g.true) cond = 1;
      else {
        register instr P;
#if 1
/*
puts("call condition");
fout(condition, wam, stdout);
*/
        H=(term)bp_prolog_call(condition,regs,H,P,A,wam);
#else
        PREP_CALL((term)condition);
        H = (term)bp(regs,H,P,A,wam);
#endif
        cond = (int)H;
      }
    } else cond = 0;
    if (cond) {
      *timestamp = tp->timestamp;
      if (modifier & DESTRUCTIVE) {
        if (tq) {
          tq->next = tp->next;
        } else {
          bb->firstterm = tp->next;
        }
        if (tp == bb->lastterm) bb->lastterm = tq;
        buddyfree(bb->buddy, tp, tp->size);
      } 
      bbunlock(bb);
      return H;
    } else {
      /* restore initial status */
      TR_TOP = unwind_trail(TR_TOP,bakTR);
      H = oldH;
    }
    tq = tp;
    tp = tp->next;
  }
  if (!(modifier & BLOCKING)) {
    bbunlock(bb);
    return 0;
  }
  H[0] = g.DOT;
  H[1] = (cell)t; 
  H[2] = (cell)condition; 
  size = term_size(H);
  datasize = sizeof(processnode_t)-4+size;
  pp = (processnode_t *)buddymalloc(bb->buddy, datasize);
  if (!pp) {
    warnmes("Blacboard memory exhausted");
    bbunlock(bb);
    return 0;
  }
  pp->next = (processnode_t *)NULL;
  pp->size = datasize;
  pp->thread = thr_self();
  cond_init(&(pp->condvar),USYNC_THREAD,NULL);
  pp->answer = (termnode_t *)NULL;
  pp->modifier = modifier;
  copytobb((term)pp->data,H,wam);

  pp->pattern = (term *)(pp->data[1]);
  pp->condition = (term *)&(pp->data[2]);
  if (bb->firstprocess) {
    bb->lastprocess->next = pp;
    bb->lastprocess = pp;
  } else {
    bb->firstprocess = bb->lastprocess = pp;
  }
  bb->owner = 0;
/*  mutex_unlock(&(bb->mutex)); 
  thr_suspend(thr_self()); 
  mutex_lock(&(bb->mutex)); */
  cond_wait(&(pp->condvar),&(bb->mutex));
  bb->owner = thr_self();
  *timestamp = pp->timestamp;  
  if (pp->answer) {
    int status = 1;
    if (!bbunify((term)pp->answer->data,t,wam,A,&H,&status)) {
      fprintf(stderr,"Blackboard error: answer field inconsistent");
      fout(t,wam,stdout); puts("");
      fout((term)pp->answer->data,wam,stdout); puts("");
      exit(1);
    }
    buddyfree(bb->buddy, pp->answer, pp->answer->size);
  }
  buddyfree(bb->buddy, pp, pp->size);
  bbunlock(bb);
  return H;
}

term bbterms(bb,H,t,wam)
register blackboard_t *bb;
register term H, *t;
register stack wam;
/******************************************
   returns the list of waiting terms in t
 *****************************************/
{
  assert(bb);
  if (bb->type == LOCALBB) {
    register termnode_t *tp;
    register int alreadyowner;

    *t = H;
    bblock(bb);
    tp = ((localbb_t *)bb)->firstterm;
    while (tp) {
      PUSH_LIST((T2C(tp->data)));
      tp = tp->next;
    }
    bbunlock(bb);
    PUSH_NIL();
    return H;
  } else 
    return rpc_bbterms(bb,H,t,wam);
}

term bbprocesses(bb,H,t,wam)
register blackboard_t *bb;
register term H, *t;
register stack wam;
/******************************************
   returns the list of waiting processes in t
 *****************************************/
{
  assert(bb);
  if (bb->type == LOCALBB) {
    register processnode_t *pp;
    register int alreadyowner;
    register term beginoflist = H;

    PUSH_NIL();
    bblock(bb);
    pp = ((localbb_t *)bb)->firstprocess;
    while (pp) {
      register term node = H;
  
      PUSHVAL(new_func(",",2));
      PUSHVAL(INPUT_INT(pp->thread));
      if (pp->modifier & DESTRUCTIVE) {
        if (pp->modifier & BLOCKING) {
          PUSHVAL(new_func("gett",2));
        } else {
          PUSHVAL(new_func("gett_nb",2)); /* can never occur */
        }
      } else {
        if (pp->modifier & BLOCKING) {
          PUSHVAL(new_func("readt",2));
        } else {
          PUSHVAL(new_func("readt_nb",2)); /* can never occur */
        }
      }
      PUSHVAL(T2C(pp->pattern));
      PUSHVAL(T2C(pp->condition));
      PUSH_LIST(T2C(node));
      PUSHVAL(T2C(beginoflist));
      beginoflist = H-3;
      pp = pp->next;
    }
    PUSH_NIL();
    bbunlock(bb);
    *t = beginoflist;
    return H;
  } else 
    return rpc_bbprocesses(bb,H,t,wam);
}

term processes(H,t)
register term H, *t;
/**************************************
   this routine puts all available thread ids
   into a prolog list (t).
 **************************************/
{
 
  wamprocess_t *wp;
  *t = H;
  mutex_lock(&(allprocesses.mutex));
  wp  = allprocesses.process;
  while (wp) {
    PUSH_LIST(INPUT_INT(wp->thread));
    wp = wp->next;
  }
  mutex_unlock(&(allprocesses.mutex));
  PUSH_NIL();
  return H;
}


term bbinfo(bb,H,t)
register blackboard_t *bb;
register term H, *t;
/******************************************
   returns information aboutn bb  in t
 *****************************************/
{
  register term nameinfo; 
  register int alreadyowner;

  assert(bb);
  nameinfo = H;
  PUSHVAL(new_func("name",1));
  bblock(bb);
  PUSHVAL(new_func(bb->name,0));
  if (bb->type == LOCALBB) {
    register term addressinfo, freeinfo, stampinfo;
    char buffer[20];

    addressinfo = H;
    sprintf(buffer, "%08x", bb);
    PUSHVAL(new_func("address",1));
    PUSHVAL(new_func(buffer, 0));
    freeinfo = H;
    PUSHVAL(new_func("free_memory",1));
    PUSHVAL(INPUT_INT(totalbuddymem(((localbb_t *)bb)->buddy)));
    stampinfo = H;
    PUSHVAL(new_func("time_stamp", 1));
    PUSHVAL((cell)(((localbb_t *)bb)->timestamp)) /* already internalized */;
    bbunlock(bb);
    *t = H;
    
    PUSH_LIST(T2C(nameinfo));
    PUSH_LIST(T2C(addressinfo));
    PUSH_LIST(T2C(freeinfo));
    PUSH_LIST(T2C(stampinfo));
    PUSH_NIL();
  } else {
  
    register term remotebbinfo, hostinfo, serverinfo;

    remotebbinfo = H;
    PUSHVAL(new_func("blackboard",1));
    PUSHVAL(((remotebb_t *)bb)->bboard);
    serverinfo = H;
    PUSHVAL(new_func("server",1));
    if (printable(((remotebb_t *)bb)->prognum)) {
      long l[2] = {0, 0};
      l[0] = ((remotebb_t *)bb)->prognum;
      PUSHVAL(new_func((char*)l, 0));
    } else
      PUSHVAL(INPUT_INT(((remotebb_t *)bb)->prognum));
    hostinfo = H;
    PUSHVAL(new_func("host",1));
    PUSHVAL(new_func(((remotebb_t *)bb)->host,0));

    bbunlock(bb); 
    *t = H;
    PUSH_LIST(T2C(nameinfo));
    PUSH_LIST(T2C(hostinfo));
    PUSH_LIST(T2C(serverinfo));
    PUSH_LIST(T2C(remotebbinfo));
    PUSH_NIL();
  }
  return H;
}

term processinfo(thread,H,t)
register thread_t thread;
register term H, *t;
/*********************************
  return info about thread in t
 ********************************/
{
  register wamprocess_t *wp;

  mutex_lock(&(allprocesses.mutex));
  wp = allprocesses.process;
  while (wp) {
    if (wp->thread == thread) {
      register term procinfo = H;

      PUSHVAL(new_func("process",2));
      PUSHVAL(INPUT_INT(wp->thread));
      PUSHVAL((cell)(wp->data));
      mutex_unlock(&(allprocesses.mutex));
      *t = H;
      PUSH_LIST(T2C(procinfo));
      PUSH_NIL();
      return H;
    } else 
      wp = wp->next;
  }
  mutex_unlock(&(allprocesses.mutex));
  return 0;
}

remotebb_t *createremotebb(client,name,host,prognum,bboard)
  register CLIENT *client;
  register u_long prognum;
  register cell name;
  register char *host;
  register cell bboard;
/***************************************
  create remote blackboard description and
  link it into the list of blackboards
 **************************************/
{ 
  if (rpc_bbcheck(client,NAME(bboard))) {
    register CLIENT *cp; 
    register remotebb_t *bb;
    register pool_t *pool = (pool_t *)hget(g.bbmark, g.bbmark);

    assert(pool);
    mutex_lock(&(pool->mutex));
    bb = (remotebb_t *)buddymalloc(pool->buddy,sizeof(remotebb_t));
    bb->nextbb = pool->blackboards;
    pool->blackboards = (blackboard_t *)bb;
    bb->type = REMOTEBB;
    mutex_init(&(bb->mutex),USYNC_THREAD,NULL);
    bb->owner = 0;
    bb->lockno = 1;
    strncpy(bb->name,NAME(name),NAMESIZE);
    bb->name[NAMESIZE-1] = 0; /* 0-terminate */
    mutex_unlock(&(pool->mutex));
    strcpy(bb->host,host);
    bb->bboard = bboard;
    bb->prognum = prognum;
    bb->cl_list = (clnt_node_t *) malloc( sizeof(clnt_node_t));
    bb->cl_list->inuse = FALSE;
    bb->cl_list->cl = client;
    bb->cl_list->next = NULL;
    mutex_unlock(&(bb->mutex));
    return bb;
  } else 
    return (remotebb_t *)NULL;
}


int make_bb_pool()
/* initialize variables */
{
  bb_pool_address = (pool_t *)malloc(BB_POOL_SIZE);
  hdef(g.bbmark, g.bbmark, bb_pool_address, g.timestamp);
  return (int)bb_pool_address;
}

void pool_init(pool,size)
register pool_t *pool;
register unsigned size;
/*************************************
Initialise pool data structures.
**************************************/
{
  assert(pool);
  buddycreate(pool->buddy,BUDDYSIZE);
  buddyinit(pool->buddy, (char *)pool+sizeof(pool_t),size-sizeof(pool_t));
  pool->blackboards = (blackboard_t *)NULL;
  mutex_init(&(pool->mutex),USYNC_THREAD,NULL);
  mutex_unlock(&(pool->mutex));
}

void delete_pool(pool)
register pool_t *pool;
/********************************
  returns all memory allocated for
  blackboard operations to the 
  memory pool, removes the mutex
  on the shared memory pool.
**********************************/
{
  register blackboard_t *bb;

  assert(pool);
  mutex_lock(&(pool->mutex));
  bb = pool->blackboards;
  while (bb) {
    deleteblackboard(pool,bb);
    bb = bb->nextbb;
  }
  mutex_destroy(&(pool->mutex));
}

int init_small_mem(wam)
register stack wam;
/****************************************
   creates separate data structures
   for the trail, heap and choice point stack
 ****************************************/
{
  register int errctr = 0;
/* make_stack gedefinieerd in ru.c */
  errctr+=!make_stack(&(wam[TrailStk]),max.TRAIL,2,"trail",NULL,1);
  errctr+=!make_stack(&(wam[HeapStk]),max.HEAP,2000,"heap",NULL,1);
  errctr+=!make_stack(&(wam[ChoiceStk]),max.CHOICE,MAXARITY+3,"choice",NULL,1);
  if(errctr)
    ERREXIT("not enough memory.")
  else 
    return 1;
}
 
void delete_small_mem(wam)
register stack wam;
/******************************
  frees the memory allocated by
  init_small_mem
 ****************************/
{
  free(wam[TrailStk].base);
  free(wam[HeapStk].base);
  free(wam[ChoiceStk].base);
}

void bb_save_orig()
{
  register pool_t *pool = (pool_t *)hget(g.bbmark, g.bbmark);

  assert(pool);
  pool_init(pool,BB_POOL_SIZE);
  thr_setconcurrency(6);
  hdef(g.bbmark, g.stdbb, createlocalbb(g.stdbb),g.timestamp);
  mutex_init(&(allprocesses.mutex),USYNC_THREAD,NULL);
  allprocesses.process = (wamprocess_t *)NULL;
  mutex_unlock(&(allprocesses.mutex));
}

void bb_restart_orig()
{
  register pool_t *pool = (pool_t *)hget(g.bbmark, g.bbmark);
  register wamprocess_t *wp;

  assert(pool);
  pool_init(pool,BB_POOL_SIZE);
  hdef(g.bbmark, g.stdbb, createlocalbb(g.stdbb), g.timestamp);
  mutex_lock(&(allprocesses.mutex));
  wp = allprocesses.process;
  mutex_unlock(&(allprocesses.mutex));
  while (wp) {
    register wamprocess_t *p = wp;
    wp = wp->next;
#if 1
    delete_small_mem(p->wam); 
    free(p->wam);
    free(p);
#else
    thr_kill(wp->thread, SIGUSR1);
#endif
  }
  allprocesses.process = (wamprocess_t *)NULL;
}

void killproc(i)
long i;
{
  deleteprocess(thr_self());
  puts("ERROR: this interrupt code should never be executed");
  thr_exit(0);
}

void *wamproc(arg)
void *arg;
{
  cell RegisterBuffer[TEMPARGS+MAXREG];

  register term regs=RegisterBuffer+TEMPARGS;
  register term H;
  register instr P;
  register term *A;
  register term xref; 
  register cell xval; int arity; 
  register stack wam = (stack)arg;
  sigset_t set;
  struct sigaction sigact;

  A=wam[ChoiceStk].base;       
  TR_TOP=wam[TrailStk].base;   
  H=(term)wam[HeapStk].oldtop;   
  addprocess(thr_self(),(term)wam[HeapStk].base,wam);

  sigact.sa_handler = killproc;
  sigemptyset(&sigact.sa_mask);
  sigact.sa_flags = SA_RESTART;
  sigaction(SIGUSR1,&sigact,NULL);

#if 1
  H=(term)bp_prolog_call(((term)wam[HeapStk].base),regs,H,P,A,wam);
#else
  FDEREF(H); 
  arity=GETARITY(xval); 
  P=(instr)hget(g.predmark,(no)(PUTARITY(xval,arity+1))); 
  if(!P) { 
    local_error(xval, 
    "metacall to unknown predicate in recursive WAM\n", wam); 
    return 0; 
  } 
  COPY_CELLS(regs,xref,arity); 
  SETREF(regs+arity+1,g.true); 
  bp(regs,H,P,A,wam);
#endif
#if THREADS
  deleteprocess(thr_self());
  printf("ERROR: could not find thread %d\n", thr_self());
  thr_exit(0);
#endif
  return (void *)NULL; /* for the sake of type correctness */
}

term bb_op(type,in_arg,out_arg,A,wam,H)
register int type;
register cell in_arg;
register term *out_arg,*A;
register stack wam;
register term H;
{
  switch (type) {
    case BB_OP_NULLPROC: {
/*      puts("call BB_OP_NULPROC"); */
      if (SYMCONST(in_arg)) {
        register localbb_t *bb = (localbb_t *)hget(g.bbmark,(no)in_arg);
        if (VALID(bb)) {
/*          puts("call rpc_nulproc"); */
          rpc_nulproc(bb, H);
          return H;
        }
      }
      return 0;
    }
    case BB_OP_LOCK: {
      if (SYMCONST(in_arg)) {
        register localbb_t *bb = (localbb_t *)hget(g.bbmark,(no)in_arg);
        if (VALID(bb)) {
          if (bb->owner != thr_self()) {
            mutex_lock(&(bb->mutex));
            bb->owner = thr_self();
          } 
          bb->lockno--;
          return H;
        }
      } else
        warnmes("Blackboard name is not a constant");
      return 0;
    } 
    case BB_OP_UNLOCK: {
      if (SYMCONST(in_arg)) {
        register localbb_t *bb = (localbb_t *)hget(g.bbmark, (no)in_arg);
        if (VALID(bb)) {
          if (!bb->lockno) {
            mutex_unlock(&(bb->mutex));
            bb->owner = 0;
          } 
          bb->lockno++;
          return H;
        }
      } else
        warnmes("Blackboard name is not a constant");
      return 0;
    } 
    case BB_OP_PROCESSINFO: {
      if (INTEGER(in_arg))
        return processinfo(OUTPUT_INT(in_arg),H,out_arg);
      else
        return 0;
    } 
    case BB_OP_PROCESSES: {
      return processes(H,out_arg);
    }
    case BB_OP_BLACKBOARDS: {
      return blackboards(H,out_arg);
    }
    case BB_OP_BBMEM: {
      if (SYMCONST(in_arg)) {
        register localbb_t *bb = (localbb_t *)hget(g.bbmark,(no)in_arg);
        if (VALID(bb)) {
          if (bb->type == LOCALBB) {
            return bbmem(bb,H,out_arg);
          } else { /* remote blackboard */
            return rpc_bbmem(bb,H,out_arg,wam);
          }
        }  
      } 
      return 0;
    }
    case BB_OP_BBFREE: {
      register cell xval;
      register term xref;

      if (*(term)in_arg != new_func("bb",2)) return 0;
      FDEREF(((term)in_arg)[1]);
      if (SYMCONST(xval)) {
        register localbb_t *bb = (localbb_t *)hget(g.bbmark,(no)xval);
        if ( VALID(bb) ) { 
          FDEREF(((term)in_arg)[2]);
          if (INTEGER(xval)) {
            if (bb->type == LOCALBB) { /* local blackboard */
            return bbfree(bb, OUTPUT_INT(xval)) ? H : 0;
            } else { /* remote blackboard */
               return rpc_bbfree(bb, OUTPUT_INT(xval)) ? H : 0;
            }
          }  
        }
      } 
      return 0;
    }
    case BB_OP_BBMALLOC: {
      register cell xval;
      register term xref;
            
      if (*(term)in_arg != new_func("bb",2)) return 0;
      FDEREF(((term)in_arg)[1]);
      if (SYMCONST(xval)) {
        register localbb_t *bb = (localbb_t *)hget(g.bbmark,(no)xval);
        if (VALID(bb)) { 
          FDEREF(((term)in_arg)[2]);
          if (INTEGER(xval)) {
            if (bb->type == LOCALBB) {
              return bbmalloc(bb, OUTPUT_INT(xval)) ? H : 0;
            } else { /* remote blackboard */
               return rpc_bbmalloc(bb, OUTPUT_INT(xval)) ? H : 0;
            }
          }    
        }
      } 
      return 0;
    }
    
    case BB_OP_POOLMEM: 
      return poolmem((pool_t *)hget(g.bbmark,g.bbmark), H, out_arg);
      
    case BB_OP_FREE_: {
      void *p = malloc(100000); /* KDB */
      printf("top = %08x\n", p); 
      free(p);
      return H;
    }
    case BB_OP_PROCREATE: {
       thread_t thread;
       register stack newwam = malloc(sizeof(struct stack)*MaxStk);

       memcpy(newwam, wam, sizeof(struct stack)*MaxStk);
       init_small_mem(newwam);  
       newwam[HeapStk].oldtop = (term *)copytobuffer((term)newwam[HeapStk].base, 
                 (term)in_arg, wam, (int)(max.HEAP>>2)); 
#if THREADS
       thr_create(NULL,STACKSIZE,wamproc,(void *)newwam,THR_DETACHED,&thread);
#else
       addprocess(-1,in_arg,newwam);
       wamproc(newwam);
#endif
       return H;
    }

    case BB_OP_CREATELOCALBB: {
    
 
      if (SYMCONST(in_arg)) {
        register localbb_t *bb;
        char buffer[10]; 
        no test;
        
	test = hget(g.bbmark,(no)in_arg);
        bb = (localbb_t *)hget(g.bbmark,(no)in_arg);
        if (VALID(bb)) {
          if (bb->type == LOCALBB) {
            sprintf(buffer, "%08x", bb);
            *out_arg = (term)new_func(buffer, 0);
            return H;
          } else {
            warnmes("cannot change existing blackboard");
            return 0;
          }
        } else {
          register localbb_t *oldbb = bb;
          bb=createlocalbb(in_arg);
          sprintf(buffer, "%08x", bb);
          *out_arg = (term)new_func(buffer, 0);
          if (!oldbb) 
            return hdef(g.bbmark,(no)in_arg,(no)bb,g.timestamp) ? H : 0;
          else /* g.empty */
            return hset(g.bbmark,(no)in_arg,(no)bb) ? H : 0;
        }
      } else if (VAR(in_arg)) {
        register localbb_t *bb=createlocalbb(new_func("_",0));
        if (bb) {
          sprintf(bb->name, "%08x", bb);
          *out_arg = (term)new_func(bb->name, 0);
          if (!hget(g.bbmark,(no)*out_arg))
            hdef(g.bbmark,(no)*out_arg,bb,g.timestamp);
          return (unify(in_arg,out_arg,wam,A) && hset(g.bbmark,(no)*out_arg,(no)bb))
                      ? H : 0;
        } else
          return 0;
      } else
        return 0;
    } 
    case BB_OP_CREATEREMOTEBB: {
      register cell xval, name, bboard, host;
      register term xref;
      register CLIENT *cl;
      register u_long prognum;
      register remotebb_t *rbb;

     if (*(term)in_arg!=new_func("bb",4)) return 0;
      FDEREF(((term)in_arg)[1]);
      name = xval;
      FDEREF(((term)in_arg)[2]);
      host = xval;
      FDEREF(((term)in_arg)[3]);
      if (IDENTIFIER(xval)) 
        prognum = nametoprog(NAME(xval));
      else if (INTEGER(xval))
        prognum = OUTPUT_INT(xval);
      else return 0;
      FDEREF(((term)in_arg)[4]);
      bboard = xval;
      if (SYMCONST(name)) {
        register remotebb_t *bb = (remotebb_t *)hget(g.bbmark,(no)name);
         
        if (VALID(bb)) {
          if (bb->type == REMOTEBB)
            return (!strcmp(bb->host,NAME(host)) &&
                    !strcmp(bb->name,NAME(name)) &&
                    bb->prognum == prognum &&
                    bb->bboard == bboard) ? H : 0;
          else {
            warnmes("Cannot change blackboard type");
            return 0;
          }
        } else /* no VALID */ {
          if (cl=clnt_create(NAME(host),prognum,BBVERS,"tcp")) {
            if (rbb = createremotebb(cl,name,NAME(host),prognum,bboard)) {
              if (!bb) 
                return hdef(g.bbmark,(no)name,rbb,g.timestamp) ? H : 0;
              else /* g.empty */
                return hset(g.bbmark,(no)name,(no)rbb) ? H : 0;
            } else { printf("clnt_create unsuccessful"); return 0;}
          } else {
            warnmes("RPC error"); return 0;
          }
        }
      } else if (VAR(name)) {
        if (cl=clnt_create(NAME(host),prognum,BBVERS,"tcp")) {
          char buffer[10];
          sprintf(buffer, "%08x", cl);
          *out_arg = (term)new_func(buffer, 0);
          /* this might cause problems if unify fails */
          if (rbb = createremotebb(cl,name,NAME(host),prognum,bboard)) {
            if (!hget(g.bbmark,(no)*out_arg))
              hdef(g.bbmark,(no)*out_arg,rbb,g.timestamp);
            return unify(((term)in_arg)[1],out_arg,wam,A) && hset(g.bbmark,(no)*out_arg,(no)rbb) ? H : 0;
          } else return 0;
        } else
          return 0;  
      }
      return 0;
    } 
    case BB_OP_BBKILL: {
      if (SYMCONST(in_arg)) {
        register blackboard_t *bb = (blackboard_t *)hget(g.bbmark,(no)in_arg);
        if (VALID(bb)) {
          register pool_t *pool = (pool_t *)hget(g.bbmark,g.bbmark);

          assert(pool);
          mutex_lock(&(pool->mutex));
          if (deleteblackboard(pool,bb)) {
            mutex_unlock(&(pool->mutex));
            return hset(g.bbmark,(no)in_arg,(no)g.empty) ? H : 0;
          } else {
            mutex_unlock(&(pool->mutex));
            return 0;
          }
        }
      }
      return 0;
    }
    case BB_OP_BBCLEAN: {
      if (SYMCONST(in_arg)) {
        register blackboard_t *bb = (blackboard_t *)hget(g.bbmark,(no)in_arg);
        if (VALID(bb)) 
          return bbclean(bb) ? H : 0;
      }
      return 0;
    } 
    case BB_OP_KILLPROC: {
      if (INTEGER(in_arg)) {
        removeprocessfromblackboard(OUTPUT_INT(in_arg));
        thr_kill(OUTPUT_INT(in_arg),SIGUSR1);
        /* to guarantee the processing of SIGUSR1 */
        thr_continue(OUTPUT_INT(in_arg)); 
        return H;
      } else
        return 0;
    } 
    case BB_OP_BBTERMS: {
      if (SYMCONST(in_arg)) {
        register blackboard_t *bb = (blackboard_t *)hget(g.bbmark,(no)in_arg);
        if (VALID(bb)) 
          return bbterms(bb,H,out_arg,wam);
      }
      return 0;
    } 
    case BB_OP_BBPROCESSES: {
      if (SYMCONST(in_arg)) {
        register blackboard_t *bb = (blackboard_t *)hget(g.bbmark,(no)in_arg);
        if (VALID(bb)) 
          return bbprocesses(bb,H,out_arg,wam);
      } 
      return 0;
    } 
    case BB_OP_BBINFO: {
      if (SYMCONST(in_arg)) {
        register blackboard_t *bb = (blackboard_t *)hget(g.bbmark,(no)in_arg);
        if (VALID(bb)) 
          return bbinfo(bb,H,out_arg);
      }
      return 0;
    } 
    case RPC_CREATESERVER: {
       if (serverprog)
         return 0;
       else 
         return (*out_arg = (term)initserver(in_arg,wam)) ?  H : 0;
    }
    case RPC_KILLSERVER: {
    
       if (SYMCONST(in_arg)) {
         if (!strncmp((char *)&serverprog,NAME(in_arg),4)) {
           killserver();
           return H;
         } else
           ERREXIT("Server does not exist");
       } else if (INTEGER(in_arg)) {
         if (serverprog == OUTPUT_INT(in_arg)) {
           killserver();
           return H;
         } else 
           ERREXIT("Server does not exist");
       }
       return 0;
    }
    case RPC_BLACKBOARDS: {
      register cell xval, host;
      register term xref;
      register u_long prognum;
      
      if (*(term)in_arg!=new_func("bb",2)) return 0;
      FDEREF(((term)in_arg)[1]);
      host = xval;
      FDEREF(((term)in_arg)[2]);
      if (IDENTIFIER(xval)) 
        prognum = nametoprog(NAME(xval));
      else if (INTEGER(xval))
        prognum = OUTPUT_INT(xval);
      else return 0;
        return rpc_blackboards(host,prognum,H,out_arg,wam);  
    }
    
    case RPC_BBMOVE: {  
      register cell      xval, host,bb_l, bb_r;
      register term      xref;
      register u_long    prognum;
      register localbb_t *lbb;

    if (*(term)in_arg!=new_func("bb",4)) return 0;  
      FDEREF(((term)in_arg)[1]);  
      if (!SYMCONST(xval)){
        fprintf(stderr,"first argument is not a symbol constant\n");
        return (term)NULL;
      }
      bb_l = xval;
      lbb = (localbb_t *)hget(g.bbmark,(no)bb_l);
      if ( !VALID(lbb) || (lbb->type != LOCALBB)) {
        fprintf(stderr,"local blackboard not valid or not local\n");
        return (term)NULL;
      }
      FDEREF(((term)in_arg)[2]);
      host = xval;
      FDEREF(((term)in_arg)[3]);
      if (IDENTIFIER(xval)) 
        prognum = nametoprog(NAME(xval));
      else if (INTEGER(xval))
        prognum = OUTPUT_INT(xval);
      FDEREF(((term)in_arg)[4]);
      if (!SYMCONST(xval)){
        fprintf(stderr,"fourth argument is not a symbol constant\n");
        return (term)NULL;
      }
      bb_r = xval; 
    return rpc_bbmove(bb_l,host,prognum,bb_r) ? H : 0; 
    }
 
     default:
      return 0;
  }
}


int addprocess(thread,t,wam)
register thread_t thread;
register term t;
register stack wam;
/*******************************************
  routine to add a newly created process
  to the list of processes. This list is 
  needed to keep track of the resources used   
  by the processes. A better solution would
  be to generalize the wam-structure to 
  keep this kind of information.
 *******************************************/
{
  register wamprocess_t *wp;
  register int datasize = sizeof(wamprocess_t) - 4 + term_size(t);

  wp = (wamprocess_t *)malloc(datasize); 
  wp->wam = wam;
  wp->thread = thread;
  copytobb((term)wp->data,t,wam);
  
  mutex_lock(&(allprocesses.mutex));
  wp->next = allprocesses.process;
  allprocesses.process = wp;
  mutex_unlock(&(allprocesses.mutex));
  return 1;
}

int removeprocessfromblackboard(thread)
register thread_t thread;
/*******************************************
  routine to remove a possible pending 
  process entry on any currently existing blackboard
  This routine should be called before killing
  a process. Otherwise, a blackbaord will
  contain a reference to a non-existing process
 *******************************************/
{
  register pool_t *pool = (pool_t *)hget(g.bbmark, g.bbmark);
  register blackboard_t *bb;

  assert(pool);
  mutex_lock(&(pool->mutex));
  bb = pool->blackboards;
  while (bb) {
    if (bb->type == LOCALBB) {
      register processnode_t *pp, *pq=(processnode_t *)NULL;
      register int alreadyowner;

      bblock(bb);
      pp = ((localbb_t *)bb)->firstprocess;
      while (pp) {
        if (pp->thread == thread) {
          if (pq) {
            pq->next = pp->next;
          } else {
            ((localbb_t *)bb)->firstprocess = pp->next;
          }
          buddyfree(((localbb_t *)bb)->buddy, pp, pp->size);
          bbunlock(bb);
          mutex_unlock(&(pool->mutex));
          return 1;
        } else {
          pq = pp;
          pp = pp->next;
        }
      }
      bbunlock(bb);
    }
    bb = bb->nextbb;
  }
  mutex_unlock(&(pool->mutex));
  return 0;
}

int deleteprocess(thread)
register thread_t thread;
/******************************************
  routine to release the resources used
  by thread, and to finally end the thread
  This routine can only be called by the 
  thread to end itself.
 ******************************************/
{
  register wamprocess_t *wp, *wq = (wamprocess_t *)NULL;

  mutex_lock(&(allprocesses.mutex));
  wp = allprocesses.process;
  while (wp) {
    if (wp->thread == thread) {
      if (wq)
        wq->next = wp->next;
      else
        allprocesses.process = wp->next;
      mutex_unlock(&(allprocesses.mutex));
      delete_small_mem(wp->wam); 
      free(wp->wam);
      free(wp);
      thr_exit(0);
    } else {
      wq = wp;
      wp = wp->next;
    }
  }
  mutex_unlock(&(allprocesses.mutex));
  return 0;
}


