#ifndef _BBOARD_H
#define _BBOARD_H

#define REMOVE 1
#define NOREMOVE 0

#define BLOCKING    (1<<RBITS)
#define DESTRUCTIVE (2<<RBITS)

#include <stdlib.h>
#include <sys/ipc.h>
#include <rpc/rpc.h>
#include <synch.h>
#include <thread.h>
#include <errno.h>
#include <sys/sem.h>
#include <unistd.h>
#include "defs.h"
#include "buddy.h"
#include "rpc.h"

#define BB_POOL_SIZE 0x30000

#define BLACKBOARDSIZE 0x00001000

typedef struct pool {
  buddycontrolblock_t buddy;       /* local memory management */
  mutex_t mutex;                   /* sharedmem mutex */
  struct blackboard *blackboards;  /* link to local blackboards */
} pool_t;

#define NAMESIZE 32
#define HOSTNAMESIZE 128

#define LOCALBB 0
#define REMOTEBB 1

/* KO */
typedef struct clnt_node {
  CLIENT *cl;
  int    inuse;
  struct clnt_node *next;
} clnt_node_t;

/* KO */
typedef clnt_node_t *clnt_list_t;

typedef struct blackboard {
  int type;
  struct blackboard *nextbb;
  char name[NAMESIZE];              /* name of blackboard */
  mutex_t mutex;                    /* blackboard mutex */
  thread_t owner;                   /* current blackboard owner */
  unsigned lockno;                  /* number of simultaneous locks */
} blackboard_t;

typedef struct localbb {
  int type;
  struct blackboard *nextbb;
  char name[NAMESIZE];              /* name of blackboard */
  mutex_t mutex;                    /* blackboard mutex */
  thread_t owner;                   /* current blackboard owner */
  unsigned lockno;                  /* number of simultaneous locks */
  /* variable part */
  buddycontrolblock_t buddy;        /* local memory management */
  struct termnode *firstterm;       /* head of term list */
  struct termnode *lastterm;        /* tail of term list */
  struct processnode *firstprocess; /* head of process list */
  struct processnode *lastprocess;  /* tail of process list */
  term timestamp;                   /* timestamp */
} localbb_t;
 
typedef struct remotebb {
  int type;
  struct blackboard *nextbb;
  char name[NAMESIZE];             /* local name of blackboard */
  mutex_t mutex;                   /* blackboard mutex */
  thread_t owner;                  /* current blackboard owner */
  unsigned lockno;                 /* number of simultaneous locks */
  /* variable part */
  char host[HOSTNAMESIZE];         /* name of the host machine */
  u_long prognum;                  /* rpc prognum */
  cell bboard;                     /* local blackboard name on remote host */
  /* KO */
  /*CLIENT *cl;*/                      /* client handle */
  clnt_list_t cl_list;             /* list of clients in use */
} remotebb_t;

typedef struct termnode {
  struct termnode *next;
  term timestamp;
  unsigned size;
  unsigned data[1];
} termnode_t;

typedef struct processnode {
  struct processnode *next;
  unsigned size;
  thread_t thread; 
  cond_t condvar;
  int modifier;
  term timestamp;
  struct termnode *answer;        
  term *pattern;
  term *condition;
  unsigned data[1];
} processnode_t;

typedef struct wamprocess {
  struct wamprocess *next;
  stack wam;
  thread_t thread;
  unsigned data[1];
} wamprocess_t;

CLIENT * get_bbclient();
localbb_t *blackboardcreate();
void blackboarddelete();
int blackboardclean();
int blackboardshow();
term putterm();
term getterm();
unsigned int termcount();
unsigned int processcount();
int make_shm();
char *shmcalloc();

#define bblock(BB)      {if ((BB)->owner != thr_self()) { \
                          mutex_lock(&((BB)->mutex)); \
                          (BB)->owner = thr_self(); alreadyowner = 0; \
                        } else alreadyowner = 1; /*printf("%08x bblock\n",BB);*/}
#define bbunlock(BB)    {/*printf("%08x bbunlock\n",BB);*/ if (!alreadyowner) {(BB)->owner = 0; \
                          mutex_unlock(&((BB)->mutex));}}


#define ON_BB(ADR) (((unsigned)(ADR)-(unsigned)(bb_pool_address)) <  \
                     (unsigned)BB_POOL_SIZE)



#define BB_OP_LOCK           1
#define BB_OP_UNLOCK         2
#define BB_OP_PROCESSINFO    3
#define BB_OP_BLACKBOARDS    4
#define BB_OP_POOLMEM        5
#define BB_OP_FREE_          6
#define BB_OP_PROCREATE      7
#define BB_OP_CREATELOCALBB  8
#define BB_OP_CREATEREMOTEBB 9
#define BB_OP_BBKILL        10
#define BB_OP_BBCLEAN       11
#define BB_OP_BBFREE        12
#define BB_OP_KILLPROC      13
#define BB_OP_BBMALLOC      14
#define BB_OP_BBTERMS       15
#define BB_OP_BBPROCESSES   16
#define BB_OP_BBINFO        17
#define BB_OP_BBMEM         18
#define BB_OP_PROCESSES     19
#define RPC_CREATESERVER    20
#define RPC_KILLSERVER      21
#define RPC_BLACKBOARDS     22
#define RPC_BBMOVE          23
#define BB_OP_NULLPROC      25
/* #define BB_OP_TESTKK        50
*/
#endif _BBOARD_H
