/**************************** instanz.c **********************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tree.h"
#include "instanz.h"

extern Welt *a_Sub;
extern O_Welt *W_Abst;



/*****************************************************/
/* FUNCTION  : SubC_O                                */
/* PARAMETER : char O[36], Welt *W, Terme *T         */
/* RETURN    : Liste                                 */
/* PURPOSE   : liefert eine Liste mit allen Konzept- */
/*             bezeichner, ihre Konzeptterme die     */
/*             Abstraction des Objekts O subsumieren.*/
/*****************************************************/



Liste *SubC_O (char name[36], Welt *W, Terme *T)
{

  Baum *abstract;
  Terme *tm, *act_tm;
  Liste *lst;
  Welt *act_Welt;
  O_Welt *OW;

  lst = NULL;

  /* W_Abst: globale Variable. W_Abst = abstraction (W, T) */
  OW = W_Abst;
 
  /* Sucht die Abstraction des Objekts O */
  /* (mit dem Bezeichner 'name') in OW.  */ 

  while (strcmp (OW->name, name) != 0 && OW != NULL)
    OW = OW->next;

  tm = copy_Terme (T);

  /* Das Objekt O ist in der Objektwelt definiert. */
  if (OW != NULL) { 
    abstract = OW->abstract;
 
    act_tm = tm;

    /* prueft nach, ob die Abstraction des Objkets O */
    /* in der Terminologie T definiert ist.          */
   
    while (act_tm != NULL) {
      if (vergleich (act_tm->expression->term, abstract) == -1)
        act_tm = act_tm->next;
      else {
   
        /* Die Abstraction des Objekts O ist in T definiert.  */
    
        /* a_Sub: globale Variable. a_Sub = all_Sub_Terme (T).*/
        /* Eine Liste mit allen Termbezeichner, deren definierenden */
        /* Terme die Abstraction des Objekts O subsummieren, wird   */
        /* in a_Sub gesucht.                                        */   
      
        act_Welt = a_Sub;
        while (act_Welt != NULL && 
               strcmp (act_Welt->ls->str, act_tm->expression->ATerm) != 0)
          act_Welt = act_Welt->next;
      
        lst = act_Welt->ls;

        break;
      }
    }

    /* Die Abstraction des Objekts O ist nicht in T definiert.  */

    if (act_tm == NULL) {
       act_tm = tm;
       while (act_tm->next != NULL)
         act_tm = act_tm->next;
       act_tm->next = Node_Terme ( );
    
       /* Eine neue TermDefinition der Form: 'neuTerm :=C abstract' */
       /* wird am Ende der Terminologie anhaengen.                  */
       /* Eine neue temporaere Terminologie entsteht. */

       act_tm->next->expression = Node_Term ( );
       act_tm->next->expression->ATerm = (char *)strdup ("neuTerm");
       act_tm->next->expression->equal = (char *)strdup (":=C");
       act_tm->next->expression->term = abstract;
       act_tm->next->next = NULL;


       /* berechnet die Liste mit allen Termbezeichner, */
       /* deren definierenden Terme die Abstraction des */
       /* Objekts O subsummieren, in der temporaeren    */
       /* Terminologie.                                 */

       lst = all_Sub (act_tm->next->expression->term, tm); 
       lst = lst->next;
    }
  }
  else {
    printf ("%s ist kein Objekt in W.\n", name);
    exit (1);
  }    


  return (lst);
}


/*****************************************************/
/* FUNCTION  : Inst                                  */
/* PARAMETER : char Cbez[36], Welt *W, Terme *T      */
/* RETURN    : Liste *                               */
/* PURPOSE   : berechnet die Instanzen eines Konzepts*/
/*             mit dem Bezeichner Cbez hinsichtlich  */
/*             der Terminologie T und Objektwelt W.  */ 
/*****************************************************/


Liste *Inst(char *Cbez, Welt *W, Terme *T)
{
   Liste *O_lst, *lst, *ASC, *act_lst;

   O_lst = NULL;

   /* lst: die Menge aller Objektbezeichner in W */
   lst = Objekt_Liste (W);

   while (lst != NULL) {

     /* ASC: die Menge aller Konzeptbeichner, ihre Konzept-    */
     /* terme die Abstraction eines Objekts mit dem Bezeichner */
     /* 'lst->str' subsummieren.                               */
    

     ASC = SubC_O(lst->str, W, T);

     while (ASC != NULL) {

       /* Wenn 'Cbez' ein Element des ASC ist, wird  */
       /* 'lst->str' an der Liste O_lst angehaenget. */

       if (strcmp (Cbez, ASC->str) == 0) {
         if (O_lst == NULL) {
           O_lst = Node_Liste ( );
           strcpy (O_lst->str, lst->str);
           O_lst->next = NULL;
         }
         else {
           act_lst = O_lst;
           while (act_lst->next != NULL)
             act_lst = act_lst->next; 
           act_lst->next = Node_Liste ( );
           act_lst = act_lst->next;
           strcpy (act_lst->str, lst->str);
           act_lst->next = NULL;
	 }

         break;
       }

       else
         ASC = ASC->next;
     }

     lst = lst->next;
   }


   return (O_lst);
}       
