/*******************  TEWIS.y  **********************/

/* Das ist das yacc_File. Yacc File ist als ein parser zu verstehen. */
/* Die hat die gleichen drei Bestandenteilen wie das lex_File.       */
 
/* In %union sind alle Datentypen enthalten. Der Datentype jedes     */
/* einzelnen Element in "rules section " wird in %type festgestellt. */

%{ 
#include <stdio.h>
#include <string.h>
#include "tree.h"
#include "yacc.h"

extern char yytext[];
extern Terme *ITF;
extern Welt *SAF;
extern int linenum;
%}

%token not and atleast all androle  aset  atmost exists ok
%token NOTHING NUMBER PROATOM PRIMPROATOM DEF GL nr
%token <str> '(' ')' ','

%union {
  Terme *tme;
  Term *tm;
  char str[256]; 
  Baum *bm;
  int num;
  Liste *lst;
  Welt *w;
}

%type <tme> TermDefinitionen 
%type <tm> TermDefinition
%type <bm> Konzept Concept Rolle Role AttributeListe
%type <str> DEF GL
%type <str> not and atleast androle all aset atmost exists nr
%type <str> NOTHING PROATOM PRIMPROATOM atomaresKonzept atomareRolle
%type <w>  Objektwelt
%type <lst> Weltbeschreibung Konzeptinstanziierung Rolleninstanziierung

%start Regel
%%

Regel:               TermDefinitionen 
                     { 
                       ITF = $1;
                     }
           |         Objektwelt
                     {
                       SAF = $1;
		     }
           ; 

Objektwelt:          Weltbeschreibung
                     { 
                       $$ = (Welt *) malloc (sizeof (Welt));
                       $$->ls = $1;
                       $$->next = NULL;
                     }

           |         Weltbeschreibung ';' Objektwelt
                     {
                       $$ = (Welt *) malloc (sizeof (Welt));
                       $$->ls = $1;
                       $$->next = $3;
                     }  
           ;

Weltbeschreibung:    Konzeptinstanziierung
                     {
                       $$ = $1;
		     }

           |         Rolleninstanziierung
                     {
                       $$ = $1;
		     } 
           ;

Konzeptinstanziierung: '(' PROATOM atomaresKonzept ')'
                       {
                         $$ = Welt4 ($1, $2, $3, $4);
		       }

           |           '(' PROATOM '(' not PRIMPROATOM ')' ')'
                       {
                         $$ = Welt7 ($1, $2, $3, $4, $5, $6, $7);
		       }
           ;

Rolleninstanziierung:  '(' PROATOM atomareRolle PROATOM ')'
                       {
                         $$ = Welt5 ($1, $2, $3, $4, $5);
		       }

           |           '(' PROATOM '('atleast nr atomareRolle ')' ')'
                       {
                         $$ = Welt8 ($1, $2, $3, $4, $5, $6, $7, $8);
		       }
 
           |           '('PROATOM '('atmost nr atomareRolle ')' ')'
                       {
                         $$ = Welt8 ($1, $2, $3, $4, $5, $6, $7, $8);
		       }
           ;

TermDefinitionen:    TermDefinition ';' TermDefinitionen 
                     {
                       $$ = Node_Terme ( );
                       $$->expression = $1; $$->next = $3 ;
                     }

           |         TermDefinition
                     {
                       $$ = Node_Terme ( );
                       $$->expression = $1; $$->next = NULL;
                     }    
           ;

TermDefinition:      atomaresKonzept DEF  Konzept
                     {
                       $$ = NodeDef ($1,$2,$3);
                     }

           |         atomareRolle GL Rolle
                     { 
                       $$ = NodeDef ($1,$2, $3);
                     }
           ;

atomaresKonzept:     PROATOM   { strcpy($$ , $1);}
           |         NOTHING   { strcpy($$ , $1);}
           ;
 
atomareRolle:        PROATOM   { strcpy($$, $1);} 
           ; 
 
Konzept:             PROATOM              
                     {
                        $$ = Aus_0 ($1);
                     }
            
            |        NOTHING
                     {
                        $$ = Aus_0 ($1);
                     }

            |        PRIMPROATOM
                     {
                        $$ = Aus_0 ($1);
		     }

            |        '(' not PRIMPROATOM ')'
                     { 
                       $$ = Aus_1 ($1, $2, $3, $4);
                                             
		     } 
    
            |        '(' and Concept ')'
                     {
                        $$ = Aus_2 ($1,$2,$3,$4);
                     }

            |        '('all Rolle Konzept ')'
                     {
                        $$ = Aus_3 ($1,$2,$3,$4,$5);
                     }

            |        '(' atleast nr Rolle ')'
                     {
                        $$ = Aus_6 ($1,$2,$3,$4,$5);
                     } 

            |        '(' atmost nr Rolle ')'
                     {
                        $$ = Aus_6 ($1,$2,$3,$4,$5); 
                     } 

            |        '(' exists Rolle Konzept ')'
                     {
                        $$ = Aus_3 ($1,$2,$3,$4,$5); 
                     }
 
            |        aset '(' AttributeListe ')' 
                     {
                       $$ = Aus_4 ($1, $2, $3,$4);
                     }
            ;
 

Concept:             Konzept
                     {
                        $$ = copy_Baum ($1, NULL);
		     }
            
            |        Konzept Concept
                     {
                        $$ = NodeAus ($1, $2);
		     }
            ;

AttributeListe:      PROATOM
                     {
                       $$ = Aus_0($1);
                     }

            |        PROATOM ',' AttributeListe
                     {
                       $$ = Aus_5 ($1, $2, $3);
		     }
            ;


Rolle:               PROATOM
                     { 
                        $$ = Aus_0 ($1);
                     }

            |        PRIMPROATOM
                     { 
                        $$ = Aus_0 ($1);
                     }

            |        '(' androle Role  ')'
                     {
                        $$ = Aus_2 ($1,$2,$3,$4);
                     }
            ;

Role:                Rolle
                     {
                        $$ = $1;
		     }
            |        Rolle Role
                     {
                        $$ = NodeAus ($1, $2);
		     }
            ;


%%
int yyerror( char *errname)
{
  fprintf(stderr, "ITL : %s in line %d at \"%s\"\n", errname, linenum, yytext);
  exit(1);
}

int warn( char *warnname)
{
  fprintf(stderr, "ITL: Warning: %s in line %d  at \"%s\"\n",warnname, 
          linenum, yytext);
  return(1);
}
