/*********************   abstract.c   ************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tree.h"
#include "abstract.h"
#define MAX 100

/* globale Variable */
O_Welt *OW;

/*****************************************************/
/* FUNCTION  : Objekt_Liste                          */
/* PARAMETER : Welt *W                               */
/* RETURN    : Liste *                               */
/* PURPOSE   : liefert eine Liste, die aller Objekt_ */
/*             bezeichner der Objektwelt W enthaehlt.*/
/*****************************************************/

Liste *Objekt_Liste (Welt *W)
{
   Welt *w1;
   Liste *s_Liste, *act_Liste;

   w1 = W;
   s_Liste = NULL;
 
   /* Erstellen der Liste */
   while (w1 != NULL) {
     if (s_Liste == NULL) {
       s_Liste = Node_Liste ( );
       strcpy (s_Liste->str, w1->ls->next->str);
       s_Liste->next = NULL;
     }
     else {
       act_Liste = s_Liste;

       /* prueft nach, ob der Objektbezeichner in der */
       /* Liste (ohne letztes Element) vorhanden ist. */ 

       while (act_Liste->next != NULL) {
         
         if (strcmp (act_Liste->str, w1->ls->next->str) != 0)
           act_Liste = act_Liste->next;
         else
           /* Der Objektbezeichner ist vorhanden */
           goto ende;
       }
       
       /* Der Objektbezeichner ist identisch */
       /* mit dem letzten Element der Liste. */
       if (strcmp (act_Liste->str, w1->ls->next->str) == 0)
          goto ende;

       else {
         /* Der Objektbezeichner ist in der Liste nicht vorhanden. */
         /* Er wird an der Liste angehaengt.*/

         act_Liste->next = Node_Liste ( );
         act_Liste = act_Liste->next;
         strcpy (act_Liste->str, w1->ls->next->str);
         act_Liste->next = NULL;

       }
     }
       
     ende:;
     w1 = w1->next;
   }
  
   return (s_Liste);
}



/*****************************************************/
/* FUNCTION  : init_OW                               */
/* PARAMETER : Welt *W                               */
/* RETURN    : Liste *                               */
/* PURPOSE   : initializiert die globale Variable OW */
/*             von der Datentype O_Welt.             */
/*****************************************************/

O_Welt *init_OW (Welt *W)
{

   Liste *O_lst;
   O_Welt *act_OW;
   
   OW = NULL;
   O_lst = Objekt_Liste (W);

   if (O_lst != NULL) {
     OW = Node_OW ();
     strcpy (OW->name, O_lst->str);
     OW->abstract = NULL;
     OW->srole = NULL;
     OW->abstChange = TRUE;
     OW->next = NULL;

     O_lst = O_lst->next;
     while (O_lst != NULL) {
       act_OW = OW;
       while (act_OW->next != NULL)
         act_OW = act_OW->next;
       act_OW->next = Node_OW ( );
       act_OW = act_OW->next;
       strcpy (act_OW->name, O_lst->str);
       act_OW->abstract = NULL;
       act_OW->srole = NULL;
       act_OW->abstChange = TRUE;
       act_OW->next = NULL;

       O_lst = O_lst->next;
     }
   }
  
   return (OW);
}




/*****************************************************/
/* FUNCTION  : init_astring                          */
/* PARAMETER : char O[36], Welt *W                   */
/* RETURN    : Baum                                  */
/* PURPOSE   : Hilfsfunktion.                        */
/*             Initializieren  des astring[O] des    */
/*             Objekts O. Jedes C von (O  C) in W    */
/*             wird an astring[O] eingefuegt.        */
/*****************************************************/

Baum *init_astring (char name[36], Welt *W)
{
   int i;
   Liste *lst;
   Baum  *bm1, *astring;
   Welt *W1;
   
   /* Erstellen des Wurzelknotens des Baums astring */
   astring = Node_Baum ();
   astring->lKlammer = (char *)strdup ("(");
   astring->Token = (char *)strdup ("and");
   astring->rKlammer = (char *)strdup (")");
   astring->child = NULL;
   astring->father = NULL;
   astring->brother = NULL;
 
   W1 = W;
   while (W1 != NULL) {
     lst = W1->ls;

     /* Die Art der Liste in der Objektwelt W wird mit der Hilfe */
     /* der Anzahl des Elements dieser Liste festgestellt.       */
     i = 0;
     while (lst != NULL) {
       i++;
       lst  = lst->next;
     }
   
     /* Die Liste hat die Form (O C) */
     if (i == 4) {
       lst = W1->ls;
       if (strcmp (lst->next->str, name) == 0) {
         
         /* C wird am Ende des astring angehaengt */
         if (astring->child == NULL) {
           astring->child = Node_Baum ( );
           astring->child->Token = (char *)strdup (lst->next->next->str);
           astring->child->lKlammer = (char *)strdup (" ");
           astring->child->rKlammer = (char *)strdup (" ");
           astring->child->child = NULL;
           astring->child->brother = NULL;
           astring->child->father = astring;
         }
         else {
           bm1 = astring->child;
           while (bm1->brother != NULL)
             bm1 = bm1->brother;
           
           bm1->brother = Node_Baum ( );
           bm1 = bm1->brother;

           bm1->Token = (char *)strdup (lst->next->next->str);
           bm1->lKlammer = (char *)strdup (" ");
           bm1->rKlammer = (char *)strdup (" ");
           bm1->child = NULL;
           bm1->brother = NULL;
           bm1->father = astring;
        }
       }    
     }

     W1 = W1->next;
   }

   return (astring);
}



/*****************************************************/
/* FUNCTION  : init_role                             */
/* PARAMETER : char O[36], Welt *W, Terme *T         */
/* RETURN    : Baume *                               */
/* PURPOSE   : Hilfsfunktion.                        */
/*             init_role liefert alle fuer das Objekt*/
/*             O relevante Rollen.                   */
/*****************************************************/


Baume *init_role (char name[36], Welt *W, Terme *T)
{
   Baum *bm, *bm1, *bm2,*BM;  
   Baume *start_bme, *act_bme;
   Terme *tm, *tm1;   

   /* bm = (and C[1] C[2] ...C[n]) */
   bm = init_astring (name, W);
   
   /* bm zeigt auf C[1] */
   bm = bm->child;

   tm = copy_Terme (T);
   start_bme = NULL;
 
   while (bm != NULL) {
     tm1 = tm;
    
     /* Sucht die TermDefinition von C[i] in der Terminologie */
     while ( strcmp (bm->Token, tm1->expression->ATerm) != 0
                && tm1 != NULL)
          tm1 = tm1->next;

     if (tm1 == NULL) {
       printf ("Term %s ist nicht definiert.\n", bm->Token);
       return NULL;
     }
     else {
        /* alle Rolle in C[i] sind die fuer */
        /* das Objekt O relevante Rolle.    */
       BM = tm1->expression->term;

       bm1 = BM->child;
       while (bm1 != NULL) {
         if (strcmp (bm1->Token, "all") == 0 ||
             strcmp (bm1->Token, "exists") == 0) {
          
           /* bm2 zeigt auf die Rolle R von (all R C) */
           bm2 = copy_Baum (bm1->child, NULL);
           loesch_Baum (bm2->brother);
           bm2->brother = NULL;
           
           if (start_bme == NULL) {
             start_bme = Node_Baume ( );
             start_bme->bm = bm2;
             start_bme->next = NULL;
           }
           else {
             /* prueft nach, ob diese Rolle schon notiert ist. */
             act_bme = start_bme;
             while (act_bme->next != NULL) {
               if (vergleich (act_bme->bm, bm2) == -1)
                 act_bme = act_bme->next;
               else
                 break;
             } 
             if (act_bme->next == NULL &&
                 vergleich (act_bme->bm, bm2) == -1) {
               act_bme->next = Node_Baume ( );
               act_bme = act_bme->next;
               act_bme->bm = bm2;
               act_bme->next = NULL;
             }
           }
	 }
         else {
           if (strcmp (bm1->Token, "atleast") == 0 ||
              strcmp (bm1->Token, "atmost") == 0) {

             /* bm2 zeigt auf die Rolle R von   */
             /* (atleast n R) oder (atmost n R) */ 
             bm2 = copy_Baum (bm1->child->brother, NULL);

             if (start_bme == NULL) {
               start_bme = Node_Baume ( );  
               start_bme->bm = bm2;   
               start_bme->next = NULL;
             }
             else {
               act_bme = start_bme;
               while (act_bme->next != NULL) {
                 if (vergleich (act_bme->bm, bm2) == -1)
                   act_bme = act_bme->next;
                 else
                   break;
               } 
               if (act_bme->next == NULL &&
                   vergleich (act_bme->bm, bm2) == -1) {
                 act_bme->next = Node_Baume ( );
                 act_bme = act_bme->next; 
                 act_bme->bm = bm2;
                 act_bme->next = NULL;
               }
             }
           }
	 }
         bm1 = bm1->brother;
       }
     }
     bm = bm->brother;
   }   

   loesch_Terme (&tm);

   return (start_bme);
}



/*****************************************************/
/* FUNCTION  : init_srole                            */
/* PARAMETER : char O[36], Welt *W, Terme *T         */
/* RETURN    : set_role *                            */
/* PURPOSE   : Hilfsfunktion.                        */
/*             Initializieren der Datentype set_role */
/*             von O_Welt fuer das Objekt O.         */
/*****************************************************/


set_role *init_srole (char name[36], Welt *W, Terme *T)
{
   Baum *bm1, *bm2;
   Baume *bme, *bme1, *rolen, *act_bme;  
   set_role *act_srole, *srole;


   srole = NULL;

   /* bme zeigt auf fuer das Objekt O relevante Rollen */
   rolen = init_role (name, W, T);
   bme = rolen;

   /* O besitzt ihre relevente Rolle */
   if (bme != NULL) {
   
     /* Initalisieren des set_role */
     srole = Node_srole ( );
     srole->role = bme->bm;
     srole->sub_role = Node_Baume ( );
     srole->sub_role->bm = bme->bm;
     srole->sub_role->next = NULL;

     srole->super_role = Node_Baume ( );
     srole->super_role->bm = bme->bm;
     srole->super_role->next = NULL;             
    
     srole->filler = NULL;     
     srole->count = 0;
     srole->max = MAX;
           
     srole->MSG = NULL; 
     srole->next = NULL;
         
     /* Initalisieren der super_role und sub_role des set_role */
     bme1 = rolen;
     while (bme1 != NULL) {
       if (vergleich (bme1->bm, bme->bm) == -1) {
            
          bm1 = copy_Baum (bme1->bm, NULL);
          bm2 = copy_Baum (bme->bm, NULL);         

          /* super_role */
          if (termsub (bm1, bm2, T) == 1) {
            act_bme = srole->super_role;
            while (act_bme->next != NULL)
              act_bme = act_bme->next;
            act_bme->next = Node_Baume ( );
            act_bme = act_bme->next;
            act_bme->bm = bme1->bm;
            act_bme->next = NULL;
          }
          else {
            /* sub_role */ 
            if (termsub (bm2, bm1, T) == 1) {
              act_bme = srole->sub_role;
              while (act_bme->next != NULL)
                act_bme = act_bme->next;
              act_bme->next = Node_Baume ( );
              act_bme = act_bme->next;
              act_bme->bm = bme1->bm;
              act_bme->next = NULL;
            }
          }
           
          /* Speicherplaetze fuer bm1 und bm2 freigeben */
          loesch_Baum (bm1);
          bm1 = NULL;
          loesch_Baum (bm2);
          bm2 = NULL;
       }
   
       bme1 = bme1->next;
     }

     bme = bme->next;
     while (bme != NULL) {
       act_srole = srole;
       while (act_srole->next != NULL)
         act_srole = act_srole->next;
       act_srole->next = Node_srole ( );
       act_srole = act_srole->next;

       act_srole->role = bme->bm;
       act_srole->sub_role = Node_Baume ( );
       act_srole->sub_role->bm = bme->bm;
       act_srole->sub_role->next = NULL;

       act_srole->super_role = Node_Baume ( );
       act_srole->super_role->bm = bme->bm;
       act_srole->super_role->next = NULL;             
 
       act_srole->filler = NULL;
       act_srole->count = 0;
       act_srole->max = MAX;       
      
       act_srole->MSG = NULL;
       act_srole->next = NULL;

     
       bme1 = rolen;
       while (bme1 != NULL) {
         if (vergleich (bme1->bm, bme->bm) == -1) {

           /* Hilfsbaeume */
           bm1 = copy_Baum (bme1->bm, NULL);
           bm2 = copy_Baum (bme->bm, NULL);
       
           /* super_role */
           if (termsub (bm1, bm2, T) == 1) {
             act_bme = act_srole->super_role;

             while (act_bme->next != NULL)
               act_bme = act_bme->next;
             act_bme->next = Node_Baume ( );
             act_bme = act_bme->next;
             act_bme->bm = bme1->bm;
             act_bme->next = NULL;
           }
           else {
             /* sub_role */
             if (termsub (bm2, bm1, T) == 1) {
               act_bme = act_srole->sub_role;
               while (act_bme->next != NULL)
                 act_bme = act_bme->next;
               act_bme->next = Node_Baume ( );
               act_bme = act_bme->next;
               act_bme->bm = bme1->bm;
               act_bme->next = NULL;
             }
	   }
           loesch_Baum (bm1);
           bm1 = NULL;
           loesch_Baum (bm2);
           bm2 = NULL;
	 }
         bme1 = bme1->next;
       }
       bme = bme->next;
     }
   }
   else {/* O besitzt keine relevante Rolle */  
     srole = NULL;
   }

   return (srole);
}


/*****************************************************/
/* FUNCTION  : least_most_astring                    */
/* PARAMETER : char O[36], Welt *W, Terme *T         */
/* RETURN    : Baum                                  */
/* PURPOSE   : Hilfsfunktion.                        */
/*             Fuer jede Superrole R' der R wird     */
/*             (atleast n R') an astring des O's     */
/*             angehaengt. Wenn W eine Liste         */
/*             (O (atleast n R)) enthaelt.           */
/*             Fuer jede Subrole R' der R wird       */
/*             (atmost n R') an astring des O's      */
/*             angehaengt. Wenn W eine Liste         */
/*             (O (atmost n R)) enthaelt.            */
/*****************************************************/


Baum *least_most_astring (char name[36], Welt *W, Terme *T)
{
   int i;
   Liste *lst;
   Baum *bm, *astring, *neu_Baum;
   Baume *bme1;
   set_role *srole, *srole1;
   Welt *w1;

  
   srole = init_srole (name, W, T);
   astring = init_astring (name, W);
  
   if (srole != NULL) {

     w1 = W;
     while (w1 != NULL) {
       i = 0;
       lst = w1->ls;
       while (lst != NULL) {
         i++;
         lst = lst->next;
       }

       if (i == 8) {
         lst = w1->ls;

         /* W enthaelt (O (atleast n R)) */ 
         if (strcmp (lst->next->str, name) == 0) {
           if (strcmp (lst->next->next->next->str, "atleast") == 0) {
             srole1 = srole;
             while (srole1 != NULL) {

               /* Sucht die Rolle R in set_role des O's */
               if (strcmp (srole1->role->Token, 
                           lst->next->next->next->next->next->str) != 0)
                 srole1 = srole1->next;
               else
                 break;
             }
           
             if (srole1 != NULL) {
          
               /* bme1 zeigt auf die Superrole der R */
               bme1 = srole1->super_role;
               
               while (bme1 != NULL) {
                 /* (atleast n R') wird an astring des O's angehaengt */
                 bm = astring->child;
                 while (bm->brother != NULL)
                   bm = bm->brother;
                 bm->brother = Node_Baum ( );
                 bm = bm->brother;
               
                 bm->Token = (char *) strdup ("atleast");
                 bm->lKlammer = (char *)strdup ("(");
                 bm->rKlammer = (char *)strdup (")");

                 bm->father = astring;
                 bm->brother = NULL;
                 bm->child = Node_Baum ( );
                 bm->child->Token = 
                     (char *)strdup (lst->next->next->next->next->str);
                 bm->child->lKlammer = (char *)strdup (" ");
                 bm->child->rKlammer = (char *)strdup (" ");
                 bm->child->child = NULL;           
                 bm->child->father = bm;
                 bm->child->brother = copy_Baum (bme1->bm, bm);

                 bme1 = bme1->next;
	       } 
	     }
	   }
            else {
             /* W enthaelt (O (atmost n R)) */ 
             if (strcmp (lst->next->next->next->str, "atmost") == 0) {
               srole1 = srole;
               while (srole1 != NULL) {

                 /* Sucht die Rolle R in set_role des O's */
                 if (strcmp (srole1->role->Token, 
                             lst->next->next->next->next->next->str) != 0)
                   srole1 = srole1->next;
                 else
                   break;
	       }
           
               if (srole1 != NULL) {
                 /* bme1 zeigt auf die Subrole der R */
                 bme1 = srole1->sub_role;
               
                 while (bme1 != NULL) {
                   /* (atmost n R') wird an astring des O's angehaengt */
                   bm = astring;
                   bm = bm->child;
                   while (bm->brother != NULL)
                     bm = bm->brother;
                   bm->brother = Node_Baum ( );
                   bm = bm->brother;
                   bm->Token = (char *) strdup ("atmost");
                   bm->lKlammer = (char *)strdup ("(");
                   bm->rKlammer = (char *)strdup (")");

                   bm->brother = NULL;
                   bm->father = astring;
                   bm->child = Node_Baum ( );
                   bm->child->Token = 
                       (char *)strdup (lst->next->next->next->next->str);
                   bm->child->lKlammer = (char *)strdup (" ");
                   bm->child->rKlammer = (char *)strdup (" ");
                   bm->child->child = NULL;           
                   bm->child->father = bm;
                   bm->child->brother = copy_Baum (bme1->bm, bm);
                   
                   bme1 = bme1->next;
		 }
	       }
             }
	   }
         }
       }
          
       w1 = w1->next;
     }
 
    }

   return (astring);
 }






/*****************************************************/
/* FUNCTION  : roll_filler                           */
/* PARAMETER : char O[36], Welt *W, Terme *T         */
/* RETURN    : set_role                              */
/* PURPOSE   : Hilfsfunktion.                        */
/*             O' wird an filler der set_role des    */
/*             O's angehaengt, wenn (O R O') in W    */
/*             enhaelt ist und O' nicht zu filler    */
/*             der R gehoert.                        */
/*****************************************************/


set_role *role_filler (char name[36], Welt *W, Terme *T)
{
   int i, k;
   Liste *lst, *lst1;
   Terme *tm;
   set_role *act_srole, *srole;
   Welt *w1;
 
   srole = init_srole (name, W, T);   

   if (srole != NULL) {
     w1 = W;
     while (w1 != NULL) {
       i = 0;
       lst = w1->ls;
       while (lst != NULL) {
         i++;
         lst = lst->next;
       }

       /* Die Liste hat  die Form (O R O') */
       if (i == 5) { 
         lst = w1->ls;
         if (strcmp (lst->next->str, name) == 0) {

           /* Sucht R in set_role von O */
           act_srole = srole;
           while (act_srole != NULL) {
             if (strcmp (act_srole->role->Token, 
                         lst->next->next->str) != 0)
                 act_srole = act_srole->next;
             else
               break;
	   }

           if (act_srole == NULL) {
             /* sucht die Termdefinition von R in T */
             tm = T;
             while (tm != NULL) {
               if (strcmp (tm->expression->ATerm, 
                           lst->next->next->str) != 0)
                 tm = tm->next;
               else
                 break;
	     }

             act_srole = srole;
             while (act_srole != NULL) {
               if (vergleich (act_srole->role, tm->expression->term)
                   == -1)
                 act_srole = act_srole->next;
               else
                 break;
	     }
           }

                     
           if (act_srole != NULL) {

             /* lst1 zeigt auf der filler_Liste der Rolle R */
             lst1 = act_srole->filler;
             k = 0;
        
             /* prueft nach, ob O' in der filler_Liste */
             /* der R vorhanden ist.                   */  
             while (lst1 != NULL) {
               if (strcmp (lst1->str, lst->next->next->next->str) == 0) {
                 k = 1;
                 break;
               }
               else
                 lst1 = lst1->next;
	     }

             /* O' ist nicht in der filler_Liste der R vorhanden */
             if (k == 0) {

               /* O' wird an der filler_Liste der R angehaengt */
               if (act_srole->filler == NULL) {
                 act_srole->filler = Node_Liste ( );
                 strcpy (act_srole->filler->str, lst->next->next->next->str);
                 act_srole->filler->next = NULL;
              
                 act_srole->count = act_srole->count +1;
               }
               else {
                 lst1 = act_srole->filler;
                 while (lst1->next != NULL)
                   lst1 = lst1->next;
                 lst1->next = Node_Liste ( );
                 lst1 = lst1->next;
                 strcpy (lst1->str, lst->next->next->next->str);
                 lst1->next = NULL;
              
                 act_srole->count = act_srole->count + 1;
	       }
	     }
	   }
	 }
      }
      w1 = w1->next;
     }
   }

   return (srole);
}



/*****************************************************/
/* FUNCTION  : max_srole                             */
/* PARAMETER : char O[36], Welt *W, Terme *T,        */
/* RETURN    : set_role *                            */
/* PURPOSE   : Hilfsfunktion.                        */
/*             Fuer jede Subrole R' der R wird       */
/*             max = n zugewiesen, wenn W eine Liste */             
/*             (O (atmost n R)) enthaelt.            */
/*****************************************************/


set_role *max_srole (char name[36], Welt *W, Terme *T)
{
   int i;
   Liste *lst;
   Baum *bm;
   Baume *bme1;
   Terme *tm;
   set_role *act_srole, *h_srole, *srole;
   Welt *w1;

   srole = role_filler (name, W, T); 
   
   if (srole != NULL) {  
     w1 = W;
     while (w1 != NULL) {
       i = 0;
       lst = w1->ls;
       while (lst != NULL) {
         i++;
         lst = lst->next;
       }

       /* W enthaelt (O (atmost n R)) */ 
       if (i == 8) {
         lst = w1->ls;
         if (strcmp (lst->next->str, name) == 0 &&
             strcmp (lst->next->next->next->str, "atmost") == 0) {
           
           /* Sucht die Rolle R in set_role von O */
           act_srole = srole;
           while (act_srole != NULL) {
             if (strcmp (act_srole->role->Token, 
                         lst->next->next->next->next->next->str) != 0)
               act_srole = act_srole->next;
             else
               break;
           }

           if (act_srole == NULL) {

             /* sucht die Termdefinition von R in T */
             tm = T;
             while (tm != NULL) {
               if (strcmp (tm->expression->ATerm, 
                           lst->next->next->str) != 0)
                 tm = tm->next;
               else
                 break;
	     }

             act_srole = srole;
             while (act_srole != NULL) {
               if (vergleich (act_srole->role, tm->expression->term)
                   == -1)
                 act_srole = act_srole->next;
               else
                 break;
	     }
           }             
 
           if (act_srole != NULL) {
          
             /* bme1 zeigt auf Subrole der R */
             bme1 = act_srole->sub_role;
     
             while (bme1 != NULL) {

               /* Sucht die Subrole R' in set_role von O */
               h_srole = srole;
               while (h_srole != NULL) {
                 if (vergleich (h_srole->role, bme1->bm) == -1)
                   h_srole = h_srole->next;
                 else
                   break;
	       }

               /* n < max von R' */
               if (atoi (lst->next->next->next->next->str) <
                   h_srole->max)
                  h_srole->max = atoi (lst->next->next->next->next->str);
	       
                bme1 = bme1->next;
	     }
	   }
	 }
       }
       w1 = w1->next;
     }
   }

   return (srole);
}


/*****************************************************/
/* FUNCTION  : itoa                                  */
/* PARAMETER : int n,  char s[10]                    */
/* RETURN    : char *                                */
/* PURPOSE   : Hilfsfunktion.                        */
/*             wandet ein Integer zu einem String um.*/
/*****************************************************/

char *itoa (int n, char s[10])
{
    int i, sign;

    if ((sign = n) < 0)
      n = -n;
    i = 0;
   
    do {
      s[i++] = n%10 +'0';
    }while ((n /= 10) > 0);
    
    if (sign < 0)
      s[i++] = '-';
  
    s[i] = '\0';
    return (s);
}



/*****************************************************/
/* FUNCTION  : fcount_astring                        */
/* PARAMETER : char name[36], Welt *W, Terme *T      */
/* RETURN    : Baum *                                */
/* PURPOSE   : Hilfsfunktion.                        */
/*             (atleast count R) wird an astring     */
/*             angehaengt, wenn fuer das Objekt      */
/*             relevant Rolle gilt: count >= 1.      */
/*****************************************************/


Baum *fcount_astring (char name[36], Welt *W, Terme *T )
{
   char str[10];
   Baum *bm, *bm1, *astring;
   set_role *srole; 

   astring = least_most_astring(name, W, T);
   srole = max_srole (name, W, T);

   if (srole != NULL) {
     while (srole != NULL) {
      if (srole->count >= 1) { 
        /* Sucht das Ende von astring */
         bm = astring->child;
         while (bm->brother != NULL)
           bm = bm->brother;
         bm->brother = Node_Baum ( );
         bm = bm->brother;
         
         /* (atleast count R) wird an astring angehaengt */
         bm->lKlammer = (char *)strdup ("(");
         bm->Token = (char *)strdup ("atleast");
         bm->rKlammer = (char *)strdup (")");
         bm->father = astring;
         bm->brother = NULL; 

         itoa (srole->count, str);

         bm->child = Node_Baum ( );;
         bm->child->Token = (char *)strdup (str);
         bm->child->lKlammer = (char *)strdup (" ");
         bm->child->rKlammer = (char *)strdup (" ");
         bm->child->child = NULL;
         bm->child->father = bm;

         bm->child->brother = copy_Baum (srole->role, bm );

       }
       srole = srole->next;
     }
   }

   return (astring);
}


/*****************************************************/
/* FUNCTION  : gen_prim                              */
/* PARAMETER : Baum *bm1, Baum *bm2                  */
/* RETURN    : Baum *C                               */
/* PURPOSE   : Funktion 'gen_prim' wird in Funktion  */
/*             'generalize' aufgerufen.              */
/*             Funktion liefert einen Baum C, der    */
/*             enthaelt alle positive oder negierte  */
/*             primtive Konzeptkomponenten in Baeume */
/*             bm1 und bm2.                          */
/*****************************************************/


Baum *gen_prim (Baum *bm1, Baum *bm2)
{   
    int k;
    Baum *bm, *act_bm, *gen_bm, *act_gen;

    gen_bm = NULL;

    act_bm = bm1->child;
    while (act_bm != NULL) {
    
      k = 0;
            
      /* positive primitive Konzeptkomponente */
      if (act_bm->Token[0] == '<') {
        bm = bm2->child;
        while (bm != NULL) {
          if (strcmp (bm->Token, act_bm->Token) == 0){
            k = 1;
            break;
          }
          else
            bm = bm->brother;
        }
        
        if (k == 1) {
          if (gen_bm == NULL) {
            gen_bm = Node_Baum ( );
            gen_bm->Token = (char *)strdup (bm->Token);
            gen_bm->rKlammer = (char *)strdup (" ");
            gen_bm->lKlammer = (char *)strdup (" ");
            gen_bm->father = NULL;
            gen_bm->brother = NULL;
            gen_bm->child = NULL;            
          }
          else {
            act_gen = gen_bm;
            while (act_gen->brother != NULL)
              act_gen = act_gen->brother;
               
            act_gen->brother = Node_Baum ( );
            act_gen->brother->Token = (char *)strdup (bm->Token);
            act_gen->brother->rKlammer = (char *)strdup (" ");
            act_gen->brother->lKlammer = (char *)strdup (" ");
            act_gen->brother->father = NULL;
            act_gen->brother->child = NULL;
            act_gen->brother->brother = NULL;
	  }
	}
      }
      else {

        /* negierte Konzeptkomponente */
        if (strcmp (act_bm->Token, "not") == 0) {

          bm = bm2->child;
          while (bm != NULL) {
            if (strcmp (bm->Token, "not") == 0 &&
                strcmp (act_bm->child->Token, bm->child->Token) == 0) {
              k = 1;
              break;
            }
            else
              bm = bm->brother;
          }
          
          if (k == 1) {
            if (gen_bm == NULL) {
              gen_bm = copy_Baum (bm, NULL);
              loesch_Baum (gen_bm->brother);
              gen_bm->brother = NULL;
            }
            else {
              act_gen = gen_bm;
              while (act_gen->brother != NULL)
                act_gen = act_gen->brother;
               
              act_gen->brother = copy_Baum (bm, NULL);
              act_gen = act_gen->brother;
              loesch_Baum (act_gen->brother);
              act_gen->brother = NULL;
	    }
	  }
	}
      }
      
      act_bm = act_bm->brother;
    }


    return (gen_bm);
}



/*****************************************************/
/* FUNCTION  : gen_least                             */
/* PARAMETER : Baum *bm1, Baum *bm2                  */
/* RETURN    : Baum *C                               */
/* PURPOSE   : Funktion 'gen_least' wird in Funktion */
/*             'generalize' aufgerufen.              */
/*             Ein Baum (atleast n R) wird geliefert,*/
/*             wenn Baeume bm1 und bm2 die "atleast" */
/*             Restriktion auf R haben.              */
/*             n = min (m1, m2). m1 ist das Maximum  */
/*             aller "atleast" Restriktion auf R im  */
/*             Baum bm1. Analog ist m2.              */ 
/*****************************************************/

Baum *gen_least (Baum *bm1, Baum *bm2)
{
   int i, k;
   Baum *gen_bm, *bm, *h1_bm, *h2_bm, *act1_bm, *act2_bm, *act_gen;

   gen_bm = NULL;

   if (bm1 != NULL && bm2 != NULL) {

     /* bm1 = (and A[1]...A[k]); bm zeigt auf A[1] */
     bm = copy_Baum (bm1->child, NULL);

     do {
       i = 0;

       /* act2_bm ist der Vorgaenger des act1_bm. */
       act1_bm = bm;
       act2_bm = NULL;
     
       while (act1_bm != NULL) {

         /* A[i] = (atleast n R), 1 <= i <= n. */
         if (strcmp (act1_bm->Token, "atleast") == 0 ) {
           i = 1;

           /* h2_bm ist der Vorgaenger des h1_bm */
           h1_bm = act1_bm->brother;
           h2_bm = act1_bm;
   
           while (h1_bm != NULL) {
         
             /* sucht das Maximum  "atleast"  */
             /* Restriktion auf R im Baum bm1 */
             /* Das Maximum vom Baum bm1 ist n*/

             /* A[j] = (atleast m R) */
             if (strcmp (h1_bm->Token, "atleast") == 0 &&
                 vergleich (act1_bm->child->brother,
                            h1_bm->child->brother) != -1) {

               if (atoi (act1_bm->child->Token) <
                   atoi (h1_bm->child->Token))
                 act1_bm->child->Token = 
                      (char *)strdup (h1_bm->child->Token);

               h2_bm->brother = h1_bm->brother;

               /* loescht den Baum A[j] */
               loesch_Baum (h1_bm);
                                 
               h1_bm = h2_bm->brother;
	     }
             else {
               h2_bm = h2_bm->brother;
               h1_bm = h1_bm->brother;
	     }
	   }

           break;
         }

         act2_bm = act1_bm;
         act1_bm = act1_bm->brother;
       }
  
       /* Es gibt Ausdruck der Form (atleast n R) in bm1 */
       if (act1_bm != NULL) {
         
         /* bm2 = (and B[1]... B[h]); h1_bm zeigt auf B[1] */
         h1_bm = bm2->child;

         while (h1_bm != NULL) {
         
           /* A[j] = (atleast m R), i < j <= h. */
           if (strcmp (h1_bm->Token, "atleast") == 0 &&
               vergleich (act1_bm->child->brother,
                          h1_bm->child->brother) != -1) {

             h2_bm = h1_bm->brother;

             while (h2_bm != NULL) {

               /* sucht das Maximum  "atleast"   */
               /* Restriktion auf R in Baum bm2. */
               /* Das Maximum von Baum bm2 ist m */
               if (strcmp (h2_bm->Token, "atleast") == 0 &&
                   vergleich (h1_bm->child->brother,
                              h2_bm->child->brother) != -1) {

                 if (atoi (h1_bm->child->Token) <
                     atoi (h2_bm->child->Token))
                     
                   h1_bm->child->Token = 
                             (char *)strdup (h2_bm->child->Token);
	       }

               h2_bm = h2_bm->brother;
	     }

             break;
	   }

           else
             h1_bm = h1_bm->brother;
	 }

      
         /* Es gibt auch Ausdruck der Form (atleast m R) in bm2. */
         if (h1_bm != NULL) {

           /* Das Minimum von n und m */
           if (atoi (act1_bm->child->Token) >
               atoi (h1_bm->child->Token))
             act1_bm->child->Token = 
               (char *) strdup (h1_bm->child->Token);
           /* Baum (atleast g R) wird am Baum gen_bm angehaengt. */
           /* g = min (n, m). */
           if (gen_bm == NULL) {
             gen_bm = copy_Baum (act1_bm, NULL);
             loesch_Baum (gen_bm->brother);
             gen_bm->brother = NULL;
           }
           else {
             act_gen  = gen_bm;
             while (act_gen->brother != NULL)
               act_gen = act_gen->brother;
             act_gen->brother = copy_Baum (act1_bm, NULL);
             act_gen = act_gen->brother;
             loesch_Baum (act_gen->brother);
             act_gen->brother = NULL;
           }
         }

         /* loescht den Baum (atleast n R) im Baum bm. */

         /* (atleast n R) ist das erste Blatt vom Baum bm. */
         if (act2_bm == NULL) {

           bm = act1_bm->brother;
           loesch_Baum (act1_bm);
      
           act1_bm = NULL;

         }

         /* (atleast n R) ist nicht das erste Blatt vom Baum bm. */
         else {

           act2_bm->brother = act1_bm->brother;
           loesch_Baum (act1_bm);
        
           act1_bm = NULL;

         }
       }
     
     } 
     /* solang noch "atleast" Restriktion im Baum bm gibt. */
     while (i == 1);

     loesch_Baum (bm);
   }
   
   return (gen_bm);
}




/*****************************************************/
/* FUNCTION  : gen_most                              */
/* PARAMETER : Baum *bm1, Baum *bm2                  */
/* RETURN    : Baum *C                               */
/* PURPOSE   : Funktion 'gen_most' wird in Funktion  */
/*             'generalize' aufgerufen.              */
/*             Baum (atmost n R) wird geliefert,     */
/*             wenn Baeume bm1 und bm2 die "atmost"  */
/*             Restriktion auf R haben.              */
/*             n = max (m1, m2). m1 ist das Minimum  */
/*             aller "atmost" Restriktion auf R im   */
/*             Baum bm1. Analog ist m2.              */ 
/*****************************************************/

Baum *gen_most (Baum *bm1, Baum *bm2)
{
   int i, k;
   Baum *gen_bm, *bm, *h1_bm, *h2_bm, *act1_bm, *act2_bm, *act_gen;

   gen_bm = NULL;

   if (bm1 != NULL && bm2 != NULL) {

     /* bm1 = (and A[1]...A[k]); bm zeigt auf A[1] */
     bm = copy_Baum (bm1->child, NULL);

     do {
       i = 0;

       /* act2_bm ist der Vorgaenger des act1_bm. */
       act1_bm = bm;
       act2_bm = NULL;
     
       while (act1_bm != NULL) {

         /* A[i] = (atmost n R), 1 <= i <= n. */
         if (strcmp (act1_bm->Token, "atmost") == 0 ) {
           i = 1;

           /* h2_bm ist der Vorgaenger des h1_bm */
           h1_bm = act1_bm->brother;
           h2_bm = act1_bm;
   
           while (h1_bm != NULL) {
         
             /* sucht das Maximum  "atmost"    */
             /* Restriktion auf R im Baum bm1  */
             /* Das Minimum vom Baum bm1 ist n */
             if (strcmp (h1_bm->Token, "atmost") == 0 &&
                 vergleich (act1_bm->child->brother,
                            h1_bm->child->brother) != -1) {
          
               if (atoi (act1_bm->child->Token) >
                   atoi (h1_bm->child->Token))
                 act1_bm->child->Token = 
                      (char *)strdup (h1_bm->child->Token);

               h2_bm->brother = h1_bm->brother;

               /* loescht den Baum A[j] */
               loesch_Baum (h1_bm);
                                   
               h1_bm = h2_bm->brother;
	     }
             else {
               h2_bm = h2_bm->brother;
               h1_bm = h1_bm->brother;
	     }
	   }

           break;
         }

         act2_bm = act1_bm;
         act1_bm = act1_bm->brother;
       }
  
       /* Es gibt Ausdruck der Form (atmost n R) in bm1 */
       if (act1_bm != NULL) {
         
         /* bm2 = (and B[1]... B[h]); h1_bm zeigt auf B[1] */
         h1_bm = bm2->child;

         while (h1_bm != NULL) {
         
           /* A[j] = (atmost m R), i < j <= h. */
           if (strcmp (h1_bm->Token, "atmost") == 0 &&
               vergleich (act1_bm->child->brother,
                          h1_bm->child->brother) != -1) {

             h2_bm = h1_bm->brother;

             while (h2_bm != NULL) {

               /* sucht das Manimum  "atmost"    */
               /* Restriktion auf R in Baum bm2. */
               /* Das Maximum von Baum bm2 ist m */
               if (strcmp (h2_bm->Token, "atmost") == 0 &&
                   vergleich (h1_bm->child->brother,
                              h2_bm->child->brother) != -1) {

                 if (atoi (h1_bm->child->Token) >
                     atoi (h2_bm->child->Token))
                     
                   h1_bm->child->Token = 
                             (char *)strdup (h2_bm->child->Token);
	       }

               h2_bm = h2_bm->brother;
	     }

             break;
	   }

           else
             h1_bm = h1_bm->brother;
	 }

      
         /* Es gibt auch Ausdruck der Form (atmost m R) in bm2. */
         if (h1_bm != NULL) {

           /* Das Minimum von n und m */
           if (atoi (act1_bm->child->Token) <
               atoi (h1_bm->child->Token))
             act1_bm->child->Token = 
               (char *) strdup (h1_bm->child->Token);

           /* Baum (atmost g R) wird am Baum gen_bm angehaengt. */
           /* g = max (n, m). */
           if (gen_bm == NULL) {
             gen_bm = copy_Baum (act1_bm, NULL);
             loesch_Baum (gen_bm->brother);
             gen_bm->brother = NULL;
           }
           else {
             act_gen  = gen_bm;
             while (act_gen->brother != NULL)
               act_gen = act_gen->brother;
             act_gen->brother = copy_Baum (act1_bm, NULL);
             act_gen = act_gen->brother;
             loesch_Baum (act_gen->brother);
             act_gen->brother = NULL;
           }
         }

         /* loescht den Baum (atmost n R) im Baum bm. */

         /* (atmost n R) ist das erste Blatt vom Baum bm. */
         if (act2_bm == NULL) {

           bm = act1_bm->brother;
           loesch_Baum (act1_bm);
 
           act1_bm = NULL;

         }

         /* (atmost n R) ist nicht das erste Blatt vom Baum bm. */
         else {

           act2_bm->brother = act1_bm->brother;
           loesch_Baum (act1_bm);
     
           act1_bm = NULL;

         }
       }
     
     } 
     /* solang noch "atmost" Restriktion im Baum bm gibt. */
     while (i == 1);

     loesch_Baum (bm);
   }
   
   return (gen_bm);
}



/*****************************************************/
/* FUNCTION  : generalize                            */
/* PARAMETER : Baum *bm1, Baum *bm2                  */
/* RETURN    : Baum *                                */
/* PURPOSE   : generaliziert zwei normalizierte      */
/*             Konzeptterme bm1 und bm2.             */
/*****************************************************/

Baum *generalize (Baum *bm1, Baum *bm2)
{
    int k = 0;
    Baum *gen, *act_gen, *bm;

    gen = NULL; 

   if (bm1 == NULL && bm2 == NULL)
     return NULL;

   if (bm2 == NULL)
     return (bm1);
   else
     if (bm1 == NULL)
       return (bm2); 
    
   if (strcmp (bm1->Token, "NOTHING") == 0)
      return bm2;
    else
      if (strcmp (bm2->Token, "NOTHING") == 0)
        return bm1;
      else
            if (comp (bm1, bm2) == 1 && comp (bm2, bm1) == 1)
              return bm1;
            else {
           
              /* Erstellen des Baums gen. */
              gen = Node_Baum ( );
              gen->Token = (char *)strdup("and");
              gen->lKlammer = (char *)strdup ("(");
              gen->rKlammer = (char *)strdup (")");
              gen->father = NULL;
              gen->brother = NULL;

              gen->child = copy_Baum (gen_prim (bm1, bm2), gen);

         
              act_gen = gen->child;
              if (act_gen == NULL)
                gen->child = copy_Baum (gen_least (bm1, bm2), gen);
              else {
                while (act_gen->brother != NULL)
                  act_gen = act_gen->brother;
                act_gen->brother = copy_Baum (gen_least(bm1, bm2), gen);
              } 
     
              act_gen = gen->child;
              if (act_gen == NULL)
                gen->child = copy_Baum (gen_most(bm1, bm2), gen);
              else {
                while (act_gen->brother != NULL)
                  act_gen = act_gen->brother;
                 act_gen->brother = copy_Baum (gen_most (bm1, bm2), gen);
              }     
          
             return (gen);

          }
}


/*****************************************************/
/* FUNCTION  : MSG_astring                           */
/* PARAMETER : char name[36], Welt *W, Terme *T      */
/* RETURN    : Baum *                                */
/* PURPOSE   : Hilfsfunktion.                        */
/*             (atleast R MSG[O, R]) wird an astring */
/*             angehaengt, wenn fuer das Objekt O    */
/*             relevante Rolle R gilt:               */
/*             count [O, R] = max [O, R].            */ 
/*****************************************************/



Baum *MSG_astring (char name[36],Welt *W, Terme *T)
{
   Baum *bm, *astring;
   set_role *srole, *h_srole, *act_srole;
   Liste *lst;
   Terme *tm;
   O_Welt  *act_OW;


   astring = fcount_astring(name, W, T);
   srole = max_srole (name, W, T);

   if (srole != NULL) {  
     act_srole = srole;
     while (act_srole != NULL) {
       if (act_srole->count == act_srole->max) {
     
         act_srole->MSG = Node_Baum ( );
         act_srole->MSG->Token = (char *)strdup ("NOTHING");
         act_srole->MSG->lKlammer = (char *)strdup (" ");
         act_srole->MSG->rKlammer = (char *)strdup (" ");
         act_srole->MSG->child = NULL;
         act_srole->MSG->brother = NULL;
         act_srole->MSG->father = NULL;   
        
         lst = act_srole->filler; 
      
         while (lst != NULL) {
           act_OW = OW;
           while (act_OW != NULL && strcmp (act_OW->name, lst->str) != 0)
             act_OW = act_OW->next;

           if (act_OW != NULL)
              act_srole->MSG = generalize (act_OW->abstract, act_srole->MSG);

            lst = lst->next;
	 }

         norm (act_srole->MSG, T);

         bm = astring->child;
         while (bm->brother != NULL)
           bm = bm->brother;
        
         bm->brother = Node_Baum ( );
         bm = bm->brother;
         bm->Token = (char *)strdup ("all");
         bm->lKlammer = (char *) strdup ("( ");
         bm->rKlammer = (char *) strdup (")");
         bm->brother = NULL;
         bm->father = astring;  
         bm->child = copy_Baum (act_srole->role, bm);
         bm->child->brother = copy_Baum (act_srole->MSG, bm);
       }
        
       act_srole = act_srole->next;
     }
   }

   return (astring); 
}




/*****************************************************/
/* FUNCTION  : MSG_srole                             */
/* PARAMETER : char O[36], Welt *W, Terme *T,        */
/* RETURN    : set_role *                            */
/* PURPOSE   : Hilfsfunktion.                        */
/*             MSG[O, R]: jede fuer das Objekt O     */
/*             relevante Rolle wird berechnet.       */ 
/*****************************************************/


set_role *MSG_srole (char name[36],Welt *W, Terme *T)
{
   Baum *bm, *astring;
   set_role *srole, *h_srole, *act_srole;
   Liste *lst;
   Terme *tm;
   O_Welt *act_OW;

 
   srole = max_srole (name, W, T);
 
   if (srole != NULL) {
     act_srole = srole;
     while (act_srole != NULL) {
       if (act_srole->count == act_srole->max) {
      
         act_srole->MSG = Node_Baum ( );
         act_srole->MSG->Token = (char *)strdup ("NOTHING");
         act_srole->MSG->lKlammer = (char *)strdup (" ");
         act_srole->MSG->rKlammer = (char *)strdup (" ");
         act_srole->MSG->child = NULL;
         act_srole->MSG->brother = NULL;
         act_srole->MSG->father = NULL;   

         lst = act_srole->filler;
         while (lst != NULL) {
           act_OW = OW;
           while (strcmp (act_OW->name, lst->str) != 0 && act_OW != NULL)
             act_OW = act_OW->next;

           act_srole->MSG = copy_Baum (
                   generalize (act_OW->abstract, act_srole->MSG)
                   , NULL);
           lst = lst->next;
	 }
       }
       act_srole = act_srole->next;
     }
   }

   return (srole); 
}




/*****************************************************/
/* FUNCTION  : abstraction                           */
/* PARAMETER : Welt *W, Terme *T                     */
/* RETURN    : O_Welt *                              */
/* PURPOSE   : bestimmt die Abstraction fuer alle    */
/*             Objekte der Objektwelt W hinsichtlich */
/*             der ihr zugrundlegende Terminologie T.*/
/*****************************************************/

O_Welt *abstraction (Welt *W, Terme *T)
{  
   int k;
   enum BOOL abstComputer;
   Baum  *astring, *abstract, *bm;
   Liste *lst;
   Terme *tm, *act_tm;
   O_Welt *act_OW, *h_OW;
   set_role *sr;
 
   OW = init_OW (W); 
   tm = copy_Terme (T);
   all_exp_Terme (tm);
   norm_Terme(tm);
   
   do {
     abstComputer = FALSE;
     act_OW = OW;

     /* Fuer jedes Objekt O[i] */
     while (act_OW != NULL) {
       if (act_OW->abstChange == TRUE) { 
         abstComputer = TRUE;

         astring = MSG_astring (act_OW->name, W, T);
         act_OW->abstract = norm (all_expand(astring, tm), tm);
         act_OW->srole = MSG_srole(act_OW->name, W, T);
         act_OW->abstChange = FALSE;

         /* Wenn O[i] ein Element des Rolefillers[O[j]] */
         /* ist, dann wird abstChange[O[j]] = TRUE.     */
         h_OW = OW;
         while (h_OW != NULL) {
           sr = h_OW->srole;
           while (sr != NULL) {
             if (sr->filler != NULL) {
               lst = sr->filler;
               while (lst != NULL) {
                 if (strcmp (act_OW->name, lst->str) == 0) {
                   h_OW->abstChange = TRUE;
                  goto I;
                }
                else 
                  lst = lst->next;
	      }
	    }
            sr = sr->next;
	  }
          I: h_OW = h_OW->next;
	 }
       }
       act_OW = act_OW->next;
     }
   } while (abstComputer == TRUE);

   return (OW);
}
