/*********************   class.c   ************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tree.h"
#include "class.h"
 
/*****************************************************/
/* FUNCTION  : all_sub                               */
/* PARAMETER : Baum *a, Terme *t                     */
/* RETURN    : Liste *                               */
/* PURPOSE   : liefert eine Liste mit allen Term-    */
/*             bezeichner, deren definierenden Terme */
/*             der Term a subsummiert.               */
/*****************************************************/


Liste *all_sub (Baum *bm, Terme *tm)
{
   Liste *lst, *first;
   Terme *ptr, *ptr1;


    first = NULL;

   /* Sucht in der Terminologie den Termbezeichner des Terms bm. */
   ptr1 = tm;
   while (ptr1 != NULL) {
     if (vergleich (ptr1->expression->term, bm) == -1)
       ptr1 = ptr1->next;
     else
       break;
   }   

   if (ptr1 != NULL) {

     /* Das erste Element der Liste 'first'  */
     /* ist der Termbezeichner des Terms bm. */ 
     first = Node_Liste ( );
     strcpy (first->str, ptr1->expression->ATerm);
     first->next = NULL;
 
     ptr = tm;
     while (ptr != NULL) {
       /* ptr zeigt auf der Termdefinition von bm selbst. */
       if (vergleich (bm, ptr->expression->term) != -1) 
         ptr = ptr->next;

       else {
         if (bm != NULL && comp (bm, ptr->expression->term) == 1) {

           /* Die Bedingung ist erfuellt. Der Termbezeichner */
           /* wird am Ende der Liste 'first' angehaengt.     */ 
           lst = first;
           while (lst->next != NULL) 
             lst = lst->next;
           lst->next = Node_Liste ( );
           lst = lst->next;
           strcpy (lst->str, ptr->expression->ATerm);
           lst->next = NULL;
         }

         ptr = ptr->next;
       }
     }
   }
   else {
     printf ("In Funktion all_sub ist der Term nicht in der Terminologie \
definiert.\n");
     exit (2);     
   }

   return (first);
}


/*****************************************************/
/* FUNCTION  : all_sub_Terme                         */
/* PARAMETER : Terme *tm                             */
/* RETURN    : Welt *                                */
/* PURPOSE   : Die Funktion 'all_sub' wird auf jede  */
/*             Termdefinition der Terminologie tm    */
/*             angewendet.                           */
/*****************************************************/


Welt *all_sub_Terme (Terme *tm)
{
    Terme *ptr;
    Welt *a_sub, *ptr1;


    if (tm != NULL) {
      ptr = tm; 
      a_sub = Node_Welt ( );
      a_sub->ls = all_sub (ptr->expression->term, tm);
      a_sub->next = NULL;

      ptr = ptr->next;
      while (ptr != NULL) {

        ptr1 = a_sub;
        while (ptr1->next != NULL)
          ptr1 = ptr1->next;

        ptr1->next = Node_Welt ( );
        ptr1 = ptr1->next;
      
        ptr1->ls = all_sub (ptr->expression->term, tm);
        ptr1->next = NULL;

        ptr = ptr->next;
      }
    }
    else {
      printf ("Die Terminologie in Funktion all_sub_Terme ist leer.\n");
      return NULL;
    }

    return (a_sub);
}



/*****************************************************/
/* FUNCTION  : imsub                                 */
/* PARAMETER : Baum *a, Terme *t                     */
/* RETURN    : Liste *                               */
/* PURPOSE   : liefert eine Liste mit Termbezeichner,*/
/*             deren definierenden Terme der Term a  */
/*             unmittelbar subsummiert.              */
/*****************************************************/

Liste *imsub (Baum *bm, Terme *tm)
{

   Liste *lst1, *lst2, *lst3, *first;
   Terme *ptr1, *ptr2;
  
   first = all_sub (bm, tm); 

   /* Jedes Element der Liste wird daraufhin ueberprueft,   */
   /* ob ihr definierenden Term einen anderen definierenden */
   /* Term eines anderen Elements der Liste sumsummiert.    */
   /* Wenn ja, wird das andere Element aus der Liste entfernt. */

  
   if (first != NULL) {
     lst1 = first->next;

     while (lst1 != NULL) {
        ptr1 = tm;  

        /* Sucht den ersten Termbezeichner in der Terminologie */
        while (strcmp (lst1->str, ptr1->expression->ATerm) != 0)
          ptr1 = ptr1->next;
                    
        lst2 = first->next;
      
        while (lst2 != NULL) {
          if (lst1 != lst2) {
            ptr2 = tm;
           
            /* Sucht den zweiten Termbezeichner in der Terminologie */
            while (strcmp (lst2->str, ptr2->expression->ATerm) != 0)
              ptr2 = ptr2->next;

            /* Die Bedingung ist erfuellt und den zweiten */
            /* Termbezeichner wird aus der Liste entfernt.*/
            if (comp (ptr1->expression->term, 
                      ptr2->expression->term ) == 1) {
              lst3 = lst2->next;
              first = del_Liste (lst2->str, first);
              lst2 = lst3;
            }
            else
              lst2 = lst2->next;
          }
          else
            lst2 = lst2->next;
        }
        lst1 = lst1->next;
     }
   }

   return (first);  
}


/*****************************************************/
/* FUNCTION  : imsub_Terme                           */
/* PARAMETER : Terme *t                              */
/* RETURN    : Welt *                                */
/* PURPOSE   : wendet die Funktion 'imsub' auf jede  */
/*             Termdefinition der Terminologie tm an.*/
/*****************************************************/

Welt *imsub_Terme (Terme *tm)
{
    Terme *ptr;
    Welt *sub, *ptr1;

    if (tm != NULL) {
      ptr = tm;
     
      sub = Node_Welt ( );
      sub->ls = imsub (ptr->expression->term, tm);
      sub->next = NULL;

      ptr = ptr->next;
      while (ptr != NULL) {
        
        ptr1 = sub;
        while (ptr1->next != NULL)
          ptr1 = ptr1->next;
        ptr1->next = Node_Welt ( );
        ptr1 = ptr1->next;
        
        ptr1->ls = imsub (ptr->expression->term, tm);
        ptr1->next = NULL;

        ptr = ptr->next;
      }
    }
    else {
      printf ("Die Terminologie in Funktion imsub_Terme ist leer.\n");
      return NULL;
    }

    return (sub);
}


/*****************************************************/
/* FUNCTION  : all_Sub                               */
/* PARAMETER : Baum *a, Terme *t                     */
/* RETURN    : Liste *                               */
/* PURPOSE   : liefert eine Liste mit allen Term-    */
/*             bezeichner, deren definierenden Terme */
/*             den Term a subsummieren.              */
/*****************************************************/


Liste *all_Sub (Baum *bm, Terme *tm)
{
   Liste *lst, *first;
   Terme *ptr, *ptr1;

   first = NULL;

   /* Sucht in der Terminologie den Termbezeichner des Terms bm. */
   ptr1 = tm;
   while (ptr1 != NULL) {
     if (vergleich (ptr1->expression->term, bm) == -1)
       ptr1 = ptr1->next;
     else
       break;
   }  


   if (ptr1 != NULL) { 
     /* Das erste Element der Liste 'first'  */
     /* ist der Termbezeichner des Terms bm. */      

     first = Node_Liste ( );
     strcpy (first->str, ptr1->expression->ATerm);
     first->next = NULL;
 
     ptr = tm;
     while (ptr != NULL) { 

       /* ptr zeigt auf der Termdefinition von bm selbst. */
       if (vergleich (bm, ptr->expression->term) != -1)
         ptr = ptr->next;
       
       else {
         /* bm != (NOTHING) */
         if (strcmp (bm->Token, "NOTHING") != 0) {
           if (comp (ptr->expression->term, bm) == 1) {  

             /* Die Bedingung ist erfuellt. Der Termbezeichner */
             /* wird am Ende der Liste 'first' angehaengt.     */ 
             lst = first;
             while (lst->next != NULL) 
               lst = lst->next;
             lst->next = Node_Liste ( );
             lst = lst->next;
             strcpy (lst->str, ptr->expression->ATerm);
             lst->next = NULL;
           }
         }
         
         /* bm = (NOTHING) */
         else {
           if (strcmp (ptr->expression->equal, ":=C") == 0) {
             lst = first;
             while (lst->next != NULL) 
               lst = lst->next;
             lst->next = Node_Liste ( );
             lst = lst->next;
             strcpy (lst->str, ptr->expression->ATerm);
             lst->next = NULL;
           }
         }

         ptr = ptr->next;
       }
     }
   }
   else {
     printf ("In Funktion all_Sub ist der Term nicht in der Terminologie 
definiert.\n");
     exit (2);
   } 
   
   return (first);
}


/*****************************************************/
/* FUNCTION  : all_Sub_Terme                         */
/* PARAMETER : Terme *tm                             */
/* RETURN    : Welt *                                */
/* PURPOSE   : wendet die Funktion 'all_Sub' auf     */
/*             jede Termdefinition der Terminologie  */
/*             tm an.                                */
/*****************************************************/

Welt *all_Sub_Terme (Terme *tm)
{
    Terme *ptr;
    Welt *a_Sub, *ptr1;

    if (tm != NULL) {
      ptr = tm;
      
      a_Sub = Node_Welt ( );
      a_Sub->ls = all_Sub (ptr->expression->term, tm);
      a_Sub->next = NULL;
    
      ptr = ptr->next;
      while (ptr != NULL) {

        ptr1 = a_Sub;
        while (ptr1->next != NULL)
          ptr1 = ptr1->next;
        ptr1->next = Node_Welt ( );
        ptr1 = ptr1->next;
        
        ptr1->ls = all_Sub (ptr->expression->term, tm);
        ptr1->next = NULL;
 
        ptr = ptr->next;
      }
    }
    else {
      printf ("Die Terminologie in Funktion all_Sub_Terme ist leer.\n");
      return NULL;
    }

    return (a_Sub);
}


/*****************************************************/
/* FUNCTION  : imSub                                 */
/* PARAMETER : Baum *a, Terme *t                     */
/* RETURN    : Liste *                               */
/* PURPOSE   : liefert eine Liste mit Termbezeichner,*/
/*             deren definierenden Terme den Term a  */
/*             unmittelbar subsummieren.             */
/*****************************************************/


Liste *imSub (Baum *bm, Terme *tm)
{
   int k;
   Liste *lst1, *lst2, *lst3, *first, *ptr3, *ptr4;
   Terme *ptr1, *ptr2;

   first = all_Sub (bm, tm);
     
   /* Jedes Element der Liste wird daraufhin ueberprueft,  */
   /* ob ihr definierenden Term einen anderen definierenden*/
   /* Term eines anderen Elements der Liste sumsummiert.   */ 
   /* Wenn ja, wird dieses Element aus der Liste entfernt. */

   if (first != NULL) {
     lst1 = first->next;

     while (lst1 != NULL) {
       ptr1 = tm;  

       /* Sucht den ersten Termbezeichner in der Terminologie. */
       while (strcmp (lst1->str, ptr1->expression->ATerm) != 0)
         ptr1 = ptr1->next;

       lst2 = first->next; 
    
       while (lst2 != NULL) {
         if (lst1 != lst2 ) {
           ptr2 = tm;
           while (strcmp (lst2->str, ptr2->expression->ATerm) != 0)
             ptr2 = ptr2->next;

           if (comp (ptr1->expression->term, 
                     ptr2->expression->term) == 1) {
             
             lst3 = lst1->next;
          
             /* Die Bedingung ist erfuellt und den ersten  */
             /* Termbezeichner wird aus der Liste entfernt.*/
             first = del_Liste (lst1->str, first);
             lst1 = lst3;
             goto Ende;
           } 
         }
         lst2 = lst2->next;
       }

       lst1 = lst1->next;
       Ende:;
     }
   }

   return (first);  
 }


/*****************************************************/
/* FUNCTION  : imSub_Terme                           */
/* PARAMETER : Terme *tm                             */
/* RETURN    : Welt *                                */
/* PURPOSE   : wendet die Funktion 'imSub' auf jede  */
/*             Termdefinition der Terminologie tm an.*/
/*****************************************************/

Welt *imSub_Terme (Terme *tm)
{
    Terme *ptr;
    Welt *Sub, *ptr1;


    if (tm != NULL) { 
      ptr = tm; 
    
      /* Sub: globale Variable */
      Sub = Node_Welt ( );
      Sub->ls = imSub (ptr->expression->term, tm);
      Sub->next = NULL;
          
      ptr = ptr->next;
      while (ptr != NULL) {
        ptr1 = Sub;
        while (ptr1->next != NULL)
          ptr1 = ptr1->next;
        
        ptr1->next = Node_Welt ( );
        ptr1 = ptr1->next;
      
        ptr1->ls = imSub (ptr->expression->term, tm);
        ptr1->next = NULL;
      
        ptr = ptr->next;
      }
    }
    else {
      printf ("Die Terminologie in Funktion imSub_Terme ist leer.\n");
      return NULL;
    }

    return (Sub);
}



