/**********************  comp.c  ************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tree.h"
#include "comp.h"

/* Vordeklaration der Funktion */
enum BOOL comp(Baum *a, Baum *b); 

/* globale Variable */
enum BOOL compare;



/*****************************************************/
/* FUNCTION  : compall                               */
/* PARAMETER : Baum *A, Baum *B                      */
/* RETURN    : void                                  */
/* PURPOSE   : prueft nach, ob Term B in allen Teil- */
/*             termen des Terms A existiert, der in  */
/*             A terminologisch enthalten ist.       */
/*****************************************************/

void compall (Baum *a, Baum *b)
{
   Baum *ptr;
  
   ptr = a->child;  
 
   while ( ptr != NULL && (comp (ptr, b) == TRUE))
     ptr = ptr->brother;
 
   if (ptr == NULL)
     compare = TRUE;
   else
     compare = FALSE;
}



/*****************************************************/
/* FUNCTION  : compexists                            */
/* PARAMETER : Baum *A, Baum *B                      */
/* RETURN    : void                                  */
/* PURPOSE   : prueft nach, ob ein Teilterm B' von B */
/*             existiert, der in A terminologisch    */
/*             enthalten ist.                        */
/*****************************************************/


void compexists (Baum *a, Baum *b)
{
   Baum *ptr;
   
   compare = TRUE;

   ptr = b->child;
   while (ptr != NULL && comp (a, ptr) == FALSE)
     ptr = ptr->brother;

   if (ptr == NULL)
     compare = FALSE;
}



/*****************************************************/
/* FUNCTION  : exist_aset                            */
/* PARAMETER : Baum *B                               */
/* RETURN    : int                                   */
/* PURPOSE   : prueft nach, ob ein Teilterm B' des B */
/*             ein 'aset' Term ist. Wenn ja, liefert */
/*             die Funktion die Anzhal aller         */
/*             Elemente, die aset enthaelt.          */
/*             Ansonsten liefert die Funktion einen  */
/*             Intergerwert -1.                      */   
/*****************************************************/

int exist_aset (Baum *b)
{
   int j = -1;
   Baum *ptr;


   ptr = b->child;
   while (ptr != NULL) {
     if (strcmp (ptr->Token, "aset") == 0) {
       j = 1;
       break;
     }
     else 
       ptr = ptr->brother;
   }

   if (j == -1)
     return -1;
   else
     return (number_aset (ptr));
}



/*****************************************************/
/* FUNCTION  : comp_most                             */
/* PARAMETER : Baum *A, Baum *B                      */
/* RETURN    : void                                  */
/* PURPOSE   : comp_most ist eine Teilfunktion der   */
/*             Funktion comp. Die wird in Funktion   */
/*             comp aufgeruft, wenn  der Teilterm    */
/*             ein 'atmost' Term ist.                */
/*****************************************************/


void comp_most (Baum *a, Baum *b)
{  
   /* a = (atmost n R) */

   /* b = (atmost m R') && n > m */
   if (strcmp (b->Token, "atmost") == 0 &&
       strcmp (a->child->Token, b->child->Token) >= 0) 

     /* compare = comp (R', R). */
     comp (b->child->brother, a->child->brother);

   /* b = (all R (and..aset (A[1],...,A[k])..)) && k <= n */
   else {
     if (strcmp (b->Token, "all") == 0 && (exist_aset (b) != -1) &&
         strcmp (b->child->Token, a->child->brother->Token) == 0 &&
         (vergleich (b->child->child, a->child->brother->child) != -1)&&
         (exist_aset (b)) <=  atoi (a->child->Token))
       
       compare = TRUE;

     /* b = (and ...) */
     else {
       if (strcmp (b->Token, "and") == 0)
         compexists (a, b);

       else
         compare = FALSE;
     }
   }
}




/*****************************************************/
/* FUNCTION  : asetb_aseta                           */
/* PARAMETER : Baum *A, Baum *B                      */
/* RETURN    : void                                  */
/* PURPOSE   : asetb_aseta ist eine Hilfsfunktion    */
/*             der Funktion comp_aset. Die prueft    */
/*             nach, ob alle Elemente des aset_Terms */
/*             B in aset_Term A enthalten sind.      */
/*             Wenn ja, liefert die Funktion einen   */
/*             Intergerwert 1. Ansonsten -1.         */   
/*****************************************************/


int asetb_aseta (Baum *a, Baum *b)
{
   int k;
   Baum *ptr1, *ptr2;

  
   if (strcmp (a->Token, "aset") == 0 && 
       strcmp (b->Token, "aset") == 0) {
  
     /* Zeiger auf das erste Element des aset_Term von b. */
     ptr2 = b->child->child; 
    
     while (ptr2 != NULL) {
       k = 0; 
       
       /* Zeiger auf das erste Element des aset_Term von a. */
       ptr1 = a->child->child;
    
       while (ptr1 != NULL) { 

         /* Ein Element von Term b ist in Term a. */
         if ( strcmp (ptr1->Token, ptr2->Token) == 0) {
           k = 1;
           break;
         }
         else 
           ptr1 = ptr1->brother;
       }

       /* Ein Element von Term b ist nicht in Term a. */
       if (k == 0)
         goto ende;
       else 
         ptr2 = ptr2->brother;
     }
   }
    
   return 1;
  
   ende : return -1;
}     



/*****************************************************/
/* FUNCTION  : comp_aset                             */
/* PARAMETER : Baum *A, Baum *B                      */
/* RETURN    : void                                  */
/* PURPOSE   : comp_aset ist eine Teilfunktion der   */
/*             Funktion comp. Die wird in Funktion   */
/*             comp aufgeruft, wenn der Teilterm     */
/*             ein 'aset' Term ist.                  */
/*****************************************************/


void comp_aset (Baum *a, Baum *b)
{
      
  /* a = aset (A[1],...,A[k]) */


  /* b = aset (B[1],...,B[j]) && (B[1],...,B[j])<= (A[1],...,A[k]). */
  if (strcmp (b->Token, "aset") == 0 &&
      asetb_aseta (a, b) == 1)
   
    compare = TRUE;

  /* b = (and ... ) */
  else { 
    if (strcmp (b->Token, "and") == 0)
      compexists (a, b);

    else 
      compare = FALSE; 
  } 
}


/*****************************************************/
/* FUNCTION  : comp                                  */
/* PARAMETER : Baum *A, Baum *B                      */
/* RETURN    : enum BOOL                             */
/* PURPOSE   : prueft nach, ob der Term B in Term A  */
/*             terminologisch enthalten ist.         */
/*****************************************************/
  
enum BOOL comp (Baum *a, Baum *b)
{ 
   Baum *bm;

   if (a != NULL && b != NULL) { 

   /* b = (NOTHING) */
   if (strcmp (b->Token, "NOTHING") == 0)
     compare = TRUE;
   else {

     /* a = (and ...) oder a = (androle ...) */
     if (strcmp (a->Token, "and") == 0 ||  
         strcmp (a->Token, "androle") == 0)
       compall (a, b);

     else {

       /* a = (all R C) */
       if (strcmp (a->Token, "all") == 0) { 
        
         /* b = (and ...) */
         if (strcmp (b->Token, "and") == 0) 
           compexists (a, b);

         else {  

           /* b = (atmost 0 R') */
           if (strcmp (b->Token, "atmost") == 0 &&
               b->child != NULL && 
               strcmp (b->child->Token, "0") == 0) { 

             /* eine Kopie fuer R */
             bm = copy_Baum (a->child, NULL);
             loesch_Baum (bm->brother);
             bm->brother = NULL;

             /* compare = comp (R R') */
             comp (bm, b->child->brother);
           
             loesch_Baum (bm);
           }
           else {

             /*  b = (all R C') */
             if (strcmp (b->Token, "all") == 0 &&
                 b->child != NULL && a->child != NULL &&
                 strcmp (a->child->Token, b->child->Token) == 0 &&
                 vergleich (a->child->child ,b->child->child) != -1)
                  
               /* compare = comp (C, C') */
               comp (a->child->brother, b->child->brother);

             else 
               compare = FALSE;
           }
         }
       }

       else { 

         /* a = (exists R C) */
         if (strcmp (a->Token, "exists") == 0) { 
                             
           /* b = (and ...) */
           if (strcmp (b->Token, "and") == 0) 
             compexists (a, b);

           else { 

             /* b = (exists R C') */
             if (strcmp (b->Token, "exists") == 0 &&
                 a->child != NULL && b->child != NULL &&
                 strcmp (a->child->Token, b->child->Token) == 0 &&
                 vergleich (a->child->child, b->child->child) != -1)
                 
               /* compare = comp (C, C') */
               comp (a->child->brother, b->child->brother);
              
             else 
               compare = FALSE;
           }
         }   

         else {

           /*  a = (atleast n R)  */
           if (strcmp (a->Token, "atleast") == 0) {
 
             /* b = (atleast m R' && n < m } */    
             if (strcmp (b->Token, "atleast") == 0 &&
                 a->child != NULL && b->child != NULL &&
                 strcmp (a->child->Token, b->child->Token) <= 0)
                  
               /* compare = comp (R, R') */
               comp (a->child->brother, b->child->brother);

             else {

               /* b = (and ... ) */
               if (strcmp (b->Token, "and") == 0)
                 compexists (a, b);

               else
                 compare = FALSE;
             }
           }

           else { 

             /* a = (atmost n R) */
             if (strcmp (a->Token, "atmost") == 0) 
               comp_most (a, b);

             else {
 
               /* a = aset (A[1],...,A[k]) */
               if (strcmp (a->Token, "aset") == 0)
                 comp_aset (a, b);

               else {
 
                 /* a ist primitive Konzepte- oder Rollenkomponent */
                 if ((a->Token)[0] == '<') {

                   /* b = (and ...) oder b = (androle ...) */ 
                   if ( strcmp (b->Token, "and") == 0 ||
                        strcmp (b->Token, "androle") == 0)
                     compexists (a, b);

                   else {
                     /* a = b */
                     if (strcmp (a->Token, b->Token) == 0)
                       compare = TRUE;
                     else 
                       compare = FALSE;
                   }
                 }

                 else { 
                   
                   /* a = (not primitive Konzeptkomponent1) */
                   if (strcmp (a->Token, "not") == 0) {
                     
                     /* b = (and ...) */
                     if ( strcmp (b->Token, "and") == 0)
                       compexists (a, b);
     
                     else {

                       /* b = (not primitive Konzeptkomponent2) && */
                       /* primitive Konzeptkomponent1 = */
                       /* primitive Konzeptkomponent2   */

                       if (strcmp (b->Token, "not") == 0 &&
                           a->child != NULL && b->child != NULL &&
                           strcmp (a->child->Token, b->child->Token) == 0)
                         compare = TRUE;           

                       else 
                         compare = FALSE;
                     }
                   }
                 }
               }
             }
           }
         }
       }
     }
   } 
   }
   else
     if (a == NULL && b != NULL)
       compare = FALSE;
     else  
       compare = TRUE;
   
   return compare; 
}
