/**************************** msc.c ***********************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tree.h"
#include "msc.h"

extern O_Welt *W_Abst; 
extern Welt *MSC;


/*****************************************************/
/* FUNCTION  : MSC_O                                 */
/* PARAMETER : char O[36], Welt *W, Terme *T         */
/* RETURN    : Liste                                 */
/* PURPOSE   : berechnet die Menge der MSC:          */
/*             'Most Specific Concepts' eines Objekts*/
/*             O hinsichtlich der Objektwelt W und   */
/*             der Teminologie T.                    */
/*****************************************************/


Liste *MSC_O (char name[36], Welt *W, Terme *T)
{

   Baum *abstract;
   Terme *tm, *act_tm;
   Liste *lst, *act_lst;
   O_Welt *OW;

   lst = NULL;

   /* W_Abst: globale Variable. W_Abst = abstraction (W, T) */
   
   OW = W_Abst;


   /* Sucht die Abstraction des Objekts O mit dem Bezeichner 'name' in OW. */ 
   while (strcmp (OW->name, name) != 0 && OW != NULL)
     OW = OW->next;
 
   if (OW != NULL) {
     abstract = OW->abstract;
     
     tm = copy_Terme (T);     
     act_tm = tm;

     /* prueft nach, ob die Abstraction des Objekts */
     /* O in der Terminologie T definiert ist.      */

     while (act_tm != NULL) {
       if (vergleich (act_tm->expression->term, abstract) == -1)
         act_tm = act_tm->next;
       else {

         /* Die Abstraction des Objekts O ist in T definiert.    */
         /* MSC(O):= der Konzeptbezeichner, Ihr definierender    */
         /*        Konzeptterm die Abstraction des Objekts O ist.*/

         lst = Node_Liste ( );
         strcpy (lst->str, name);
         lst->next = Node_Liste ( ); 
         strcpy (lst->next->str, act_tm->expression->ATerm);
         lst->next->next = NULL;
      
         break;
       }
     }
  
     /* Die Abstraction des Objekts O ist    */
     /* nicht in der Terminologie definiert. */

     if (act_tm == NULL) {

       /* Eine neue Termdefinition in der Form: 'name :=C abstract' */
       /* wid am Ende der Terminologie anhaengen. Eine neue         */
       /* termporaere Terminologie tm1 entsteht.                    */
       act_tm = tm;
       while (act_tm->next != NULL)
         act_tm = act_tm->next;
       act_tm->next = Node_Terme ( );
     
       act_tm->next->expression = Node_Term ( );
       act_tm->next->expression->ATerm = (char *)strdup (name);
       act_tm->next->expression->equal = (char *)strdup (":=C");
       act_tm->next->expression->term = abstract;
       act_tm->next->next = NULL;

       /* berechnet imSub(abstract) in der neue Terminologie. */
       lst = imSub (act_tm->next->expression->term, tm);

     }
   }
   else {
     printf ("Das Objekt %s ist nicht in W definiert.\n");
     return NULL;
   }

 
   return (lst);
}


/*****************************************************/
/* FUNCTION  : MSC_Welt                              */
/* PARAMETER : Welt *W, Terme *T                     */
/* RETURN    : Liste                                 */
/* PURPOSE   : berechnet das MSC fuer jeden Objekt   */
/*             der Objektwelt W hinsichtlich der ihr */
/*             zugrundliegenden Termiologie T.       */
/*****************************************************/

Welt *MSC_Welt (Welt *W, Terme *T)
{

   Liste *lst;
   Welt *act_Welt;
       
   /* MSC: globale Variable */
   MSC = NULL;

   /* lst: die Menge aller Objektbezeichner in W */
   lst = Objekt_Liste (W);
   
   while (lst != NULL) {
     if (MSC == NULL) {
       MSC = Node_Welt ( );
       MSC->ls = MSC_O (lst->str, W, T);
       MSC->next = NULL;
     }
     else {
       act_Welt = MSC;
       while (act_Welt->next != NULL)
         act_Welt = act_Welt->next;
       act_Welt->next = Node_Welt ( );
       act_Welt = act_Welt->next;
       act_Welt->ls = MSC_O (lst->str, W, T);
       act_Welt->next = NULL;
     }

     lst = lst->next;
   }
  
   return (MSC);
}