/**************************norm.c************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include "tree.h"
#include "norm.h"

enum BOOL change;
Baum *norm (Baum *bm, Terme *tm);

/*****************************************************/
/* FUNCTION  : number_aset                           */
/* PARAMETER : Baum *                                */
/* RETURN    : int                                   */
/* PURPOSE   : berechnet die Anzahl aller Elemente,  */ 
/*             die aset enthaelt.                    */
/*****************************************************/

int number_aset (Baum *bm)
{
  int i = 0;
  Baum *ptr;

 
  if (bm != NULL && strcmp (bm->Token, "aset") == 0) {

    /* ptr zeigt auf das erste Element von aset_Term */

    if (bm->child != NULL) {

      ptr = bm->child->child;
      while (ptr != NULL) {
        i++;
        ptr = ptr->brother;
      }
    }

  }
  return i;
}



/*****************************************************/
/* FUNCTION  : vergleich                             */
/* PARAMETER : Baum *a, Baum *b                      */
/* RETURN    : int                                   */
/* PURPOSE   : Hilfsfunktion                         */
/*             prueft nach, ob zwei Konzepte (Rolle) */
/*             gleich sind. Wenn nicht, liefert die  */
/*             Funktion einen Integerwert -1.        */
/*****************************************************/

int vergleich (Baum *bm1, Baum *bm2)
{

    if (bm1 == NULL && bm2 == NULL)
      return 1;
    
    if ((bm1 == NULL && bm2 != NULL) ||
        (bm1 != NULL && bm2 == NULL))
      return -1;

    /* zwei Baeume rekursiv vergleichen */ 

    /* Die Nachfolger 'child' von zwei Baeumen werden verglichen. */

    if (bm1->child != NULL && bm2->child != NULL) {
      vergleich (bm1->child, bm2->child); 
      if (vergleich (bm1->child, bm2->child) == -1)
         return -1;
    }
    else {
      if ((bm1->child != NULL && bm2->child == NULL) ||
          (bm1->child == NULL) && bm2->child != NULL)
           /* {Zwei Baeume haben unterschiedlichen Struktur} */
        return -1;
    }

    /* gleicher Durchlauf fuer die Nachfolger 'brother' der Baeume */

    if (bm1->brother != NULL && bm2->brother != NULL){
      vergleich (bm1->brother, bm2->brother); 
      if (vergleich (bm1->brother, bm2->brother) == -1)
        return -1;
    }
    else {
      if ((bm1->brother != NULL && bm2->brother == NULL) ||
          (bm1->brother == NULL) && bm2->brother != NULL)
        return -1;
    }
    
    /* zwei Baeume sind ungleich. */
    if (strcmp (bm1->Token, bm2->Token) != 0)
      return -1;
}



/*****************************************************/
/* FUNCTION  : and_and                               */
/* PARAMETER : Baum *                                */
/* RETURN    : void                                  */
/* PURPOSE   : eliminiert geschachtelte 'and'        */
/*****************************************************/


void and_and (Baum *bm)
{   
   Baum *ptr, *ptr1, *ptr2;

   if (bm != NULL) {

   /* rekursive Anwendung der Funktion auf den Baum mit der Wurzel bm */
   if (bm->child != NULL)
     and_and (bm->child);
   if (bm->brother != NULL)
     and_and (bm->brother);
   
   /* Die Wurzel des Baums. */
   if (strcmp (bm->Token, "and") == 0) {

     /* sucht nach 'and' bei den Blaetter des Baums. */

     /* Das erste Blatt des Baums */
     if (bm->child != NULL ) {
       if (strcmp (bm->child->Token, "and") == 0) {
                                         
         change = TRUE;

         /* Der Baumknoten des Blattes 'and' wird geloescht.*/
         /* Die neue Verbindungen zu seiner Umgebung werden hergestellt.*/
         if (bm->child->child != NULL) {
           ptr = bm->child->child; 
    
           while (ptr->brother != NULL) {
             ptr->father = bm;
             ptr = ptr->brother;
           }
       
           ptr->father = bm;
           ptr->brother = bm->child->brother;
         }     
  
         ptr = bm->child;
         bm->child = bm->child->child;
         free (ptr->lKlammer);
         free (ptr->rKlammer);
         free (ptr->Token);
         free (ptr);
      
       }
     
       /* weiterer Blaette des Baums */
       /* Durchlauf analog zu oben   */
       ptr = bm->child;
       ptr1 = bm->child->brother;
     
       while (ptr1 != NULL) {
         if (strcmp (ptr1->Token, "and") == 0) {
           change = TRUE;
           ptr2 = ptr1->child;
           ptr->brother = ptr2;
           while (ptr2->brother != NULL) {
             ptr2->father = bm;
             ptr2 = ptr2->brother;
           }
          
           ptr2->father = bm;          
           ptr2->brother = ptr1->brother;
           free (ptr1->lKlammer);
           free (ptr1->rKlammer);
           free (ptr1->Token);
           free (ptr1);

           ptr = ptr2;
           ptr1 = ptr2->brother;
         }
       
         else {
           ptr = ptr1;
           ptr1 = ptr1->brother;
         }
       }
     }
   }
 }
}

/*****************************************************/
/* FUNCTION  : role_role                             */
/* PARAMETER : Baum *                                */
/* RETURN    : void                                  */
/* PURPOSE   : eliminiert geschachtelte 'androle'    */
/*****************************************************/

void role_role (Baum *bm)
{   
   Baum *ptr, *ptr1, *ptr2;

   /* aenlicher Prozess wie bei der Funktion 'and_and' */

   if (bm != NULL) {

   if (bm->child != NULL)
     role_role (bm->child);
   if (bm->brother != NULL)
     role_role (bm->brother);
   
   if (strcmp (bm->Token, "androle") == 0) {
     if (bm->child != NULL) {

       if (strcmp (bm->child->Token, "androle") == 0) {
         change = TRUE;
         ptr = bm->child->child;
         while (ptr->brother != NULL) {
           ptr->father = bm;
           ptr = ptr->brother;
         }
         ptr->father = bm;
         ptr->brother = bm->child->brother;
       
         ptr = bm->child;
         bm->child = bm->child->child;
         free (ptr->lKlammer);
         free (ptr->rKlammer);
         free (ptr->Token);
         free (ptr);
       }


       ptr = bm->child;
       ptr1 = bm->child->brother;
      
        while (ptr1 != NULL) {
         if (strcmp (ptr1->Token, "androle") == 0) {
           change = TRUE;
           ptr2 = ptr1->child;
           ptr->brother = ptr2;
           while (ptr2->brother != NULL) {
             ptr2->father = bm;
             ptr2 = ptr2->brother;
           }
           ptr2->father = bm;          
           ptr2->brother = ptr1->brother;

           free (ptr1->lKlammer);
           free (ptr1->rKlammer);
           free (ptr1->Token);
           free (ptr1);

           ptr = ptr2;
           ptr1 = ptr2->brother;
         }
 
         else {
           ptr = ptr1;
           ptr1 = ptr1->brother;
         }
       }
     }
   } 
 }
}

/*****************************************************/
/* FUNCTION  : C_notC                                */
/* PARAMETER : Baum *                                */
/* RETURN    : void                                  */
/* PURPOSE   : Ein Ausdruck der Form                 */
/*             (..(and ..<C>.. (not <C>)..)..) wird  */
/*             als (..NOTHING ..) ausgewertet.       */
/*****************************************************/

void C_notC (Baum *bm)
{
   Baum *ptr, *ptr1;

   if (bm != NULL) {

   /* rekursive Anwendung der Funktion auf den Baum mit der Wurzel bm */  
   if (bm->child != NULL)
     C_notC (bm->child);
   if (bm->brother != NULL)
     C_notC (bm->brother);

   /* bm = (and A[1] ... A[n]) */
   if (strcmp (bm->Token, "and") == 0) {
           
     /* ptr zeigt auf A[1] */
     ptr = bm->child;
     
     while (ptr != NULL) {

       /* A[i] = (not <C>), 1 <= i <= n */
       if (strcmp (ptr->Token, "not") == 0) { 
                  
         /* Die Suche nach <C> beginnt bei A[1] */
         ptr1 = bm->child;
         while (ptr1 != NULL) { 
             
           /* A[j] = <C>, 1 <= j <= n, i <> j */
           if (strcmp (ptr->child->Token, ptr1->Token) == 0) {
             change = TRUE;

             /* Den Ausdruck als 'NOTHING' auswerten */
             bm->Token = (char *)strdup ("NOTHING");
             ptr = bm->child;
             bm->child = NULL;
             
             /* Nicht benoetigten Speicher freigeben */
             loesch_Baum (ptr);
           
             goto Ende; 
           
           }
           /* weitere Suche nach <C> */
           ptr1 = ptr1->brother;
         }
       }
       /* weitere Suche nach 'not' */ 
       ptr = ptr->brother;
    }
    Ende:;
   }   
 }
}

/*****************************************************/
/* FUNCTION  : least_most                            */
/* PARAMETER : Baum *                                */
/* RETURN    : void                                  */
/* PURPOSE   : wertet einen Ausdruck, der Teile      */
/*             (atleast n R1) und (atmost m R2)      */
/*             enthaelt und die Bedingungen n > m und*/
/*             termsub (R2, R1, tm) = 1 erfuellt,    */
/*             als 'NOTHING' aus.                    */
/*****************************************************/


void least_most (Baum *bm, Terme *tm)
{
  Baum *ptr, *ptr1;

  if (bm != NULL && tm != NULL){
  
  /* rekursive Anwendung der Funktion auf den Baum mit der Wurzel bm */  
  if (bm->child != NULL)
    least_most (bm->child, tm);
  if (bm->brother != NULL)
     least_most (bm->brother, tm);

  /* bm = (and A[1]... A[n]) */
  if (strcmp (bm->Token, "and") == 0) {
 
    /* ptr zeigt auf A[1] */
    ptr = bm->child;
    while (ptr != NULL) {

      if (strcmp (ptr->Token, "atleast") != 0)
        ptr = ptr->brother;
      else {
        /* A[i] = (atleast n R), 1 <= i <= n */

        /*Die Suche nach 'atmost' beginnt bei A[1] */
        ptr1 = bm->child;
 
        while (ptr1 != NULL) { 
          if (strcmp (ptr1->Token, "atmost") != 0)
            ptr1 = ptr1->brother;
          else {
             if ( ptr->child != NULL && ptr1->child != NULL &&
                  atoi (ptr->child->Token) > atoi(ptr1->child->Token) &&
                  comp(norm (all_expand (ptr1->child->brother, tm), tm), 
                       norm (all_expand (ptr->child->brother, tm), tm)) == 1) {
              change = TRUE;
           
              bm->Token = (char *)strdup ("NOTHING");
              loesch_Baum (bm->child);
              bm->child = NULL; 
              goto Ende;
            }
            ptr1 = ptr1->brother;
          }
        }   
        ptr = ptr->brother;
      }
    }
  
   Ende:;
  }
 }    
}


    
/*****************************************************/
/* FUNCTION  : and_NOTHING                           */
/* PARAMETER : Baum *                                */
/* RETURN    : void                                  */
/* PURPOSE   : wertet einen Ausdruck, der 'NOTHING'  */
/*             enthaelt, als 'NOTHING' aus.          */
/*****************************************************/

void and_NOTHING (Baum *bm)
{
   Baum *ptr, *ptr1;

   if (bm != NULL){

   /* rekursive Anwendung der Funktion auf den Baum mit der Wurzel bm */  
   if (bm->child != NULL)
     and_NOTHING (bm->child);
   if (bm->brother != NULL)
     and_NOTHING (bm->brother);

   /* bm = (and A[1] ... A[n]) */
   if (strcmp (bm->Token, "and") == 0) { 

     /* ptr zeigt auf A[1] */
     ptr = bm->child;

     while (ptr != NULL) {

       /* A[i] = NOTHING */
       if (strcmp (ptr->Token, "NOTHING") == 0) {
      
          change= TRUE;

          /* Den Ausdruck als 'NOTHING' auswerten */
          bm->Token = (char *)strdup ("NOTHING");
          ptr = bm->child;
          bm->child = NULL; 
          
          /* Nicht benoetigten Speicher freigeben */
          loesch_Baum (ptr);
           
          break;
       }
       else
         ptr = ptr->brother;
     }
   } 
  }
}


/*****************************************************/
/* FUNCTION  : all_and                               */
/* PARAMETER : Baum *                                */
/* RETURN    : void                                  */
/* PURPOSE   : fasst den Teilausdruck                */
/*             (all R (and C(1)...C(i))) und         */
/*             (all R (and C(i+1)...C(n))            */
/*             in (all R (and C(1)...C(n))) zusammen.*/
/*****************************************************/


void all_and (Baum *bm)
{

    Baum *ptr, *ptr1, *ptr2, *ptr3, *ptr4, *sp, *ap;
   
    if (bm != NULL) {
  
      
    /* rekursive Anwendung der Funktion auf den Baum mit der Wurzel bm */
    if (bm->child != NULL)
      all_and (bm->child);
    if (bm->brother != NULL)
      all_and (bm->brother);

    /* bm = (and A[1] ... A[n]) */
    if (strcmp (bm->Token, "and") == 0) {

      /* ptr zeigt auf A[1] */ 
      ptr = bm->child;

      while (ptr != NULL) {

        /* A[i] = (all R C) */
        if (strcmp (ptr->Token, "all") == 0) { 

          /* ptr4 zeigt auf den Vorgaenger von ptr1 */
          ptr4 = ptr;
          ptr1 = ptr->brother; 
                            
          while (ptr1 != NULL) {

             /* A[j] = (all R C'), j > i */ 
             if (strcmp (ptr1->Token, "all") == 0 &&
                 ptr->child != NULL && ptr1->child !=NULL &&
                 strcmp (ptr->child->Token, ptr1->child->Token) == 0 &&
                 vergleich (ptr->child->child, ptr1->child->child) != -1) {
                 
               change = TRUE;

               /* Gewinnen eines zusammenfassenden Ausdrucks sp */

               /* ptr2 zeigt auf C, ptr3 zeigt auf C' */
               ptr2 = ptr->child->brother;
               ptr3 = ptr1->child->brother;

               if ( ptr2->child != NULL) {              
                 sp = copy_Baum (ptr2, NULL);
                 ap = sp->child;              
                 while (ap->brother != NULL)
                   ap = ap->brother;
                              
                 if (ptr3->child != NULL)
                   ap->brother = copy_Baum (ptr3->child, sp);
                 else
                   ap->brother = copy_Baum (ptr3, sp);
               }
               else {
                 if (ptr3->child != NULL) {
                   sp = copy_Baum (ptr3, NULL);
                   ap = sp->child;              
                   while (ap->brother != NULL)
                     ap = ap->brother;
           
                   ap->brother = copy_Baum (ptr2, sp);
                 }
                 else {
                   sp = Node_Baum ( );
                   sp->Token = (char *)strdup ("and");
                   sp->lKlammer = (char *) strdup ("(");
                   sp->rKlammer = (char *) strdup (")");
                   sp->brother = NULL;
                   sp->father = NULL;
                   sp->child = copy_Baum (ptr2, sp);
                   sp->child->brother = copy_Baum (ptr3, sp);
                 }
               }
 
               /* Die Verbindung zu sp wird hergestellt */
               ptr->child->brother = sp;
               sp->father = ptr;
         
               /* Speicher freigeben fuer die Teile,*/
               /* die zusammen gefasst werden. */
         
               loesch_Baum (ptr2);
               ptr4->brother = ptr1->brother;
         
               loesch_Baum (ptr1->child);
               free (ptr1->Token);
               free (ptr1->lKlammer);
               free (ptr1->rKlammer);
               free (ptr1);
             
               ptr1 = ptr4->brother;
             }

             else {
               ptr4 = ptr4->brother;
               ptr1 = ptr1->brother;
             }
           }
        }

        ptr = ptr->brother;
      }
    }    
  }
}



/*****************************************************/
/* FUNCTION  : all_NOTHING                           */
/* PARAMETER : Baum *                                */
/* RETURN    : void                                  */
/* PURPOSE   : Transformation von (all R NOTHING)    */
/*             in (atmost 0 R).                      */
/*****************************************************/


void all_NOTHING (Baum *bm)
{
   Baum *ptr, *ptr1;

   if (bm != NULL) {

   /* rekursive Anwendung der Funktion auf den Baum mit der Wurzel bm */
   if (bm->child != NULL)
      all_NOTHING (bm->child);
   if (bm->brother != NULL)
      all_NOTHING (bm->brother);
 

   /* bm = (all R NOTHING) */
   if (strcmp (bm->Token, "all") == 0 && 
       bm->child != NULL && bm->child->brother != NULL &&
       strcmp (bm->child->brother->Token, "NOTHING") == 0) {

      change = TRUE;
     
      /* Transformation in (atmost 0 R) */

      bm->Token = (char *)strdup ("atmost");

      /* ptr1 = R */
      ptr1 = bm->child;
      loesch_Baum (ptr1->brother);
      ptr1->brother = NULL;

      ptr = Node_Baum ( );
      ptr->Token = (char *) strdup ("0");
      ptr->lKlammer = (char *) strdup (" ");
      ptr->rKlammer = (char *) strdup (" ");
      ptr->father = bm;
      bm->child = ptr;

      ptr->child = NULL;
      ptr->brother = ptr1;
  
    }
  }
}


/*****************************************************/
/* FUNCTION  : all_least                             */
/* PARAMETER : Baum *                                */
/* RETURN    : void                                  */
/* PURPOSE   : wertet einen Ausdruck, der Teile      */
/*             (all R1 NOTHING) (atleast n2 R2)      */
/*             enthaelt und die Bedingung:           */
/*             termsub (R1, R2, tm) = 1 erfuellt,    */
/*             als 'NOTHING' aus.                    */
/*****************************************************/


void all_least (Baum *bm, Terme *tm)
{
  Baum *ptr, *ptr1, *ptr2, *ptr3;

  if (bm != NULL && tm != NULL) {


  /* rekursive Anwendung der Funktion auf den Baum mit der Wurzel bm */  
  if (bm->child != NULL)
    all_least (bm->child, tm);
  if (bm->brother != NULL)
    all_least (bm->brother, tm);

  /* bm = (and A[1]... A[n]) */ 
  if (strcmp (bm->Token, "and") == 0) {
    ptr = bm->child;
    while (ptr != NULL) {

      /* A[i] = (all R1 NOTHING) */
      if (strcmp (ptr->Token, "all") == 0 && 
          ptr->child != NULL && ptr->child->brother != NULL &&
          strcmp (ptr->child->brother->Token, "NOTHING") == 0) {

        ptr1 = bm->child;
        while (ptr1 != NULL) { 
 
          if (strcmp (ptr1->Token, "atleast") != 0)
            ptr1 = ptr1->brother;
          else {

            /* A[j] = (atleast n2 R2) */

            /* Eine Kopie fuer R1 */
            ptr2 = copy_Baum (ptr->child, NULL);
            loesch_Baum (ptr2->brother);
            ptr2->brother = NULL;
           
            /* Eine Kopie fuer R2 */
            ptr3 = copy_Baum (ptr1->child->brother, NULL);
      
            /* termsub (R1, R2, tm) == 1 */
            if (comp (norm (all_expand (ptr2, tm), tm), 
                      norm ( all_expand (ptr3, tm), tm)) == 1) {
              change = TRUE;

              /* Den Ausdruck als 'NOTHING' auswerten. */
              bm->Token = (char *)strdup ("NOTHING");

              loesch_Baum (bm->child);
              bm->child = NULL; 

              loesch_Baum (ptr2);
              loesch_Baum (ptr3);

              goto ende;
            }
            ptr1 = ptr1->brother;
          }
        }
        ptr = ptr->brother;
      }      
      else  
        ptr = ptr->brother;
    }
    ende:;
   }
 }
}

/*****************************************************/
/* FUNCTION  : least_least                           */
/* PARAMETER : Baum *                                */
/* RETURN    : void                                  */
/* PURPOSE   : fasst die Teilausdruecke (atleast n R)*/
/*             und (atleast m R) in (atleast l R)    */
/*             zusammen. l = max (n,m).              */
/*****************************************************/

void least_least (Baum *bm)
{
   Baum *ptr, *ptr1, *ptr2;


   if (bm != NULL) {

   /* rekursive Anwendung der Funktion auf den Baum mit der Wurzel bm */
   if (bm->child != NULL)
     least_least (bm->child);
   if (bm->brother != NULL)
     least_least (bm->brother);


   /* bm = (and A[1]...A[u]) */
   if (strcmp (bm->Token, "and") == 0) {

      ptr = bm->child;
      while (ptr != NULL) {

        /* A[i] = (atleast n R) */ 
        if (strcmp (ptr->Token, "atleast") == 0) { 
            
           /* ptr2 zeigt auf den Vorgaenger von ptr1. */ 
           /* Die Suche nach weitere 'atleast' beginnt mit dem    */
           /*  unmittelbaren Nachfolger des ersten 'atleast' A[i] */
      
           ptr2 = ptr;
           ptr1 = ptr->brother;

           while (ptr1 != NULL) {

             /* A[j] = (atleast m R) */
             if (strcmp (ptr1->Token, "atleast") == 0 &&
                 ptr->child != NULL && ptr1->child != NULL &&
                 vergleich (ptr->child->brother, 
                            ptr1->child->brother) != -1) {
              
               change = TRUE;
              
               /* n < m */
               if (atoi (ptr->child->Token) < atoi (ptr1->child->Token))
                  ptr->child->Token = (char *)strdup (ptr1->child->Token);   
        
               ptr2->brother = ptr1->brother;
        
               loesch_Baum (ptr1->child);
               free (ptr1->lKlammer);
               free (ptr1->rKlammer);
               free (ptr1->Token);
               free (ptr1);
   
               ptr1 = ptr2->brother;
             }
             else {
               ptr1 = ptr1->brother;
               ptr2 = ptr2->brother;
             }
           }
         }

         ptr = ptr->brother; 
      }
    }
  }
}


/*****************************************************/
/* FUNCTION  : most_most                             */
/* PARAMETER : Baum *                                */
/* RETURN    : void                                  */
/* PURPOSE   : fasst die Teilausdruck (atmost n R)   */
/*             und (atmost m R) in (atmost l R)      */
/*             zusammen. l = min (n,m).              */
/*****************************************************/


void most_most (Baum *bm)
{
   Baum *ptr, *ptr1, *ptr2;

   if (bm != NULL) {


   /* Durchlauf analog zu Funktion least_least */

   if (bm->child != NULL)
     most_most (bm->child);
   if (bm->brother != NULL)
     most_most (bm->brother);

   if (strcmp (bm->Token, "and") == 0) {

      ptr = bm->child;
      while (ptr != NULL) {

        if (strcmp (ptr->Token, "atmost") == 0) {

           ptr2 = ptr;
           ptr1 = ptr->brother;

           while (ptr1 != NULL) {
             if (strcmp (ptr1->Token, "atmost") == 0 &&
                 ptr->child != NULL && ptr1->child != NULL &&
                 vergleich (ptr->child->brother, 
                            ptr1->child->brother) != -1) {

               change = TRUE;

               if (atoi (ptr->child->Token) > atoi (ptr1->child->Token))
                 ptr->child->Token = (char *)strdup (ptr1->child->Token);   

               ptr2->brother = ptr1->brother;

               loesch_Baum (ptr1->child);
               free (ptr1->lKlammer);
               free (ptr1->rKlammer);
               free (ptr1->Token);
               free (ptr1);

               ptr1 = ptr2->brother;
             }
             else {
               ptr1 = ptr1->brother;
               ptr2 = ptr2->brother;
             }
           }
         }
         ptr = ptr->brother;
      }
    }
  }
}


/*****************************************************/
/* FUNCTION  : term_term                             */
/* PARAMETER : Baum *                                */
/* RETURN    : void                                  */
/* PURPOSE   : formt den Teilausdruck (and ..A.. A..)*/
/*             oder aset(..a,..,a,..) zu (and ..A..) */
/*             oder aset(..a,..) um.                 */
/*****************************************************/

void term_term (Baum *bm)
{
   Baum *ptr, *ptr1, *ptr2;
  
   if (bm != NULL) {
 
   /* rekursive Anwendung der Funktion auf den Baum mit der Wurzel bm */
   if (bm->child != NULL)
     term_term (bm->child);
   if (bm->brother != NULL)
     term_term (bm->brother);

   /* bm = (and A[1]...A[n]) oder bm = (androle A[1]...A[n]) */  
   if (strcmp (bm->Token, "and") == 0 ||
       strcmp (bm->Token, "androle") == 0) {

     ptr = bm->child;
     while (ptr != NULL) {
       
       /* ptr2: Vorgaenger von ptr1 */
       ptr2 = ptr; 
       ptr1 = ptr->brother;
      
       while (ptr1 != NULL) {
      
         /* A[i] = A[j], i <> j */         
         if (strcmp (ptr->Token, ptr1->Token) == 0 &&
             vergleich (ptr->child, ptr1->child) != -1) {
        
           change = TRUE;

           /* Die Verbindung versetzen und Speicher fuer */
           /* einen des identischen Ausdrucks freigeben. */        
           ptr2->brother = ptr1->brother;
      
           loesch_Baum (ptr1->child);
           free (ptr1->lKlammer);
           free (ptr1->rKlammer);
           free (ptr1->Token);
           free (ptr1);

           ptr1 = ptr2->brother;
         }
  
         else {
           ptr2 = ptr2->brother;
           ptr1 = ptr1->brother;
         }
       }
       ptr = ptr->brother;
     }
   }
   else {
     /* A[i] = aset (a1, a2, ..., ak) */
     if (strcmp (bm->Token, "aset") == 0 && bm->child->child != NULL) {
       
       /* ptr zeigt auf a1 */
       ptr = bm->child->child;

       while (ptr != NULL) {
         
         /* ptr2 zeigt auf den Vorgaenger von ptr1 */
         ptr2 = ptr;
         ptr1 = ptr->brother;
          
         while (ptr1 != NULL) {

           /* am = an, m <> n */
           if (strcmp (ptr->Token, ptr1->Token) == 0) {
             change = TRUE;
             ptr2->brother = ptr1->brother;
             free (ptr1);
             ptr1 = ptr2->brother;
           }
           else {
             ptr2 = ptr2->brother;
             ptr1 = ptr1->brother;
           }
         }
         ptr = ptr->brother;
       }
     }
   }      
 }
}



/*****************************************************/
/* FUNCTION  : aset_aset                             */
/* PARAMETER : Baum *                                */
/* RETURN    : void                                  */
/* PURPOSE   : fasst den Teilausdruck                */
/*             (and..aset(a, b, d)..aset(d, b)..)..) */
/*             in  (and..aset(b, d) zusammen.        */
/*****************************************************/

void aset_aset (Baum *bm)
{
   int n;  
   Baum *ptr, *ptr1, *ptr2, *ptr3, *ptr4, *ptr5;

   if (bm != NULL) {


  /* rekursive Anwendung der Funktion auf den Baum mit der Wurzel bm */
   if (bm->child != NULL)
     aset_aset (bm->child);
   if (bm->brother != NULL) 
     aset_aset (bm->brother);
   
   /* bm = (and A[1] ... A[n]) */
   if (strcmp (bm->Token, "and") == 0) {

     ptr = bm->child;
     while (ptr != NULL) {

        /* Die Suche nach 'aset'. */
       if (strcmp (ptr->Token, "aset") != 0)
          ptr = ptr->brother;
       else
         break;
     }

     /* A[i] = aset (a1, ..., ak) */
     if (ptr != NULL) { 

        /* ptr2 zeigt auf den Vorgaenger von ptr1. */
        ptr2 = ptr; 
        ptr1 = ptr->brother;

        while (ptr1 != NULL) {

          /* A[j] = aset (b1, ...bm) */
          if (strcmp (ptr1->Token, "aset") == 0) { 
            change = TRUE;

           /* ptr4 zeigt auf das erste Element des 1. 'aset'. */
           ptr4 = ptr->child->child;

           /* Es wird nachgeprueft, ob ein Element vom      */
           /* ersten 'aset auch zum zweiten 'aset' gehoehrt.*/   
           while (ptr4 != NULL) {

             /* ptr5 zeigt auf das zweiten Element des 2. 'aset'. */
             ptr5 = ptr1->child->child;
            
             n = 0;
             while (ptr5 != NULL) {

               /* gemeinsames Element */
               if (strcmp (ptr4->Token, ptr5->Token) == 0) {
                 n = 1;
                 break; 
               }
               else 
                 ptr5 = ptr5->brother;
             }

             /* Das ist ein gemeinsames Element. */
             if (n == 1)
               ptr4 = ptr4->brother; 

             /* Das ist kein gemeinsames Element und  wird */
             /* aus dem ersten 'aset' entfernt. */
             else {
               
               /* Das Element, das entfernet soll, ist */
               /* das erste Element vom ersten 'aset'. */
               if (ptr4 == ptr->child->child){ 
                 ptr3 = ptr4->brother;
                 ptr->child->child = ptr3;
                 free (ptr4);
                 ptr4 = ptr3; 
               }

               else { 
                 ptr3 = ptr->child->child;

                 while (ptr3->brother != ptr4)
                   ptr3 = ptr3->brother;
               
                 ptr3->brother = ptr4->brother; 
                 free (ptr4);
                 ptr4 = ptr3->brother;
              }
            }
          }

          /* Speicher fuer das zweite 'aset' freigeben */
          ptr2->brother = ptr1->brother;
          loesch_Baum (ptr1->child);
          free (ptr1->lKlammer);
          free (ptr1->rKlammer);
          free (ptr1->Token);
          free (ptr1);

          ptr1 = ptr2->brother;
       }
       else {
         ptr1 = ptr1->brother;
         ptr2 = ptr2->brother;
       }
     }
    }
  }
 }
}

 





/*****************************************************/
/* FUNCTION  : form_aset                             */
/* PARAMETER : Baum *                                */
/* RETURN    : void                                  */
/* PURPOSE   : umformt aset (a,b,c,) zu aset (a,b,c).*/
/*             Nach der Anwendung der Funktion       */
/*             aset_aset oder Term_Term koennt die   */
/*             Form aset (a,b,c,) entstehen.         */
/*****************************************************/

void form_aset (Baum *bm)
{
   Baum *ptr;
 
   if (bm != NULL) {


   /* rekursive Anwendung der Funktion auf den Baum mit der Wurzel bm */
   if (bm->child != NULL)
     form_aset (bm->child);
   if (bm->brother != NULL)
     form_aset (bm->brother);

   if (strcmp (bm->Token, "aset") == 0 && bm->child != NULL && bm->child->child != NULL) {
     ptr = bm->child->child;
     while (ptr->brother != NULL)
       ptr = ptr->brother;
  
     if (strcmp (ptr->rKlammer, ",") == 0) {
       change = TRUE;
       ptr->rKlammer = (char *)strdup (" ");
     }
   }
}
}

/*****************************************************/
/* FUNCTION  : aset_most                             */
/* PARAMETER : Baum *                                */
/* RETURN    : void                                  */
/* PURPOSE   : wertet einen Ausdruck als 'NOTHING'   */
/*             aus, wenn er Teilausdruecke           */
/*             (all R (and ..aset (a(1)..a(m)))..)   */
/*             und (atmost n R) mit n < m enthaelt.  */
/*****************************************************/ 

void aset_most (Baum *bm)
{ 
   
   int m;
   Baum *ptr, *ptr1, *ptr2;
 
  
  if (bm != NULL) {

   /* rekursive Anwendung der Funktion auf den Baum mit der Wurzel bm */
   if (bm->child != NULL)
     aset_most (bm->child);
   if (bm->brother != NULL)
     aset_most (bm->brother);

   /* bm = (and A[1] ... A[n]) */
   if (strcmp (bm->Token, "and") == 0 ) {

     ptr = bm->child;
     while (ptr != NULL) {

       /* A[i] = (all R (and C[1]... C[u])) */
       if (strcmp (ptr->Token, "all") == 0 && ptr->child != NULL && 
           ptr->child->brother != NULL) { 

         /* ptr1 zeigt auf C[1] */
         ptr1 = ptr->child->brother->child; 
         
         while (ptr1 != NULL) {
           /* C[k] = aset (a1, ..., am) */
           if (strcmp (ptr1->Token, "aset") == 0) {
             m = number_aset (ptr1);

             /* Die Suche nach 'atmost' beginnt bei A[1]*/
             ptr2 = bm->child;
             while (ptr2 != NULL) {

               /* A[j] = (atmost v R), i <> j , v < m */
               if (strcmp (ptr2->Token, "atmost") == 0 && ptr2->child != NULL){
                 if (atoi (ptr2->child->Token) < m &&
                     ptr2->child->brother != NULL &&
                     strcmp (ptr2->child->brother->Token, 
                             ptr->child->Token) == 0 &&
                     vergleich (ptr2->child->brother->child, 
                                ptr->child->child) != -1) {
       
                   /* Den Ausdruck als 'NOTHING' auswerten. */
                   /* Nicht benoetigten Speicher freigeben */
       
                   change = TRUE;
       
                   bm->Token = (char *)strdup ("NOTHING");
                   loesch_Baum (bm->child);
                   bm->child = NULL;
       
                   goto Ende;
                 }              
               }

               /* Die weitere Suche nach 'atmost' */
               ptr2 = ptr2->brother;
             }
             break;
          }
          /* Die Weitere Suche nach 'aset' */
          else
            ptr1 = ptr1->brother;
        }
      }
      ptr = ptr->brother;
    }
  }

   Ende:;
 }
}



/*****************************************************/
/* FUNCTION  : aset_least                            */
/* PARAMETER : Baum *                                */
/* RETURN    : void                                  */
/* PURPOSE   : wertet einen Ausdruck als 'NOTHING'   */
/*             aus, wenn er die Teilausdruecke       */
/*             (all R (and ..aset (a(1)..a(m)))..)   */
/*             und (atleast n R) mit m < n enthaelt. */
/*****************************************************/ 

void aset_least (Baum *bm)
{ 
   
   int m;
   Baum *ptr, *ptr1, *ptr2;
   

   if (bm != NULL) {

   /* Durchlauf analog zur Funktion aset_most */

   if (bm->child != NULL)
     aset_least (bm->child);
   if (bm->brother != NULL)
     aset_least (bm->brother);

   if (strcmp (bm->Token, "and") == 0) {

     ptr = bm->child;
     while (ptr != NULL) {

      if (strcmp (ptr->Token, "all") == 0 &&
          ptr->child != NULL && ptr->child->brother != NULL && ptr->child->brother->child != NULL) {
        ptr1 = ptr->child->brother->child;
        while (ptr1 != NULL) {

          if (strcmp (ptr1->Token, "aset") == 0) {
             m = number_aset (ptr1);
             ptr2 = bm->child;

             while (ptr2 != NULL) {
               if (strcmp (ptr2->Token, "atleast") == 0){
                 if (ptr2->child != NULL &&
                     atoi (ptr2->child->Token) > m &&
                     ptr2->child->brother != NULL &&
                     strcmp (ptr2->child->brother->Token, 
                             ptr->child->Token) == 0 &&
                     vergleich (ptr2->child->brother->child, 
                                ptr->child->child) != -1) {
                   change = TRUE;
                
                   bm->Token = (char *)strdup ("NOTHING");
                   loesch_Baum (bm->child);
                   bm->child = NULL;
                
                   goto Ende;
                 }              
               }
               ptr2 = ptr2->brother;
             }
             break;
          }
          else
            ptr1 = ptr1->brother;
        }
      }
      ptr = ptr->brother;
    }
  }

   Ende:;
 }

}

/*****************************************************/
/* FUNCTION  : norm                                  */
/* PARAMETER : Baum *                                */
/* RETURN    : void                                  */
/* PURPOSE   : wendet alle obigen Funktionen auf     */
/*             eine Termdefinition an, bis keine     */      
/*             Veraenderung mehr stattfindet.        */
/*****************************************************/ 


Baum *norm (Baum *bm, Terme *tm)
{
   if (bm != NULL  && tm != NULL) {
 
     do {
       change = FALSE;
   
       and_and (bm);
       role_role (bm);
       C_notC (bm);
       all_least (bm, tm);
       term_term (bm);
       and_NOTHING (bm); 
       all_NOTHING (bm);
       all_and (bm);
       least_least (bm);
       most_most (bm);
       least_most (bm, tm);
       aset_aset (bm); 
       form_aset (bm);
       aset_least (bm);  
       aset_most (bm);
   
     } while (change == TRUE);

  }
   
  return bm;
}


/*****************************************************/
/* FUNCTION  : norm_Terme                            */
/* PARAMETER : Terme *tm                             */
/* RETURN    : void                                  */
/* PURPOSE   : wendet Funktion norm auf die ganzen   */
/*             Terminologie tm an.                   */
/*****************************************************/
 
void norm_Terme (Terme *tm)
{ 

   if (tm != NULL) {

     if (tm->expression->term !=NULL)
       norm (tm->expression->term, tm);
     if (tm->next != NULL) 
       norm_Terme (tm->next);
   }

}
