
int yyparse ();
FILE *yyin;

/****************************************************/
/*                                                  */
/*  urspruengliche Funktion Sub_Term                */
/*  Dokumentation im Originalquellcode              */
/*                                                  */
/****************************************************/

char* tewis_Sub_Term (char *str1, char *str2, Terme *tm)
{
   Terme *ptr1, *ptr2;

   /* Suche den erste Termbezeichner in der Terminologie */
   ptr1 = tm;

   while (ptr1 != NULL) {
     if (strcmp (ptr1->expression->ATerm, str1) != 0)
       ptr1 = ptr1->next;

     else {
       /* Suche den zweite Termbezeichner in der Terminologie */
       ptr2 = tm;

       while (ptr2 != NULL) {
         if (strcmp (ptr2->expression->ATerm, str2) != 0)
           ptr2 = ptr2->next;
         else {
           /* Beide Termbezeichner sind in */
           /* der Terminologie definiert.  */

           if (comp (ptr1->expression->term, 
                     ptr2->expression->term) == 1) {

             if (strcmp (ptr1->expression->equal, ":=C") == 0)
                 return("Das Konzept A subsummiert das Konzept B.");
             else              
               return ("Die Rolle A subsummiert die Rolle B.");
	   }

           else { 

             if (strcmp (ptr1->expression->equal, ":=C") == 0)
               return ("Das Konzept A subsummiert nicht das Konzept B.");
             else              
               return ("Die Rolle A subsummiert nicht die Rolle B.");
	   }
         }
       }

       if (ptr2 == NULL) {
         return ("B ist nicht in der Terminologie definiert.");
       }
     }
   }
    
   if (ptr1 == NULL) {
        return ("A ist nicht in der Terminologie definiert.");      
   }
}


/******************************************************/
/* FUNKTION  : tewis_Subsumption_AneuBneu             */
/* ARGUMENTE : 2 Termdefinitionen                     */
/* ZWECK     : Ueberpruefung, ob der definierende     */
/*             Term von A den von B subsummiert.      */
/******************************************************/

int tewis_Subsumption_AneuBneu(ClientData clientData, Tcl_Interp *interp,int argc,char *argv[])
{
    FILE *fp;
    Terme *tm1, *ptr;

    if (T == NULL){
      interp->result = "Die Terminologiedatei existiert noch nicht.";
      return TCL_ERROR;
    }
    if (argc != 3) {
      interp->result = "Falsche Anzahl von Argumenten.";
      return TCL_ERROR;
    }

    /* Die Termdefinitionen eingeben und die */
    /* Datei im File 'tmpfile' speichern.    */

    fp = fopen ("tmpfile", "w+");
    if (fp == NULL) {
      interp->result = "tmpfile in Subsumption_AneuBneu kann nicht zum Schreiben geoeffnet werden.";
      return TCL_ERROR;
    } 

    /* Aenderung gegenueber Sub1: Die Definitionen von A und B 
       werden gleichzeitig in die Datei geschrieben */
    fprintf (fp, "%s;", argv[1]);
    fprintf (fp, "%s", argv[2]);
    fclose (fp);

    /* Die neuen Termdefinitionen werden vom lex und yacc geprueft */
    yyin = fopen ("tmpfile", "r");
    yyparse();
    
    sort1 (ITF);
    sort2 (ITF);
 
 
    /* Die neue Termdefinitionen am Ende der Terminologie anhaengen. */ 
    tm1 = copy_Terme (T);
    ptr = tm1;

    while (ptr->next != NULL) 
      ptr = ptr->next;


    ptr->next = (void*)Node_Terme ( );
    ptr->next = ITF;
    ptr = ptr->next;  
    
    form_Terme (tm1);
    all_exp_Terme (tm1);
    norm_Terme (tm1);
    
    /* Subsumptionstest */
    if (tm1 == NULL){printf("NULL");}
    interp->result = tewis_Sub_Term (ptr->expression->ATerm, ptr->next->expression->ATerm, tm1);

    loesch_Terme (&tm1);
    return TCL_OK;
}       

/******************************************************/
/* FUNKTION  : tewis_Subsumption_ABneu                */
/* ARGUMENTE : Termbezeichner, Termdefinition         */
/* ZWECK     : Ueberpruefung, ob der definierende     */
/*             Term von A den von B subsummiert.      */
/******************************************************/

int tewis_Subsumption_ABneu(ClientData clientData, Tcl_Interp *interp,int argc,char *argv[])
{
    char str[36]; 
    FILE *fp;
    Terme *tm1, *ptr;
    if (T == NULL){
      interp->result = "Die Terminologiedatei existiert noch nicht.";
      return TCL_ERROR;
    }
    if (argc != 3) {
      interp->result = "Falsche Anzahl von Argumenten.";
      return TCL_ERROR;
    }
    strncpy(str,argv[1],36);

    /* Die Termdefinition eingeben und die */
    /* Datei im File 'tmpfile' speichern.  */
    fp = fopen ("tmpfile", "w+");
    if (fp == NULL) {
      interp->result = "tmpfile in Subsumption_ABneu kann nicht zum Schreiben geoeffnet werden.";
      return TCL_ERROR;
    } 

    /* Aenderung gegenber Sub2: Die ganze Definition wird auf einmal in das File geschrieben */
    fprintf (fp, "%s", argv[2]);
    fclose (fp);

    /* Die neue Termdefinition wird vom lex und yacc geprueft */
    yyin = fopen ("tmpfile", "r");
    yyparse();

    sort1 (ITF);
    sort2 (ITF);

    /* Die neue Termdefinitionen am Ende der Terminologie anhaengen. */ 
    tm1 = copy_Terme (T);
    ptr = tm1;
    while (ptr->next != NULL)
      ptr = ptr->next;
    ptr->next = (void*)Node_Terme ( );
    ptr->next = ITF;
    ptr = ptr->next;  
    
    form_Terme (tm1);
    all_exp_Terme (tm1);
    norm_Terme (tm1);

    /* Subsumptionstest */    
    interp->result = tewis_Sub_Term (str, ptr->expression->ATerm, tm1);

    loesch_Terme (&tm1);
    return TCL_OK;
}       


/******************************************************/
/* FUNKTION  : tewis_Subsumption_AneuB                */
/* ARGUMENTE : Termdefinition, Termbezeichner         */
/* ZWECK     : Ueberpruefung, ob der definierende     */
/*             Term von A den von B subsummiert.      */
/******************************************************/

int tewis_Subsumption_AneuB(ClientData clientData, Tcl_Interp *interp,int argc,char *argv[])
{
    char str[36]; 
    FILE *fp;
    Terme *tm1, *ptr;
    if (T == NULL){
      interp->result = "Die Terminologiedatei existiert noch nicht.";
      return TCL_ERROR;
    }
    if (argc != 3) {
      interp->result = "Falsche Anzahl von Argumenten.";
      return TCL_ERROR;
    }

    strncpy(str,argv[2],36);

    /* Die Termdefinition eingeben und die */
    /* Datei im File 'tmpfile' speichern.  */
    fp = fopen ("tmpfile", "w+");
    if (fp == NULL) {
      interp->result = "tmpfile in Subsumption_AneuB kann nicht zum Schreiben geoeffnet werden.";
      return TCL_ERROR;   
    } 

    /* Aenderung gegenber Sub3: Die ganze Definition wird auf einmal in das File geschrieben */
    fprintf(fp,"%s",argv[1]);
    fclose (fp);

    /* Die neue Termdefinition wird vom lex und yacc geprueft */
    yyin = fopen ("tmpfile", "r");
    yyparse();

    sort1 (ITF);
    sort2 (ITF);

    /* Die neue Termdefinitionen am Ende der Terminologie anhaengen. */ 
    tm1 = copy_Terme (T);
    ptr = tm1;
    while (ptr->next != NULL)
      ptr = ptr->next;
    ptr->next = (void*)Node_Terme ( );
    ptr->next = ITF;
    ptr = ptr->next;  
    
    form_Terme (tm1);
    all_exp_Terme (tm1);
    norm_Terme (tm1);

    /* Subsumptionstest */    
    interp->result = tewis_Sub_Term (ptr->expression->ATerm, str, tm1);

    loesch_Terme (&tm1);
    return TCL_OK;
}       

/******************************************************/
/* FUNKTION  : tewis_Subsumption_AB                   */
/* ARGUMENTE : 2 Termbezeichner                       */
/* ZWECK     : Ueberpruefung, ob der definierende     */
/*             Term von A den von B subsummiert.      */
/******************************************************/

int tewis_Subsumption_AB(ClientData clientData, Tcl_Interp *interp,int argc,char *argv[]){
  char str1[36], str2[36];
  if (T == NULL){
    interp->result = "Die Terminologiedatei existiert noch nicht.";
    return TCL_ERROR;
  }
  if (argc != 3) {
    interp->result = "Falsche Anzahl von Argumenten.";
    return TCL_ERROR;
  }
  strncpy(str1,argv[1],36);
  strncpy(str2,argv[2],36);
  interp->result = tewis_Sub_Term (str1,str2,T);
  return TCL_OK;
}
