/*********************  tree.c  ******************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tree.h"
#include "tree1.h"




/*********************************************************************/
/*   Die Funktionen von tree.c werden in yacc_File 'ITL.y' verwendet,*/
/*   um eine Kopie von Datenstrukturen herzustellen, die waehrend    */
/*   der Ausfuehrung von 'ITL.y' erzeugt werden.                     */
/*********************************************************************/




/* Ein Konzept oder eine Rolle wird als Datentyp Baum definiert.  */
/* Folgende Funktionen werden verwendet, um verschiedene Form des */
/* Konzepts oder der Rolle als Datentyp Baum zu speichern.        */


/*****************************************************/
/* FUNCTION  : Aus_0                                 */
/* PARAMETER : char *                                */
/* RETURN    : Baum *                                */
/* PURPOSE   : Fuer den Fall:                        */
/*               <Konzept> ::= <atomaresKonzept> |   */
/*                     <primitiveKonzeptkomponente>  */
/*             oder                                  */
/*               <Rolle> ::= <atomareRolle> |        */
/*                     <primitiveKonzeptkomponente>  */
/*****************************************************/
 
Baum *Aus_0 (char *arg1)
{
   Baum *ptr;

   /* Der Baum hat einen Knoten (ptr). */

   /* Erstellen eines Baums */

   ptr = Node_Baum ( );
   ptr->lKlammer = (char *) strdup (" ");
   ptr->rKlammer = (char *) strdup (" ");
   ptr->Token = (char *) strdup (arg1);
            
   ptr->child = NULL;                                         
   ptr->brother = NULL;
   ptr->father = NULL;
          
   return ptr;
}



/*****************************************************/
/* FUNCTION  : Aus_1                                 */
/* PARAMETER : char * char * char * char *           */
/* RETURN    : Baum *                                */
/* PURPOSE   : Fuer den Fall:                        */
/*             <Konzept> ::=                         */
/*                (not <primitiveKonzeptkomponente>) */ 
/*****************************************************/

Baum *Aus_1 (char *arg1, char *arg2, char *arg3, char *arg4)
{
   Baum *ptr;

   /* Der Baum hat zwei Knoten. */
   /* Knoten 1 (ptr) hat child Knoten 2 (ptr->child). */
 
   /* Erstellen des Knotens 1 */
   ptr = Node_Baum ( );
   ptr->lKlammer = (char *)strdup (arg1);
   ptr->Token = (char *)strdup (arg2);
   ptr->rKlammer = (char *) strdup (arg4);
   ptr->father = NULL;
   ptr->brother = NULL;
   
   /* Erstellen des Knotens 2 */
   ptr->child = Node_Baum ( ); 
   ptr->child->lKlammer = (char *)strdup (" ");
   ptr->child->rKlammer = (char *)strdup (" ");
   ptr->child->Token = (char *)strdup (arg3);
   ptr->child->child = NULL;
   ptr->child->brother = NULL;
   ptr->child->father = ptr;
  

   return ptr;
}



/*****************************************************/
/* FUNCTION  : NodeAus                               */
/* PARAMETER : Baum * Baum *                         */
/* RETURN    : Baum *                                */
/* PURPOSE   : In dem Fall:                          */
/*                Konzept ::= (and <Konzept>+)       */
/*             oder                                  */
/*                Rolle ::= (androle <Rolle>+)       */
/*             wird <Konzept>+ oder <Rolle>+ als     */
/*             Datentyp Baum gespeichert.            */
/*****************************************************/

Baum *NodeAus (Baum *arg1, Baum *arg2)
{
   Baum *ptr;

   /* Drei Baumknoten werden erzeugt. */
   /* KnoDten 1 hat 'child' Knoten 2 (ptr->child). */ 
   /* Knoten 2 hat 'brother' Knoten 3 (ptr->child->brother). */
 
   /* Erstellen des Knotens 1 */
   ptr = Node_Baum ( );
   ptr->lKlammer = (char *)strdup (" ");
   ptr->Token = (char *)strdup (" ");
   ptr->rKlammer = (char *)strdup (" ");
   ptr->father = NULL;
   ptr->brother = NULL;
   
   /* Erstellen des Knotens 2 */
   ptr->child = copy_Baum (arg1, ptr);
          
   /* Erstellen des Knotens 3 */
   ptr->child->brother = copy_Baum (arg2, ptr);
          
   return ptr;
}



/*****************************************************/
/* FUNCTION  : Aus_2                                 */
/* PARAMETER : char * char * Baum * char *           */
/* RETURN    : Baum *                                */
/* PURPOSE   : Fuer den Fall:                        */
/*               <Konzept> ::= ( and <Konzept>+ )    */
/*             oder                                  */
/*               <Rolle> ::= ( androle <Rolle>+ )    */
/*****************************************************/

Baum *Aus_2 
   (char *arg1, char *arg2, Baum *arg3, char *arg4)
{
   Baum *ptr;
 
   /* Zwei Baumknoten, Knoten 1 (ptr) und sein 'child' */
   /* Knoten 2 (ptr->child) werden erzeugt.            */
 
   /* Erstellen des Knotens 1 */
   ptr = Node_Baum ( );
   ptr->lKlammer = (char *)strdup(arg1);
   ptr->Token =(char *)strdup(arg2);
   ptr->rKlammer = (char *)strdup(arg4);
   ptr->father = NULL;
   ptr->brother = NULL;

   /* Erstellen des Knotens 2 */
   ptr->child = copy_Baum (arg3, ptr);
    
   return ptr;
}



/*****************************************************/
/* FUNCTION  : Aus_3                                 */
/* PARAMETER : char * char * Baum * Baum * char *    */
/* RETURN    : Baum *                                */
/* PURPOSE   : Fuer den Fall                         */
/*             <Konzept> ::= (all <Rolle> <Konzept>)|*/
/*                        (exists <Rolle> <Konzept>) */
/*****************************************************/

Baum *Aus_3
    (char *arg1, char *arg2, Baum *arg3, Baum *arg4, char *arg5)
{
   Baum *ptr;

   /* Drei Baumknoten werden erzeugt. */
   /* Knoten 1 (ptr) hat 'child' Knoten 2 (ptr->child). */ 
   /* Knoten 2 hat 'brother' Knoten 3 (ptr->child->brother). */

   /* Erstellen des Knotens 1 */ 
   ptr =  Node_Baum ( );
   ptr->lKlammer = (char *)strdup (arg1);
   ptr->Token = (char *)strdup (arg2);
   ptr->rKlammer = (char *)strdup (arg5);
   ptr->father = NULL;
   ptr->brother = NULL;
   
   /* Erstellen des Knotens 2 */
   ptr->child = copy_Baum (arg3, ptr);
        
   /* Erstellen des Knotens 3 */
   ptr->child->brother = copy_Baum (arg4, ptr);
   
   return ptr;  
}



/*****************************************************/
/* FUNCTION  : Aus_4                                 */
/* PARAMETER : char * char * Baum * char *           */
/* RETURN    : Baum *                                */
/* PURPOSE   : Fuer den Fall:                        */
/*             <Konzept> ::= aset (<AttributeListe>) */
/*****************************************************/

Baum *Aus_4 (char *arg1, char *arg2, Baum *arg3, char *arg4)
{

   Baum *ptr;

   /* Drei Baumknoten werden erzeugt. */
   /* Knoten 1 hat 'child' Knoten 2 (ptr->child). */ 
   /* Knoten 2 hat 'child' Knoten 3 (ptr->child->child). */
 
   /* Erstellen des Knotens 1 */ 
   ptr = Node_Baum ( );
   ptr->lKlammer = (char *)strdup (" ");
   ptr->Token = (char *)strdup (arg1);
   ptr->rKlammer = (char *)strdup (" ");
   ptr->brother = NULL;
   ptr->father = NULL;

   /* Erstellen des Knotens 2 */ 
   ptr->child = Node_Baum ( );
   ptr->child->lKlammer = (char *)strdup (arg2);
   ptr->child->rKlammer = (char *)strdup (arg4);
   ptr->child->Token = (char *)strdup (" ");
   ptr->child->brother = NULL;
   ptr->child->father = ptr;

   /* Erstellen des Knotens 1 */ 
   ptr->child->child = copy_Baum (arg3, ptr->child);
      
   return ptr;
}


/*****************************************************/
/* FUNCTION  : Aus_5                                 */
/* PARAMETER : char * char * Baum *                  */
/* RETURN    : Baum *                                */
/* PURPOSE   : Fuer den Fall:                        */
/*             <AttributeListe> ::= <Attribute> |    */
/*                    <Attribute>, <AttributeListe>  */
/*             um <Attribute>, <AttributeListe> als  */
/*             Baum zu speichern.                    */
/*****************************************************/

Baum *Aus_5 (char *arg1, char *arg2, Baum *arg3)
{
   Baum *ptr;

   /* Zwei Baumknoten werden erzeugt. */
   /* Knoten 1  und sein 'brother' Knoten 2. */
   
   /* Erstellen des Knotens 1 */ 
   ptr = Node_Baum ( );
   ptr->lKlammer = (char *)strdup (" ");
   ptr->Token = (char *)strdup (arg1);
   ptr->rKlammer = (char *)strdup ( arg2);
   ptr->child = NULL;
   ptr->father = NULL;

   /* Erstellen des Knotens 2 */ 
   ptr->brother = (Baum *)copy_Baum (arg3, NULL);

   return ptr;
}


/*****************************************************/
/* FUNCTION  : Aus_6                                 */
/* PARAMETER : char * char * char * Baum * char *    */
/* RETURN    : Baum *                                */
/* PURPOSE   : Fuer den Fall:                        */
/*              <Konzept> ::=(atleast <nr> <Rolle>)| */
/*                           (atmost <znr> <Rolle>)  */
/*****************************************************/

Baum *Aus_6
    (char *arg1, char *arg2, char *arg3, Baum *arg4, char *arg5)
{
   Baum *ptr;

   /* Drei Baumknoten werden erzeugt. */
   /* Knoten 1 (ptr) hat 'child' Knoten 2 (ptr->child). */ 
   /* Knoten 2 hat 'brother' Knoten 3 (ptr->child->brother). */
 
   /* Erstellen des Knotens 1 */ 
   ptr =  Node_Baum ( );
   ptr->lKlammer = (char *)strdup (arg1);
   ptr->Token = (char *)strdup (arg2);
   ptr->rKlammer = (char *)strdup (arg5);
   ptr->father = NULL;
   ptr->brother = NULL;
 
   /* Erstellen des Knotens 2 */ 
   ptr->child = Node_Baum ( );
   ptr->child->lKlammer = (char *)strdup (" ");
   ptr->child->rKlammer = (char *)strdup (" ");
   ptr->child->Token = (char *) strdup (arg3);
   ptr->child->child = NULL; 
   ptr->child->father = ptr; 

   /* Erstellen des Knotens 3 */ 
   ptr->child->brother = copy_Baum (arg4, ptr);
     
   return ptr;  
}


/*****************************************************/
/* FUNCTION  : NodeDef                               */
/* PARAMETER : char * char * Baum *                  */
/* RETURN    : Term *                                */
/* PURPOSE   : Termdefinition wird als Datentyp Term */
/*             gespeichert.                          */
/*****************************************************/

Term *NodeDef (char *arg1, char *arg2, Baum *arg3)
{
  Term *ptr;

  /* Erstellen des Datentyps Term */
  ptr = Node_Term ( );
  ptr->ATerm = (char *)strdup(arg1);
  ptr->equal = (char *)strdup(arg2);
  ptr->term = arg3;

  return ptr;
}




/* Eine Weltbeschreibung wird als Datentyp Liste gespeichert. */
/* Folgende Funktionen werden verwendet, um die verschiedene  */
/* Form der Weltbeschreibung als Liste zu speichern.          */


/*****************************************************/
/* FUNCTION  : Welt4                                 */
/* PARAMETER : char * char * char * char *           */
/* RETURN    : Liste *                               */
/* PURPOSE   : Fuer den Fall:                        */
/*               <Konzeptinstantiierung> ::=         */
/*                    (<Objekt> <atomaresKonzept>)   */
/*****************************************************/

Liste *Welt4 (char *arg1, char *arg2, char *arg3, char *arg4)
{
  Liste *first, *last; 

  
  /* Erstellen des ersten Knotens der Liste */
  first = Node_Liste ();     
  strcpy (first->str, arg1);

  /* Erstellen des zweiten Knotens der Liste */
  last = first;
  last->next = Node_Liste ();
  last = last->next;
  strcpy (last->str, arg2);

  /* Erstellen des dritten Knotens der Liste */
  last->next = Node_Liste ();
  last = last->next;
  strcpy (last->str, arg3);

  /* Erstellen des vierten Knotens der Liste */
  last->next = Node_Liste ();
  last = last->next;
  strcpy (last->str, arg4);

  /* Ende der Liste */
  last->next = NULL;
 
  return first;
}



/*****************************************************/
/* FUNCTION  : Welt5                                 */
/* PARAMETER : char * char * char * char * char *    */
/* RETURN    : Liste *                               */
/* PURPOSE   : Fuer den Fall:                        */
/*              <Konzeptinstantiierung> ::=          */
/*              (<Objekt> <atomaresKonzept><Objekt>) */
/*****************************************************/

Liste *Welt5 (char *arg1, char *arg2, char *arg3, char *arg4,
              char *arg5)
{
  Liste *first, *last;
  
  /* Erstellen einer Liste mit vier Knoten */
  first = Welt4 (arg1, arg2, arg3, arg4);

  /* Suchen des Endes der erstellten Liste */
  last = first;
  while (last->next != NULL)
    last = last->next;

  /* Ein weiterer Knoten wird am Ende der Liste angehaengt. */
  last->next = Node_Liste ( );
  last = last->next;
  strcpy (last->str, arg5);

  /* Ende der Liste */
  last->next = NULL;
   
  return first;
}



/*****************************************************/
/* FUNCTION  : Welt7                                 */
/* PARAMETER : char * char * char * char * char *    */
/*             char * char *                         */
/* RETURN    : Liste *                               */
/* PURPOSE   : Fuer den Fall:                        */
/*              <Konzeptinstantiierung> ::=          */
/*    (<Objekt> (not <primitiveKonzeptkomponente>))  */
/*****************************************************/

Liste *Welt7 (char *arg1, char *arg2, char *arg3, char *arg4,
              char *arg5, char *arg6,  char *arg7)
{
  Liste *first, *last;

  /* Erstellen einer Liste mit fuenf Knoten */  
  first = Welt5 (arg1, arg2, arg3, arg4, arg5);

  /* Suchen des Endes der erstellten Liste */
  last = first;
  while (last->next != NULL)
    last = last->next;

  /* Zwei weitere Knoten werden am Ende der Liste angehaengt. */

  last->next = Node_Liste ( );
  last = last->next;
  strcpy (last->str, arg6);

  last->next = Node_Liste ( );
  last = last->next;
  strcpy (last->str, arg7);

  /* Ende der Liste */
 
  last->next = NULL;
 
  return first;
}


/*****************************************************/
/* FUNCTION  : Welt8                                 */
/* PARAMETER : char * char * char * char * char *    */
/*             char * char * char *                  */
/* RETURN    : Liste *                               */
/* PURPOSE   : Fuer den Fall:                        */
/*              <Konzeptinstantiierung> ::=          */  
/*      (<Objekt> (<NumRestriktion> <atomareRolle>)) */
/*****************************************************/

Liste *Welt8 (char *arg1, char *arg2, char *arg3, char *arg4,
              char *arg5, char *arg6, char *arg7, char *arg8)
{
  Liste *first, *last;
   
  /* Erstellen einer Liste mit sieben Knoten */  
  first = Welt7 (arg1, arg2, arg3, arg4, arg5, arg6, arg7);

  /* Suchen des Endes der erstellten Liste */
  last = first;
  while (last->next != NULL)
    last = last->next;
   
  /* Ein weiterer Knoten wird am Ende der Liste angehaengt. */
  last->next = Node_Liste ( );
  last = last->next;
  strcpy (last->str, arg8);

  /* Ende der Liste */
  last->next = NULL;
  
  return first;
}
